/*
 *   Copyright (c) International Business Machines  Corp., 2001
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: dm.c
 */

#include <stdio.h>
#include <string.h>
#include <sys/ioctl.h>

#include "fullengine.h"
#include "engine.h"
#include "memman.h"
#include "dm.h"
#include "dm-ioctl.h"
#include "remote.h"

#define ALIGNMENT		sizeof(u_int64_t)
#define IOCTL_MIN_LENGTH	16 * 1024

/**
 * build_linear_params
 *
 * Generate an ASCII string for a linear target. A linear string has the form:
 * <major>:<minor> <start_lba>
 **/
static int build_linear_params(dm_target_t *target)
{
	dm_device_t *dev = target->data.linear;
	int rc = ENOMEM;

	LOG_PROC_ENTRY();

	target->params = engine_alloc(40);
	if (target->params) {
		snprintf(target->params, 40, "%x:%x %"PRIu64"",
			 dev->major, dev->minor, dev->start);
		rc = 0;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * translate_linear_params
 *
 * Fill in a linear target structure based on an ASCII string of the form:
 * <major>:<minor> <start_lba>
 *   or
 * name(<major>,<minor>) <start_lba>
 **/
static int translate_linear_params(dm_target_t *target)
{
	dm_device_t *dev = target->data.linear;
	int rc;

	LOG_PROC_ENTRY();

	rc = sscanf(target->params, "%x:%x %"SCNu64,
		    &dev->major, &dev->minor, &dev->start);
	if (rc != 3) {
		rc = sscanf(target->params, "%*[^(](%d,%d) %"SCNu64,
			&dev->major, &dev->minor, &dev->start);
	}
	rc = (rc != 3) ? EINVAL : 0;

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * build_stripe_params
 *
 * Generate an ASCII string for a stripe target. A stripe string has the form:
 * <stripes> <chunk_size> [<major>:<minor> <start_lba>]+
 **/
static int build_stripe_params(dm_target_t *target)
{
	dm_target_stripe_t *stripe = target->data.stripe;
	char params[40];
	int max_len = 25 + stripe->num_stripes * 40;
	int i, rc = ENOMEM;

	LOG_PROC_ENTRY();

	target->params = engine_alloc(max_len);
	if (target->params) {
		snprintf(target->params, max_len, "%u %u",
			 stripe->num_stripes, stripe->chunk_size);
		for (i = 0; i < stripe->num_stripes; i++) {
			snprintf(params, 40, " %x:%x %"PRIu64"",
				 stripe->devices[i].major,
				 stripe->devices[i].minor,
				 stripe->devices[i].start);
			strncat(target->params, params,
				max_len - strlen(target->params));
		}
		rc = 0;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * translate_stripe_params
 *
 * Fill in a stripe target structure based on an ASCII string of the form:
 *  <stripes> <chunk_size> [<major>:<minor> <start_lba>]+
 *    or
 *  <stripes> <chunk_size> [name(<major>,<minor>) <start_lba>]+
 **/
static int translate_stripe_params(dm_target_t *target)
{
	dm_target_stripe_t *stripe = target->data.stripe;
	char *params = target->params;
	int i, rc;

	LOG_PROC_ENTRY();

	rc = sscanf(params, "%u %u", &stripe->num_stripes, &stripe->chunk_size);
	if (rc != 2) {
		rc = EINVAL;
		goto out;
	}

	for (i = 0; i < stripe->num_stripes; i++) {
		params = strstr(params, " ");
		if (!params) {
			rc = EINVAL;
			goto out;
		}
		params++;
		params = strstr(params, " ");
		if (!params) {
			rc = EINVAL;
			goto out;
		}
		params++;
		rc = sscanf(params, "%x:%x %"SCNu64,
			    &(stripe->devices[i].major),
			    &(stripe->devices[i].minor),
			    &(stripe->devices[i].start));
		if (rc != 3) {
			rc = sscanf(params, "%*[^(](%d,%d) %"SCNu64,
				    &(stripe->devices[i].major),
				    &(stripe->devices[i].minor),
				    &(stripe->devices[i].start));
		}
		if (rc != 3) {
			rc = EINVAL;
			goto out;
		}
	}

	rc = 0;
out:
	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/**
 * build_mirror_params
 *
 * Generate an ASCII string for a mirror target. A mirror string has the form:
 * <major1>:<minor1> <start_lba1> <major2>:<minor2> <start_lba2> <chunk_size> [<priority>]
 **/
static int build_mirror_params(dm_target_t *target)
{
	dm_target_mirror_t *mirror = target->data.mirror;
	char params[15];
	int rc = ENOMEM;

	LOG_PROC_ENTRY();

	target->params = engine_alloc(100);
	if (target->params) {
		snprintf(target->params, 100, "%x:%x %"PRIu64" %x:%x %"PRIu64" %u",
			 mirror->dev1.major, mirror->dev1.minor,
			 mirror->dev1.start, mirror->dev2.major,
			 mirror->dev2.minor, mirror->dev2.start,
			 mirror->chunk_size);
		if (mirror->priority >= 0) {
			snprintf(params, 15, " %u", mirror->priority);
			strncat(target->params, params,
				100 - strlen(target->params));
		}
		rc = 0;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * translate_mirror_params
 *
 * Fill in a mirror target structure based on an ASCII string of the form:
 * <major1>:<minor1> <start_lba> <major2>:<minor2> <start_lba> <chunk_size> <priority>
 *   or
 * name(<major1>,<minor1>) <start_lba1> name(<major2>,<minor2>) <start_lba2> <chunk_size> <priority>
 **/
static int translate_mirror_params(dm_target_t *target)
{
	dm_target_mirror_t *mirror = target->data.mirror;
	int rc;

	LOG_PROC_ENTRY();

	rc = sscanf(target->params, "%x:%x %"SCNu64" %x:%x %"SCNu64" %u %u",
		    &mirror->dev1.major, &mirror->dev1.minor,
		    &mirror->dev1.start, &mirror->dev2.major,
		    &mirror->dev2.minor, &mirror->dev2.start,
		    &mirror->chunk_size, &mirror->priority);
	if (rc != 7 || rc != 8) {
		rc = sscanf(target->params, "%*[^(](%d,%d) %"SCNu64" %*[^(](%d,%d) %"SCNu64" %u %u",
			    &mirror->dev1.major, &mirror->dev1.minor,
			    &mirror->dev1.start, &mirror->dev2.major,
			    &mirror->dev2.minor, &mirror->dev2.start,
			    &mirror->chunk_size, &mirror->priority);
	}
	rc = (rc != 7 || rc != 8) ? EINVAL : 0;

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * build_snapshot_params
 *
 * Generate an ASCII string for a snapshot target. A snapshot string has the form:
 * <org_major>:<org_minor> <snap_major>:<snap_minor> <p|n> <chunk_size> <org_parent_major>:<org_parent_minor>
 **/
static int build_snapshot_params(dm_target_t *target)
{
	dm_target_snapshot_t *snapshot = target->data.snapshot;
	int rc = ENOMEM;

	LOG_PROC_ENTRY();

	target->params = engine_alloc(50);
	if (target->params) {
		snprintf(target->params, 50, "%x:%x %x:%x %c %u %x:%x",
			 snapshot->origin.major, snapshot->origin.minor,
			 snapshot->snapshot.major, snapshot->snapshot.minor,
			 (snapshot->persistent) ? 'p' : 'u', snapshot->chunk_size,
			 snapshot->origin_parent.major,
			 snapshot->origin_parent.minor);
		rc = 0;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * translate_snapshot_params
 *
 * Fill in a snapshot target structure based on an ASCII string of the form:
 * <org_major>:<org_minor> <snap_major>:<snap_minor> <p|n> <chunk_size>
 *   or
 * name(<org_major>,<org_minor>) name(<snap_major>,<snap_minor>) <p|n> <chunk_size>
 **/
static int translate_snapshot_params(dm_target_t *target)
{
	dm_target_snapshot_t *snapshot = target->data.snapshot;
	char persistent;
	int rc;

	LOG_PROC_ENTRY();

	rc = sscanf(target->params, "%x:%x %x:%x %c %u",
		    &snapshot->origin.major, &snapshot->origin.minor,
		    &snapshot->snapshot.major, &snapshot->snapshot.minor,
		    &persistent, &snapshot->chunk_size);
	if (rc != 6) {
		rc = sscanf(target->params, "%*[^(](%d,%d) %*[^(](%d,%d) %c %u",
			    &snapshot->origin.major, &snapshot->origin.minor,
			    &snapshot->snapshot.major, &snapshot->snapshot.minor,
			    &persistent, &snapshot->chunk_size);
	}
	rc = (rc != 6) ? EINVAL : 0;
	if (persistent == 'p' || persistent == 'P') {
		snapshot->persistent = 1;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * build_origin_params
 *
 * Generate an ASCII string for a snapshot origin target. A snapshot origin
 * string has the form:
 * <org_major>:<org_minor>
 **/
static int build_origin_params(dm_target_t *target)
{
	dm_device_t *org = target->data.linear;
	int rc = ENOMEM;

	LOG_PROC_ENTRY();

	target->params = engine_alloc(25);
	if (target->params) {
		snprintf(target->params, 25, "%x:%x",
			 org->major, org->minor);
		rc = 0;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * translate_origin_params
 *
 * Fill in a snapshot origin target structure based on an ASCII string of the
 * form:
 * <org_major>:<org_minor>
 *  or
 * name(<org_major>,<org_minor>)
 **/
static int translate_origin_params(dm_target_t *target)
{
	dm_device_t *org = target->data.linear;
	int rc;

	LOG_PROC_ENTRY();

	rc = sscanf(target->params, "%x:%x",
		    &org->major, &org->minor);
	if (rc != 2) {
		rc = sscanf(target->params, "%*[^(](%d,%d)",
			    &org->major, &org->minor);
	}
	rc = (rc != 2) ? EINVAL : 0;

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * build_bbr_params
 *
 * Generate an ASCII string for a bbr target. A bbr string has the form:
 * <major>:<minor> <start> <table1_lba> <table2_lba> <table_size> <replacement_blks_lba> <num_replacement_blks> <blk_size>
 **/
static int build_bbr_params(dm_target_t *target)
{
	dm_target_bbr_t *bbr = target->data.bbr;
	int rc = ENOMEM;

	LOG_PROC_ENTRY();

	target->params = engine_alloc(150);
	if (target->params) {
		snprintf(target->params, 150, "%x:%x %"PRIu64" %"PRIu64" %"PRIu64" %"PRIu64" %"PRIu64" %"PRIu64" %u",
			 bbr->device.major, bbr->device.minor,
			 bbr->device.start,
			 bbr->table1_lba, bbr->table2_lba, bbr->table_size,
			 bbr->replacement_blocks_lba,
			 bbr->num_replacement_blocks,
			 bbr->block_size);
		rc = 0;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * translate_bbr_params
 *
 * Fill in a bbr target structure based on an ASCII string of the form:
 * <major>:<minor> <table1_lba> <table2_lba> <table_size> <replacement_blks_lba> <num_replacement_blks> <blk_size>
 *   or
 * name(<major>,<minor>) <table1_lba> <table2_lba> <table_size> <replacement_blks_lba> <num_replacement_blks> <blk_size>
 **/
static int translate_bbr_params(dm_target_t *target)
{
	dm_target_bbr_t *bbr = target->data.bbr;
	int rc;

	LOG_PROC_ENTRY();

	rc = sscanf(target->params, "%x:%x %"SCNu64" %"SCNu64" %"SCNu64" %"SCNu64" %"SCNu64" %"SCNu64" %u",
		    &bbr->device.major, &bbr->device.minor, &bbr->device.start,
		    &bbr->table1_lba, &bbr->table2_lba, &bbr->table_size,
		    &bbr->replacement_blocks_lba,
		    &bbr->num_replacement_blocks,
		    &bbr->block_size);
	if (rc != 9) {
		rc = sscanf(target->params, "%*[^(](%d,%d) %"SCNu64" %"SCNu64" %"SCNu64" %"SCNu64" %"SCNu64" %"SCNu64" %u",
			    &bbr->device.major, &bbr->device.minor, &bbr->device.start,
			    &bbr->table1_lba, &bbr->table2_lba, &bbr->table_size,
			    &bbr->replacement_blocks_lba,
			    &bbr->num_replacement_blocks,
			    &bbr->block_size);
	}
	rc = (rc != 9) ? EINVAL : 0;

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * build_sparse_params
 *
 * Generate an ASCII string for a sparse target. A sparse string has the form:
 * <major>:<minor> <start_lba> <chunk_size> <num_chunks>
 **/
static int build_sparse_params(dm_target_t *target)
{
	dm_target_sparse_t *sparse = target->data.sparse;
	int rc = ENOMEM;

	LOG_PROC_ENTRY();

	target->params = engine_alloc(65);
	if (target->params) {
		snprintf(target->params, 65, "%x:%x %"PRIu64" %u %u",
			 sparse->device.major, sparse->device.minor,
			 sparse->device.start, sparse->chunk_size,
			 sparse->num_chunks);
		rc = 0;
		
	}

	LOG_PROC_EXIT_INT(rc)
	return rc;
}

/**
 * translate_sparse_params
 *
 * Fill in a sparse target structure based on an ASCII string of the form:
 * <major>:<minor> <start_lba> <chunk_size> <num_chunks>
 *   or
 * name(<major>,<minor>) <start_lba> <chunk_size> <num_chunks>
 **/
static int translate_sparse_params(dm_target_t *target)
{
	dm_target_sparse_t *sparse = target->data.sparse;
	int rc;

	LOG_PROC_ENTRY();

	rc = sscanf(target->params, "%x:%x %"SCNu64" %u %u",
		    &sparse->device.major, &sparse->device.minor,
		    &sparse->device.start, &sparse->chunk_size,
		    &sparse->num_chunks);
	if (rc != 5) {
		rc = sscanf(target->params, "%*[^(](%d,%d) %"SCNu64" %u %u",
			    &sparse->device.major, &sparse->device.minor,
			    &sparse->device.start, &sparse->chunk_size,
			    &sparse->num_chunks);
	}
	rc = (rc != 5) ? EINVAL : 0;

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * build_error_params
 *
 * Error targets do not have any additional ASCII parameters, so just
 * allocate an empty string.
 **/
static int build_error_params(dm_target_t *target)
{
	int rc = ENOMEM;

	LOG_PROC_ENTRY();

	target->params = engine_alloc(1);
	if (target->params) {
		*target->params = '\0';
		rc = 0;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * translate_error_params
 *
 * Error targets do not have additional ASCII parameters, so nothing to do.
 **/
static int translate_error_params(dm_target_t *target)
{
	LOG_PROC_ENTRY();
	LOG_PROC_EXIT_INT(0);
	return 0;
}

/**
 * target_type_info
 *
 * This structure is indexed based on dm_target_type value, and can be used
 * to retrieve target-type-specific information without needing switch()
 * statements. Keep this structure initialization in the same order as the
 * dm_target_type definition in include/devmapper.h.
 **/
struct {
	char *string;
	int (*build_params)(dm_target_t *target);
	int (*translate_params)(dm_target_t *target);
	unsigned long struct_size;
} target_type_info[] = {
	{ "linear",		build_linear_params,	translate_linear_params,	sizeof(dm_device_t)		},
	{ "striped",		build_stripe_params,	translate_stripe_params,	sizeof(dm_target_stripe_t)	},
	{ "mirror",		build_mirror_params,	translate_mirror_params,	sizeof(dm_target_mirror_t)	},
	{ "snapshot",		build_snapshot_params,	translate_snapshot_params,	sizeof(dm_target_snapshot_t)	},
	{ "snapshot-origin",	build_origin_params,	translate_origin_params,	sizeof(dm_device_t)		},
	{ "bbr",		build_bbr_params,	translate_bbr_params,		sizeof(dm_target_bbr_t)		},
	{ "sparse",		build_sparse_params,	translate_sparse_params,	sizeof(dm_target_sparse_t)	},
	{ "error",		build_error_params,	translate_error_params,		0				}
};

static inline void *align(void *ptr, unsigned long a)
{
	register unsigned long align = --a;
	return (void *)(((unsigned long)ptr + align) & ~align);
}

/**
 * add_ioctl_target
 *
 * @target:	The target information to add to the ioctl.
 * @begin:	The location to put the target info.
 * @end:	The end of the ioctl packet, for boundary checking.
 *
 * Add information about a single target to the end of an ioctl packet.
 **/
static void *add_ioctl_target(dm_target_t *target,
			      void *begin,
			      void *end)
{
	dm_ioctl_target_t i_target;
	void *params = begin;
	unsigned long params_length = strlen(target->params);

	LOG_PROC_ENTRY();

	params += sizeof(dm_ioctl_target_t);
	if (params + params_length + 1 >= end) {
		/* Ran off the end of the ioctl packet. */
		LOG_PROC_EXIT_PTR(NULL);
		return NULL;
	}

	/* Initialize the ioctl target. */
	i_target.start = target->start;
	i_target.length = target->length;
	i_target.status = 0;
	strncpy(i_target.target_type, target_type_info[target->type].string,
		sizeof(i_target.target_type));

	/* Copy the parameter string. */
	strcpy((char *)params, target->params);
	params += params_length + 1;
	params = align(params, ALIGNMENT);

	i_target.next = params - begin;
	memcpy(begin, &i_target, sizeof(dm_ioctl_target_t));

	LOG_PROC_EXIT_PTR(params);
	return params;
}

/**
 * build_ioctl_packet
 *
 * @name:	Name of device
 * @target_list:List of targets to include with this packet.
 * @new_name:	New device name. Don't specify both target_list and new_name.
 *
 * Construct a DM ioctl packet with the given information. Each DM ioctl
 * packet starts with a header (dm_ioctl_t), followed immediately by padding
 * to ensure 64-bit alignment of the remaining data. After the header and
 * the padding can be one of two things:
 * 1) An array of targets. Each target begins with a dm_ioctl_target_t
 *    structure, and is followed immediately by an ASCII string representing
 *    the target-type-specific data. After this string is enough padding to
 *    ensure 64-bit alignment of the next target.
 * 2) A new device name, as an ASCII string.
 *
 * This function returns a pointer to the beginning of the constructed
 * packet. The caller is responsible for filling in extra information
 * in the ioctl header (flags and dev) based on the command being sent.
 **/
static dm_ioctl_t *build_ioctl_packet(unsigned char *name,
				      dm_target_t *target_list,
				      unsigned char *new_name)
{
	dm_ioctl_t *dmi;
	dm_target_t *target;
	void *begin, *end;
	unsigned long packet_length = sizeof(dm_ioctl_t) + ALIGNMENT;
	int target_count = 0;

	LOG_PROC_ENTRY();

	/* Can't specify both target_list and new_name. */
	if (target_list != NULL && new_name != NULL) {
		LOG_PROC_EXIT_PTR(NULL);
		return NULL;
	}

	/* Count the number of targets. */
	for (target = target_list; target; target = target->next) {
		packet_length += sizeof(dm_ioctl_target_t);
		packet_length += strlen(target->params) + 1 + ALIGNMENT;
		target_count++;
	}

	if (new_name) {
		packet_length += strlen(new_name) + 1;
	}

	/* Make sure packet is minimum length, in case target information
	 * needs to be returned from the kernel.
	 */
	if (packet_length < IOCTL_MIN_LENGTH) {
		packet_length = IOCTL_MIN_LENGTH;
	}

	/* Allocate the ioctl packet. */
	dmi = engine_alloc(packet_length);
	if (!dmi) {
		LOG_ERROR("Error allocating memory for ioctl packet. ");
		LOG_ERROR("Name = %s\n", name);
		LOG_PROC_EXIT_PTR(NULL);
		return NULL;
	}

	/* Initialize the ioctl header. */
	dmi->version[0] = DM_VERSION_MAJOR;
	dmi->version[1] = DM_VERSION_MINOR;
	dmi->version[2] = DM_VERSION_PATCHLEVEL;
	dmi->data_size = packet_length;
	dmi->data_start = (u_int32_t)(unsigned long)align((void *)sizeof(dm_ioctl_t), ALIGNMENT);
	dmi->target_count = target_count;
	if (name) {
		strncpy(dmi->name, name, sizeof(dmi->name));
	}

	/* Add extra information after the ioctl header. */
	begin = (void *)((char *)dmi + dmi->data_start);
	end = (void *)((char *)dmi + dmi->data_size);

	/* Add each target to the packet. */
	for (target = target_list; target; target = target->next) {
		begin = add_ioctl_target(target, begin, end);
		if (!begin) {
			engine_free(dmi);
			LOG_PROC_EXIT_PTR(NULL);
			return NULL;
		}
	}

	/* Add the new name to the packet. */
	if (new_name) {
		strcpy(begin, new_name);
	}

	LOG_PROC_EXIT_PTR(dmi);
	return dmi;
}

/**
 * dm_ioctls
 *
 * This array is indexed by dm_ioctl_cmd values, and is used by run_command()
 * to convert a dm_ioctl_cmd to an actual ioctl command.
 **/
unsigned long dm_ioctls[] = {
	DM_VERSION,
	DM_REMOVE_ALL,
	DM_DEV_CREATE,
	DM_DEV_REMOVE,
	DM_DEV_RELOAD,
	DM_DEV_RENAME,
	DM_DEV_SUSPEND,
	DM_DEV_DEPS,
	DM_DEV_STATUS,
	DM_TARGET_STATUS,
	DM_TARGET_WAIT
};

/**
 * run_command
 *
 * Execute the ioctl command.
 **/
static int run_command(dm_ioctl_t *dmi, dm_ioctl_cmd command)
{
	int rc = 0;

	LOG_PROC_ENTRY();
	LOG_DEBUG("Issuing DM ioctl %d for device %s.\n", command, dmi->name);

	if (dm_control_fd != 0) {
		rc = ioctl(dm_control_fd, dm_ioctls[command], dmi);
		if (rc) {
			rc = errno;
			LOG_ERROR("Error returned from ioctl call: %d: %s.\n",
				  rc, strerror(rc));
		}
	} else {
		LOG_ERROR("Device-Mapper control file not open.");
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * deallocate_target_type_params
 *
 * Walk a list of targets and deallocate the memory for the params string.
 **/
static void deallocate_target_type_params(dm_target_t *target_list)
{
	dm_target_t *target;

	LOG_PROC_ENTRY();
	for (target = target_list; target; target = target->next) {
		if (target->params) {
			engine_free(target->params);
			target->params = NULL;
		}
	}
	LOG_PROC_EXIT_VOID();
}

/**
 * build_target_type_params
 *
 * @target_list: List of targets to process.
 *
 * Walk a list of targets and fill in the appropriate ASCII string in the
 * params fields.
 **/
static int build_target_type_params(dm_target_t *target_list)
{
	dm_target_t *target;
	int rc = 0;

	LOG_PROC_ENTRY();

	for (target = target_list; target; target = target->next) {
		rc = target_type_info[target->type].build_params(target);
		if (rc) {
			LOG_ERROR("Error allocating memory for target parameter string.\n");
			LOG_ERROR("   Target Type: %d, Start: %"PRIu64", Length: %"PRIu64"\n",
				  target->type, target->start, target->length);
			deallocate_target_type_params(target_list);
			break;
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * build_target_list
 *
 * @dmi: Pointer to ioctl packet.
 *
 * Convert an ioctl packet to a list of targets.
 **/
static dm_target_t *build_target_list(dm_ioctl_t *dmi)
{
	dm_target_t *target, *target_list = NULL;
	dm_ioctl_target_t *i_target;
	dm_target_type type;
	char *data_start = (char*)dmi + dmi->data_start;
	unsigned int stripes;
	char *params;
	int i, rc;

	LOG_PROC_ENTRY();

	i_target = (dm_ioctl_target_t*)data_start;
	for (i = 0; i < dmi->target_count; i++) {
		params = (char*)i_target + sizeof(dm_ioctl_target_t);
		stripes = 0;

		/* Determine the target type. */
		for (type = 0; type <= DM_TARGET_MAX; type++) {
			if (!strncmp(i_target->target_type,
				     target_type_info[type].string,
				     DM_MAX_TYPE_NAME)) {
				break;
			}
		}
		if (type > DM_TARGET_MAX) {
			/* Invalid target identifier. */
			LOG_ERROR("Invalid target type (%d) in ioctl packet.\n",
				  type);
			goto error;
		} else if (type == DM_TARGET_STRIPE) {
			sscanf(params, "%u", &stripes);
		}

		/* Allocate a target structure. */
		target = dm_allocate_target(type, i_target->start,
					    i_target->length, stripes);
		if (!target) {
			LOG_ERROR("Error allocating target for type \"%s\"\n",
				  target_type_info[type].string);
			goto error;
		}
		target->params = params;
		rc = target_type_info[type].translate_params(target);
		if (rc) {
			LOG_ERROR("Invalid parameter string for target type "
				  "\"%s\"\n", target_type_info[type].string);
			LOG_ERROR("   Returned parameter string is: %s\n",
				  params);
			goto error;
		}

		dm_add_target(target, &target_list);

		i_target = (dm_ioctl_target_t*)(data_start + i_target->next);
	}

	/* Need to NULL the params fields in each target. */
	for (target = target_list; target; target = target->next) {
		target->params = NULL;
	}

	LOG_PROC_EXIT_PTR(target_list);
	return target_list;

error:
	LOG_ERROR("Error building target list. Name = %s\n", dmi->name);
	dm_deallocate_targets(target_list);
	LOG_PROC_EXIT_PTR(NULL);
	return NULL;
}


/*
 * DM Common-services APIs
 */

/**
 * dm_allocate_target
 *
 * @type:	Type of target to allocate. DM_TARGET_*
 * @start:	Logical start of this target within the DM device.
 * @length:	Length of this target.
 * @num_stripes: Only use when allocating a DM_TARGET_STRIPE. Otherwise set to 0.
 *
 * Allocate a DM target and initialize the common fields.
 **/
dm_target_t *dm_allocate_target(dm_target_type type,
				u_int64_t start,
				u_int64_t length,
				u_int32_t num_stripes)
{
	dm_target_t *target = NULL;
	unsigned long size;

	LOG_PROC_ENTRY();

	if (type > DM_TARGET_MAX) {
		/* Bad parameters. */
		goto out;
	}

	LOG_DEBUG("Request to allocate a %s target.\n",
		  target_type_info[type].string);

	target = engine_alloc(sizeof(dm_target_t));
	if (!target) {
		goto out;
	}

	size = target_type_info[type].struct_size;
	if (type == DM_TARGET_STRIPE) {
		size += num_stripes * sizeof(dm_device_t);
	}
	if (size) {
		target->data.linear = engine_alloc(size);
		if (!target->data.linear) {
			engine_free(target);
			target = NULL;
			goto out;
		}
	}
	target->start = start;
	target->length = length;
	target->type = type;

out:
	if (!target) {
		LOG_ERROR("Error allocating memory for target.\n");
		LOG_ERROR("   Type: %d, Start %"PRIu64", Length %"PRIu64"\n",
			  type, start, length);
	}
	LOG_PROC_EXIT_PTR(target);
	return target;
}

/**
 * dm_add_target
 *
 * @target:		New target to insert in the list.
 * @target_list:	Pointer to the head of the target list.
 *
 * Add a DM target to a list of targets in preparation for an activate or
 * reactivate. The target is inserted in order based on the "start" field.
 **/
void dm_add_target(dm_target_t *target, dm_target_t **target_list)
{
	dm_target_t **t;

	LOG_PROC_ENTRY();

	for (t = target_list; *t; t = &(*t)->next) {
		if ((*t)->start > target->start) {
			break;
		}
	}

	target->next = (*t) ? (*t)->next : NULL;
	*t = target;

	LOG_PROC_EXIT_VOID();
}

/**
 * dm_deallocate_targets
 *
 * Free the memory for a list of DM targets. This assumes the targets were
 * allocated with dm_allocate_target() and linked together with dm_add_target().
 **/
void dm_deallocate_targets(dm_target_t *target_list)
{
	dm_target_t *target;

	LOG_PROC_ENTRY();

	for (target = target_list; target; target = target_list) {
		target_list = target->next;
		if (target->data.linear) {
			engine_free(target->data.linear);
		}
		engine_free(target);
	}

	LOG_PROC_EXIT_VOID();
}

/**
 * dm_suspend
 *
 * @object:	EVMS storage object to suspend/resume.
 * @suspend:	TRUE for suspend. FALSE for resume.
 *
 * Suspend or resume the Device-Mapper device representing an active EVMS
 * object.
 **/
int dm_suspend(storage_object_t *object, int suspend)
{
	dm_ioctl_t * dmi = NULL;
	int rc = EINVAL;

	LOG_PROC_ENTRY();

	if (!object) {
		/* Bad parameter. */
		goto out;
	}

	LOG_DEBUG("Request to %s object %s\n",
		  suspend ? "suspend" : "resume", object->name);

	if (!(object->flags & SOFLAG_ACTIVE)) {
		/* Can't suspend inactive objects. */
		goto out;
	}

	if (suspend && (object->flags & SOFLAG_SUSPENDED)) {
		/* Object is already suspended. */
		rc = 0;
		goto out;
	}

	if (!suspend && !(object->flags & SOFLAG_SUSPENDED)) {
		/* Object is already resumed. */
		rc = 0;
		goto out;
	}

	/* Build the ioctl packet. */
	dmi = build_ioctl_packet(object->name, NULL, NULL);
	if (!dmi) {
		rc = ENOMEM;
		goto out;
	}

	/* Suspend and resume are the same ioctl, with different flags. */
	if (suspend) {
		dmi->flags |= DM_SUSPEND_FLAG;
	}

	/* Run the command. */
	if (local_focus) {
		rc = run_command(dmi, DM_DEV_SUSPEND_CMD);
	} else {
		rc = remote_dm_command(&object->node, dmi, DM_DEV_SUSPEND_CMD);
	}
	if (rc) {
		goto out;
	}

	/* Update the object's state. */
	if (suspend) {
		object->flags |= SOFLAG_SUSPENDED;
	} else {
		object->flags &= ~SOFLAG_SUSPENDED;
	}

out:
	if (dmi) {
		engine_free(dmi);
	}
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * dm_activate
 *
 * @object:	EVMS storage object to activate.
 * @target_list:List of targets that will make up the mapped device.
 *
 * Activate (or re-activate) an EVMS object as a Device-Mapper device. After
 * activating, update the object's device major:minor and mark object as active.
 **/
int dm_activate(storage_object_t *object, dm_target_t *target_list)
{
	dm_ioctl_t *dmi = NULL;
	dm_ioctl_cmd command;
	int rc = EINVAL;

	LOG_PROC_ENTRY();

	if (!object || !target_list) {
		/* Bad parameters. */
		goto out;
	}

	LOG_DEBUG("Request to %sactivate object %s\n",
		  (object->flags & SOFLAG_ACTIVE) ? "re" : "", object->name);

	/* Create the ASCII parameter strings for all targets. */
	rc = build_target_type_params(target_list);
	if (rc) {
		goto out;
	}

	/* Build the ioctl packet. */
	dmi = build_ioctl_packet(object->name, target_list, NULL);
	if (!dmi) {
		rc = ENOMEM;
		goto out;
	}

	if (object->flags & SOFLAG_READ_ONLY) {
		dmi->flags |= DM_READONLY_FLAG;
	}

	/* Run the command. */
	if (object->flags & SOFLAG_ACTIVE) {
		command = DM_DEV_RELOAD_CMD;
		dmi->flags |= DM_SUSPEND_FLAG;
	} else {
		command = DM_DEV_CREATE_CMD;
	}

	if (local_focus) {
		rc = run_command(dmi, command);
	} else {
		rc = remote_dm_command(&object->node, dmi, command);
	}
	if (rc) {
		goto out;
	}

	/* On an activate, update the object's information */
	if (!(object->flags & SOFLAG_ACTIVE)) {
		object->dev_major = major(dmi->dev);
		object->dev_minor = minor(dmi->dev);
		object->flags |= SOFLAG_ACTIVE;
	}
	
out:
	/* Free the ioctl packet and target parameter strings. */
	deallocate_target_type_params(target_list);
	if (dmi) {
		engine_free(dmi);
	}
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * dm_activate_volume
 *
 * @volume:	EVMS volume to activate.
 * @target_list:List of targets that will make up the mapped device.
 *
 * Activate (or re-activate) an EVMS volume as a Device-Mapper device. After
 * activating, update the volume's device major:minor and mark volume as active.
 **/
int dm_activate_volume(logical_volume_t *volume, dm_target_t *target_list)
{
	dm_ioctl_t *dmi = NULL;
	dm_ioctl_cmd command;
	char *base_name;
	int rc = EINVAL;

	LOG_PROC_ENTRY();

	if (!volume || !target_list) {
		/* Bad parameters. */
		goto out;
	}

	LOG_DEBUG("Request to %sactivate volume %s\n",
		  (volume->flags & VOLFLAG_ACTIVE) ? "re" : "", volume->name);

	/* Create the ASCII parameter strings for all targets. */
	rc = build_target_type_params(target_list);
	if (rc) {
		goto out;
	}

	if (memcmp(volume->name, EVMS_DEV_NODE_PATH,
		   EVMS_DEV_NODE_PATH_LEN) != 0) {
		LOG_ERROR("Volume %s does not have the \"%s\" prefix.\n",
			  volume->name, EVMS_DEV_NODE_PATH);
		rc = EINVAL;
		goto out;
	}

	/* Build the ioctl packet. */
	base_name = volume->name + EVMS_DEV_NODE_PATH_LEN;
	dmi = build_ioctl_packet(base_name, target_list, NULL);
	if (!dmi) {
		rc = ENOMEM;
		goto out;
	}

	if (volume->flags & VOLFLAG_READ_ONLY) {
		dmi->flags |= DM_READONLY_FLAG;
	}

	/* Run the command. */
	if (volume->flags & VOLFLAG_ACTIVE) {
		command = DM_DEV_RELOAD_CMD;
		dmi->flags |= DM_SUSPEND_FLAG;
	} else {
		command = DM_DEV_CREATE_CMD;
	}
	if (local_focus) {
		rc = run_command(dmi, command);
	} else {
		rc = remote_dm_command(&volume->node, dmi, command);
	}
	if (rc) {
		goto out;
	}

	/* On an activate, update the object's information */
	if (!(volume->flags & VOLFLAG_ACTIVE)) {
		volume->dev_major = major(dmi->dev);
		volume->dev_minor = minor(dmi->dev);
		volume->flags |= VOLFLAG_ACTIVE;
	}

out:
	/* Free the ioctl packet and target parameter strings. */
	deallocate_target_type_params(target_list);
	if (dmi) {
		engine_free(dmi);
	}
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * dm_deactivate
 *
 * @object: EVMS storage object to deactivate.
 *
 * Deactivate the Device-Mapper device representing the EVMS storage object.
 * Update the object's information.
 **/
int dm_deactivate(storage_object_t *object)
{
	dm_ioctl_t *dmi = NULL;
	int rc = EINVAL;

	LOG_PROC_ENTRY();

	if (!object) {
		/* Bad parameters. */
		goto out;
	}

	LOG_DEBUG("Request to deactivate object %s\n", object->name);

	if (!(object->flags & SOFLAG_ACTIVE)) {
		/* Object is not active. */
		rc = 0;
		goto out;
	}

	/* Build the ioctl packet. */
	dmi = build_ioctl_packet(object->name, NULL, NULL);
	if (!dmi) {
		rc = ENOMEM;
		goto out;
	}

	/* Run the command. */
	if (local_focus) {
		rc = run_command(dmi, DM_DEV_REMOVE_CMD);
	} else {
		rc = remote_dm_command(&object->node, dmi, DM_DEV_REMOVE_CMD);
	}
	if (rc) {
		goto out;
	}

	/* Update the object's information. */
	object->dev_major = 0;
	object->dev_minor = 0;
	object->flags &= ~SOFLAG_ACTIVE;

out:
	if (dmi) {
		engine_free(dmi);
	}
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * dm_deactivate_volume
 *
 * @volume: EVMS logical volume to deactivate.
 *
 * Deactivate the Device-Mapper device representing the EVMS logical volume.
 * Update the volume's information.
 **/
int dm_deactivate_volume(logical_volume_t *volume)
{
	dm_ioctl_t *dmi = NULL;
	char *base_name;
	int rc = EINVAL;

	LOG_PROC_ENTRY();

	if (!volume) {
		/* Bad parameters. */
		goto out;
	}

	LOG_DEBUG("Request to deactivate volume %s\n", volume->name);

	if (!(volume->flags & VOLFLAG_ACTIVE)) {
		/* Object is not active. */
		rc = 0;
		goto out;
	}

	if (memcmp(volume->name, EVMS_DEV_NODE_PATH,
		   EVMS_DEV_NODE_PATH_LEN) != 0) {
		LOG_ERROR("Volume %s does not have the \"%s\" prefix.\n",
			  volume->name, EVMS_DEV_NODE_PATH);
		rc = EINVAL;
		goto out;
	}

	/* Build the ioctl packet. */
	base_name = volume->name + EVMS_DEV_NODE_PATH_LEN;
	dmi = build_ioctl_packet(base_name, NULL, NULL);
	if (!dmi) {
		rc = ENOMEM;
		goto out;
	}

	/* Run the command. */
	if (local_focus) {
		rc = run_command(dmi, DM_DEV_REMOVE_CMD);
	} else {
		rc = remote_dm_command(&volume->node, dmi, DM_DEV_REMOVE_CMD);
	}
	if (rc) {
		goto out;
	}

	/* Update the volume's information. */
	volume->dev_major = 0;
	volume->dev_minor = 0;
	volume->flags &= ~VOLFLAG_ACTIVE;

out:
	if (dmi) {
		engine_free(dmi);
	}
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * dm_rename
 *
 * @object:	The object whose name is changing
 * @old_name:	The object's old name.
 * @new_name:	The object's new name.
 *
 * Tell Device-Mapper to change the name for the specified active object. Both
 * the old name and the new name are parameters to this function, so as to not
 * force a particular ordering on when the object's actual name is changed.
 * This function does not change the name in the object, and the caller can
 * change that name before or after calling this API.
 **/
int dm_rename(storage_object_t * object,
	      char * old_name,
	      char * new_name)
{
	dm_ioctl_t * dmi = NULL;
	int rc = EINVAL;

	LOG_PROC_ENTRY();

	if (!old_name || !new_name) {
		/* Bad parameters. */
		goto out;
	}

	LOG_DEBUG("Request to rename object %s from %s to %s\n",
		  object->name, old_name, new_name);

	if (!(object->flags & SOFLAG_ACTIVE)) {
		/* Object is not active. */
		goto out;
	}

	dmi = build_ioctl_packet(old_name, NULL, new_name);
	if (!dmi) {
		rc = ENOMEM;
		goto out;
	}

	/* Run the command. */
	if (local_focus) {
		rc = run_command(dmi, DM_DEV_RENAME_CMD);
	} else {
		rc = remote_dm_command(&object->node, dmi, DM_DEV_RENAME_CMD);
	}
	if (rc) {
		goto out;
	}

out:
	if (dmi) {
		engine_free(dmi);
	}
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * dm_rename_volume
 *
 * @object:	The volume whose name is changing
 * @old_name:	The volume's old name.
 * @new_name:	The volume's new name.
 *
 * Tell Device-Mapper to change the name for the specified active volume. Both
 * the old name and the new name are parameters to this function, so as to not
 * force a particular ordering on when the volume's actual name is changed.
 * This function does not change the name in the volume, and the caller can
 * change that name before or after calling this API.
 **/
int dm_rename_volume(logical_volume_t * volume,
		     char * old_name,
		     char * new_name)
{
	dm_ioctl_t * dmi = NULL;
	char *old_base_name;
	char *new_base_name;
	int rc = EINVAL;

	LOG_PROC_ENTRY();

	if (!old_name || !new_name) {
		/* Bad parameters. */
		goto out;
	}

	LOG_DEBUG("Request to rename volume %s from %s to %s\n",
		  volume->name, old_name, new_name);

	if (!(volume->flags & VOLFLAG_ACTIVE)) {
		/* Volume is not active. */
		goto out;
	}

	if (memcmp(old_name, EVMS_DEV_NODE_PATH, EVMS_DEV_NODE_PATH_LEN) != 0) {
		LOG_ERROR("Name %s does not have the \"%s\" prefix.\n",
			  old_name, EVMS_DEV_NODE_PATH);
		rc = EINVAL;
		goto out;
	}
        old_base_name = old_name + EVMS_DEV_NODE_PATH_LEN;
	
	if (memcmp(new_name, EVMS_DEV_NODE_PATH, EVMS_DEV_NODE_PATH_LEN) != 0) {
		LOG_ERROR("Name %s does not have the \"%s\" prefix.\n",
			  new_name, EVMS_DEV_NODE_PATH);
		rc = EINVAL;
		goto out;
	}
        new_base_name = new_name + EVMS_DEV_NODE_PATH_LEN;

	dmi = build_ioctl_packet(old_base_name, NULL, new_base_name);
	if (!dmi) {
		rc = ENOMEM;
		goto out;
	}

	/* Run the command. */
	if (local_focus) {
		rc = run_command(dmi, DM_DEV_RENAME_CMD);
	} else {
		rc = remote_dm_command(&volume->node, dmi, DM_DEV_RENAME_CMD);
	}
	if (rc) {
		goto out;
	}

out:
	if (dmi) {
		engine_free(dmi);
	}
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * dm_update_status
 *
 * @object:	EVMS object to retrieve status for.
 *
 * Query device-mapper for the status of this EVMS object. If the device
 * exists in the kernel, it is marked active, and the remaining object
 * information is updated. If the device does not exist in the kernel, the
 * object is marked inactive.
 **/
int dm_update_status(storage_object_t *object)
{
	dm_ioctl_t *dmi = NULL;
	int rc = EINVAL;

	LOG_PROC_ENTRY();

	if (!object) {
		/* Bad parameters. */
		goto out;
	}

	LOG_DEBUG("Request to update the status of object %s\n", object->name);

	/* Build the ioctl packet. */
	dmi = build_ioctl_packet(object->name, NULL, NULL);
	if (!dmi) {
		rc = ENOMEM;
		goto out;
	}

	/* Run the command. */
	if (local_focus) {
		rc = run_command(dmi, DM_DEV_STATUS_CMD);
	} else {
		rc = remote_dm_command(&object->node, dmi, DM_DEV_STATUS_CMD);
	}
	if (rc) {
		goto out;
	}

	/* Update object's information. */
	if (dmi->flags & DM_EXISTS_FLAG) {
		object->flags |= SOFLAG_ACTIVE;
		object->dev_major = major(dmi->dev);
		object->dev_minor = minor(dmi->dev);
		if (dmi->flags & DM_READONLY_FLAG) {
			object->flags |= SOFLAG_READ_ONLY;
		} else {
			object->flags &= ~SOFLAG_READ_ONLY;
		}
	} else {
		object->flags &= ~SOFLAG_ACTIVE;
		object->dev_major = 0;
		object->dev_minor = 0;
	}

out:
	if (dmi) {
		engine_free(dmi);
	}
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * dm_update_volume_status
 *
 * @volume: EVMS volume to retrieve status for.
 *
 * Query device-mapper for the status of this EVMS volume. If the device
 * exists in the kernel, it is marked active, and the remaining volume
 * information is updated. If the device does not exist in the kernel, the
 * volume is marked inactive.
 **/
int dm_update_volume_status(logical_volume_t *volume)
{
	dm_ioctl_t *dmi = NULL;
	char *base_name;
	int rc = EINVAL;

	LOG_PROC_ENTRY();

	if (!volume) {
		/* Bad parameters. */
		goto out;
	}

	LOG_DEBUG("Request to update the status of volume %s\n", volume->name);

	if (memcmp(volume->name, EVMS_DEV_NODE_PATH,
		   EVMS_DEV_NODE_PATH_LEN) != 0) {
		LOG_ERROR("Volume %s does not have the \"%s\" prefix.\n",
			  volume->name, EVMS_DEV_NODE_PATH);
		rc = EINVAL;
		goto out;
	}

	/* Build the ioctl packet. */
	base_name = volume->name + EVMS_DEV_NODE_PATH_LEN;
	dmi = build_ioctl_packet(base_name, NULL, NULL);
	if (!dmi) {
		rc = ENOMEM;
		goto out;
	}

	/* Run the command. */
	if (local_focus) {
		rc = run_command(dmi, DM_DEV_STATUS_CMD);
	} else {
		rc = remote_dm_command(&volume->node, dmi, DM_DEV_STATUS_CMD);
	}
	if (rc) {
		goto out;
	}

	/* Update volume's information. */
	if (dmi->flags & DM_EXISTS_FLAG) {
		volume->flags |= VOLFLAG_ACTIVE;
		volume->dev_major = major(dmi->dev);
		volume->dev_minor = minor(dmi->dev);
		if (dmi->flags & DM_READONLY_FLAG) {
			volume->flags |= VOLFLAG_READ_ONLY;
		} else {
			volume->flags &= ~VOLFLAG_READ_ONLY;
		}
	} else {
		volume->flags &= ~VOLFLAG_ACTIVE;
		volume->dev_major = 0;
		volume->dev_minor = 0;
	}

out:
	if (dmi) {
		engine_free(dmi);
	}
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * dm_get_targets
 *
 * @object:	Object to retrieve target information for.
 * @target_list:Return pointer to list of targets.
 *
 * Query device-mapper for the set of targets that comprise the specified
 * object. This function sets target_list to point at a list of targets. When
 * the caller is done using this list, it must be freed with a call to
 * dm_deallocate_targets().
 **/
int dm_get_targets(storage_object_t *object, dm_target_t **target_list)
{
	dm_ioctl_t *dmi = NULL;
	dm_target_t *targets = NULL;
	int rc = EINVAL;

	LOG_PROC_ENTRY();

	if (!object || !target_list) {
		/* Bad parameters. */
		goto out;
	}

	LOG_DEBUG("Request to get the targets for object %s\n", object->name);

	/* Build the ioctl packet. */
	dmi = build_ioctl_packet(object->name, NULL, NULL);
	if (!dmi) {
		rc = ENOMEM;
		goto out;
	}

	dmi->flags |= DM_STATUS_TABLE_FLAG;

	/* Run the command. */
	if (local_focus) {
		rc = run_command(dmi, DM_TARGET_STATUS_CMD);
	} else {
		rc = remote_dm_command(&object->node, dmi, DM_TARGET_STATUS_CMD);
	}
	if (rc) {
		goto out;
	}

	targets = build_target_list(dmi);
	if (!targets) {
		rc = EINVAL;
	}

out:
	if (target_list) {
		*target_list = targets;
	}
	if (dmi) {
		engine_free(dmi);
	}
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * dm_get_info
 *
 * @object:	Object to retrieve target information for.
 * @info:	Return pointer to info string.
 *
 * Query device-mapper for the latest status info for the specified object.
 * This function sets info to point at a newly-allocated string. When the
 * caller is done using that string, it must be freed with a call to the
 * engine_free service. Status info is target-dependent. The caller should
 * know how to interpret the returned string.

 * Currently, this function assumes the desired device has only one target.
 * Thus, only one string will be returned. May change this later to return
 * an array of strings, one for each target in the device. However, since
 * currently only snapshot provides any meaningful info for this call, there
 * is no need for multi-target info.
 **/
int dm_get_info(storage_object_t *object, char ** info)
{
	dm_ioctl_t *dmi = NULL;
	char *info_str;
	int rc = EINVAL;

	LOG_PROC_ENTRY();

	if (!object || !info) {
		/* Bad parameters. */
		goto out;
	}

	LOG_DEBUG("Request to get info for object %s\n", object->name);

	/* Build the ioctl packet. */
	dmi = build_ioctl_packet(object->name, NULL, NULL);
	if (!dmi) {
		rc = ENOMEM;
		goto out;
	}

	/* Run the command. */
	if (local_focus) {
		rc = run_command(dmi, DM_TARGET_STATUS_CMD);
	} else {
		rc = remote_dm_command(&object->node, dmi, DM_TARGET_STATUS_CMD);
	}
	if (rc) {
		goto out;
	}

	info_str = (char *)dmi + dmi->data_start + sizeof(dm_ioctl_target_t);
	*info = engine_alloc(strlen(info_str)+1);
	if (!*info) {
		rc = ENOMEM;
		goto out;
	}

	strcpy(*info, info_str);

out:
	if (dmi) {
		engine_free(dmi);
	}
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * dm_check_version
 *
 * Query device-mapper for the current version of the ioctl interface. If the
 * kernel's version of the interface does not match what EVMS requires, return
 * an error. This function should only be called by the engine core, so it is
 * not exported as a common-service API.
 **/
int dm_check_version(void)
{
	dm_ioctl_t *dmi = NULL;
	int major, minor, patch;
	int rc = EINVAL;

	LOG_PROC_ENTRY();

	dmi = build_ioctl_packet(NULL, NULL, NULL);
	if (!dmi) {
		rc = ENOMEM;
		goto out;
	}

	/* Run the command. */
	rc = run_command(dmi, DM_VERSION_CMD);
	if (rc) {
		goto out;
	}

	major = dmi->version[0];
	minor = dmi->version[1];
	patch = dmi->version[2];

	if (!(major == DM_VERSION_MAJOR &&
	      minor == DM_VERSION_MINOR &&
	      patch == DM_VERSION_PATCHLEVEL)) {
		LOG_WARNING("Device-Mapper ioctl version mismatch.\n");
		LOG_WARNING("  Device-Mapper provides: %d.%d.%d\n",
			    major, minor, patch);
		LOG_WARNING("  EVMS requires: %d.%d.%d\n", DM_VERSION_MAJOR,
			    DM_VERSION_MINOR, DM_VERSION_PATCHLEVEL);
		rc = EINVAL;
	}

out:
	if (dmi) {
		engine_free(dmi);
	}
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

