/*
 *
 *   Copyright (c) International Business Machines  Corp., 2002
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: remote.c
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <semaphore.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <pthread.h>

#include "fullengine.h"
#include "dmonmsg.h"
#include "dm-ioctl.h"
#include "remote.h"
#include "cluster.h"
#include "engine.h"
#include "daemon.h"
#include "memman.h"
#include "message.h"


#define SEND_MSG(msg, rc)                                                                     \
LOG_DEBUG("Send %s to node %s.\n", msg_name((msg)->cmd), nodeid_to_string(&(msg)->node)); \
rc = ece_funcs->send_msg(msg);                                                                \
if (rc != 0) {                                                                                \
	LOG_SERIOUS("send_msg() returned error code %d: %s\n", rc, evms_strerror(rc));    \
}


#define REQUEST_TIMEOUT 600	/* In seconds */

sem_t * alarm_sem = NULL;

static void alarm_handler(int sig_num) {

	if (sig_num == SIGALRM) {
		LOG_DETAILS("Alarm signal\n");

		if (alarm_sem != NULL) {
			sem_post(alarm_sem);
			alarm_sem = NULL;
		}
	}
}


static int sem_wait_timeout(sem_t * sem, unsigned int seconds) {

	int rc;

	alarm_sem = sem;
	signal(SIGALRM, alarm_handler);
	alarm(seconds);
	rc = sem_wait(sem);
	alarm(0);
	signal(SIGALRM, SIG_DFL);
	alarm_sem = NULL;

	return rc;
}


typedef struct get_version_s {
	LIST_FIELDS(struct get_version_s);
	ece_nodeid_t             * node;
	u_int32_t                  status;
        msg_get_version_response_t results;
	BOOLEAN                    got_response;
} get_version_t;

typedef struct remote_get_version_s {
	LIST_FIELDS(struct remote_get_version_s);
	ece_msg_t     msg;
	sem_t         sem;
	u_int32_t     num_nodes;
	get_version_t get_version_list;
} remote_get_version_t;

static remote_get_version_t version_cb_list = LIST_DECL(&version_cb_list);
static pthread_mutex_t      version_cb_list_mutex = PTHREAD_MUTEX_INITIALIZER;

int remote_verify_version(void) {

	int rc = 0;
	int i;
	get_version_t * get_version_cb;
	remote_get_version_t * version_cb;

	LOG_PROC_ENTRY();

	version_cb = engine_alloc(sizeof(remote_get_version_t));
	if (version_cb == NULL) {
		LOG_CRITICAL("Error allocating memory for a remote_get_version_t.\n");
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	LIST_INIT(&version_cb->get_version_list);

	for (i = 0; (i < membership->num_entries) && (rc == 0); i++) {

		get_version_cb = engine_alloc(sizeof(remote_get_version_t));

		if (get_version_cb != NULL) {
			get_version_cb->node = &membership->node[i];
			get_version_cb->status = ETIMEDOUT;
			LIST_INSERT_BEFORE(get_version_cb, &version_cb->get_version_list);
			version_cb->num_nodes++;

		} else {
			LOG_CRITICAL("Error allocating memory for a get_version_t.\n");
			rc = ENOMEM;
		}
	}

	if (rc == 0) {

		version_cb->msg.cmd = MSG_GET_VERSION;
		version_cb->msg.size = 0;
		version_cb->msg.msg = NULL;

		sem_init(&version_cb->sem, 0, 0);

		pthread_mutex_lock(&version_cb_list_mutex);
		LIST_INSERT_BEFORE(version_cb, &version_cb_list);
		pthread_mutex_unlock(&version_cb_list_mutex);

		for (get_version_cb = version_cb->get_version_list.next;
		    (get_version_cb != &version_cb->get_version_list);
		    get_version_cb = get_version_cb->next) {

			memcpy(&version_cb->msg.node, get_version_cb->node, sizeof(ece_nodeid_t));

			pthread_mutex_lock(&version_cb_list_mutex);
			SEND_MSG(&version_cb->msg, rc);

			if (rc != 0) {
				get_version_cb->status = rc;
				version_cb->num_nodes--;
			}
			pthread_mutex_unlock(&version_cb_list_mutex);
		}

		if (version_cb->num_nodes) {
			LOG_DEBUG("Wait for responses...\n");
			sem_wait_timeout(&version_cb->sem, REQUEST_TIMEOUT);
			LOG_DEBUG("Return from sem_wait_timeout().\n");
		}
	}

	for (get_version_cb = version_cb->get_version_list.next;
	    (get_version_cb != &version_cb->get_version_list);
	    get_version_cb = version_cb->get_version_list.next) {

		int tmp_rc = get_version_cb->status;

		if (tmp_rc == 0) {
			tmp_rc = check_version(get_version_cb->results.version, engine_daemon_msg_version);

			if (tmp_rc != 0) {
				if (tmp_rc < 0) {
					engine_user_message(NULL, NULL, "The daemon on node %s is at version (%d.%d.%d) which is greater than this Engine's daemon message version (%d.%d.%d).\n",
							    nodeid_to_string(get_version_cb->node),
							    get_version_cb->results.version.major, get_version_cb->results.version.minor, get_version_cb->results.version.patchlevel,
							    engine_daemon_msg_version.major, engine_daemon_msg_version.minor, engine_daemon_msg_version.patchlevel);

				} else {
					engine_user_message(NULL, NULL, "The daemon on node %s is at version (%d.%d.%d) which is less than this Engine's daemon message version (%d.%d.%d).\n",
							    nodeid_to_string(get_version_cb->node),
							    get_version_cb->results.version.major, get_version_cb->results.version.minor, get_version_cb->results.version.patchlevel,
							    engine_daemon_msg_version.major, engine_daemon_msg_version.minor, engine_daemon_msg_version.patchlevel);

				}

				tmp_rc = ENOSYS;
			}
		}

		/* Use the first bad error code. */
		if (rc == 0) {
			rc = tmp_rc;
		}

		LIST_REMOVE(get_version_cb);
		engine_free(get_version_cb);
	}

	pthread_mutex_lock(&version_cb_list_mutex);
	LIST_REMOVE(version_cb);
	pthread_mutex_unlock(&version_cb_list_mutex);

	engine_free(version_cb);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static void get_version_response(const ece_msg_t * msg) {
	msg_get_version_response_t * get_version_msg = msg->msg;
	remote_get_version_t * version_cb;
	get_version_t * get_version_cb;

	LOG_PROC_ENTRY();
	
	pthread_mutex_lock(&version_cb_list_mutex);
	for (version_cb = version_cb_list.next;
	    (version_cb != &version_cb_list) &&
	    (version_cb->msg.corrolator != msg->corrolator);
	    version_cb = version_cb->next);
	pthread_mutex_unlock(&version_cb_list_mutex);

	if (version_cb != &version_cb_list) {

		for (get_version_cb = version_cb->get_version_list.next;
		    (get_version_cb != &version_cb->get_version_list) &&
		    (memcmp(get_version_cb->node, &msg->node, sizeof(ece_nodeid_t)));
		    get_version_cb = get_version_cb->next);

		if (get_version_cb != &version_cb->get_version_list) {

			if (!get_version_cb->got_response) {
				get_version_cb->got_response = TRUE;

				if (msg->size == 0) {
					/*
					 * Remote node couldn't even allocate memory
					 * for a response.
					 */
					engine_user_message(NULL, NULL,
							    "Node %s returned a response with no version number in it.\n",
							    nodeid_to_string(&msg->node));

					get_version_cb->status = ENOSYS;

				} else {
					get_version_cb->results.version = get_version_msg->version;
					get_version_cb->status = 0;
				}

				version_cb->num_nodes--;
				LOG_DEBUG("Number of responses pending: %d\n", version_cb->num_nodes);
				if (version_cb->num_nodes == 0) {
					LOG_DEBUG("Post the semaphore.\n");
					sem_post(&version_cb->sem);
				}

			} else {
				LOG_WARNING("Already received a response from node %s.  This one is ignored.\n", nodeid_to_string(get_version_cb->node));
			}

		} else {
			LOG_WARNING("Could not find a get_version_cb for node %s.\n", nodeid_to_string(&msg->node));
		}

	} else {
		LOG_WARNING("Could not find a version_cb for corrolator %d.\n", msg->corrolator);
	}

	LOG_PROC_EXIT_VOID();
}


typedef struct lock_engine_s {
	LIST_FIELDS(struct lock_engine_s);
	ece_nodeid_t * node;
	u_int32_t      status;
	BOOLEAN        got_response;
} lock_engine_t;


typedef struct remote_lock_engine_s {
	LIST_FIELDS(struct remote_lock_engine_s);
	ece_msg_t     msg;
	sem_t         sem;
	u_int32_t     num_nodes;
	lock_engine_t lock_engine_list;
} remote_lock_engine_t;

static remote_lock_engine_t lock_cb_list = LIST_DECL(&lock_cb_list);
static pthread_mutex_t      lock_cb_list_mutex = PTHREAD_MUTEX_INITIALIZER;

int remote_lock_engine(void) {

	int rc = 0;
	int i;
	msg_lock_engine_t msg_lock_engine;
	lock_engine_t * lock_engine_cb;
	remote_lock_engine_t * lock_cb;

	LOG_PROC_ENTRY();

	lock_cb = engine_alloc(sizeof(remote_lock_engine_t));
	if (lock_cb == NULL) {
		LOG_CRITICAL("Error allocating memory for a remote_lock_engine_t.\n");
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	LIST_INIT(&lock_cb->lock_engine_list);

	for (i = 0; (i < membership->num_entries) && (rc == 0); i++) {

		/* Don't send a lock message to myself. */
		if (memcmp(&membership->node[i], &my_nodeid, sizeof(ece_nodeid_t)) != 0) {
			lock_engine_cb = engine_alloc(sizeof(remote_lock_engine_t));

			if (lock_engine_cb != NULL) {
				lock_engine_cb->node = &membership->node[i];
				lock_engine_cb->status = ETIMEDOUT;
				LIST_INSERT_BEFORE(lock_engine_cb, &lock_cb->lock_engine_list);
				lock_cb->num_nodes++;

			} else {
				LOG_CRITICAL("Error allocating memory for a lock_engine_t.\n");
				rc = ENOMEM;
			}
		}
	}

	if (rc == 0) {
		LOG_DEBUG("Lock mode  is %d.\n", engine_mode);
		msg_lock_engine.mode = engine_mode;

		lock_cb->msg.cmd = MSG_LOCK_ENGINE;
		lock_cb->msg.size = sizeof(msg_lock_engine_t);
		lock_cb->msg.msg = &msg_lock_engine;

		sem_init(&lock_cb->sem, 0, 0);

		pthread_mutex_lock(&lock_cb_list_mutex);
		LIST_INSERT_BEFORE(lock_cb, &lock_cb_list);
		pthread_mutex_unlock(&lock_cb_list_mutex);

		for (lock_engine_cb = lock_cb->lock_engine_list.next;
		    (lock_engine_cb != &lock_cb->lock_engine_list);
		    lock_engine_cb = lock_engine_cb->next) {

			memcpy(&lock_cb->msg.node, lock_engine_cb->node, sizeof(ece_nodeid_t));

			pthread_mutex_lock(&lock_cb_list_mutex);
			SEND_MSG(&lock_cb->msg, rc);

			if (rc != 0) {
				lock_engine_cb->status = rc;
				lock_cb->num_nodes--;
			}
			pthread_mutex_unlock(&lock_cb_list_mutex);
		}

		if (lock_cb->num_nodes > 0) {
			LOG_DEBUG("Wait for responses...\n");
			sem_wait_timeout(&lock_cb->sem, REQUEST_TIMEOUT);
			LOG_DEBUG("Return from sem_wait_timeout().\n");
		}
	}

	for (lock_engine_cb = lock_cb->lock_engine_list.next;
	    (lock_engine_cb != &lock_cb->lock_engine_list);
	    lock_engine_cb = lock_cb->lock_engine_list.next) {

		/* Use the first bad error code. */
		if (rc == 0) {
			rc = lock_engine_cb->status;
		}

		LIST_REMOVE(lock_engine_cb);
		engine_free(lock_engine_cb);
	}

	pthread_mutex_lock(&lock_cb_list_mutex);
	LIST_REMOVE(lock_cb);
        pthread_mutex_unlock(&lock_cb_list_mutex);

	engine_free(lock_cb);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static void lock_engine_response(const ece_msg_t * msg) {
	msg_lock_engine_response_t * lock_msg = msg->msg;
	remote_lock_engine_t * lock_cb;
	lock_engine_t * lock_engine_cb;

	LOG_PROC_ENTRY();

	
	pthread_mutex_lock(&lock_cb_list_mutex);
	for (lock_cb = lock_cb_list.next;
	    (lock_cb != &lock_cb_list) &&
	    (lock_cb->msg.corrolator != msg->corrolator);
	    lock_cb = lock_cb->next);
	pthread_mutex_unlock(&lock_cb_list_mutex);

	if (lock_cb != &lock_cb_list) {

		for (lock_engine_cb = lock_cb->lock_engine_list.next;
		    (lock_engine_cb != &lock_cb->lock_engine_list) &&
		    (memcmp(lock_engine_cb->node, &msg->node, sizeof(ece_nodeid_t)));
		    lock_engine_cb = lock_engine_cb->next);

		if (lock_engine_cb != &lock_cb->lock_engine_list) {
			if (!lock_engine_cb->got_response) {
				lock_engine_cb->got_response = TRUE;
				lock_engine_cb->status = lock_msg->status;

				lock_cb->num_nodes--;
				LOG_DEBUG("Number of responses pending: %d\n", lock_cb->num_nodes);
				if (lock_cb->num_nodes == 0) {
					LOG_DEBUG("Post the semaphore.\n");
					sem_post(&lock_cb->sem);
				}

			} else {
				LOG_WARNING("Already received a response from node %s.  This one is ignored.\n", nodeid_to_string(lock_engine_cb->node));
			}

		} else {
			LOG_WARNING("Could not find a lock_engine_cb for node %s.\n", nodeid_to_string(&msg->node));
		}

	} else {
		LOG_WARNING("Could not find a lock_cb for corrolator %d.\n", msg->corrolator);
	}

	LOG_PROC_EXIT_VOID();
}


int remote_unlock_engine() {

	int rc = 0;
	int i;
	ece_msg_t msg;

	LOG_PROC_ENTRY();

	msg.corrolator = 0;
	msg.cmd = MSG_UNLOCK_ENGINE;
	msg.size = 0;
	msg.msg = NULL;

	if (membership != NULL) {
		for (i = 0; (i < membership->num_entries) && (rc == 0); i++) {

			/* Don't send a lock message to myself. */
			if (memcmp(&membership->node[i], &my_nodeid, sizeof(ece_nodeid_t)) != 0) {
				memcpy(&msg.node, &membership->node[i], sizeof(ece_nodeid_t));
				SEND_MSG(&msg, rc);
			}
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static void unlock_engine_response(const ece_msg_t * msg) {

	LOG_PROC_ENTRY();

	/*
	 * We don't care if the remote daemon succeeded or failed.
	 * There is nothing we can do about a remote failure
	 * from this end.
	 */

	LOG_PROC_EXIT_VOID();
}


typedef struct remote_chknod_s {
	LIST_FIELDS(struct remote_chknod_s);
	ece_msg_t             msg;
	sem_t                 sem;
	msg_chknod_response_t results;
} remote_chknod_t;

static remote_chknod_t chknod_cb_list = LIST_DECL(&chknod_cb_list);
static pthread_mutex_t chknod_cb_list_mutex = PTHREAD_MUTEX_INITIALIZER;

int remote_hasa_dev_node(ece_nodeid_t * node,
			 char         * name,
			 u_int32_t      major,
			 u_int32_t      minor) {

	int rc = 0;
	msg_chknod_t    * chknod_msg;
	uint              msg_size = sizeof(msg_chknod_t) + strlen(name) + 1;
	remote_chknod_t * chknod_cb;

	LOG_PROC_ENTRY();

	chknod_msg = engine_alloc(msg_size);
	if (chknod_msg == NULL) {
		LOG_CRITICAL("Error allocating memory for a remote chknod message.\n");
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	chknod_cb = engine_alloc(sizeof(remote_chknod_t));
	if (chknod_cb == NULL) {
		LOG_CRITICAL("Error allocating memory for a remote chknod control block.\n");
		engine_free(chknod_msg);
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	memcpy(&chknod_cb->msg.node, node, sizeof(ece_nodeid_t));
	chknod_cb->msg.cmd  = MSG_CHKNOD;
	chknod_cb->msg.msg  = chknod_msg;
	chknod_cb->msg.size = msg_size;

	chknod_msg->major = major;
	chknod_msg->minor = minor;
	strcpy(chknod_msg->name, name);

	chknod_cb->results.status = ETIMEDOUT;
	sem_init(&chknod_cb->sem, 0, 0);

	pthread_mutex_lock(&chknod_cb_list_mutex);
	LIST_INSERT_BEFORE(chknod_cb, &chknod_cb_list);

	SEND_MSG(&chknod_cb->msg, rc);
	pthread_mutex_unlock(&chknod_cb_list_mutex);

	if (rc == 0) {
		sem_wait_timeout(&chknod_cb->sem, REQUEST_TIMEOUT);

		rc = chknod_cb->results.status;
	}

	pthread_mutex_lock(&chknod_cb_list_mutex);
	LIST_REMOVE(chknod_cb);
	pthread_mutex_unlock(&chknod_cb_list_mutex);

	engine_free(chknod_msg);
	engine_free(chknod_cb);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static void chknod_response(const ece_msg_t * msg) {

	msg_chknod_response_t * chknod_msg = msg->msg;
	remote_chknod_t * chknod_cb;

	LOG_PROC_ENTRY();

	pthread_mutex_lock(&chknod_cb_list_mutex);
	for (chknod_cb = chknod_cb_list.next;
	    (chknod_cb != &chknod_cb_list) &&
	    (chknod_cb->msg.corrolator != msg->corrolator);
	    chknod_cb = chknod_cb->next);
	pthread_mutex_unlock(&chknod_cb_list_mutex);

	if (chknod_cb != &chknod_cb_list) {
		chknod_cb->results = *chknod_msg;
                sem_post(&chknod_cb->sem);

	} else {
		LOG_WARNING("Could not find a chknod_cb for node %s.\n", nodeid_to_string(&msg->node));
	}

	LOG_PROC_EXIT_VOID();
}


typedef struct remote_mknod_s {
	LIST_FIELDS(struct remote_mknod_s);
	ece_msg_t            msg;
	sem_t                sem;
	msg_mknod_response_t results;
} remote_mknod_t;

static remote_mknod_t  mknod_cb_list = LIST_DECL(&mknod_cb_list);
static pthread_mutex_t mknod_cb_list_mutex = PTHREAD_MUTEX_INITIALIZER;

int remote_ensure_dev_node(ece_nodeid_t * node,
			   char         * name,
			   u_int32_t      major,
			   u_int32_t      minor) {

	int rc = 0;
	msg_mknod_t    * mknod_msg;
	uint             msg_size = sizeof(msg_mknod_t) + strlen(name) + 1;
	remote_mknod_t * mknod_cb;

	LOG_PROC_ENTRY();

	mknod_msg = engine_alloc(msg_size);
	if (mknod_msg == NULL) {
		LOG_CRITICAL("Error allocating memory for a remote mknod message.\n");
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	mknod_cb = engine_alloc(sizeof(remote_mknod_t));
	if (mknod_cb == NULL) {
		LOG_CRITICAL("Error allocating memory for a remote mknod control block.\n");
		engine_free(mknod_msg);
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	memcpy(&mknod_cb->msg.node, node, sizeof(ece_nodeid_t));
	mknod_cb->msg.cmd  = MSG_MKNOD;
	mknod_cb->msg.msg  = mknod_msg;
	mknod_cb->msg.size = msg_size;

	mknod_msg->major = major;
	mknod_msg->minor = minor;
	strcpy(mknod_msg->name, name);

	mknod_cb->results.status = ETIMEDOUT;
	sem_init(&mknod_cb->sem, 0, 0);

	pthread_mutex_lock(&mknod_cb_list_mutex);
	LIST_INSERT_BEFORE(mknod_cb, &mknod_cb_list);

	SEND_MSG(&mknod_cb->msg, rc);
	pthread_mutex_unlock(&mknod_cb_list_mutex);

	if (rc == 0) {
		sem_wait_timeout(&mknod_cb->sem, REQUEST_TIMEOUT);

		rc = mknod_cb->results.status;
	}

	pthread_mutex_lock(&mknod_cb_list_mutex);
	LIST_REMOVE(mknod_cb);
	pthread_mutex_unlock(&mknod_cb_list_mutex);

	engine_free(mknod_msg);
	engine_free(mknod_cb);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static void mknod_response(const ece_msg_t * msg) {

	msg_mknod_response_t * mknod_msg = msg->msg;
	remote_mknod_t * mknod_cb;

	LOG_PROC_ENTRY();

	pthread_mutex_lock(&mknod_cb_list_mutex);
	for (mknod_cb = mknod_cb_list.next;
	    (mknod_cb != &mknod_cb_list) &&
	    (mknod_cb->msg.corrolator != msg->corrolator);
	    mknod_cb = mknod_cb->next);
	pthread_mutex_unlock(&mknod_cb_list_mutex);

	if (mknod_cb != &mknod_cb_list) {
		mknod_cb->results = *mknod_msg;
                sem_post(&mknod_cb->sem);

	} else {
		LOG_WARNING("Could not find a mknod_cb for node %s.\n", nodeid_to_string(&msg->node));
	}

	LOG_PROC_EXIT_VOID();
}


typedef struct remote_discover_disks_s {
	LIST_FIELDS(struct remote_discover_disks_s);
	ece_msg_t                     msg;
	sem_t                         sem;
	msg_discover_disks_response_t results;
	storage_object_t            * disks;
} remote_discover_disks_t;

static remote_discover_disks_t discover_disks_cb_list = LIST_DECL(&discover_disks_cb_list);
static pthread_mutex_t         discover_disks_cb_list_mutex = PTHREAD_MUTEX_INITIALIZER;

int remote_discover_disks(ece_nodeid_t       * node,
			  u_int32_t          * num_disks,
			  storage_object_t * * disks) {

	int rc = 0;
	remote_discover_disks_t * discover_disks_cb;

	LOG_PROC_ENTRY();
	*num_disks = 0;
	*disks = NULL;

	discover_disks_cb = engine_alloc(sizeof(remote_discover_disks_t));
	if (discover_disks_cb == NULL) {
		LOG_CRITICAL("Error allocating memory for a remote discover disks control block.\n");
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	memcpy(&discover_disks_cb->msg.node, node, sizeof(ece_nodeid_t));
	discover_disks_cb->msg.cmd  = MSG_DISCOVER_DISKS;
	discover_disks_cb->msg.size = 0;
	discover_disks_cb->msg.msg  = NULL;

	discover_disks_cb->results.status = ETIMEDOUT;
	sem_init(&discover_disks_cb->sem, 0, 0);

	pthread_mutex_lock(&discover_disks_cb_list_mutex);
	LIST_INSERT_BEFORE(discover_disks_cb, &discover_disks_cb_list);

	SEND_MSG(&discover_disks_cb->msg, rc);
	pthread_mutex_unlock(&discover_disks_cb_list_mutex);

	if (rc == 0) {
		sem_wait_timeout(&discover_disks_cb->sem, REQUEST_TIMEOUT);

		rc = discover_disks_cb->results.status;
		*num_disks = discover_disks_cb->results.num_disks;

		if (rc == 0) {
			*disks = discover_disks_cb->disks;
		}
	}

	pthread_mutex_lock(&discover_disks_cb_list_mutex);
	LIST_REMOVE(discover_disks_cb);
	pthread_mutex_unlock(&discover_disks_cb_list_mutex);

	engine_free(discover_disks_cb);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static void discover_disks_response(const ece_msg_t * msg) {
	msg_discover_disks_response_t * discover_disks_msg = msg->msg;
	remote_discover_disks_t * discover_disks_cb;

	LOG_PROC_ENTRY();

	pthread_mutex_lock(&discover_disks_cb_list_mutex);
	for (discover_disks_cb = discover_disks_cb_list.next;
	    (discover_disks_cb != &discover_disks_cb_list) &&
	    (discover_disks_cb->msg.corrolator != msg->corrolator);
	    discover_disks_cb = discover_disks_cb->next);
	pthread_mutex_unlock(&discover_disks_cb_list_mutex);

	if (discover_disks_cb != &discover_disks_cb_list) {
		discover_disks_cb->results = *discover_disks_msg;

		if (discover_disks_msg->num_disks > 0) {
			/* Allocate a new disk array to pass back to the caller. */
			discover_disks_cb->disks = engine_alloc(sizeof(storage_object_t) * discover_disks_msg->num_disks);
			if (discover_disks_cb->disks != NULL) {
				memcpy(discover_disks_cb->disks, discover_disks_msg->disks, discover_disks_msg->num_disks * sizeof(storage_object_t));

			} else {
				discover_disks_cb->results.status = ENOMEM;
			}
		}

		sem_post(&discover_disks_cb->sem);

	} else {
		LOG_WARNING("Could not find a discover_disks_cb for node %s.\n", nodeid_to_string(&msg->node));
	}

	LOG_PROC_EXIT_VOID();
}


typedef struct remote_get_dev_node_names_s {
	LIST_FIELDS(struct remote_get_dev_node_names_s);
	ece_msg_t                         msg;
	sem_t                             sem;
	msg_get_dev_node_names_response_t results;
	char                          * * name_list;
} remote_get_dev_node_names_t;

static remote_get_dev_node_names_t get_dev_node_names_cb_list = LIST_DECL(&get_dev_node_names_cb_list);
static pthread_mutex_t             get_dev_node_names_cb_list_mutex = PTHREAD_MUTEX_INITIALIZER;

int remote_get_dev_node_names(ece_nodeid_t * node,
			      char         * dir_name,
			      char     * * * name_list,
			      int          * count) {

	int rc = 0;
	msg_get_dev_node_names_t    * get_dev_node_names_msg;
	remote_get_dev_node_names_t * get_dev_node_names_cb;
	int msg_size = strlen(dir_name) + 1;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Get dev node names for directory %s.\n", dir_name);

	get_dev_node_names_cb = engine_alloc(sizeof(remote_get_dev_node_names_t));
	if (get_dev_node_names_cb == NULL) {
		LOG_CRITICAL("Error allocating memory for a remote get_dev_node_names control block.\n");
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	get_dev_node_names_msg = engine_alloc(msg_size);
	if (get_dev_node_names_msg == NULL) {
		LOG_CRITICAL("Error allocating memory for a remote get_dev_node_names control block.\n");
		engine_free(get_dev_node_names_cb);
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	memcpy(&get_dev_node_names_cb->msg.node, node, sizeof(ece_nodeid_t));
	get_dev_node_names_cb->msg.cmd  = MSG_GET_DEV_NODE_NAMES;
	get_dev_node_names_cb->msg.size = msg_size;
	get_dev_node_names_cb->msg.msg  = get_dev_node_names_msg;

	strcpy(get_dev_node_names_msg->dir_name, dir_name);

	get_dev_node_names_cb->results.status = ETIMEDOUT;
	sem_init(&get_dev_node_names_cb->sem, 0, 0);

	pthread_mutex_lock(&get_dev_node_names_cb_list_mutex);
	LIST_INSERT_BEFORE(get_dev_node_names_cb, &get_dev_node_names_cb_list);

	SEND_MSG(&get_dev_node_names_cb->msg, rc);
	pthread_mutex_unlock(&get_dev_node_names_cb_list_mutex);

	if (rc == 0) {
		sem_wait_timeout(&get_dev_node_names_cb->sem, REQUEST_TIMEOUT);

		*count = get_dev_node_names_cb->results.count;
		*name_list = get_dev_node_names_cb->name_list;
		rc = get_dev_node_names_cb->results.status;
	}

	pthread_mutex_lock(&get_dev_node_names_cb_list_mutex);
	LIST_REMOVE(get_dev_node_names_cb);
	pthread_mutex_unlock(&get_dev_node_names_cb_list_mutex);

	engine_free(get_dev_node_names_msg);
	engine_free(get_dev_node_names_cb);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static void get_dev_node_names_response(const ece_msg_t * msg) {

	msg_get_dev_node_names_response_t * get_dev_node_names_msg = msg->msg;
	remote_get_dev_node_names_t * get_dev_node_names_cb;

	LOG_PROC_ENTRY();

	pthread_mutex_lock(&get_dev_node_names_cb_list_mutex);
	for (get_dev_node_names_cb = get_dev_node_names_cb_list.next;
	    (get_dev_node_names_cb != &get_dev_node_names_cb_list) &&
	    (get_dev_node_names_cb->msg.corrolator != msg->corrolator);
	    get_dev_node_names_cb = get_dev_node_names_cb->next);
	pthread_mutex_unlock(&get_dev_node_names_cb_list_mutex);

	if (get_dev_node_names_cb != &get_dev_node_names_cb_list) {
		get_dev_node_names_cb->results = *get_dev_node_names_msg;
		if (get_dev_node_names_msg->status == 0) {
			char * * name_list;
			uint count = get_dev_node_names_cb->results.count;

			name_list = calloc(1, sizeof(char *) * count);
			if (name_list != NULL) {
				int i;
				char * pch;

				for (i = 0, pch = get_dev_node_names_msg->names;
				     i < count;
				     i++, pch += strlen(pch) + 1) {
					name_list[i] = strdup(pch);
				}

			} else {
				LOG_CRITICAL("Error getting memory for the name list.\n");
				get_dev_node_names_cb->results.status = ENOMEM;
			}

			get_dev_node_names_cb->name_list = name_list;
		}

		sem_post(&get_dev_node_names_cb->sem);

	} else {
		LOG_WARNING("Could not find a get_dev_node_names_cb for node %s.\n", nodeid_to_string(&msg->node));
	}

	LOG_PROC_EXIT_VOID();
}


typedef struct remote_open_s {
	LIST_FIELDS(struct remote_open_s);
	ece_msg_t           msg;
	sem_t               sem;
	msg_open_response_t results;
} remote_open_t;

static remote_open_t   open_cb_list = LIST_DECL(&open_cb_list);
static pthread_mutex_t open_cb_list_mutex = PTHREAD_MUTEX_INITIALIZER;

int remote_open(ece_nodeid_t * node,
		char         * name,
		int            flags,
		u_int32_t      mode,
		int          * fd) {

	int rc = 0;
	msg_open_t    * open_msg;
	uint            msg_size = sizeof(msg_open_t) + strlen(name) + 1;
	remote_open_t * open_cb;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Open %s on node %s with flags %#x.\n", name, nodeid_to_string(node), flags);

	open_msg = engine_alloc(msg_size);
	if (open_msg == NULL) {
		LOG_CRITICAL("Error allocating memory for a remote open message block.\n");
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	open_cb = engine_alloc(sizeof(remote_open_t));
	if (open_cb == NULL) {
		LOG_CRITICAL("Error allocating memory for a remote open control block.\n");
		engine_free(open_msg);
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	memcpy(&open_cb->msg.node, node, sizeof(ece_nodeid_t));
	open_cb->msg.cmd  = MSG_OPEN;
        open_cb->msg.size = msg_size;
	open_cb->msg.msg  = open_msg;

	open_msg->flags = (u_int32_t) flags;
	open_msg->mode  = (u_int32_t) mode;
	strcpy(open_msg->name, name);

	open_cb->results.status = ETIMEDOUT;
	sem_init(&open_cb->sem, 0, 0);

	pthread_mutex_lock(&open_cb_list_mutex);
	LIST_INSERT_BEFORE(open_cb, &open_cb_list);

	SEND_MSG(&open_cb->msg, rc);
	pthread_mutex_unlock(&open_cb_list_mutex);

	if (rc == 0) {
		sem_wait_timeout(&open_cb->sem, REQUEST_TIMEOUT);

		rc = open_cb->results.status;
		if (rc == 0) {
			*fd = open_cb->results.fd;
		}
	}

	pthread_mutex_lock(&open_cb_list_mutex);
	LIST_REMOVE(open_cb);
	pthread_mutex_unlock(&open_cb_list_mutex);

	engine_free(open_msg);
	engine_free(open_cb);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static void open_response(const ece_msg_t * msg) {

	msg_open_response_t * open_msg = msg->msg;
	remote_open_t * open_cb;

	LOG_PROC_ENTRY();

	pthread_mutex_lock(&open_cb_list_mutex);
	for (open_cb = open_cb_list.next;
	    (open_cb != &open_cb_list) &&
	    (open_cb->msg.corrolator != msg->corrolator);
	    open_cb = open_cb->next);
	pthread_mutex_unlock(&open_cb_list_mutex);

	if (open_cb != &open_cb_list) {
		open_cb->results = *open_msg;
                sem_post(&open_cb->sem);

	} else {
		LOG_WARNING("Could not find a open_cb for node %s.\n", nodeid_to_string(&msg->node));
	}

	LOG_PROC_EXIT_VOID();
}


typedef struct remote_read_s {
	LIST_FIELDS(struct remote_read_s);
	ece_msg_t           msg;
	sem_t               sem;
	msg_read_response_t results;
	void              * buffer;
} remote_read_t;

static remote_read_t   read_cb_list = LIST_DECL(&read_cb_list);
static pthread_mutex_t read_cb_list_mutex = PTHREAD_MUTEX_INITIALIZER;

int remote_read(ece_nodeid_t * node,
		int            fd,
		u_int64_t      offset,
		u_int32_t      length,
		void *         buffer,
		int32_t      * bytes_read) {

	int rc = 0;
	msg_read_t read_msg;
	remote_read_t * read_cb;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Read %u bytes from fd %d at offset %"PRIu64" on node %s.\n", length, fd, offset, nodeid_to_string(node));

	read_cb = engine_alloc(sizeof(remote_read_t));
	if (read_cb == NULL) {
		LOG_CRITICAL("Error allocating memory for a remote read control block.\n");
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	memcpy(&read_cb->msg.node, node, sizeof(ece_nodeid_t));
	read_cb->msg.cmd  = MSG_READ;
	read_cb->msg.size = sizeof(msg_read_t);
	read_cb->msg.msg  = &read_msg;

	read_msg.fd = (u_int32_t) fd;
	read_msg.offset = offset;
	read_msg.length = length;

	read_cb->results.status = ETIMEDOUT;
	read_cb->buffer = buffer;
	sem_init(&read_cb->sem, 0, 0);

	pthread_mutex_lock(&read_cb_list_mutex);
	LIST_INSERT_BEFORE(read_cb, &read_cb_list);

	SEND_MSG(&read_cb->msg, rc);
	pthread_mutex_unlock(&read_cb_list_mutex);

	if (rc == 0) {
		sem_wait_timeout(&read_cb->sem, REQUEST_TIMEOUT);

		*bytes_read = read_cb->results.bytes_read;
		rc = read_cb->results.status;
	}

	pthread_mutex_lock(&read_cb_list_mutex);
	LIST_REMOVE(read_cb);
	pthread_mutex_unlock(&read_cb_list_mutex);

	engine_free(read_cb);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static void read_response(const ece_msg_t * msg) {

	msg_read_response_t * read_msg = msg->msg;
	remote_read_t * read_cb;

	LOG_PROC_ENTRY();

	pthread_mutex_lock(&read_cb_list_mutex);
	for (read_cb = read_cb_list.next;
	    (read_cb != &read_cb_list) &&
	    (read_cb->msg.corrolator != msg->corrolator);
	    read_cb = read_cb->next);
	pthread_mutex_unlock(&read_cb_list_mutex);

	if (read_cb != &read_cb_list) {
		read_cb->results = *read_msg;
		if (read_msg->status == 0) {
			memcpy(read_cb->buffer, &(read_msg->data), read_cb->results.bytes_read);
		}

		sem_post(&read_cb->sem);

	} else {
		LOG_WARNING("Could not find a read_cb for node %s.\n", nodeid_to_string(&msg->node));
	}

	LOG_PROC_EXIT_VOID();
}


typedef struct remote_write_s {
	LIST_FIELDS(struct remote_write_s);
	ece_msg_t            msg;
	sem_t                sem;
	msg_write_response_t results;
} remote_write_t;

static remote_write_t  write_cb_list = LIST_DECL(&write_cb_list);
static pthread_mutex_t write_cb_list_mutex = PTHREAD_MUTEX_INITIALIZER;

int remote_write(ece_nodeid_t * node,
		 int            fd,
		 u_int64_t      offset,
		 u_int32_t      length,
		 void *         buffer,
		 int32_t      * bytes_written) {

	int rc = 0;
	msg_write_t * write_msg;
	uint             msg_size = sizeof(msg_write_t) + length;
	remote_write_t * write_cb;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Write %u bytes to fd %d at offset %"PRIu64" on node %s.\n", length, fd, offset, nodeid_to_string(node));

	write_cb = engine_alloc(sizeof(remote_write_t));
	if (write_cb == NULL) {
		LOG_CRITICAL("Error allocating memory for a remote write control block.\n");
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	write_msg = engine_alloc(msg_size);
	if (write_msg == NULL) {
		LOG_CRITICAL("Error allocating memory for a remote write control block.\n");
		engine_free(write_cb);
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	memcpy(&write_cb->msg.node, node, sizeof(ece_nodeid_t));
	write_cb->msg.cmd  = MSG_WRITE;
	write_cb->msg.size = msg_size;
	write_cb->msg.msg  = write_msg;

	write_msg->fd = fd;
	write_msg->offset = offset;
	write_msg->length = length;
	memcpy(write_msg->data, buffer, length);

	write_cb->results.status = ETIMEDOUT;
	sem_init(&write_cb->sem, 0, 0);

	pthread_mutex_lock(&write_cb_list_mutex);
	LIST_INSERT_BEFORE(write_cb, &write_cb_list);

	SEND_MSG(&write_cb->msg, rc);
	pthread_mutex_unlock(&write_cb_list_mutex);

	if (rc == 0) {
		sem_wait_timeout(&write_cb->sem, REQUEST_TIMEOUT);

		*bytes_written = write_cb->results.bytes_written;
		rc = write_cb->results.status;
	}

	pthread_mutex_unlock(&write_cb_list_mutex);
	LIST_REMOVE(write_cb);
	pthread_mutex_unlock(&write_cb_list_mutex);

	engine_free(write_cb);
	engine_free(write_msg);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static void write_response(const ece_msg_t * msg) {
	msg_write_response_t * write_msg = msg->msg;
	remote_write_t * write_cb;

	LOG_PROC_ENTRY();

	pthread_mutex_lock(&write_cb_list_mutex);
	for (write_cb = write_cb_list.next;
	    (write_cb != &write_cb_list) &&
	    (write_cb->msg.corrolator != msg->corrolator);
	    write_cb = write_cb->next);
	pthread_mutex_unlock(&write_cb_list_mutex);

	if (write_cb != &write_cb_list) {
		write_cb->results = *write_msg;
		sem_post(&write_cb->sem);

	} else {
		LOG_WARNING("Could not find a write_cb for node %s.\n", nodeid_to_string(&msg->node));
	}

	LOG_PROC_EXIT_VOID();
}


typedef struct remote_ioctl_s {
	LIST_FIELDS(struct remote_ioctl_s);
	ece_msg_t            msg;
	sem_t                sem;
	msg_ioctl_response_t results;
} remote_ioctl_t;

static remote_ioctl_t  ioctl_cb_list = LIST_DECL(&ioctl_cb_list);
static pthread_mutex_t ioctl_cb_list_mutex = PTHREAD_MUTEX_INITIALIZER;

int remote_ioctl(ece_nodeid_t * node,
		 int            fd,
		 unsigned long  cmd,
		 void         * arg) {

	int rc = 0;
	msg_ioctl_t * ioctl_msg;
	uint             msg_size = sizeof(msg_ioctl_t) + _IOC_SIZE(cmd);
	remote_ioctl_t * ioctl_cb;

	LOG_PROC_ENTRY();

	LOG_DEBUG("ioctl command %lx to fd %d on node %s.\n", cmd, fd, nodeid_to_string(node));

	ioctl_cb = engine_alloc(sizeof(remote_ioctl_t));
	if (ioctl_cb == NULL) {
		LOG_CRITICAL("Error allocating memory for a remote ioctl control block.\n");
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	ioctl_msg = engine_alloc(msg_size);
	if (ioctl_msg == NULL) {
		LOG_CRITICAL("Error allocating memory for a remote ioctl message.\n");
		engine_free(ioctl_cb);
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	ioctl_msg->fd = fd;
	ioctl_msg->cmd = cmd;
	if (_IOC_SIZE(cmd) != 0) {
		memcpy(ioctl_msg->arg, arg, _IOC_SIZE(cmd));
	}

	memcpy(&ioctl_cb->msg.node, node, sizeof(ece_nodeid_t));
	ioctl_cb->msg.cmd  = MSG_IOCTL;
	ioctl_cb->msg.msg  = ioctl_msg;
	ioctl_cb->msg.size = msg_size;

	ioctl_cb->results.status = ETIMEDOUT;
	sem_init(&ioctl_cb->sem, 0, 0);

	pthread_mutex_lock(&ioctl_cb_list_mutex);
	LIST_INSERT_BEFORE(ioctl_cb, &ioctl_cb_list);

	SEND_MSG(&ioctl_cb->msg, rc);
	pthread_mutex_unlock(&ioctl_cb_list_mutex);

	if (rc == 0) {
		sem_wait_timeout(&ioctl_cb->sem, REQUEST_TIMEOUT);

		rc = ioctl_cb->results.status;

		if (_IOC_SIZE(cmd) != 0) {
			memcpy(arg, ioctl_cb->results.arg, _IOC_SIZE(cmd));
		}
	}

	pthread_mutex_lock(&ioctl_cb_list_mutex);
	LIST_REMOVE(ioctl_cb);
	pthread_mutex_unlock(&ioctl_cb_list_mutex);

	engine_free(ioctl_msg);
	engine_free(ioctl_cb);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static void ioctl_response(const ece_msg_t * msg) {
	msg_ioctl_response_t * ioctl_msg = msg->msg;
	remote_ioctl_t * ioctl_cb;

	LOG_PROC_ENTRY();

	pthread_mutex_lock(&ioctl_cb_list_mutex);
	for (ioctl_cb = ioctl_cb_list.next;
	    (ioctl_cb != &ioctl_cb_list) &&
	    (ioctl_cb->msg.corrolator != msg->corrolator);
	    ioctl_cb = ioctl_cb->next);
	pthread_mutex_unlock(&ioctl_cb_list_mutex);

	if (ioctl_cb != &ioctl_cb_list) {
		ioctl_cb->results = *ioctl_msg;
		memcpy(ioctl_cb->results.arg, ioctl_msg->arg, ioctl_msg->arg_size);
		sem_post(&ioctl_cb->sem);

	} else {
		LOG_WARNING("Could not find a ioctl_cb for node %s.\n", nodeid_to_string(&msg->node));
	}

	LOG_PROC_EXIT_VOID();
}


typedef struct remote_close_s {
	LIST_FIELDS(struct remote_close_s);
	ece_msg_t            msg;
	sem_t                sem;
	msg_close_response_t results;
} remote_close_t;

static remote_close_t  close_cb_list = LIST_DECL(&close_cb_list);
static pthread_mutex_t close_cb_list_mutex = PTHREAD_MUTEX_INITIALIZER;

int remote_close(ece_nodeid_t * node,
		 int            fd) {

	int rc = 0;
	msg_close_t close_msg;
	remote_close_t * close_cb;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Close fd %d on node %s.\n", fd, nodeid_to_string(node));

	close_cb = engine_alloc(sizeof(remote_close_t));
	if (close_cb == NULL) {
		LOG_CRITICAL("Error allocating memory for a remote close control block.\n");
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	close_msg.fd = fd;

	memcpy(&close_cb->msg.node, node, sizeof(ece_nodeid_t));
	close_cb->msg.cmd  = MSG_CLOSE;
	close_cb->msg.size = sizeof(msg_close_t);
	close_cb->msg.msg  = &close_msg;

	close_cb->results.status = ETIMEDOUT;
	sem_init(&close_cb->sem, 0, 0);

	pthread_mutex_lock(&close_cb_list_mutex);
	LIST_INSERT_BEFORE(close_cb, &close_cb_list);

	SEND_MSG(&close_cb->msg, rc);
	pthread_mutex_unlock(&close_cb_list_mutex);

	if (rc == 0) {
		sem_wait_timeout(&close_cb->sem, REQUEST_TIMEOUT);

		rc = close_cb->results.status;
	}

	pthread_mutex_lock(&close_cb_list_mutex);
	LIST_REMOVE(close_cb);
	pthread_mutex_unlock(&close_cb_list_mutex);

	engine_free(close_cb);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static void close_response(const ece_msg_t * msg) {
	msg_close_response_t * close_msg = msg->msg;
	remote_close_t * close_cb;

	LOG_PROC_ENTRY();

	pthread_mutex_lock(&close_cb_list_mutex);
	for (close_cb = close_cb_list.next;
	    (close_cb != &close_cb_list) &&
	    (close_cb->msg.corrolator != msg->corrolator);
	    close_cb = close_cb->next);
	pthread_mutex_unlock(&close_cb_list_mutex);

	if (close_cb != &close_cb_list) {
		close_cb->results = *close_msg;
		sem_post(&close_cb->sem);

	} else {
		LOG_WARNING("Could not find a close_cb for node %s.\n", nodeid_to_string(&msg->node));
	}

	LOG_PROC_EXIT_VOID();
}


typedef struct remote_unlink_s {
	LIST_FIELDS(struct remote_unlink_s);
	ece_msg_t             msg;
	sem_t                 sem;
	msg_unlink_response_t results;
} remote_unlink_t;

static remote_unlink_t unlink_cb_list = LIST_DECL(&unlink_cb_list);
static pthread_mutex_t unlink_cb_list_mutex = PTHREAD_MUTEX_INITIALIZER;

int remote_unlink(ece_nodeid_t  * node,
		  char          * file_name) {

	int rc = 0;
	msg_unlink_t * unlink_msg;
	remote_unlink_t * unlink_cb;

	size_t msg_size = sizeof(msg_unlink_t) + strlen(file_name) + 1;

	LOG_PROC_ENTRY();

	unlink_cb = engine_alloc(sizeof(remote_unlink_t));
	if (unlink_cb == NULL) {
		LOG_CRITICAL("Error allocating memory for a remote unlink control block.\n");
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	unlink_msg = engine_alloc(msg_size);
	if (unlink_msg == NULL) {
		LOG_CRITICAL("Error allocating memory for a remote unlink message.\n");
		engine_free(unlink_cb);
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	strcpy(unlink_msg->file_name, file_name);

	memcpy(&unlink_cb->msg.node, node, sizeof(ece_nodeid_t));
	unlink_cb->msg.cmd  = MSG_UNLINK;
	unlink_cb->msg.size = msg_size;
	unlink_cb->msg.msg  = unlink_msg;

	unlink_cb->results.status = ETIMEDOUT;
	sem_init(&unlink_cb->sem, 0, 0);

	pthread_mutex_lock(&unlink_cb_list_mutex);
	LIST_INSERT_BEFORE(unlink_cb, &unlink_cb_list);

	SEND_MSG(&unlink_cb->msg, rc);
	pthread_mutex_unlock(&unlink_cb_list_mutex);

	if (rc == 0) {
		sem_wait_timeout(&unlink_cb->sem, REQUEST_TIMEOUT);

		rc = unlink_cb->results.status;
	}

	pthread_mutex_lock(&unlink_cb_list_mutex);
	LIST_REMOVE(unlink_cb);
	pthread_mutex_unlock(&unlink_cb_list_mutex);

	engine_free(unlink_msg);
	engine_free(unlink_cb);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static void unlink_response(const ece_msg_t * msg) {
	msg_unlink_response_t * unlink_msg = msg->msg;
	remote_unlink_t * unlink_cb;

	LOG_PROC_ENTRY();

	pthread_mutex_lock(&unlink_cb_list_mutex);
	for (unlink_cb = unlink_cb_list.next;
	    (unlink_cb != &unlink_cb_list) &&
	    (unlink_cb->msg.corrolator != msg->corrolator);
	    unlink_cb = unlink_cb->next);
	pthread_mutex_unlock(&unlink_cb_list_mutex);

	if (unlink_cb != &unlink_cb_list) {
		unlink_cb->results = *unlink_msg;
		sem_post(&unlink_cb->sem);

	} else {
		LOG_WARNING("Could not find a unlink_cb for node %s.\n", nodeid_to_string(&msg->node));
	}

	LOG_PROC_EXIT_VOID();
}


typedef struct remote_check_mount_status_s {
	LIST_FIELDS(struct remote_check_mount_status_s);
	ece_msg_t                         msg;
	sem_t                             sem;
	msg_check_mount_status_response_t results;
} remote_check_mount_status_t;

static remote_check_mount_status_t check_mount_status_cb_list = LIST_DECL(&check_mount_status_cb_list);
static pthread_mutex_t             check_mount_status_cb_list_mutex = PTHREAD_MUTEX_INITIALIZER;

int remote_check_mount_status(ece_nodeid_t * node,
			      char         * volume_name,
			      u_int32_t    * mounted) {

	int rc = 0;
	msg_check_mount_status_t * check_mount_status_msg;
	size_t msg_size = sizeof(msg_check_mount_status_t) + strlen(volume_name) + 1;
	remote_check_mount_status_t * check_mount_status_cb;

	LOG_PROC_ENTRY();

	check_mount_status_cb = engine_alloc(sizeof(remote_check_mount_status_t));
	if (check_mount_status_cb == NULL) {
		LOG_CRITICAL("Error allocating memory for a remote check mount status control block.\n");
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	check_mount_status_msg = engine_alloc(msg_size);
	if (check_mount_status_msg == NULL) {
		LOG_CRITICAL("Error allocating memory for a remote check mount status control block.\n");
		engine_free(check_mount_status_cb);
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	memcpy(&check_mount_status_cb->msg.node, node, sizeof(ece_nodeid_t));
	check_mount_status_cb->msg.cmd  = MSG_CHECK_MOUNT_STATUS;
	check_mount_status_cb->msg.size = msg_size;
	check_mount_status_cb->msg.msg  = check_mount_status_msg;

	check_mount_status_cb->results.status = ETIMEDOUT;
	sem_init(&check_mount_status_cb->sem, 0, 0);

	pthread_mutex_lock(&check_mount_status_cb_list_mutex);
	LIST_INSERT_BEFORE(check_mount_status_cb, &check_mount_status_cb_list);

	SEND_MSG(&check_mount_status_cb->msg, rc);
	pthread_mutex_unlock(&check_mount_status_cb_list_mutex);

	if (rc == 0) {
		sem_wait_timeout(&check_mount_status_cb->sem, REQUEST_TIMEOUT);

		rc = check_mount_status_cb->results.status;
		*mounted = check_mount_status_cb->results.mounted;
	}

	pthread_mutex_lock(&check_mount_status_cb_list_mutex);
	LIST_REMOVE(check_mount_status_cb);
	pthread_mutex_unlock(&check_mount_status_cb_list_mutex);

	engine_free(check_mount_status_cb);
	engine_free(check_mount_status_msg);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static void check_mount_status_response(const ece_msg_t * msg) {
	msg_check_mount_status_response_t * check_mount_status_msg = msg->msg;
	remote_check_mount_status_t * check_mount_status_cb;

	LOG_PROC_ENTRY();

	pthread_mutex_lock(&check_mount_status_cb_list_mutex);
	for (check_mount_status_cb = check_mount_status_cb_list.next;
	    (check_mount_status_cb != &check_mount_status_cb_list) &&
	    (check_mount_status_cb->msg.corrolator != msg->corrolator);
	    check_mount_status_cb = check_mount_status_cb->next);
	pthread_mutex_unlock(&check_mount_status_cb_list_mutex);

	if (check_mount_status_cb != &check_mount_status_cb_list) {
		check_mount_status_cb->results = *check_mount_status_msg;
		sem_post(&check_mount_status_cb->sem);

	} else {
		LOG_WARNING("Could not find a check_mount_status_cb for node %s.\n", nodeid_to_string(&msg->node));
	}

	LOG_PROC_EXIT_VOID();
}


typedef struct remote_check_open_status_s {
	LIST_FIELDS(struct remote_check_open_status_s);
	sem_t                            sem;
	ece_msg_t                        msg;
	msg_check_open_status_response_t results;
} remote_check_open_status_t;

static remote_check_open_status_t check_open_status_cb_list = LIST_DECL(&check_open_status_cb_list);
static pthread_mutex_t            check_open_status_cb_list_mutex = PTHREAD_MUTEX_INITIALIZER;

int remote_check_open_status(ece_nodeid_t * node,
			     char         * volume_name,
			     u_int32_t    * opened) {

	int rc = 0;
	msg_check_open_status_t * check_open_status_msg;
	size_t msg_size = sizeof(msg_check_open_status_t) + strlen(volume_name) + 1;
	remote_check_open_status_t * check_open_status_cb;

	LOG_PROC_ENTRY();

	check_open_status_cb = engine_alloc(sizeof(remote_check_open_status_t));
	if (check_open_status_cb == NULL) {
		LOG_CRITICAL("Error allocating memory for a remote check open status control block.\n");
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	check_open_status_msg = engine_alloc(msg_size);
	if (check_open_status_msg == NULL) {
		LOG_CRITICAL("Error allocating memory for a remote check open status control block.\n");
		engine_free(check_open_status_cb);
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	memcpy(&check_open_status_cb->msg.node, node, sizeof(ece_nodeid_t));
	check_open_status_cb->msg.cmd  = MSG_CHECK_MOUNT_STATUS;
	check_open_status_cb->msg.size = msg_size;
	check_open_status_cb->msg.msg  = check_open_status_msg;

	check_open_status_cb->results.status = ETIMEDOUT;
	sem_init(&check_open_status_cb->sem, 0, 0);

	pthread_mutex_lock(&check_open_status_cb_list_mutex);
	LIST_INSERT_BEFORE(check_open_status_cb, &check_open_status_cb_list);

	SEND_MSG(&check_open_status_cb->msg, rc);
	pthread_mutex_unlock(&check_open_status_cb_list_mutex);

	if (rc == 0) {
		sem_wait_timeout(&check_open_status_cb->sem, REQUEST_TIMEOUT);

		rc = check_open_status_cb->results.status;
		*opened = check_open_status_cb->results.opened;
	}

	pthread_mutex_lock(&check_open_status_cb_list_mutex);
	LIST_REMOVE(check_open_status_cb);
	pthread_mutex_unlock(&check_open_status_cb_list_mutex);

	engine_free(check_open_status_cb);
	engine_free(check_open_status_msg);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static void check_open_status_response(const ece_msg_t * msg) {
	msg_check_open_status_response_t * check_open_status_msg = msg->msg;
	remote_check_open_status_t * check_open_status_cb;

	LOG_PROC_ENTRY();

	pthread_mutex_lock(&check_open_status_cb_list_mutex);
	for (check_open_status_cb = check_open_status_cb_list.next;
	    (check_open_status_cb != &check_open_status_cb_list) &&
	    (check_open_status_cb->msg.corrolator != msg->corrolator);
	    check_open_status_cb = check_open_status_cb->next);
	pthread_mutex_unlock(&check_open_status_cb_list_mutex);

	if (check_open_status_cb != &check_open_status_cb_list) {
		check_open_status_cb->results = *check_open_status_msg;
		sem_post(&check_open_status_cb->sem);

	} else {
		LOG_WARNING("Could not find a check_open_status_cb for node %s.\n", nodeid_to_string(&msg->node));
	}

	LOG_PROC_EXIT_VOID();
}


typedef struct remote_copy_s {
	LIST_FIELDS(struct remote_copy_s);
	ece_msg_t           msg;
	sem_t               sem;
	pthread_mutex_t     results_mutex;
        msg_copy_response_t results;
} remote_copy_t;

static remote_copy_t   copy_cb_list = LIST_DECL(&copy_cb_list);
static pthread_mutex_t copy_cb_list_mutex = PTHREAD_MUTEX_INITIALIZER;

int remote_copy(ece_nodeid_t  * node,
		int             src_fd,
		u_int64_t       src_offset,
		int             trg_fd,
		u_int64_t       trg_offset,
		u_int64_t       len,
		int32_t         chunk_size,
		void          * context,
		copy_callback_t callback) {

	int rc = 0;
	msg_copy_t copy_msg;
	remote_copy_t * copy_cb;
	u_int64_t chunks_to_copy = (len + 1) / chunk_size;

	LOG_PROC_ENTRY();

	copy_cb = engine_alloc(sizeof(remote_copy_t));
	if (copy_cb == NULL) {
		LOG_CRITICAL("Error allocating memory for a remote_copy_t.\n");
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	pthread_mutex_init(&copy_cb->results_mutex, NULL);

	memcpy(&copy_cb->msg.node, node, sizeof(ece_nodeid_t));
	copy_cb->msg.cmd  = MSG_COPY;
	copy_cb->msg.size = sizeof(msg_copy_t);
	copy_cb->msg.msg  = &copy_msg;

	copy_msg.src_fd     = src_fd;
	copy_msg.src_offset = src_offset;
	copy_msg.trg_fd     = trg_fd;
	copy_msg.trg_offset = trg_offset;
	copy_msg.len        = len;
	copy_msg.chunk_size = chunk_size;
	copy_msg.context    = (u_int64_t) ((unsigned long)context);

	sem_init(&copy_cb->sem, 0, 0);

	pthread_mutex_lock(&copy_cb_list_mutex);
	LIST_INSERT_BEFORE(copy_cb, &copy_cb_list);

	SEND_MSG(&copy_cb->msg, rc);
	pthread_mutex_unlock(&copy_cb_list_mutex);

	if (rc == 0) {

		while ((copy_cb->results.status == 0) &&
		       (copy_cb->results.chunks_copied < chunks_to_copy)) {
			u_int32_t status;
			u_int64_t chunks_copied;

			LOG_DEBUG("Wait for responses...\n");
			sem_wait_timeout(&copy_cb->sem, REQUEST_TIMEOUT);
			LOG_DEBUG("Return from sem_wait_timeout().\n");

			pthread_mutex_lock(&copy_cb->results_mutex);
			/*
			 * Copy to local variables so the the results_mutex
			 * is not held too long.
			 */
			status = copy_cb->results.status;
			chunks_copied = copy_cb->results.chunks_copied;
			pthread_mutex_unlock(&copy_cb->results_mutex);

			if ((status == 0) && (callback != NULL)) {
				callback(chunks_copied, context);
			}
		}
	}

	pthread_mutex_lock(&copy_cb_list_mutex);
	LIST_REMOVE(copy_cb);
	pthread_mutex_unlock(&copy_cb_list_mutex);

	rc = copy_cb->results.status;
	engine_free(copy_cb);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static void copy_response(const ece_msg_t * msg) {
	msg_copy_response_t * copy_msg = msg->msg;
	remote_copy_t * copy_cb;

	LOG_PROC_ENTRY();
	
	pthread_mutex_lock(&copy_cb_list_mutex);
	for (copy_cb = copy_cb_list.next;
	    (copy_cb != &copy_cb_list) &&
	    (copy_cb->msg.corrolator != msg->corrolator);
	    copy_cb = copy_cb->next);
	pthread_mutex_unlock(&copy_cb_list_mutex);

	if (copy_cb != &copy_cb_list) {
		int sem_value = 0;

		pthread_mutex_lock(&copy_cb->results_mutex);
		copy_cb->results = *copy_msg;
		pthread_mutex_unlock(&copy_cb->results_mutex);

		/*
		 * Don't do additional posts if the sem value is 1 or greater.
		 * sem_post() increments the sem value.  sem_wait() decrements
		 * the sem value and waits if the new sem value is not 0.  So,
		 * if the sem value goes to 2, a sem_wait() will decrement it
		 * and then wait because the sem value is not 0.  Wrong thing to
		 * do if multiple responses have come in since it last woke up
		 * from a sem_wait().  (Why does sem_wait() check for non-zero
		 * instead of < 0 ?)
		 */
		sem_getvalue(&copy_cb->sem, &sem_value);
		if (sem_value < 1) {
			sem_post(&copy_cb->sem);
		}

	} else {
		LOG_WARNING("Could not find a copy_cb for corrolator %d.\n", msg->corrolator);
	}

	LOG_PROC_EXIT_VOID();
}


typedef struct remote_dm_cmd_s {
	LIST_FIELDS(struct remote_dm_cmd_s);
	ece_msg_t             msg;
	sem_t                 sem;
        msg_dm_cmd_response_t results;
	dm_ioctl_t          * dmi;
} remote_dm_cmd_t;

static remote_dm_cmd_t dm_cmd_cb_list = LIST_DECL(&dm_cmd_cb_list);
static pthread_mutex_t dm_cmd_cb_list_mutex = PTHREAD_MUTEX_INITIALIZER;

int remote_dm_command(ece_nodeid_t * node,
		      dm_ioctl_t   * dmi,
		      dm_ioctl_cmd   command) {

	int rc = 0;
	remote_dm_cmd_t * dm_cmd_cb;
	msg_dm_cmd_t    * dm_cmd_msg;

	LOG_PROC_ENTRY();

	dm_cmd_cb = engine_alloc(sizeof(remote_dm_cmd_t));
	if (dm_cmd_cb == NULL) {
		LOG_CRITICAL("Error allocating memory for a remote_dm_cmd_t.\n");
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	dm_cmd_msg = engine_alloc(sizeof(msg_dm_cmd_t) + dmi->data_size);
	if (dm_cmd_msg == NULL) {
		LOG_CRITICAL("Error allocating memory for a msg_dm_cmd_t.\n");
		engine_free(dm_cmd_cb);
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	memcpy(&dm_cmd_cb->msg.node, node, sizeof(ece_nodeid_t));
	dm_cmd_cb->msg.cmd  = MSG_DM_CMD;
	dm_cmd_cb->msg.size = sizeof(msg_dm_cmd_t) + dmi->data_size;
	dm_cmd_cb->msg.msg  = dm_cmd_msg;

	dm_cmd_cb->dmi = dmi;

	dm_cmd_msg->cmd = command;
	memcpy(dm_cmd_msg->ioctl_data, dmi, dmi->data_size);

	sem_init(&dm_cmd_cb->sem, 0, 0);

	pthread_mutex_lock(&dm_cmd_cb_list_mutex);
	LIST_INSERT_BEFORE(dm_cmd_cb, &dm_cmd_cb_list);

	SEND_MSG(&dm_cmd_cb->msg, rc);
	pthread_mutex_unlock(&dm_cmd_cb_list_mutex);

	if (rc == 0) {
		sem_wait_timeout(&dm_cmd_cb->sem, REQUEST_TIMEOUT);

		rc = dm_cmd_cb->results.status;

		/*
		 * Response buffer was copied back to *dmi by
		 * dm_cmd_response();
		 */
	}

	pthread_mutex_lock(&dm_cmd_cb_list_mutex);
	LIST_REMOVE(dm_cmd_cb);
	pthread_mutex_unlock(&dm_cmd_cb_list_mutex);

	engine_free(dm_cmd_msg);
	engine_free(dm_cmd_cb);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static void dm_cmd_response(const ece_msg_t * msg) {
	msg_dm_cmd_response_t * dm_cmd_msg = msg->msg;
	remote_dm_cmd_t       * dm_cmd_cb;
	dm_ioctl_t            * dmi;

	LOG_PROC_ENTRY();
	
	pthread_mutex_lock(&dm_cmd_cb_list_mutex);
	for (dm_cmd_cb = dm_cmd_cb_list.next;
	    (dm_cmd_cb != &dm_cmd_cb_list) &&
	    (dm_cmd_cb->msg.corrolator != msg->corrolator);
	    dm_cmd_cb = dm_cmd_cb->next);
	pthread_mutex_unlock(&dm_cmd_cb_list_mutex);

	if (dm_cmd_cb != &dm_cmd_cb_list) {
		dmi = (dm_ioctl_t *) &dm_cmd_msg->ioctl_data;

		dm_cmd_cb->results.status = dm_cmd_msg->status;
		memcpy(dm_cmd_cb->dmi, dm_cmd_msg->ioctl_data, dmi->data_size);

		sem_post(&dm_cmd_cb->sem);

	} else {
		LOG_WARNING("Could not find a dm_cmd_cb for corrolator %d.\n", msg->corrolator);
	}

	LOG_PROC_EXIT_VOID();
}


typedef struct remote_execvp_s {
	LIST_FIELDS(struct remote_execvp_s);
	ece_msg_t                msg;
	int                      response_count;
	sem_t                    sem;
	pthread_mutex_t          mutex;         /* For seraializing I/O traffic. */
	pid_t                    child_pid;
	int                      child_pipe[2];
	int                      stdin_pipe[2];
	pthread_t                stdin_thread;
	int                      stdout_pipe[2];
	int                      stderr_pipe[2];
        msg_execvp_response_t    results;
} remote_execvp_t;

static remote_execvp_t execvp_cb_list = LIST_DECL(&execvp_cb_list);
static pthread_mutex_t execvp_cb_list_mutex = PTHREAD_MUTEX_INITIALIZER;

static void * execvp_stdin_io(void * arg) {

	remote_execvp_t * execvp_cb = (remote_execvp_t *) arg;
	ece_msg_t io_msg;
	void * buffer = NULL;
	int    bytes_read;

	/*
	 * Install a cleanup routine for when this thread is canceled.
	 * It is fortunate that engine_free() takes one argument, the
	 * buffer to be freed. pthread_cleanup_push() takes a routine
	 * to call and a arg to pass to that routine.
	 */
	pthread_cleanup_push(engine_free, buffer);

	/* Thread cancelation should not be deferred to a cancellation point. */
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	buffer = engine_alloc(MAX_USER_MESSAGE_LEN);
	if (buffer == NULL) {
		pthread_cleanup_pop(0);
		LOG_CRITICAL("Error allocating memory for an I/O buffer.\n");
		LOG_PROC_EXIT_PTR(NULL);
		return NULL;
	}

	memcpy(&io_msg.node, &execvp_cb->msg.node, sizeof(ece_nodeid_t));
	io_msg.corrolator = execvp_cb->msg.corrolator;
	io_msg.cmd = MSG_EXECVP_STDIN_IO;

	while (TRUE) {
		bytes_read = read(execvp_cb->stdin_pipe[0], buffer, MAX_USER_MESSAGE_LEN);
		if (bytes_read > 0) {
			int rc;

			io_msg.size = bytes_read;
			io_msg.msg = buffer;

			LOG_DEBUG("Send %d bytes of execvp I/O to node %s\n", bytes_read, nodeid_to_string(&io_msg.node));
			SEND_MSG(&io_msg, rc);
			if (rc != 0) {
				LOG_SERIOUS("send_msg() for execvp I/O data returned error code %d: %s\n", rc, evms_strerror(rc));
			}
		}
	}
}


/*
 * This function *must* be run as the child process launced by fork().
 * Bad things can happen if you call it on the main process, since
 * it will exit or signal itself.
 */
static void execvp_child_process(int fd) {

	int status;

	LOG_PROC_ENTRY();

	/*
	 * Wait for the signal from execvp_cleanup() indicating that the child
	 * process on the remote node exited.
	 */
	read(fd, &status, sizeof(status));

	LOG_DEBUG("status: %#x\n", status);

	if (WIFEXITED(status)) {
		LOG_DEBUG("_exit(%d)\n", WEXITSTATUS(status));
		_exit(WEXITSTATUS(status));
	}

	if (WIFSIGNALED(status)) {
		/*
		 * Send a signal to myself so that this process terminates like
		 * the remote child process did.
		 */
		LOG_DEBUG("raise(%d)\n", WTERMSIG(status));
		raise(WTERMSIG(status));
	}
}


static void execvp_cleanup(remote_execvp_t * execvp_cb) {
	
	LOG_PROC_ENTRY();

	/* Resume the child process. */
	write(execvp_cb->child_pipe[1], &execvp_cb->results.status, sizeof(execvp_cb->results.status));

	close(execvp_cb->child_pipe[0]);
	close(execvp_cb->child_pipe[1]);

	if (execvp_cb->stdin_thread != 0) {
		pthread_cancel(execvp_cb->stdin_thread);
		execvp_cb->stdin_thread = 0;
	}

	/* Make sure execvp_response() is not using the execvp_cb. */
	pthread_mutex_lock(&execvp_cb->mutex);
	pthread_mutex_unlock(&execvp_cb->mutex);

	pthread_mutex_lock(&execvp_cb_list_mutex);
	LIST_REMOVE(execvp_cb);
	pthread_mutex_unlock(&execvp_cb_list_mutex);

	engine_free(execvp_cb);

	LOG_PROC_EXIT_VOID();
}


pid_t remote_fork_and_execvp(ece_nodeid_t * node,
			     char         * argv[],
			     int            stdin_pipe[2],
			     int            stdout_pipe[2],
			     int            stderr_pipe[2]) {

	int rc = 0;
	int i;
	remote_execvp_t * execvp_cb;
	msg_execvp_t    * execvp_msg;
	size_t            msg_size;
	char *            flat_argv;
	pid_t             child_pid = -1;
	int               child_pipe[2];

	LOG_PROC_ENTRY();

	rc = pipe(child_pipe);

	if (rc != 0) {
		LOG_SERIOUS("Error creating a pipe for communicating with the child.\n");
		LOG_PROC_EXIT_INT(child_pid);
		return child_pid;
	}

	execvp_cb = engine_alloc(sizeof(remote_execvp_t));
	if (execvp_cb == NULL) {
		LOG_CRITICAL("Error allocating memory for a remote_execvp_t.\n");
		errno = ENOMEM;
		LOG_PROC_EXIT_INT(child_pid);
		return child_pid;
	}

	sem_init(&execvp_cb->sem, 0, 0);
	execvp_cb->child_pipe[0] = child_pipe[0];
	execvp_cb->child_pipe[1] = child_pipe[1];

	pthread_mutex_init(&execvp_cb->mutex, NULL);

	if (stdin_pipe != NULL) {
		execvp_cb->stdin_pipe[0]  = stdin_pipe[0];
		LOG_DEBUG("stdin pipe descriptors: read: %d  write: %d\n", stdin_pipe[0], stdin_pipe[1]);
		execvp_cb->stdin_pipe[1]  = stdin_pipe[1];
	} else {
		LOG_DEBUG("No stdin pipe given.\n");
	}
	if (stdout_pipe) {
		LOG_DEBUG("stdout pipe descriptors: read: %d  write: %d\n", stdout_pipe[0], stdout_pipe[1]);
		execvp_cb->stdout_pipe[0] = stdout_pipe[0];
		execvp_cb->stdout_pipe[1] = stdout_pipe[1];
	} else {
		LOG_DEBUG("No stdout pipe given.\n");
	}
	if (stderr_pipe) {
		LOG_DEBUG("stderr pipe descriptors: read: %d  write: %d\n", stderr_pipe[0], stderr_pipe[1]);
		execvp_cb->stderr_pipe[0] = stderr_pipe[0];
		execvp_cb->stderr_pipe[1] = stderr_pipe[1];
	} else {
		LOG_DEBUG("No stderr pipe given.\n");
	}

	LOG_DEBUG("Count the number of arguments.\n");
	for (i = 0, msg_size = 0; argv[i] != NULL; i++) {
		LOG_DEBUG("argv[%d] is \"%s\"\n", i, argv[i]);
		msg_size += strlen(argv[i]) + 1;
	}

	LOG_DEBUG("There are %d arguments.\n", i);
	/* Plus one for the final terminating nul */
	msg_size ++;
	LOG_DEBUG("Message size is %d.\n", msg_size);

	execvp_msg = engine_alloc(msg_size);
	if (execvp_msg == NULL) {
		LOG_CRITICAL("Error allocating memory for a msg_execvp_t.\n");
		engine_free(execvp_cb);
		errno = ENOMEM;
		LOG_PROC_EXIT_INT(child_pid);
		return child_pid;
	}

	memcpy(&execvp_cb->msg.node, node, sizeof(ece_nodeid_t));
	execvp_cb->msg.cmd  = MSG_EXECVP;
	execvp_cb->msg.size = msg_size;
	execvp_cb->msg.msg  = execvp_msg;

	/* Flatten the argv strings into the message buffer. */
	for (i = 0, flat_argv = execvp_msg->flat_argv; argv[i] != NULL; i++) {
		strcpy(flat_argv, argv[i]);
		flat_argv += strlen(argv[i]) + 1;
	}

	/* Add the final terminating character. */
	*flat_argv = END_OF_TEXT;

	if (debug_level >= DEBUG) {
		char * str;
		char * msg = engine_alloc(msg_size + i + 12);
		char * pch = msg;

		if (msg!= NULL) {
			strcpy(pch, "Message: ");
			pch += strlen(pch);
			for (str = execvp_msg->flat_argv; *str != END_OF_TEXT;) {
				int len = strlen(str);

				strcpy(pch, str);
				str += len + 1;
				pch += len;
				strcat(pch, "\\0");
				pch += 2;
			}

			strcat(pch, "\\03");
			LOG_DEBUG("%s\n", msg);

			engine_free(msg);
		}
	}

	/*
	 * Start the child process before sending the request to the remote
	 * node.  The response may come back before the send completes.
	 */
	child_pid = fork();
	switch (child_pid) {
	case -1:
		LOG_SERIOUS("fork() failed with error code %d: %s\n", errno, strerror(errno));
		engine_free(execvp_msg);
		engine_free(execvp_cb);
		LOG_PROC_EXIT_INT(child_pid);
		return child_pid;

	case 0:
		execvp_child_process(execvp_cb->child_pipe[0]);
	}

	execvp_cb->child_pid = child_pid;

	pthread_mutex_lock(&execvp_cb_list_mutex);
	LIST_INSERT_BEFORE(execvp_cb, &execvp_cb_list);
	pthread_mutex_unlock(&execvp_cb_list_mutex);

	SEND_MSG(&execvp_cb->msg, rc);

	if (rc == 0) {

		/*
		 * Wait for first response from daemon which indiciates that
		 * execvp has been kicked off or has errored out.
		 */
		LOG_DEBUG("Wait for execvp response.\n");
		sem_wait(&execvp_cb->sem);

		rc = execvp_cb->results.rc;

		if (rc == 0) {
			
			if (execvp_cb->stdin_pipe[0] != 0) {
				
				/*
				 * Launch a thread which will send the caller's
				 * input for stdin to the daemon.
				 */
				pthread_create(&execvp_cb->stdin_thread,
					       NULL,
					       execvp_stdin_io,
					       execvp_cb);
			}

		} else {
			/*
			 * Something went wrong on the remote node before the
			 * fork().  There will be no second response.
                         */
			execvp_cleanup(execvp_cb);
			errno = rc;
			child_pid = -1;
		}

	} else {
		/*
		 * Something went wrong on the remote node before the
		 * fork().  There will be no second response.
		 */
		execvp_cleanup(execvp_cb);
		LOG_WARNING("Error code %d from ECE send_msg().\n", rc);
		errno = rc;
		child_pid = -1;
	}

	engine_free(execvp_msg);

	LOG_PROC_EXIT_INT(child_pid);
	return child_pid;
}


static void execvp_response(const ece_msg_t * msg) {
	msg_execvp_response_t * execvp_msg = msg->msg;
	remote_execvp_t       * execvp_cb;

	LOG_PROC_ENTRY();
	
	pthread_mutex_lock(&execvp_cb_list_mutex);
	for (execvp_cb = execvp_cb_list.next;
	    (execvp_cb != &execvp_cb_list) &&
	    (execvp_cb->msg.corrolator != msg->corrolator);
	    execvp_cb = execvp_cb->next);
	pthread_mutex_unlock(&execvp_cb_list_mutex);

	if (execvp_cb != &execvp_cb_list) {
		execvp_cb->response_count++;
		LOG_DEBUG("Response #%d\n", execvp_cb->response_count);
		
		execvp_cb->results = *execvp_msg;
		LOG_DEBUG("execvp() rc: %d  status: %d\n", execvp_msg->rc, execvp_msg->status);
		
		if (execvp_cb->response_count == 1) {
			LOG_DEBUG("Post the semaphore.\n");
			sem_post(&execvp_cb->sem);
		} else if (execvp_cb->response_count == 2) {
			LOG_DEBUG("Cleaup execvp %d.\n", execvp_cb->child_pid);
			execvp_cleanup(execvp_cb);

			/*
			 * execvp_cleanup() frees the execvp_cb,
			 * so don't access it after this point.
			 */
		}

	} else {
		LOG_WARNING("Could not find a execvp_cb for corrolator %d.\n", msg->corrolator);
	}

	LOG_PROC_EXIT_VOID();
}


void msg_execvp_stdin_io(const ece_msg_t * msg) {
        remote_execvp_t        * execvp_cb;
	ece_msg_t              * response_msg;
	msg_execvp_io_response_t response;
	int rc;

	LOG_PROC_ENTRY();
	
	response_msg = get_msg(msg);
	response_msg->cmd |= COMMAND_RESPONSE;
	response_msg->msg = &response;
	response_msg->size = sizeof(response);
	response.status = 0;

	pthread_mutex_lock(&execvp_cb_list_mutex);
	for (execvp_cb = execvp_cb_list.next;
	    (execvp_cb != &execvp_cb_list) &&
	    (execvp_cb->msg.corrolator != msg->corrolator);
	    execvp_cb = execvp_cb->next);
	pthread_mutex_unlock(&execvp_cb_list_mutex);

	if (execvp_cb != &execvp_cb_list) {
		pthread_mutex_lock(&execvp_cb->mutex);
		LOG_DEBUG("Received %d bytes from node %s.\n",
			  msg->size, nodeid_to_string(&msg->node), execvp_cb->stdin_pipe[1]);
		if (execvp_cb->stdin_pipe[1] != 0) {
			int bytes_written;

			LOG_DEBUG("Write %ld bytes to file stdin pipe descriptor %d\n", msg->size, execvp_cb->stdin_pipe[1]);
			bytes_written = write(execvp_cb->stdin_pipe[1], msg->msg, msg->size);
			if (bytes_written >= 0) {
				LOG_DEBUG("Bytes written: %ld\n", bytes_written);
			} else {
				LOG_WARNING("write() failed with error code %d: %s\n", errno, strerror(errno));
				response.status = errno;
			}

		} else {
			LOG_WARNING("There is no file descriptor for writing the data.\n");
			response.status = EINVAL;
		}
		pthread_mutex_unlock(&execvp_cb->mutex);

	} else {
		LOG_WARNING("Could not find a execvp_cb for corrolator %d.\n", msg->corrolator);
		response.status = EINVAL;
	}

	SEND_MSG(response_msg, rc);
	free_msg(response_msg);

	LOG_PROC_EXIT_VOID();
}


void msg_execvp_stdout_io(const ece_msg_t * msg) {
        remote_execvp_t        * execvp_cb;
	ece_msg_t              * response_msg;
	msg_execvp_io_response_t response;
	int rc;

	LOG_PROC_ENTRY();
	
	response_msg = get_msg(msg);
	response_msg->cmd |= COMMAND_RESPONSE;
	response_msg->msg = &response;
	response_msg->size = sizeof(response);
	response.status = 0;

	pthread_mutex_lock(&execvp_cb_list_mutex);
	for (execvp_cb = execvp_cb_list.next;
	    (execvp_cb != &execvp_cb_list) &&
	    (execvp_cb->msg.corrolator != msg->corrolator);
	    execvp_cb = execvp_cb->next);
	pthread_mutex_unlock(&execvp_cb_list_mutex);

	if (execvp_cb != &execvp_cb_list) {
		pthread_mutex_lock(&execvp_cb->mutex);
		LOG_DEBUG("Received %d bytes from node %s.\n",
			  msg->size, nodeid_to_string(&msg->node), execvp_cb->stdout_pipe[1]);
		if (execvp_cb->stdout_pipe[1] != 0) {
			int bytes_written;

			LOG_DEBUG("Write %ld bytes to stdout pipe file descriptor %d\n", msg->size, execvp_cb->stdout_pipe[1]);
			bytes_written = write(execvp_cb->stdout_pipe[1], msg->msg, msg->size);
			if (bytes_written >= 0) {
				LOG_DEBUG("Bytes written: %ld\n", bytes_written);
			} else {
				LOG_WARNING("write() failed with error code %d: %s\n", errno, strerror(errno));
				response.status = errno;
			}

		} else {
			LOG_WARNING("There is no file descriptor for writing the data.\n");
			response.status = EINVAL;
		}
		pthread_mutex_unlock(&execvp_cb->mutex);

	} else {
		LOG_WARNING("Could not find a execvp_cb for corrolator %d.\n", msg->corrolator);
		response.status = EINVAL;
	}

	SEND_MSG(response_msg, rc);
	free_msg(response_msg);

	LOG_PROC_EXIT_VOID();
}


void msg_execvp_stderr_io(const ece_msg_t * msg) {
        remote_execvp_t        * execvp_cb;
	ece_msg_t              * response_msg;
	msg_execvp_io_response_t response;
	int rc;

	LOG_PROC_ENTRY();
	
	response_msg = get_msg(msg);
	response_msg->cmd |= COMMAND_RESPONSE;
	response_msg->msg = &response;
	response_msg->size = sizeof(response);
	response.status = 0;

	pthread_mutex_lock(&execvp_cb_list_mutex);
	for (execvp_cb = execvp_cb_list.next;
	    (execvp_cb != &execvp_cb_list) &&
	    (execvp_cb->msg.corrolator != msg->corrolator);
	    execvp_cb = execvp_cb->next);
	pthread_mutex_unlock(&execvp_cb_list_mutex);

	if (execvp_cb != &execvp_cb_list) {
		pthread_mutex_lock(&execvp_cb->mutex);
		LOG_DEBUG("Received %d bytes from node %s.\n",
			  msg->size, nodeid_to_string(&msg->node), execvp_cb->stderr_pipe[1]);
		if (execvp_cb->stderr_pipe[1] != 0) {
			int bytes_written;

			LOG_DEBUG("Write %ld bytes to stderr pipe file descriptor %d\n", msg->size, execvp_cb->stderr_pipe[1]);
			bytes_written = write(execvp_cb->stderr_pipe[1], msg->msg, msg->size);
			if (bytes_written >= 0) {
				LOG_DEBUG("Bytes written: %ld\n", bytes_written);
			} else {
				LOG_WARNING("write() failed with error code %d: %s\n", errno, strerror(errno));
				response.status = errno;
			}

		} else {
			LOG_WARNING("There is no file descriptor for writing the data.\n");
			response.status = EINVAL;
		}
		pthread_mutex_unlock(&execvp_cb->mutex);

	} else {
		LOG_WARNING("Could not find a execvp_cb for corrolator %d.\n", msg->corrolator);
		response.status = EINVAL;
	}

	SEND_MSG(response_msg, rc);
	free_msg(response_msg);

	LOG_PROC_EXIT_VOID();
}


static void msg_write_log(ece_msg_t * msg) {
	int rc;
	msg_write_log_response_t response;
	ece_msg_t              * response_msg;

	LOG_PROC_ENTRY();
	
	/*
	 * The daemon already determined that this
	 * message should go to the log (msg_level <=
	 * debug_level) otherwise it wouldn't have sent
	 * it.  Write it to our log at debug level
	 * CRITICAL so that it will be written to the
	 * log regardless of the debug level on this
	 * machine.
	 */
	engine_write_log_entry(CRITICAL, msg->msg);
	response.status = 0;
	response_msg = get_msg(msg);
	response_msg->cmd |= COMMAND_RESPONSE;
	response_msg->msg = &response;
	response_msg->size = sizeof(response);

	SEND_MSG(response_msg, rc);
	free_msg(response_msg);

	LOG_PROC_EXIT_VOID();
}


static void msg_user_msg(ece_msg_t * msg) {
	int rc;
	msg_user_msg_response_t response;
	ece_msg_t * response_msg;

	LOG_PROC_ENTRY();
	
	engine_user_message(NULL, NULL, msg->msg);

	response.status = 0;

	response_msg = get_msg(msg);
	response_msg->cmd |= COMMAND_RESPONSE;
	response_msg->msg = &response;
	response_msg->size = sizeof(response);

	SEND_MSG(response_msg, rc);
	free_msg(response_msg);

	LOG_PROC_EXIT_VOID();
}


void engine_router(ece_msg_t * msg) {

	u_int32_t cmd = msg->cmd;

	LOG_PROC_ENTRY();

	LOG_DEBUG("%s from node %s\n", msg_name(cmd), nodeid_to_string(&msg->node));

	if (cmd & COMMAND_RESPONSE) {
		/* Incoming response */

		cmd &= COMMAND_MASK;

		switch (cmd) {
		case MSG_GET_VERSION:
			get_version_response(msg);
			break;

		case MSG_LOCK_ENGINE:
			lock_engine_response(msg);
			break;

		case MSG_UNLOCK_ENGINE:
			unlock_engine_response(msg);
			break;

		case MSG_CHKNOD:
			chknod_response(msg);
			break;

		case MSG_MKNOD:
			mknod_response(msg);
			break;

		case MSG_DISCOVER_DISKS:
			discover_disks_response(msg);
			break;

		case MSG_GET_DEV_NODE_NAMES:
                        get_dev_node_names_response(msg);
			break;

		case MSG_OPEN:
			open_response(msg);
			break;

		case MSG_READ:
			read_response(msg);
			break;

		case MSG_WRITE:
			write_response(msg);
			break;

		case MSG_IOCTL:
			ioctl_response(msg);
			break;

		case MSG_CLOSE:
			close_response(msg);
			break;

		case MSG_UNLINK:
			unlink_response(msg);
			break;

		case MSG_CHECK_MOUNT_STATUS:
			check_mount_status_response(msg);
			break;

		case MSG_CHECK_OPEN_STATUS:
			check_open_status_response(msg);
			break;

		case MSG_COPY:
			copy_response(msg);
			break;

		case MSG_DM_CMD:
			dm_cmd_response(msg);
			break;

		case MSG_EXECVP:
			execvp_response(msg);
			break;

		case MSG_EXECVP_STDIN_IO:
		case MSG_EXECVP_STDOUT_IO:
		case MSG_EXECVP_STDERR_IO:
			/* Don't care. */
			break;
		default:
			/* Unknown response.  Ignore it. */
			LOG_WARNING("Unknown message command: %#x\n", msg->cmd);
			break;
		}

	} else {
		/* Incoming request */

		switch (cmd) {
		case MSG_WRITE_LOG:
			msg_write_log(msg);
			break;

		case MSG_USER_MSG:
			msg_user_msg(msg);
			break;

		case MSG_EXECVP_STDIN_IO:
			msg_execvp_stdin_io(msg);
			break;

		case MSG_EXECVP_STDOUT_IO:
			msg_execvp_stdout_io(msg);
			break;

		case MSG_EXECVP_STDERR_IO:
			msg_execvp_stderr_io(msg);
			break;

		default:
			{
				int rc;

				if (msg->cmd & COMMAND_RESPONSE) {
					LOG_WARNING("Invalid response from node %s: cmd is %#x\n", nodeid_to_string(&msg->node), msg->cmd);
					msg->cmd = MSG_INVALID_CMD | COMMAND_RESPONSE;
				} else {
					LOG_WARNING("Invalid request from node %s: cmd is %#x\n", nodeid_to_string(&msg->node), msg->cmd);
					msg->cmd = MSG_INVALID_CMD | COMMAND_RESPONSE;
				}

				SEND_MSG(msg, rc);
			}
		}
	}

	LOG_PROC_EXIT_VOID();
}
