/*
 *   Copyright (c) International Business Machines  Corp., 2001
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: aixregmgr
 * File: aix_containers.c
 *
 * Description: This file contains functions related to managing AIX containers.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/utsname.h>
#include <plugin.h>
#include "aixregmgr.h"


static unsigned long aix_container_name = 0;


/* aix_find_container_by_id
 *
 *	Search through the global AIX container list for the specified VG id.
 */
storage_container_t * aix_find_container_by_id( unique_id vg_id )
{
	storage_container_t	* container;
	aix_container_data_t	* c_data;
	int			rc;

	LOG_ENTRY;

	DLIST_FOR_EACH(container, aix_container_list, rc) {
		c_data = container->private_data;
		if ( COMPARE_UNIQUE_IDS(c_data->vgda->vg_head->vg_id, vg_id) ) {
			RETURN(container);
		}
	}

	RETURN(NULL);
}


/* aix_update_container_timestamp
 *
 *	Every time a change is made to something in a container, that
 *	container's timestamp must be updated.
 */
int aix_update_container_timestamp( storage_container_t * container )
{
	aix_container_data_t	* c_data = container->private_data;
	aix_vgda_t		* vgda = c_data->vgda;
	vgsa_area		* vgsa = c_data->vgsa;
	struct timeval		current_time;
	struct timezone		zone;

	LOG_ENTRY;

	gettimeofday(&current_time, &zone);

	if ( current_time.tv_sec < vgda->vg_head->vg_timestamp.tv_sec ||
	     (current_time.tv_sec == vgda->vg_head->vg_timestamp.tv_sec &&
	      current_time.tv_usec < vgda->vg_head->vg_timestamp.tv_nsec) ) {
		vgda->vg_head->vg_timestamp.tv_sec++;
	}
	else {
		vgda->vg_head->vg_timestamp.tv_sec	= current_time.tv_sec;
		vgda->vg_head->vg_timestamp.tv_nsec	= current_time.tv_usec;
	}

	vgda->vg_tail->timestamp = vgda->vg_head->vg_timestamp;
	vgsa->b_tmstamp = vgda->vg_head->vg_timestamp;
	vgsa->e_tmstamp = vgda->vg_head->vg_timestamp;

	RETURN(0);
}


/* aix_deallocate_vgda
 *
 *	Free all memory used by the Volume Group Desriptor Area structure.
 */
void aix_deallocate_vgda( aix_vgda_t * vgda )
{
	unsigned int i;

	LOG_ENTRY;

	if ( vgda->vg_head ) {
		aix_engine->engine_free(vgda->vg_head);
		vgda->vg_head = NULL;
	}
	if ( vgda->lv_array ) {
		aix_engine->engine_free(vgda->lv_array);
		vgda->lv_array = NULL;
	}
	for ( i = 0; i < AIX_MAX_PVS; i++ ) {
		if ( vgda->pv_headers[i] ) {
			aix_engine->engine_free(vgda->pv_headers[i]);
			vgda->pv_headers[i] = NULL;
			vgda->pp_array[i] = NULL;
		}
	}
	if ( vgda->lv_names ) {
		aix_engine->engine_free(vgda->lv_names);
		vgda->lv_names = NULL;
	}
	if ( vgda->vg_tail ) {
		aix_engine->engine_free(vgda->vg_tail);
		vgda->vg_tail = NULL;
	}

	aix_engine->engine_free(vgda);

	LOG_EXIT(0);
}


/* aix_deallocate_container
 *
 *	Free all memory for the specified AIX container.
 */
void aix_deallocate_container(	storage_container_t	* container,
				boolean			delete_evms_objects )
{
	aix_container_data_t	* c_data = container->private_data;
	storage_object_t	* object;
	storage_object_t	* region;
	TAG			tag;
	unsigned int		i;

	LOG_ENTRY;
	LOG_DEBUG("Deallocating container %s.\n", container->name);

	aix_remove_container_from_list(container);

	// Free all the regions.
	GoToStartOfList(container->objects_produced);
	while ( ! BlindExtractObject(container->objects_produced, &tag, NULL, (void**)&region) ) {
		region->producing_container = NULL;
		aix_deallocate_region(region, delete_evms_objects);
	}

	// Free all the PVs.
	GoToStartOfList(container->objects_consumed);
	while ( ! BlindExtractObject(container->objects_consumed, &tag, NULL, (void**)&object) ) {
		object->consuming_container = NULL;
		aix_deallocate_pv(object);
	}

	// Free the container private data.
	if ( c_data ) {
		for ( i = 0; i < AIX_MAX_LVS; i++ ) {
			c_data->regions[i] = NULL;
		}
		c_data->freespace = NULL;

		if ( c_data->vgda ) {
			aix_deallocate_vgda(c_data->vgda);
			c_data->vgda = NULL;
		}

		if ( c_data->vgsa ) {
			aix_engine->engine_free(c_data->vgsa);
			c_data->vgsa = NULL;
		}

		aix_engine->engine_free(c_data);
		container->private_data = NULL;
	}

	if ( delete_evms_objects ) {
		aix_engine->free_container(container);
	}

	LOG_EXIT(0);
}


/* aix_allocate_container
 *
 *	Allocate a storage container and all of the necessary private data. The
 *	names of AIX volume groups are not stored anywhere in the metadata, so
 *	we will just come up with a numerical-based name for each container as
 *	it is created.
 */
storage_container_t * aix_allocate_container(	vg_header	* vgh,
						vg_trailer	* vgt )
{
	storage_container_t	* new_container;
	aix_container_data_t	* c_data;
	unsigned char		name[EVMS_NAME_SIZE] = {0};
	int			rc;

	LOG_ENTRY;

	// Generate a name for this container. AIX does not store VG names in
	// its metadata, so these container will not have persistent names.
	snprintf(name, EVMS_NAME_SIZE, "%s/container-%ld", AIX_NAME, aix_container_name++);

	// Allocate the EVMS container.
	rc = aix_engine->allocate_container(name, &new_container);
	if (rc) {
		LOG_CRITICAL("Error allocating EVMS container %s.\n", name);
		RETURN(NULL);
	}

	// Allocate the container private data.
	c_data = aix_engine->engine_alloc(sizeof(aix_container_data_t));
	if ( ! c_data ) {
		LOG_CRITICAL("Error allocating private data for container %s.\n", name);
		aix_deallocate_container(new_container, TRUE);
		RETURN(NULL);
	}
	new_container->private_data = c_data;

	// Allocate a VGDA parent structure
	c_data->vgda = aix_engine->engine_alloc(sizeof(aix_vgda_t));
	if ( ! c_data->vgda ) {
		LOG_CRITICAL("Memory error creating VGDA for container %s.\n", name);
		aix_deallocate_container(new_container, TRUE);
		RETURN(NULL);
	}

	// Create the freespace region for this container.
	rc = aix_create_freespace_region(new_container);
	if (rc) {
		LOG_SERIOUS("Error creating freespace region for container %s.\n", name);
		aix_deallocate_container(new_container, TRUE);
		RETURN(NULL);
	}

	// Initialize remaining entries.
	new_container->plugin = aix_plugin;
	c_data->vgda->vg_head = vgh;
	c_data->vgda->vg_tail = vgt;

	rc = aix_add_container_to_list(new_container);
	if (rc) {
		aix_deallocate_container(new_container, TRUE);
		RETURN(NULL);
	}

	LOG_DETAILS("Allocated new container %s\n", new_container->name);
	RETURN(new_container);
}


/* aix_initialize_new_vg
 *
 *	Allocate a new VG header and VG trailer and initialize them with
 *	the appropriate values.
 *
 *	The "pp_size" argument to this function is in vsectors, NOT in
 *	exponent form.
 */
static int aix_initialize_new_vg(unsigned long	pp_size,
				vg_header	** vgh,
				vg_trailer	** vgt )
{
	vg_header	* new_vgh;
	vg_trailer	* new_vgt;
	int		rc;

	LOG_ENTRY;

	*vgh = NULL;
	*vgt = NULL;

	// Allocate a VG header and trailer.
	new_vgh = aix_engine->engine_alloc(sizeof(vg_header));
	if ( ! new_vgh ) {
		MESSAGE("Memory error creating new VG header.\n");
		RETURN(ENOMEM);
	}

	new_vgt = aix_engine->engine_alloc(sizeof(vg_trailer));
	if ( ! new_vgt ) {
		aix_engine->engine_free(new_vgh);
		MESSAGE("Memory error creating new VG trailer.\n");
		RETURN(ENOMEM);
	}

	// Create a new unique ID for this VG.
	rc = aix_create_unique_id(&new_vgh->vg_id);
	if (rc) {
		aix_engine->engine_free(new_vgh);
		aix_engine->engine_free(new_vgt);
		MESSAGE("Error creating new Unique ID for new VG header.\n");
		RETURN(rc);
	}

	// Fill in the new structures. Don't set the timestamps yet.
	new_vgh->maxlvs		= AIX_MAX_LVS;
	new_vgh->pp_size	= VSECTORS_TO_PP_SIZE(pp_size);
	new_vgh->total_vgdas	= AIX_MAX_VGDAS;
	new_vgh->vgda_size	= AIX_DEFAULT_VGDA_SIZE;
	new_vgh->quorum		= 1;
	new_vgh->auto_varyon	= 1;

	*vgh = new_vgh;
	*vgt = new_vgt;

	RETURN(0);
}


/* aix_allocate_new_container
 *
 *	This function is used when creating a new container from scratch
 *	(not during discovery). It will allocate and initialize all of the
 *	necessary VGDA and VGSA structures, along with the container itself.
 *
 *	The "pp_size" argument to this function is in vsectors, NOT in
 *	exponent form.
 */
storage_container_t * aix_allocate_new_container( unsigned long pp_size )
{
	storage_container_t	* new_container;
	aix_container_data_t	* c_data;
	aix_vgda_t		* vgda;
	vg_header		* vgh;
	vg_trailer		* vgt;
	int			rc;

	LOG_ENTRY;

	// Allocate a VG header and trailer, with the user's options.
	rc = aix_initialize_new_vg(pp_size, &vgh, &vgt);
	if (rc) {
		MESSAGE("Error creating new VG header and trailer.\n");
		RETURN(NULL);
	}

	// Create the new container.
	new_container = aix_allocate_container(vgh, vgt);
	if ( ! new_container ) {
		MESSAGE("Memory error creating new container.\n");
		RETURN(NULL);
	}
	c_data = new_container->private_data;
	vgda = c_data->vgda;

	// Allocate a blank VGSA.
	c_data->vgsa = aix_engine->engine_alloc(sizeof(vgsa_area));
	if ( ! c_data->vgsa ) {
		aix_deallocate_container(new_container, TRUE);
		MESSAGE("Memory error creating new VGSA for new container.\n");
		RETURN(NULL);
	}

	// Allocate a blank LV array and namelist.
	vgda->lv_array = aix_engine->engine_alloc(MAX_SECTORS_LV_ENTRIES * EVMS_VSECTOR_SIZE);
	vgda->lv_names = aix_engine->engine_alloc(MAX_SECTORS_NAMELIST * EVMS_VSECTOR_SIZE);
	if ( !vgda->lv_array || !vgda->lv_names ) {
		aix_deallocate_container(new_container, TRUE);
		MESSAGE("Memory error creating LV array and namelist for VGDA for new container.\n");
		RETURN(NULL);
	}

	RETURN(new_container);
}


/* aix_find_container_for_pv
 *
 *	Check the VG header on the specified object, and determine if the
 *	container has already been created. If not, create a new one.
 */
static storage_container_t * aix_find_container_for_pv( storage_object_t * object )
{
	aix_pv_data_t		* pv = object->consuming_private_data;
	vg_header		* vgh = NULL;
	vg_trailer		* vgt = NULL;
	storage_container_t	* container;
	aix_container_data_t	* c_data;

	LOG_ENTRY;

    LOG_DEBUG("pv = %p pv_num:%d\n", pv, pv->lvm->pv_num);

	// Determine which VG header to get the VG ID from.
	switch(pv->pv_state) {

	case AIX_PV_STATE_VALID:
	case AIX_PV_STATE_FIRST_VGDA:
		vgh = pv->vg_head[0];
		vgt = pv->vg_tail[0];
		aix_engine->engine_free(pv->vg_head[1]);
		aix_engine->engine_free(pv->vg_tail[1]);
		pv->vg_head[0] = NULL;
		pv->vg_tail[0] = NULL;
		pv->vg_head[1] = NULL;
		pv->vg_tail[1] = NULL;
		break;

	case AIX_PV_STATE_SECOND_VGDA:
		vgh = pv->vg_head[1];
		vgt = pv->vg_tail[1];
		aix_engine->engine_free(pv->vg_head[0]);
		aix_engine->engine_free(pv->vg_tail[0]);
		pv->vg_head[0] = NULL;
		pv->vg_tail[0] = NULL;
		pv->vg_head[1] = NULL;
		pv->vg_tail[1] = NULL;
		break;

	case AIX_PV_STATE_EITHER_VGDA:
		if ( COMPARE_UNIQUE_IDS(pv->vg_head[0]->vg_id, pv->vg_head[1]->vg_id) ) {
			vgh = pv->vg_head[0];
			vgt = pv->vg_tail[0];
		}
		else {
			// Not sure where this PV belongs. It thinks it is
			// supposed to be in two different containers. We will
			// probably need to put this on a separate, temporary
			// list, and determine later which container is missing
			// a PV.
		}
		break;

	default:
		LOG_ERROR("Invalid PV state (%d) for %s\n", pv->pv_state, object->name);
		RETURN(NULL);
	}

	// Search for the container using the VG ID.
	container = aix_find_container_by_id(vgh->vg_id);

	if ( ! container ) {
		// Didn't find the container. Create a new one.
		container = aix_allocate_container(vgh, vgt);
		if ( ! container ) {
			RETURN(NULL);
		}

		if ( pv->pv_state == AIX_PV_STATE_EITHER_VGDA ) {
			c_data = container->private_data;
			c_data->flags |= AIX_CONTAINER_INVALID_VG_HEADER;
		}
	}
	else {
		// Found the container. Make sure its copy of the VG header
		// is valid.
		c_data = container->private_data;
		if ( c_data->flags & AIX_CONTAINER_INVALID_VG_HEADER &&
		     pv->pv_state != AIX_PV_STATE_EITHER_VGDA ) {
			aix_engine->engine_free(c_data->vgda->vg_head);
			c_data->vgda->vg_head = vgh;
			c_data->flags &= ~AIX_CONTAINER_INVALID_VG_HEADER;
		}
		else {
			aix_engine->engine_free(vgh);
			aix_engine->engine_free(vgt);
		}
	}

	RETURN(container);
}


/* aix_insert_pv_in_container
 *
 *	Insert the specified PV into the specified container. The aix_pv_t
 *	must be inserted into the aix_container_t pv_list in order by
 *	pv->lvm->pv_num. pv->object can be inserted into the container's
 *	objects_consumed list in any order, since there is no way to insert
 *	into a dlist in a specified order.
 */
static int aix_insert_pv_in_container(	storage_object_t	* object,
					storage_container_t	* container )
{
	storage_object_t	* target_object;
	aix_pv_data_t		* target_pv;
	aix_pv_data_t		* pv = object->consuming_private_data;
	void			* handle;
	int			rc;

	LOG_ENTRY;

	// Insert the object into the objects_consumed list in the container.
	// This insertion must be done in order by pv_num.
	DLIST_FOR_EACH(target_object, container->objects_consumed, rc) {
		target_pv = target_object->consuming_private_data;
		if ( target_pv->lvm->pv_num > pv->lvm->pv_num ) {
			rc = InsertObject(container->objects_consumed,
					object,
					  get_tag_for_object(object),
					NULL,
					InsertBefore,
					FALSE,
					&handle);
			if (rc) {
				LOG_SERIOUS("Error inserting PV %s in container %s\n", object->name, container->name);
				RETURN(rc);
			}

			object->consuming_container = container;
			RETURN(0);
		}
	}

	// Insert at end of the list
	rc = aix_add_object_to_list(object, container->objects_consumed, AppendToList);
	if (rc) {
		LOG_SERIOUS("Error inserting PV %s in container %s\n", object->name, container->name);
		RETURN(rc);
	}

	object->consuming_container = container;
	RETURN(0);
}


/* aix_discover_containers
 *
 *	Examine the list of input storage objects. For each object, read the IPL
 *	and LVM records to see if it is an AIX PV. If so, find/create the
 *	appropriate container for that PV, and add the PV to the container's
 *	list.
 */
int aix_discover_containers(	dlist_t	input_objects,
				dlist_t	output_objects )
{
	storage_object_t	* object;
	storage_container_t	* container;
	TAG			tag;
	int			rc;

	LOG_ENTRY;
	LOG_DEBUG("Searching for AIX PVs.\n");

	GoToStartOfList(input_objects);
	while ( ! BlindExtractObject(input_objects, &tag, NULL, (void**)&object) ) {

		if ( object->data_type != DATA_TYPE ) {
			// Filter out metadata and freespace objects.
			LOG_DEBUG("Skipping %s - not DATA_TYPE.\n", object->name);
			aix_add_object_to_list(object, output_objects, AppendToList);
			continue;
		}

		rc = aix_read_pv_metadata(object);
		if (rc) {
			// This object is not a PV, or an error occurred.
			aix_add_object_to_list(object, output_objects, AppendToList);
			continue;
		}

		container = aix_find_container_for_pv(object);
		if ( ! container ) {
			// Error finding container for this PV.
			aix_deallocate_pv(object);
			aix_add_object_to_list(object, output_objects, AppendToList);
			continue;
		}

		rc = aix_insert_pv_in_container(object, container);
		if (rc) {
			// Error adding this PV to the container list.
			aix_deallocate_pv(object);
			aix_add_object_to_list(object, output_objects, AppendToList);
			continue;
		}
        LOG_DEBUG("object:%p object->consuming_private_data :%p \n", object, object->consuming_private_data);
	}

	LOG_DEBUG("Container discovery complete.\n");
	RETURN(0);
}


/* aix_read_vgdas
 *
 *	Read the Volume Group Descriptor Area for each container.
 */
int aix_read_vgdas( void )
{
	storage_container_t	* container;
	aix_container_data_t	* c_data;
	storage_object_t	* object;
	aix_pv_data_t		* pv;
	boolean			done;
	int			copy;
	int			i, rc, rc2;

	LOG_ENTRY;

	DLIST_FOR_EACH(container, aix_container_list, rc) {
		c_data = container->private_data;
		done = FALSE;

		// Get the first valid PV from the container.
		DLIST_FOR_EACH(object, container->objects_consumed, rc2) {
			pv = object->consuming_private_data;

			switch (pv->pv_state) {
			case AIX_PV_STATE_VALID:
			case AIX_PV_STATE_FIRST_VGDA:
				copy = 0;
				break;
			case AIX_PV_STATE_SECOND_VGDA:
				copy = 1;
				break;
			default:
				continue;
			}

			// Read in the VGSA, and all remaining sections of the VGDA.
			rc = aix_read_vgsa(object, copy);
			if (!rc) rc = aix_read_lv_array(object, copy);
			if (!rc) rc = aix_read_pvs(object, copy);
			if (!rc) rc = aix_read_lv_names(object, copy);
			if (!rc) {
				done = TRUE;
				break;
			}
		}

		if ( ! done ) {
			// All PVs are STATE_EITHER_VGDA, or there was some
			// IO error.
			LOG_SERIOUS("Missing metadata for container %s\n", container->name);
			continue;
		}

		// Set up the remaining pointers in the PV structs.
		DLIST_FOR_EACH(object, container->objects_consumed, rc2) {
			pv = object->consuming_private_data;
			if ( ! pv->pv_head ) {
				pv->pv_head = c_data->vgda->pv_headers[pv->lvm->pv_num - 1];
			}
			if ( ! pv->pp_map ) {
				pv->pp_map = c_data->vgda->pp_array[pv->lvm->pv_num - 1];
			}
		}

		// Calculate the size of the container, since that info is
		// not stored in the VG header.
		if ( c_data->pp_count == 0 ) {
			for ( i = 0; i < AIX_MAX_PVS; i++ ) {
				if ( c_data->vgda->pv_headers[i] ) {
					c_data->pp_count += c_data->vgda->pv_headers[i]->pp_count;
				}
			}
			container->size = c_data->pp_count * PP_SIZE_TO_VSECTORS(c_data->vgda->vg_head->pp_size);

			rc = aix_update_freespace_region(container);
			if (rc) {
				LOG_SERIOUS("Error updating size of freespace region for container %s.\n", container->name);
			}
		}
	}

	RETURN(0);
}


/* aix_add_new_pv_to_container
 *
 *	Add this new PV to this existing container. Updated all
 *	necessary metadata fields.
 */
int aix_add_new_pv_to_container(storage_object_t	* object,
				storage_container_t	* container )
{
	aix_container_data_t	* c_data = container->private_data;
	aix_vgda_t		* vgda = c_data->vgda;
	aix_pv_data_t		* pv = object->consuming_private_data;
	int			rc;

	LOG_ENTRY;

	// Add the PV's header and PP map to the container metadata.
	vgda->pv_headers[pv->pv_head->pv_num - 1] = pv->pv_head;
	vgda->pp_array[pv->pv_head->pv_num -1] = pv->pp_map;
	
	// Add the PV to the container's consumed objects list.
	rc = aix_insert_pv_in_container(object, container);
	if (rc) {
		RETURN(rc);
	}

	// Update the VG info.
	vgda->vg_head->numpvs++;

	// Increment the container size
	c_data->pp_count += pv->pv_head->pp_count;
	container->size  += pv->pv_head->pp_count * PP_SIZE_TO_VSECTORS(vgda->vg_head->pp_size);

	RETURN(0);
}


/* aix_remove_pv_from_container
 *
 *	Remove the specified object from its container, and update the
 *	container metadata.
 */
int aix_remove_pv_from_container( storage_object_t * object )
{
	storage_container_t	* container = object->consuming_container;
	aix_container_data_t	* c_data = container->private_data;
	aix_pv_data_t		* pv = object->consuming_private_data;
	int			rc;

	LOG_ENTRY;

	// Remove the object from the container list.
	rc = aix_remove_object_from_list(object, container->objects_consumed);
	if (rc) {
		LOG_ERROR("Error removing object %s from child-list in container %s.\n",
			object->name, container->name);
		RETURN(rc);
	}
	object->consuming_container = NULL;

	// Update the container metadata.
	c_data->vgda->vg_head->numpvs--;
	c_data->pp_count -= pv->pv_head->pp_count;

	// Remove the PV header and PP map entry for
	// this object from the container's VGDA.
	c_data->vgda->pv_headers[pv->pv_head->pv_num-1] = NULL;
	c_data->vgda->pp_array[pv->pv_head->pv_num-1] = NULL;
	aix_engine->engine_free(pv->pv_head);
	pv->pv_head = NULL;
	pv->pp_map = NULL;

	// Erase the PV/VG metadata from this object.
	aix_erase_container_metadata(object);
	aix_erase_pv_metadata(object);

	RETURN(0);
}


/* aix_check_available_partitions
 *
 *	Check the specified container to make sure there are enough unallocated
 *	PPs to meet the desired number.
 *
 *	Eventually this function will need to be changed to check based on
 *	desired type of allocation as well (linear, striped, mirrored, etc).
 */
int aix_check_available_partitions(	storage_container_t	* container,
					aix_region_options_t	* region_opts )
{
	aix_container_data_t	* c_data = container->private_data;
	aix_region_data_t	* r_data = c_data->freespace->private_data;

	LOG_ENTRY;

	// Just a simple free-PPs check for now.
	if ( r_data->lv->num_lps < region_opts->num_lps ) {
		MESSAGE("Requested %ld LPs for new region \"%s\". Only %ld available.\n",
			region_opts->num_lps, region_opts->name, r_data->lv->num_lps);
		RETURN(ENOSPC);
	}

	RETURN(0);
}

