/*
 *   Copyright (c) International Business Machines  Corp., 2001
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: aixregmgr
 * File: aix_dlist.c
 *
 * Description: Managing the various dlists that are used by the AIX region
 *		manager.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <plugin.h>
#include "aixregmgr.h"


/* aix_get_list_item
 *
 *	Get the current item from the specified d-list.
 */
void * aix_get_list_item( dlist_t list )
{
	void	* object;
	TAG	tag;

	LOG_ENTRY;
	BlindGetObject(list, &tag, NULL, TRUE, &object);

    LOG_DEBUG("Returning object %p \n", object);

	return(object);
}


/* aix_add_object_to_list
 *
 *	Add the specified object to the end of the specified d-list.
 */
int aix_add_object_to_list(	storage_object_t	* object,
				dlist_t			                  list,
                Insertion_Modes                   InsertMode)
{
	void	* handle;
	int	rc;

	LOG_ENTRY;

    LOG_DEBUG("add_object_to_list InsertMode:%d for object:%s\n", InsertMode, object->name);

	rc = InsertObject(list,
			object,
			  get_tag_for_object(object),
			NULL,
			InsertMode,
			TRUE,
			&handle);

	if (rc) {
		LOG_SERIOUS("Error adding %s to list\n", object->name);
	}

	RETURN(rc);
}


/* aix_remove_object_from_list
 *
 *	Remove the specified object from the specified d-list.
 */
int aix_remove_object_from_list(storage_object_t	* object,
				dlist_t			list )
{
	int	rc;

	LOG_ENTRY;

	rc = DeleteObject(list, object);

	if (rc) {
		LOG_SERIOUS("Error removing %s from list\n", object->name);
	}

	RETURN(rc);
}


/* aix_add_container_to_list
 *
 *	Add the specified AIX container to the global list.
 */
int aix_add_container_to_list( storage_container_t * container )
{
	void	* handle;
	int	rc;

	LOG_ENTRY;

	rc = InsertObject(aix_container_list,
			container,
			CONTAINER_TAG,
			NULL,
			AppendToList,
			TRUE,
			&handle);

	if (rc) {
		LOG_SERIOUS("Error adding %s to list\n", container->name);
	}

	RETURN(rc);
}


/* aix_remove_container_from_list
 *
 *	Remove the specified AIX container from the global list.
 */
int aix_remove_container_from_list( storage_container_t * container )
{
	int rc;

	LOG_ENTRY;

	rc = DeleteObject(aix_container_list, container);

	if (rc) {
		LOG_SERIOUS("Error removing %s from list\n", container->name);
	}

	RETURN(rc);
}


/* aix_append_region_to_object
 *
 *	Associate the specified region and object as parent and child. Add the
 *	region to the object's "parent" list, and add the object to the region's
 *	"child" list.
 */
int aix_append_region_to_object(storage_object_t	* region,
				                storage_object_t	* object,
                                int                   expand)

{
	int	rc;
	void	* handle;

    LOG_ENTRY;

	rc = ExclusiveInsertObject(object->parent_objects,
				region,
				   get_tag_for_object(region),
				NULL,
				AppendToList,
				TRUE,
				&handle);

	if (rc != DLIST_SUCCESS || (rc == DLIST_OBJECT_ALREADY_IN_LIST && !expand)) {
		LOG_SERIOUS("Error adding region %s as a parent of object %s.\n",
			region->name, object->name);
		return rc;
	}

    LOG_DEBUG("region:%s object:%s \n", region->name, object->name);

	rc = ExclusiveInsertObject(region->child_objects,
				object,
				   get_tag_for_object(object),
				NULL,
				AppendToList,
				TRUE,
				&handle);
	if (rc != DLIST_SUCCESS || (rc == DLIST_OBJECT_ALREADY_IN_LIST && !expand)) {
		LOG_SERIOUS("Error adding object %s as a child of region %s.\n",
			object->name, region->name);
		DeleteObject(object->parent_objects, region);
	}

    RETURN(rc);

}


/* aix_remove_region_from_object
 *
 *	Dis-associate the specified region and object as parent and child.
 *	Remove the region from the object's "parent" list, and remove the
 *	object from the region's "child" list.
 */
int aix_remove_region_from_object(storage_object_t	* region,
				storage_object_t	* object )
{
	int	rc;

    LOG_ENTRY;

	rc = DeleteObject(object->parent_objects, region);
	if (rc) {
		LOG_SERIOUS("Error removing region %s as a parent of object %s.\n",
			region->name, object->name);
	}

	rc = DeleteObject(region->child_objects, object);
	if (rc) {
		LOG_SERIOUS("Error removing object %s as a child of region %s.\n",
			object->name, region->name);
	}

	RETURN(rc);
}


/* aix_clear_region_child_list
 *
 *	Remove all of the child objects from this region, and remove this
 *	region from the parent lists of all child objects.
 */
int aix_clear_region_child_list( storage_object_t * region )
{
	storage_object_t	* object;
	dlist_t			temp_child_list;
	int			rc;

	LOG_ENTRY;

	// Need to create a temporary duplicate of the child_objects list,
	// since we will be removing objects from it during the FOR_EACH loop.
	temp_child_list = CreateList();
	if ( ! temp_child_list ) {
		RETURN(ENOMEM);
	}
	CopyList(temp_child_list, region->child_objects, AppendToList);

	DLIST_FOR_EACH(object, temp_child_list, rc) {
		aix_remove_region_from_object(region, object);
	}

	DestroyList(&temp_child_list, FALSE);

	RETURN(0);
}

