/*
 *   Copyright (c) International Business Machines  Corp., 2001
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: aixregmgr
 * File: aix_io.c
 *
 * Description: This file contains functions related to I/O in the AIX
 *		region manager.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <plugin.h>
#include "aixregmgr.h"


/* Function: bytes_to_sectors
 *
 *	Convert a number of bytes into a number of sectors, rounding up.
 */
inline u_int64_t bytes_to_sectors( u_int64_t bytes )
{
	return( (bytes >> EVMS_VSECTOR_SIZE_SHIFT) + ((bytes & (EVMS_VSECTOR_SIZE-1)) ? 1 : 0) );
}


inline u_int64_t sectors_to_bytes( u_int64_t sectors )
{
	return( sectors << EVMS_VSECTOR_SIZE_SHIFT );
}
/*
 * Function: remap sector
 *  Common function to remap volume lba to partition lba in appropriate PE
 */

int AIX_remap_sector(storage_object_t        * region,
		     lsn_t                     org_sector, // logical sector to remap
		     sector_count_t            size,	   // size (in sectors) of request to remap
		     lsn_t                   * new_sector, // remapped sector
		     sector_count_t          * new_size,   // new size (in sectors)
		     storage_object_t       ** pv_entry)   // new node for which new_sector is relative

{

	aix_region_data_t           * r_data = region->private_data;
	aix_container_data_t    * c_data = region->producing_container->private_data;
	u_int32_t               sectors_per_stripe;
	u_int32_t               partition_to_use;
	u_int32_t               column;
	u_int32_t               stripe_in_column;
	u_int32_t               le;
	u_int32_t               offset_in_le;
	u_int32_t                   mirror_copy;
	u_int32_t               stripe_size = 0;
	u_int32_t               pp_size_sectors = PP_SIZE_TO_VSECTORS(c_data->vgda->vg_head->pp_size);
	u_int32_t               org_sector32;	// Until striping is 64-bit enabled.

	LOG_DEBUG("********** %s lvname:%d **********\n",__FUNCTION__, r_data->lv->lvname);
	LOG_DEBUG(" org_sector:%Ld size:%d pp_size_sectors:%d\n",org_sector, size, c_data->vgda->vg_head->pp_size);
	LOG_DEBUG("c_data->vgda->pv_headers[0]->pv_num:%d\n", c_data->vgda->pv_headers[0]->pv_num);

	org_sector32 = org_sector;

	*(new_size) = size;

	// Check if volume is striped. Reset the size if the request
	// crosses a stripe boundary.
	if ( r_data->lv->striping_width > 1 ) {
		LOG_DEBUG(" *** STRIPED ***\n");
		LOG_DEBUG(" ------- volume->stripe_size:%d org_sector:%d volume_stripes:%d\n",stripe_size, org_sector32, r_data->lv->striping_width);

		le              = org_sector / pp_size_sectors;
		offset_in_le  = org_sector % pp_size_sectors;

		LOG_DEBUG("OLD - le:%d -- offset_in_le:%d \n",le, offset_in_le);

		stripe_size        = GET_PHYSICAL_PART_SIZE(r_data->lv->stripe_exp);
		sectors_per_stripe =  stripe_size / AIX_SECTOR_SIZE;
		partition_to_use   = (org_sector32 / sectors_per_stripe) % r_data->lv->striping_width;
		stripe_in_column   = ((((org_sector32 / stripe_size) / r_data->lv->striping_width) * stripe_size) + (org_sector32 % sectors_per_stripe));
		column             = ((org_sector32 / sectors_per_stripe) / r_data->lv->striping_width) * sectors_per_stripe;

		LOG_DEBUG("offset_in_le:%d org_sector:%Ld pe_shift:%d \n",offset_in_le, org_sector, r_data->lv->stripe_exp);

		LOG_DEBUG(" org_sector:%d  sectors_per_stripe:%d partition_to_use:%d stripe_in_column:%d column:%d\n",org_sector32, sectors_per_stripe, partition_to_use,stripe_in_column,column);
		LOG_DEBUG(" offset_in_le + size:%Ld volume->pe_size:%d volume->lv_size:%Ld\n",(offset_in_le+size),pp_size_sectors ,r_data->lv->num_lps);

		if ( offset_in_le + size > pp_size_sectors ) {
			*new_size = pp_size_sectors - offset_in_le;
			LOG_DEBUG("  new_size %Ld\n",*new_size);
		}

	}
	// Non-striped volume. Just find LE and offset. Reset the size
	// if the request crosses an LE boundary.
	else {
		LOG_DEBUG(" *** NON-STRIPED ***\n");

		le              = org_sector / pp_size_sectors;
		offset_in_le  = org_sector % pp_size_sectors;

		if ( offset_in_le + size >  pp_size_sectors) {
			*new_size = pp_size_sectors - offset_in_le;
		}


	}

	LOG_DEBUG(" le:%d offset_in_le:%d org_sector:%Ld shift:%d\n",le, offset_in_le, org_sector, r_data->lv->stripe_exp);
	LOG_DEBUG(" pp_size_sectors:%d \n", pp_size_sectors);

	if (le >= r_data->lv->num_lps) {
		LOG_DEBUG(" le Memory Overwrite !! le:%d vs volume->num_le:%d\n",le,r_data->lv->num_lps);
		return EINVAL;
	}

	LOG_DEBUG("c_data->vgda->pp_array[0]->copy:%d\n", c_data->vgda->pp_array[0]->copy);

	switch (c_data->vgda->pp_array[0]->copy)
	{
	case LVM_PRIMARY:
	case LVM_PRIMOF2:
	case LVM_PRIMOF3:
		mirror_copy = 0;
		break;
	case LVM_SCNDOF2:
	case LVM_SCNDOF3:
		mirror_copy = 1;
		break;
	case LVM_TERTOF3:
		mirror_copy = 2;
		break;
	default:
		LOG_SERIOUS("ERROR: Invalid value for mirror copy.\n");
		mirror_copy = 0;
		break;
	}

	LOG_DEBUG("mirror_copy:%d\n", mirror_copy);

	*(new_sector)       = (r_data->lp_map[le].pp_number * pp_size_sectors) + c_data->vgda->pv_headers[0]->psn_part1 + offset_in_le;
	*(pv_entry)         = r_data->lp_map[le].object;

	LOG_DEBUG(" new_sector:%Ld\n", *(new_sector));
	LOG_DEBUG(" Owning pv_entry %p\n",*(pv_entry));

	return(0);
}

/* aix_read_ipl
 *
 *	Read the AIX IPL record from the specified object, and check for a valid
 *	IPL signature. The IPL record is always the first sector of an object.
 */
int aix_read_ipl(storage_object_t       * object,
		 aix_ipl_rec_t           ** ipl )
{
	aix_ipl_rec_t   * ipl_buffer;
	int             rc;

	LOG_ENTRY;

	// Allocate a new IPL record. This structure is padded to one
	// sector, so we don't need a separate buffer.
	ipl_buffer = aix_engine->engine_alloc(sizeof(aix_ipl_rec_t));
	if ( ! ipl_buffer ) {
		LOG_CRITICAL("Memory error creating buffer to read IPL record for %s\n", object->name);
		RETURN(ENOMEM);
	}

	// Read the first sector.
	rc = READ(object, PSN_IPL_REC, 1, ipl_buffer);
	if (rc) {
		aix_engine->engine_free(ipl_buffer);
		LOG_SERIOUS("Error reading IPL record from %s\n", object->name);
		RETURN(EIO);
	}

	// Check for a valid signature.
	if ( ipl_buffer->IPL_record_id != IPL_REC_ID ) {
		aix_engine->engine_free(ipl_buffer);
		LOG_EXTRA("%s is not an AIX PV.\n", object->name);
		RETURN(EINVAL);
	}

	// Return the IPL structure.
	*ipl = ipl_buffer;

	RETURN(0);
}


/* aix_write_ipl
 *
 *	Write the IPL structure to the first sector of the specified object.
 */
int aix_write_ipl( storage_object_t * object )
{
	aix_pv_data_t   * pv = object->consuming_private_data;
	int             rc;

	LOG_ENTRY;


	// Write the first sector. The IPL structure is sector-size padded,
	// so we don't need a separate I/O buffer.
	rc = WRITE(object, PSN_IPL_REC, 1, pv->ipl);
	if (rc) {
		LOG_SERIOUS("Error writing IPL record to %s\n", object->name);
		RETURN(EIO);
	}

	RETURN(0);
}


/* aix_read_lvm
 *
 *	Read the AIX LVM record from the specified object, and check for a valid
 *	LVM signature. The LVM record is always at sector 7 of an object.
 */
int aix_read_lvm(storage_object_t       * object,
		 aix_lvm_rec_t           ** lvm )
{
	aix_lvm_rec_t   * lvm_buffer;
	int             rc;

	LOG_ENTRY;

	// Allocate a new LVM record. This structure is padded to one
	// sector, so we only need one alloction.
	lvm_buffer = aix_engine->engine_alloc(sizeof(aix_lvm_rec_t));
	if ( ! lvm_buffer ) {
		LOG_CRITICAL("Memory error creating buffer to read LVM record for %s\n", object->name);
		RETURN(ENOMEM);
	}

	// Read the LVM sector.
	rc = READ(object, PSN_LVM_REC, 1, lvm_buffer);
	if (rc) {
		aix_engine->engine_free(lvm_buffer);
		LOG_SERIOUS("Error reading LVM record from %s\n", object->name);
		RETURN(EIO);
	}

	// Check for a valid signature
	if ( lvm_buffer->lvm_id != AIX_LVM_LVMID ) {
		aix_engine->engine_free(lvm_buffer);
		LOG_EXTRA("%s is not managed by the AIX LVM.\n", object->name);
		RETURN(EINVAL);
	}

	// Return the LVM structure.
	*lvm = lvm_buffer;

	RETURN(0);
}


/* aix_write_lvm
 *
 *	Write the LVM sector to the specified object. The LVM is located at
 *	sector 7.
 */
int aix_write_lvm( storage_object_t * object )
{
	aix_pv_data_t   * pv = object->consuming_private_data;
	int             rc;

	LOG_ENTRY;

	// Write the LVM sector. The LVM structure is sector-size padded,
	// so we don't need a separate I/O buffer.
	rc = WRITE(object, PSN_LVM_REC, 1, pv->lvm);
	if (rc) {
		LOG_SERIOUS("Error writing LVM record to %s\n", object->name);
		RETURN(EIO);
	}

	RETURN(0);
}


/* aix_read_vg_header
 *
 *	Read the VG header from the specified PV. The LVM record in the PV has
 *	a pointer to the beginning of the VGDA, and the VG header is the first
 *	sector in the VGDA. Copy specifies the first (0) or second (1) VGDA.
 */
int aix_read_vg_header( storage_object_t        * object,
			vg_header               ** vgh,
			unsigned int            copy )
{
	aix_pv_data_t   * pv = object->consuming_private_data;
	vg_header       * vgh_buffer;
	int             rc;

	LOG_ENTRY;

	// Allocate a buffer to read the header. The header structure is not
	// sector-size padded, so we need a separate buffer.
	vgh_buffer = aix_engine->engine_alloc(EVMS_VSECTOR_SIZE);
	if ( ! vgh_buffer ) {
		LOG_CRITICAL("Memory error creating buffer to read VG header from %s\n", object->name);
		RETURN(ENOMEM);
	}

	// Read the header.
	rc = READ(object, pv->lvm->vgda_psn[copy], 1, vgh_buffer);
	if (rc) {
		aix_engine->engine_free(vgh_buffer);
		LOG_SERIOUS("Error reading VG header (copy %d) from %s\n", copy, object->name);
		RETURN(rc);
	}

	// Allocate a VG header structure to return.
	*vgh = aix_engine->engine_alloc(sizeof(vg_header));
	if ( ! *vgh ) {
		aix_engine->engine_free(vgh_buffer);
		LOG_CRITICAL("Memory error creating new VG header for %s\n", object->name);
		RETURN(ENOMEM);
	}

	// Copy the structure
	memcpy(*vgh, vgh_buffer, sizeof(vg_header));
	aix_engine->engine_free(vgh_buffer);
	RETURN(0);
}


/* aix_read_vg_headers
 *
 *	Read both copies of the VG header from the specified PV.
 */
int aix_read_vg_headers( storage_object_t * object )
{
	aix_pv_data_t   * pv = object->consuming_private_data;
	vg_header       * vgh;
	int             rc;

	rc = aix_read_vg_header(object, &vgh, 0);
	if (rc) {
		RETURN(rc);
	}
	pv->vg_head[0] = vgh;

	rc = aix_read_vg_header(object, &vgh, 1);
	if (rc) {
		RETURN(rc);
	}
	pv->vg_head[1] = vgh;

	RETURN(0);
}


/* aix_write_vg_header
 *
 *	Write the VG header to the specified copy of the VGDA on this object.
 */
int aix_write_vg_header(storage_object_t        * object,
			unsigned int            copy )
{
	aix_pv_data_t           * pv = object->consuming_private_data;
	storage_container_t     * container = object->consuming_container;
	aix_container_data_t    * c_data = container->private_data;
	vg_header               * vgh_buffer;
	int                     rc;

	LOG_ENTRY;

	// Need a separate I/O buffer for the VG header.
	vgh_buffer = aix_engine->engine_alloc(EVMS_VSECTOR_SIZE);
	if ( ! vgh_buffer ) {
		LOG_CRITICAL("Memory error creating buffer to write VG header to %s\n", object->name);
		RETURN(ENOMEM);
	}

	// Copy the data to the buffer.
	memcpy(vgh_buffer, c_data->vgda->vg_head, sizeof(vg_header));

	LOG_DEBUG("vgh_buffer maxlvs:%d numlvs:%d numpvs:%d\n", vgh_buffer->maxlvs, vgh_buffer->numlvs, vgh_buffer->numpvs);

	// Write the buffer..
	rc = WRITE(object, pv->lvm->vgda_psn[copy], 1, vgh_buffer);
	if (rc) {
		aix_engine->engine_free(vgh_buffer);
		LOG_SERIOUS("Error writing VG header (copy %d) to %s\n", copy, object->name);
		RETURN(rc);
	}

	aix_engine->engine_free(vgh_buffer);
	RETURN(0);
}


/* aix_read_vg_trailer
 *
 *	Read the VG trailer from the specified PV. The VG trailer is the
 *	last sector in the VGDA. Copy specifies the first (0) or second (1)
 *	VGDA.
 */
int aix_read_vg_trailer(storage_object_t        * object,
			vg_trailer              ** vgt,
			unsigned int            copy )
{
	aix_pv_data_t   * pv = object->consuming_private_data;
	vg_trailer      * vgt_buffer;
	int             rc;

	LOG_ENTRY;

	// Allocate a buffer to read the trailer. The trailer structure is not
	// sector-size padded, so we need a separate buffer.
	vgt_buffer = aix_engine->engine_alloc(EVMS_VSECTOR_SIZE);
	if ( ! vgt_buffer ) {
		LOG_CRITICAL("Memory error creating buffer to read VG trailer from %s\n", object->name);
		RETURN(ENOMEM);
	}

	// Read the trailer.
	rc = READ(object, pv->lvm->vgda_psn[copy] + pv->lvm->vgda_len - 1, 1, vgt_buffer);
	if (rc) {
		aix_engine->engine_free(vgt_buffer);
		LOG_SERIOUS("Error reading VG trailer (copy %d) from %s\n", copy, object->name);
		RETURN(EIO);
	}

	// Allocate a VG trailer structure to return.
	*vgt = aix_engine->engine_alloc(sizeof(vg_trailer));
	if ( ! *vgt ) {
		aix_engine->engine_free(vgt_buffer);
		LOG_CRITICAL("Memory error creating new VG trailer for %s\n", object->name);
		RETURN(ENOMEM);
	}

	// Copy the structure
	memcpy(*vgt, vgt_buffer, sizeof(vg_trailer));
	aix_engine->engine_free(vgt_buffer);
	RETURN(0);
}


/* aix_read_vg_trailers
 *
 *	Read both copies of the VG trailer from the specified PV.
 */
int aix_read_vg_trailers( storage_object_t * object )
{
	aix_pv_data_t   * pv = object->consuming_private_data;
	vg_trailer      * vgt;
	int             rc;

	rc = aix_read_vg_trailer(object, &vgt, 0);
	if (rc) {
		RETURN(rc);
	}
	pv->vg_tail[0] = vgt;

	rc = aix_read_vg_trailer(object, &vgt, 1);
	if (rc) {
		RETURN(rc);
	}
	pv->vg_tail[1] = vgt;

	RETURN(0);
}


/* aix_write_vg_trailer
 *
 *	Write the VG trailer to the specified copy of the VGDA on this object.
 */
int aix_write_vg_trailer(storage_object_t       * object,
			 unsigned int            copy )
{
	aix_pv_data_t           * pv = object->consuming_private_data;
	storage_container_t     * container = object->consuming_container;
	aix_container_data_t    * c_data = container->private_data;
	vg_trailer              * vgt_buffer;
	int                     rc;

	LOG_ENTRY;

	// Need a separate I/O buffer for the VG trailer.
	vgt_buffer = aix_engine->engine_alloc(EVMS_VSECTOR_SIZE);
	if ( ! vgt_buffer ) {
		LOG_CRITICAL("Memory error creating buffer to write VG trailer to %s\n", object->name);
		RETURN(ENOMEM);
	}

	// Copy the data to the buffer.
	memcpy(vgt_buffer, c_data->vgda->vg_tail, sizeof(vg_trailer));

	// Write the buffer..
	rc = WRITE(object, pv->lvm->vgda_psn[copy] + pv->lvm->vgda_len - 1, 1, vgt_buffer);
	if (rc) {
		aix_engine->engine_free(vgt_buffer);
		LOG_SERIOUS("Error writing VG trailer (copy %d) to %s\n", copy, object->name);
		RETURN(EIO);
	}

	aix_engine->engine_free(vgt_buffer);
	RETURN(0);
}


/* aix_read_lv_array
 *
 *	Read the LV array from the specified VGDA copy on the specified PV. The
 *	LV array is the second item in the VGDA, and is 16 sectors long. There
 *	are 16 LV structs per sector, for a max of 256 LVs per VG.
 */
int aix_read_lv_array(  storage_object_t        * object,
			unsigned int            copy )
{
	aix_pv_data_t           * pv = object->consuming_private_data;
	storage_container_t     * container = object->consuming_container;
	aix_container_data_t    * c_data = container->private_data;
	aix_vgda_t              * vgda = c_data->vgda;
	int                     rc,i;

	LOG_ENTRY;

	// Has the LV array already been read?
	if ( vgda->lv_array ) {
		LOG_EXTRA("Already read LV array for container %s\n", container->name);
		RETURN(0);
	}

	// Allocate the buffer to hold the array
	vgda->lv_array = aix_engine->engine_alloc(MAX_SECTORS_LV_ENTRIES * EVMS_VSECTOR_SIZE);
	if ( ! vgda->lv_array ) {
		LOG_CRITICAL("Memory error creating buffer for LV array for container %s\n", container->name);
		RETURN(ENOMEM);
	}

	// Read in the entire array.
	rc = READ(object, pv->lvm->vgda_psn[copy] + PSN_LVE_OFFSET, MAX_SECTORS_LV_ENTRIES, vgda->lv_array);
	for (i = 0; i < AIX_MAX_LVS; i++) {
		if (vgda->lv_array != NULL) {
			LOG_DEBUG("lv:%d num_lps:%d mirrors:%d\n", vgda->lv_array[i].lvname, vgda->lv_array[i].num_lps,vgda->lv_array[i].mirror);
			aix_calc_disk_lp_num(&vgda->lv_array[i], READ_DATA);
		}
	}
	if (rc) {
		LOG_SERIOUS("Error reading LV array from %s\n", object->name);
		aix_engine->engine_free(vgda->lv_array);
		vgda->lv_array = NULL;
		RETURN(EIO);
	}


	RETURN(0);
}


/* aix_write_lv_array
 *
 *	Write the LV array to the specified VGDA copy on the specified PV. The
 *	LV array is the second item in the VGDA, and is 16 sectors long. There
 *	are 16 LV structs per sector, for a max of 256 LVs per VG.
 */
int aix_write_lv_array( storage_object_t        * object,
			unsigned int            copy )
{
	aix_pv_data_t           * pv = object->consuming_private_data;
	storage_container_t     * container = object->consuming_container;
	aix_container_data_t    * c_data = container->private_data;
	int                     rc,i;

	LOG_ENTRY;

	LOG_DEBUG("lv_array:%d \n", c_data->vgda->lv_array->num_lps);

	// Write the entire array. The array in memory is max-allocated, so
	// no separate I/O buffer is needed.
	for (i = 0; i < AIX_MAX_LVS; i++) {
		if (c_data->vgda->lv_array != NULL) {
			LOG_DEBUG("lv:%d num_lps:%d mirrors:%d \n", c_data->vgda->lv_array[i].lvname,
				  c_data->vgda->lv_array[i].num_lps,
				  c_data->vgda->lv_array[i].mirror);

			aix_calc_disk_lp_num(&c_data->vgda->lv_array[i], WRITE_DATA);
		}
	}

	rc = WRITE(object, pv->lvm->vgda_psn[copy] + PSN_LVE_OFFSET, MAX_SECTORS_LV_ENTRIES, c_data->vgda->lv_array);
	if (rc) {
		LOG_SERIOUS("Error writing LV array to %s\n", object->name);
		RETURN(EIO);
	}

	for (i = 0; i < AIX_MAX_LVS; i++) {
		if (c_data->vgda->lv_array != NULL) {
			LOG_DEBUG("lv:%d num_lps:%d mirrors:%d \n", c_data->vgda->lv_array[i].lvname,
				  c_data->vgda->lv_array[i].num_lps,
				  c_data->vgda->lv_array[i].mirror);

			aix_calc_disk_lp_num(&c_data->vgda->lv_array[i], READ_DATA);
		}
	}

	RETURN(0);
}


/* aix_read_lv_names
 *
 *	Read the LV namelist from the specified copy of the VGDA on the
 *	specified PV. The LV namelist is the second-to-last item in the VGDA,
 *	and is 32 sectors long. There are 256 LVs per VG, so each name is 64
 *	characters.
 */
int aix_read_lv_names(  storage_object_t        * object,
			unsigned int            copy )
{
	aix_pv_data_t           * pv = object->consuming_private_data;
	storage_container_t     * container = object->consuming_container;
	aix_container_data_t    * c_data = container->private_data;
	aix_vgda_t              * vgda = c_data->vgda;
	int                     rc;

	LOG_ENTRY;

	// Has the namelist already been read?
	if ( vgda->lv_names ) {
		LOG_EXTRA("Already read LV namelist for container %s\n", container->name);
		RETURN(0);
	}

	// Allocate the buffer to hold the array
	vgda->lv_names = aix_engine->engine_alloc(MAX_SECTORS_NAMELIST * EVMS_VSECTOR_SIZE);
	if ( ! vgda->lv_names ) {
		LOG_CRITICAL("Memory error creating buffer for LV namelist for container %s\n", container->name);
		RETURN(ENOMEM);
	}

	// Read in the entire list
	rc = READ(object, pv->lvm->vgda_psn[copy] + pv->lvm->vgda_len - 1 - MAX_SECTORS_NAMELIST, MAX_SECTORS_NAMELIST, vgda->lv_names);
	if (rc) {
		LOG_SERIOUS("Error reading LV namelist from %s\n", object->name);
		aix_engine->engine_free(vgda->lv_names);
		vgda->lv_names = NULL;
		RETURN(rc);
	}

	RETURN(0);
}


/* aix_write_lv_names
 *
 *	Write the LV namelist to the specified copy of the VGDA on the
 *	specified PV. The LV namelist is the second-to-last item in the VGDA,
 *	and is 32 sectors long. There are 256 LVs per VG, so each name is 64
 *	characters.
 */
int aix_write_lv_names( storage_object_t        * object,
			unsigned int            copy )
{
	aix_pv_data_t           * pv = object->consuming_private_data;
	storage_container_t     * container = object->consuming_container;
	aix_container_data_t    * c_data = container->private_data;
	int                     rc;

	LOG_ENTRY;

	// Write the entire list. Like the lv_array, this is stored max-allocated in
	// memory, so no extra I/O buffer is necessary.
	rc = WRITE(object, pv->lvm->vgda_psn[copy] + pv->lvm->vgda_len - 1 - MAX_SECTORS_NAMELIST, MAX_SECTORS_NAMELIST, c_data->vgda->lv_names);
	if (rc) {
		LOG_SERIOUS("Error writing LV namelist to %s\n", object->name);
		RETURN(rc);
	}

	RETURN(0);
}


/* aix_read_pvs
 *
 *	Read all of the PV headers and PP maps from the specified copy of the
 *	VGDA on the specified PV. Each PV header/PP map is unknown sectors long,
 *	and the first header starts 17 sectors from the start of the VGDA.
 */
int aix_read_pvs(storage_object_t       * object,
		 unsigned int            copy )
{
	aix_pv_data_t           * pv = object->consuming_private_data;
	storage_container_t     * container = object->consuming_container;
	aix_container_data_t    * c_data = container->private_data;
	aix_vgda_t              * vgda = c_data->vgda;
	pv_header               * pvh_buffer;
	unsigned int            pv_num;
	int                     rc;
	unsigned int  pvh_posn[AIX_MAX_PVS];

	LOG_ENTRY;


	// Create a buffer for reading the PV headers.
	pvh_buffer = aix_engine->engine_alloc(AIX_MAX_PV_SECTORS * AIX_SECTOR_SIZE);
	if ( ! pvh_buffer ) {
		LOG_CRITICAL("Memory error creating buffer to read PV headers for container %s\n", container->name);
		RETURN(ENOMEM);
	}

	rc = AIX_pvh_data_posn_read(pv->lvm->vgda_psn[copy],pvh_posn,
				    object, c_data->vgda->vg_head->numpvs);

	for ( pv_num = 0; pv_num < c_data->vgda->vg_head->numpvs; pv_num++ ) {
		if ( vgda->pv_headers[pv_num] ) {
			LOG_EXTRA("Already read PV header %d for container %s\n", pv_num, container->name);
			continue;
		}

		// Read the PV header for the current PV number.
		rc = READ(object, pvh_posn[pv_num+1], AIX_MAX_PV_SECTORS, pvh_buffer);
		if (rc) {
			LOG_SERIOUS("Error reading PV header %d from %s\n", pv_num, object->name);
			continue;
		}

		LOG_DEBUG("pvh_buffer pvh_posn:%d pv_num:%d pv_state:%d \n", pvh_posn[pv_num+1], pvh_buffer->pv_num, pvh_buffer->pv_state);

		// Check to see if this PV header is valid.
		if ( ! pvh_buffer->pv_state ) {
			continue;
		}

		// Allocate a new PV header entry in the VGDA.
		vgda->pv_headers[pv_num] = aix_engine->engine_alloc(AIX_MAX_PV_SECTORS * EVMS_VSECTOR_SIZE);
		if ( ! vgda->pv_headers[pv_num] ) {
			LOG_CRITICAL("Memory error creating buffer for PV header for container %s\n", container->name);
			continue;
		}

		// Copy the data.
		memcpy(vgda->pv_headers[pv_num], pvh_buffer, AIX_MAX_PV_SECTORS * EVMS_VSECTOR_SIZE);
		vgda->pp_array[pv_num] = (pp_entries*)&(vgda->pv_headers[pv_num][1]);
	}

	aix_engine->engine_free(pvh_buffer);

	RETURN(0);
}


/* aix_write_pvs
 *
 *	Write all of the PV headers and PP maps to the specified copy of the
 *	VGDA on the specified PV. Each PV header/PP map is 34 sectors long,
 *	and the first header starts 17 sectors from the start of the VGDA.
 */
int aix_write_pvs(      storage_object_t        * object,
			unsigned int            copy )
{
	aix_pv_data_t           * pv = object->consuming_private_data;
	storage_container_t     * container = object->consuming_container;
	aix_container_data_t    * c_data = container->private_data;
	aix_vgda_t              * vgda = c_data->vgda;
	pv_header               * pvh_buffer;
	unsigned int      pv_num;
	unsigned int      pvh_posn[AIX_MAX_PVS];
	int                     rc, count = 0;

	LOG_ENTRY;


	rc = GetListSize(container->objects_consumed, &count);

	LOG_DEBUG("objects_consumed:%d \n", count);

	// Allocate a buffer to use for I/O.
	pvh_buffer = aix_engine->engine_alloc(AIX_MAX_PV_SECTORS * AIX_SECTOR_SIZE);
	if ( ! pvh_buffer ) {
		LOG_CRITICAL("Memory error creating buffer to write PV headers for container %s\n", container->name);
		RETURN(ENOMEM);
	}


	rc = AIX_pvh_data_posn_write(pv->lvm->vgda_psn[copy],pvh_posn,
				     vgda->pv_headers, count);

	// Go through each PV header entry in the VGDA.
	for ( pv_num = 0; pv_num < count; pv_num++ ) {

		// If this PV exists, copy the PV data to the buffer. Otherwise,
		// zero the buffer.
		if ( vgda->pv_headers[pv_num] ) {
			memcpy(pvh_buffer, vgda->pv_headers[pv_num], AIX_MAX_PV_SECTORS * EVMS_VSECTOR_SIZE);
		}
		else {
			memset(pvh_buffer, 0, AIX_MAX_PV_SECTORS * EVMS_VSECTOR_SIZE);
		}


		LOG_DEBUG("pvh_buffer pvh_posn:%d pv_num:%d pv_state:%d \n", pvh_posn[pv_num+1], pvh_buffer->pv_num, pvh_buffer->pv_state);


		// Write the PV header for the current PV number.
		rc = WRITE(object, pvh_posn[pv_num+1], AIX_MAX_PV_SECTORS, pvh_buffer);
		if (rc) {
			LOG_SERIOUS("Error writing PV header %d to %s\n", pv_num, object->name);
			continue;
		}
	}

	aix_engine->engine_free(pvh_buffer);

	RETURN(0);
}


/* aix_read_vgsa
 *
 *	Read the specified copy of the Volume Group Status Area from the
 *	specified PV.
 */
int aix_read_vgsa(      storage_object_t        * object,
			unsigned int            copy )
{
	aix_pv_data_t           * pv = object->consuming_private_data;
	storage_container_t     * container = object->consuming_container;
	aix_container_data_t    * c_data = container->private_data;
	int rc;

	LOG_ENTRY;

	if ( c_data->vgsa ) {
		LOG_EXTRA("Already read VGSA for container %s\n", container->name);
		RETURN(0);
	}

	// The VGSA structure is padded to 8 sectors, so we don't need
	// a separate buffer.
	c_data->vgsa = aix_engine->engine_alloc(sizeof(vgsa_area));
	if ( ! c_data->vgsa ) {
		LOG_CRITICAL("Memory error creating VGSA for container %s\n", container->name);
		RETURN(ENOMEM);
	}

	rc = READ(object, pv->lvm->vgsa_psn[copy], pv->lvm->vgsa_len, c_data->vgsa);
	if (rc) {
		LOG_SERIOUS("Error reading VGSA from %s\n", object->name);
		aix_engine->engine_free(c_data->vgsa);
		c_data->vgsa = NULL;
		RETURN(rc);
	}

	RETURN(0);
}


/* aix_write_vgsa
 *
 *	Write the specified copy of the Volume Group Status Area to the
 *	specified PV.
 */
int aix_write_vgsa(     storage_object_t        * object,
			unsigned int            copy )
{
	aix_pv_data_t           * pv = object->consuming_private_data;
	storage_container_t     * container = object->consuming_container;
	aix_container_data_t    * c_data = container->private_data;
	int                     rc;

	LOG_ENTRY;

	// The VGSA structure is padded to 8 sectors, so we don't need
	// a separate I/O buffer.
	rc = WRITE(object, pv->lvm->vgsa_psn[copy], pv->lvm->vgsa_len, c_data->vgsa);
	if (rc) {
		LOG_SERIOUS("Error writing VGSA to %s\n", object->name);
		RETURN(rc);
	}

	RETURN(0);
}


/* aix_erase_vgda
 *
 *	Queue up KILL_SECTORS for the specified copy of the VGDA on this PV.
 */
static int aix_erase_vgda( storage_object_t * object, int copy )
{
	aix_pv_data_t   * pv = object->consuming_private_data;

	LOG_ENTRY;

	KILL_SECTORS(object, pv->lvm->vgda_psn[copy], pv->lvm->vgda_len);

	RETURN(0);
}


/* aix_erase_vgsa
 *
 *	Queue up KILL_SECTORS for the specified copy of the VGSA on this PV.
 */
static int aix_erase_vgsa( storage_object_t * object, int copy )
{
	aix_pv_data_t   * pv = object->consuming_private_data;

	LOG_ENTRY;

	KILL_SECTORS(object, pv->lvm->vgsa_psn[copy], pv->lvm->vgsa_len);

	RETURN(0);
}


/* aix_erase_container_metadata
 */
int aix_erase_container_metadata( storage_object_t * object )
{
	int i;

	LOG_ENTRY;

	for ( i = 0; i < AIX_MAX_VGDAS; i++ ) {
		aix_erase_vgsa(object, i);
		aix_erase_vgda(object, i);
	}

	RETURN(0);
}


/* aix_erase_pv_metadata
 */
int aix_erase_pv_metadata( storage_object_t * object )
{
	LOG_ENTRY;

	// Erase the IPL and LVM sectors.
	KILL_SECTORS(object, PSN_IPL_REC, 1);
	KILL_SECTORS(object, PSN_LVM_REC, 1);

	RETURN(0);
}

int AIX_pvh_data_posn_read(u_int32_t vgda_psn, u_int32_t * pvh_posn, storage_object_t * object, u_int32_t numpvs)
{
	pv_header           * AIXpvh;
	int posn = 0;
	int num_pps;
	int tmp,i;

	LOG_DEBUG("APDP - vgda_psn:%d numpvs:%d object:%s\n", vgda_psn, numpvs, object->name);


	AIXpvh = aix_engine->engine_alloc(AIX_SECTOR_SIZE);
	if (!AIXpvh) {
		return -ENOMEM;
	}

	memset(AIXpvh, 0 , sizeof(pv_header));

	// Adjust this because when AIX VGs/Volumes are created on Intel platforms, the
	// pp_count could be anything since we don't give up the entire physical drive.
	// This is for calculation purposes only.

	pvh_posn[0] = 0;

	for (i = 1; i <= numpvs; i++) {
		//for (pv = object->consuming_private_data; pv->pv_head[i].pv_num != i; pv = pv->next );

		LOG_DEBUG("APDP line:%d pp_count:%d \n", __LINE__,  AIXpvh->pp_count);

		num_pps = AIXpvh->pp_count;
		num_pps++; // Account for the pv_header on the front

		while ((num_pps * sizeof(pp_entries)) % AIX_SECTOR_SIZE) {
			LOG_EXTRA("num_pps:%d \n", num_pps);
			num_pps++;
		}

		tmp = (num_pps * sizeof(pp_entries)) / AIX_SECTOR_SIZE;

		LOG_DEBUG("APDP tmp:%d num_pps:%d \n", tmp,num_pps);

		posn = ((vgda_psn + PSN_PVH_OFFSET) + ((i -1) * tmp));

		pvh_posn[i] = posn;

		LOG_DEBUG("APDP pvh_posn[%d]:%d \n", i,pvh_posn[i]);

		if (READ(object, posn, 1, AIXpvh)) {
			aix_engine->engine_free(AIXpvh);
			return -EIO;
		}

		LOG_DEBUG("Num_pps:%d @ pvh_posn[%d]:%d\n",AIXpvh->pp_count, i, pvh_posn[i]);

	}

	aix_engine->engine_free(AIXpvh);


	return 0;
}
int AIX_pvh_data_posn_write(u_int32_t vgda_psn, u_int32_t * pvh_posn, pv_header * * pv_headers, u_int32_t numpvs)
{
	pv_header       empty_header;
	pv_header           * AIXpvh = &empty_header;
	int posn = 0;
	int num_pps;
	int tmp,i;

	LOG_DEBUG("APDP - vgda_psn:%d numpvs:%d\n", vgda_psn, numpvs);

	memset(AIXpvh, 0 , sizeof(pv_header));

	// Adjust this because when AIX VGs/Volumes are created on Intel platforms, the
	// pp_count could be anything since we don't give up the entire physical drive.
	// This is for calculation purposes only.

	pvh_posn[0] = 0;

	for (i = 1; i <= numpvs; i++) {
		//for (pv = object->consuming_private_data; pv->pv_head[i].pv_num != i; pv = pv->next );

		LOG_DEBUG("APDP line:%d pp_count:%d \n", __LINE__,  AIXpvh->pp_count);

		num_pps = AIXpvh->pp_count;
		num_pps++; // Account for the pv_header on the front

		while ((num_pps * sizeof(pp_entries)) % AIX_SECTOR_SIZE) {
			LOG_EXTRA("num_pps:%d \n", num_pps);
			num_pps++;
		}

		tmp = (num_pps * sizeof(pp_entries)) / AIX_SECTOR_SIZE;

		LOG_DEBUG("APDP tmp:%d num_pps:%d \n", tmp,num_pps);

		posn = ((vgda_psn + PSN_PVH_OFFSET) + ((i -1) * tmp));

		pvh_posn[i] = posn;

		LOG_DEBUG("APDP pvh_posn[%d]:%d \n", i,pvh_posn[i]);

		AIXpvh = pv_headers[i-1];
		if (AIXpvh == NULL) {
			return EFAULT;
		}

		LOG_DEBUG("Num_pps:%d @ pvh_posn[%d]:%d\n",AIXpvh->pp_count, i, pvh_posn[i]);

	}

	return 0;
}
int aix_calc_disk_lp_num(lv_entries * lv, int type)
{


	LOG_ENTRY;

	if (type == READ_DATA) {
		lv->num_lps = lv->num_lps * lv->mirror;
	}
	else if (type == WRITE_DATA && lv->num_lps > 0) {
		lv->num_lps = lv->num_lps / lv->mirror;
	}
	else {
		RETURN(-1);
	}

	RETURN(0);
}
