/*
 *   Copyright (c) International Business Machines  Corp., 2001
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or 
 *   (at your option) any later version.
 * 
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software 
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: aixregmgr
 * File: aix_options.h
 */ 

#ifndef _AIX_OPTIONS_H_
#define _AIX_OPTIONS_H_ 1

// Macros used by several option handling functions.

/* Macro: SET_STRING
 * Allocate space for a string and copy. a and b must be char*
 */
#define SET_STRING(a, b) \
{ \
	if (a) { \
		aix_engine->engine_free(a); \
		(a) = NULL; \
	} \
	(a) = aix_engine->engine_alloc(strlen(b)+1); \
	if (a) { \
		strncpy((a),(b),strlen(b)); \
	} else { \
		RETURN(ENOMEM); \
	} \
}


/* Macro: SET_RANGE32
 * Set a ui32 constraint range. b is min, c is max, d is increment.
 */
#define SET_RANGE32(a, b, c, d) \
{ \
	if (a) { \
		aix_engine->engine_free(a); \
		(a) = NULL; \
	} \
	(a) = aix_engine->engine_alloc(sizeof(value_range_t)); \
	if (a) { \
		(a)->min.ui32 = b; \
		(a)->max.ui32 = c; \
		(a)->increment.ui32 = d; \
	} else { \
		RETURN(ENOMEM); \
	} \
}


/* Macro: SET_POWER2_LIST
 * Set a contraint list that increments each item by a power of 2.
 * b is the starting value, c is ending value (both must be power of 2).
 */
#define SET_POWER2_LIST(a, b, c) \
{ \
	if ( ((b) & ((b)-1)) || ((c) & ((c)-1)) ) { \
		RETURN(EINVAL); \
	} \
	if (a) { \
		aix_engine->engine_free(a); \
		(a) = NULL; \
	} \
	(a) = aix_engine->engine_alloc(aix_log2(c)*sizeof(value_t)+1); \
	if (a) { \
		long i = b; \
		(a)->count = 0; \
		while ( i <= c ) { \
			(a)->value[(a)->count].ui32 = i; \
			(a)->count++; \
			i <<= 1; \
		} \
	} else { \
		RETURN(ENOMEM); \
	} \
}


// Definitions of the options available in LVM

// Container creation options.
#define	AIX_OPTION_CONTAINER_PP_SIZE_INDEX	0
#define	AIX_OPTION_CONTAINER_PP_SIZE_STR	"pp_size"


// Region creation options
#define AIX_OPTION_REGION_NAME_INDEX		0
#define AIX_OPTION_REGION_NAME_STR		"name"
#define AIX_OPTION_REGION_NUM_LPS_INDEX		1
#define AIX_OPTION_REGION_NUM_LPS_STR		"num_pps"

#define AIX_OPTION_REGION_EXPAND_SIZE_INDEX 0
#define AIX_OPTION_REGION_EXPAND_SIZE_STR   "size"

// Container creation function prototypes.
extern int aix_create_container_parse_options(	option_array_t	* options,
						unsigned long	* pp_size );
extern int aix_create_container_verify_options( unsigned long	* pp_size );
extern int aix_create_container_allocate_option_descriptor( option_desc_array_t * od );
extern int aix_create_container_get_acceptable( dlist_t acceptable_segments );
extern int aix_create_container_set_objects(	task_context_t	* context,
						dlist_t		declined_objects,
						task_effect_t	* effect );
extern int aix_create_container_set_option(	task_context_t	* context,
						u_int32_t	index,
						value_t		* value,
						task_effect_t	* effect );


// Region creation function prototypes.
extern int aix_create_region_parse_options(	option_array_t		* options,
						aix_region_options_t	* region_opts );
extern int aix_create_region_verify_options( aix_region_options_t * region_opts );
extern int aix_create_region_allocate_option_descriptor( option_desc_array_t * od );
extern int aix_create_region_get_acceptable( dlist_t acceptable_segments );
extern int aix_create_region_set_objects(	task_context_t	* context,
						dlist_t		declined_objects,
						task_effect_t	* effect );
extern int aix_create_region_set_option(	task_context_t	* context,
						u_int32_t	index,
						value_t		* value,
						task_effect_t	* effect );


extern int aix_expand_region_parse_options(option_array_t   * options,
                                    storage_object_t        * region,
			                    	aix_region_options_t	* region_opts );

extern int aix_expand_region_verify_options( aix_region_options_t * region_opts,
                                                 storage_object_t * region);
extern int aix_shrink_region_parse_options(option_array_t		    * options,
                                    storage_object_t    * region,
			                    	aix_region_options_t	* region_opts );

extern int aix_shrink_region_verify_options( aix_region_options_t * region_opts,
                                      storage_object_t * region);

extern int aix_expand_container_allocate_option_descriptor( option_desc_array_t * od );

extern int aix_expand_container_get_acceptable( storage_container_t * container,
                                                dlist_t               acceptable_objects );

extern int aix_expand_region_allocate_option_descriptor( option_desc_array_t * od );

extern int aix_expand_region_get_acceptable( storage_object_t    * region,
                                             dlist_t               acceptable_objects );

extern int aix_shrink_region_allocate_option_descriptor( option_desc_array_t * od );

extern int aix_shrink_region_get_acceptable( storage_object_t    * region,
                                             dlist_t               acceptable_objects );
extern int aix_expand_region_set_objects(	task_context_t	* context,
					dlist_t		declined_objects,
					task_effect_t	* effect );

extern int aix_expand_container_set_objects(	task_context_t	* context,
					dlist_t		declined_objects,
					task_effect_t	* effect );

extern int aix_expand_container_set_option(	task_context_t	* context,
					u_int32_t	index,
					value_t		* value,
					task_effect_t	* effect );

extern int aix_expand_region_set_option(	task_context_t	* context,
					u_int32_t	index,
					value_t		* value,
					task_effect_t	* effect );
extern int aix_expand_container_parse_options(	option_array_t	* options,
					unsigned long	* pp_size );


extern int aix_list_acceptable_mirrors(storage_object_t *master_object,
                                       dlist_t          available_objects);

extern int aix_split_region_set_option(	task_context_t	* context,
					u_int32_t	index,
					value_t		* value,
					task_effect_t	* effect );
extern int aix_split_region_allocate_option_descriptor( option_desc_array_t * od );

extern int aix_get_available_objects(storage_object_t *region, int * count);

extern int aix_shrink_region_set_option(	task_context_t	* context,
					u_int32_t	index,
					value_t		* value,
					task_effect_t	* effect );

extern int aix_shrink_region_set_objects(	task_context_t	* context,
					dlist_t		declined_objects,
					task_effect_t	* effect );
#endif

