/*
 *   Copyright (c) International Business Machines  Corp., 2001
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: aixregmgr
 * File: aix_regions.c
 *
 * Description: This file contains functions related to region management
 *		in AIX.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <plugin.h>
#include "aixregmgr.h"


/* aix_convert_lv_name_to_region_name
 *
 *	Prepend the correct prefix and container name to the specified LV name.
 */
static int aix_convert_lv_name_to_region_name(  unsigned char           * lv_name,
						unsigned char           * region_name,
						storage_container_t     * container )
{
	LOG_ENTRY;

	memset(region_name, 0, EVMS_VOLUME_NAME_SIZE+1);
	strncpy(region_name, "aix/", strlen("aix/"));
	strncat(region_name, lv_name, EVMS_VOLUME_NAME_SIZE-strlen(region_name));

	RETURN(0);
}


/* aix_check_lv_name
 *
 *	Check that the specified name is not in use by any regions in any of the
 *	containers.
 */
int aix_check_lv_name( unsigned char * lv_name )
{
	storage_container_t     * container;
	storage_object_t        * region;
	unsigned char           region_name[EVMS_VOLUME_NAME_SIZE+1];
	int                     rc, rc2;

	LOG_ENTRY;

	DLIST_FOR_EACH(container, aix_container_list, rc) {
		aix_convert_lv_name_to_region_name(lv_name, region_name, container);

		DLIST_FOR_EACH(region, container->objects_produced, rc2) {
			rc = strncmp(region->name, region_name, EVMS_VOLUME_NAME_SIZE);
			if (!rc) {
				MESSAGE("\"%s\" is already in use as an AIX LV name.\n", lv_name);
				RETURN(EEXIST);
			}
		}
	}

	RETURN(0);
}


/* aix_get_freespace_region
 *
 *	This function gets the freespace region from the selected_objects
 *	list in a Task Context.
 */
storage_object_t * aix_get_freespace_region( dlist_t selected_objects )
{
	storage_object_t        * region;
	int                     rc;

	LOG_ENTRY;

	// The freespace should be the first item in the list.
	rc = GoToStartOfList(selected_objects);
	if (rc) {
		MESSAGE("Error finding the start of the selected-objects list.\n");
		RETURN(NULL);
	}

	// Get the freespace region from the list.
	region = aix_get_list_item(selected_objects);
	if ( ! region ) {
		MESSAGE("Error getting the first object from the selected-objects list.\n");
		RETURN(NULL);
	}

	// Make sure it is owned by AIX.
	if ( region->plugin != aix_plugin ) {
		MESSAGE("Selected object is not owned by AIX.\n");
		RETURN(NULL);
	}

	// Make sure it is a freespace object.
	if ( region->object_type != REGION ||
	     region->data_type != FREE_SPACE_TYPE ) {
		MESSAGE("Selected object is not a freespace region.\n");
		RETURN(NULL);
	}

	RETURN(region);
}


/* aix_deallocate_lp_map
 *
 *	Delete all of the LP maps for this region.
 */
int aix_deallocate_lp_map( storage_object_t * region )
{
	aix_region_data_t   * r_data = region->private_data;

	LOG_ENTRY;

	if ( r_data->lp_map) {
		aix_engine->engine_free(r_data->lp_map);
		r_data->lp_map = NULL;
	}

	RETURN(0);
}


/* aix_reallocate_lp_map
 *
 *	Reallocate the LP maps for this region.
 */
int aix_reallocate_lp_map( storage_object_t *region, unsigned long new_lps )
{

	aix_region_data_t   * r_data = region->private_data;
	aix_lp_map_entry_t  *tmp_lp_map;

	LOG_ENTRY;

	tmp_lp_map = r_data->lp_map;

	r_data->lp_map = aix_engine->engine_alloc( (new_lps) * sizeof(aix_lp_map_entry_t) );

	if ( ! r_data->lp_map ) {
		aix_deallocate_lp_map(region);
		LOG_CRITICAL("Memory error creating LP map %d for region %s\n", new_lps, region->name);
		RETURN(ENOMEM);
	} else {
		memset(r_data->lp_map, 0, (new_lps) * sizeof(aix_lp_map_entry_t));
	}

	LOG_DEBUG("Reallocating new_lps:%d lp_map %p to %d bytes \n", new_lps, r_data->lp_map, ((new_lps) * sizeof(aix_lp_map_entry_t)));

	if (  ! r_data->lp_map ) {
		aix_deallocate_lp_map(region);
		LOG_CRITICAL("Memory error creating resized LP map for region %s\n", region->name);
		RETURN(0);
	}

	if (tmp_lp_map) {
		memcpy(r_data->lp_map, tmp_lp_map, ((new_lps) * sizeof(aix_lp_map_entry_t)));
		aix_engine->engine_free(tmp_lp_map);
	}


	RETURN(0);
}

/* aix_allocate_lp_map
 *
 *	Allocate the LP maps for this region.
 */
int aix_allocate_lp_map( storage_object_t * region )
{
	aix_region_data_t   * r_data = region->private_data;
	lv_entries      * lv = r_data->lv;

	LOG_ENTRY;

	if ( lv->num_lps == 0 ) {
		aix_deallocate_lp_map(region);
		RETURN(0);
	}

	r_data->lp_map = aix_engine->engine_alloc(lv->num_lps * sizeof(aix_lp_map_entry_t));
	if ( ! r_data->lp_map ) {
		aix_deallocate_lp_map(region);
		LOG_CRITICAL("Memory error creating LP maps for region %s\n", region->name);
		RETURN(ENOMEM);
	}

	RETURN(0);
}


/* aix_deallocate_region
 *
 *	Free up all memory used by this region.
 */
int aix_deallocate_region(      storage_object_t        * region,
				boolean                 delete_evms_objects )
{
	aix_region_data_t       * r_data = region->private_data;

	LOG_ENTRY;

	if ( r_data ) {
		// Clear the parent/child associations with the PVs.
		aix_clear_region_child_list(region);

		// Free the LP maps.
		aix_deallocate_lp_map(region);

		// For freespace regions, need to free the LV entry.
		if ( region->data_type == FREE_SPACE_TYPE && r_data->lv ) {
			aix_engine->engine_free(r_data->lv);
		}
		r_data->lv = NULL;

		// Free the private data.
		aix_engine->engine_free(r_data);
		region->private_data = NULL;
	}

	// Free the region.
	if ( delete_evms_objects ) {
		aix_engine->free_region(region);
	}

	RETURN(0);
}


/* aix_allocate_region
 *
 *	Allocate a new AIX region and all necessary private data fields.
 *
 *	The "pp_size" argument to this function is in exponent form,
 *	NOT vsectors.
 */
storage_object_t * aix_allocate_region( lv_entries              * lv,
					char                    * lv_name,
					unsigned int            pp_size )
{
	storage_object_t        * region;
	aix_region_data_t       * r_data;
	char                    name[EVMS_VOLUME_NAME_SIZE+1] = {0};
	int                     rc;

	LOG_ENTRY;

	// Convert the AIX LV-name to an EVMS region-name.
	strncpy(name, AIX_NAME, EVMS_VOLUME_NAME_SIZE);
	strncat(name, "/", EVMS_VOLUME_NAME_SIZE - strlen(name));
	strncat(name, lv_name, EVMS_VOLUME_NAME_SIZE - strlen(name));

	// Create the region structure.
	rc = aix_engine->allocate_region(name, &region);
	if (rc) {
		LOG_CRITICAL("Memory error creating region %s\n", lv_name);
		RETURN(NULL);
	}

	// Create the region private data
	r_data = aix_engine->engine_alloc(sizeof(aix_region_data_t));
	if ( ! r_data ) {
		LOG_CRITICAL("Memory error creating private data for region %s\n", lv_name);
		aix_deallocate_region(region, TRUE);
		RETURN(NULL);
	}

	// Initialize remaining fields
	region->private_data    = r_data;
	region->object_type     = REGION;
	region->data_type       = DATA_TYPE;
	region->plugin          = aix_plugin;
	region->flags           = 0;
	r_data->lv                  = lv;
	region->size            = lv->num_lps * PP_SIZE_TO_VSECTORS(pp_size) / r_data->lv->mirror;

	// Allocate the LP-to-PP mapping tables. A freespace region will
	// initially wind up with no LP maps.
	rc = aix_allocate_lp_map(region);
	if (rc) {
		LOG_CRITICAL("Memory error creating LP maps for region %s.\n", lv_name);
		aix_deallocate_region(region, TRUE);
		RETURN(NULL);
	}

	LOG_DETAILS("Created region %s\n", name);

	RETURN(region);
}


/* aix_find_free_lv
 *
 *	Find the first available slot for a new LV in the specified container.
 */
int aix_find_free_lv( storage_container_t * container )
{
	aix_container_data_t    * c_data = container->private_data;
	int                     i;

	LOG_ENTRY;

	for ( i = 0; i < AIX_MAX_LVS; i++ ) {
		if ( ! c_data->regions[i] ) {
			RETURN(i);
		}
	}

	RETURN(-1);
}


/* aix_clear_lv_entry
 */
void aix_clear_lv_entry( lv_entries * lv )
{
	memset(lv, 0, sizeof(lv_entries));
}


/* aix_initialize_new_lv_entry
 *
 *	Find a free LV entry in the specified container and initialize. This
 *	also initializes the corresponding lv_name entry.
 */
lv_entries * aix_initialize_new_lv_entry( storage_container_t   * container,
					  aix_region_options_t    * region_opts )
{
	aix_container_data_t    * c_data = container->private_data;
	lv_entries              * new_lv;
	int                     lv_num;

	LOG_ENTRY;

	lv_num = aix_find_free_lv(container);
	if ( lv_num < 0 ) {
		MESSAGE("Container %s already has maximum number of regions.\n", container->name);
		RETURN(NULL);
	}

	new_lv = &(c_data->vgda->lv_array[lv_num]);
	aix_clear_lv_entry(new_lv);

	// All defaults for now (except num_lps). Much of this will change
	// when new creation options are implemented.
	new_lv->lvname          = lv_num;
	new_lv->maxsize         = AIX_MAX_LPS;
	new_lv->lv_state        = LVM_LVDEFINED;
	new_lv->mirror          = LVM_NOMIRROR;
	new_lv->mirror_policy   = LVM_PARALLEL;
	new_lv->num_lps         = region_opts->num_lps;
	new_lv->permissions     = LVM_RDWR;
	new_lv->bb_relocation   = LVM_RELOC;
	new_lv->write_verify    = LVM_NOVERIFY;
	new_lv->mirwrt_consist  = LVM_CONSIST;
	new_lv->stripe_exp      = 0;
	new_lv->striping_width  = 0;
	new_lv->lv_avoid        = 0;
	new_lv->child_minor_num = 0;

	// Set the name.
	memset(c_data->vgda->lv_names->name[lv_num], 0, AIX_NAMESIZ);
	strncpy(c_data->vgda->lv_names->name[lv_num], region_opts->name, AIX_NAMESIZ-1);

	RETURN(new_lv);
}


/* aix_calculate_freespace_lps
 *
 *	Go through all real regions in the container and calculate the number of
 *	LPs that should be allocated to the freespace region.
 */
static unsigned long aix_calculate_free_lps( storage_container_t * container )
{
	aix_container_data_t    * c_data = container->private_data;
	unsigned long           used_lps = 0;
	int                     i, freespace = 0;

	LOG_ENTRY;

	for ( i = 0; i < AIX_MAX_LVS; i++ ) {
		if ( c_data->vgda->lv_array[i].lv_state != LVM_LVUNDEF) {
			used_lps += c_data->vgda->lv_array[i].num_lps;
		}
	}

	if (c_data->pp_count - used_lps > 0) {
		freespace = c_data->pp_count - used_lps;

		if (freespace < 0) {
			freespace = 0;
		}
	}


	RETURN(freespace);
}


/* aix_create_freespace_region
 *
 *	Create a region to represent the freespace in the specified container,
 *	and place it in the container's list.
 */
int aix_create_freespace_region( storage_container_t * container )
{
	aix_container_data_t    * c_data = container->private_data;
	storage_object_t        * free_region;
	lv_entries              * free_lv;
	char                    free_name[EVMS_VOLUME_NAME_SIZE+1] = {0};

	LOG_ENTRY;

	if ( c_data->freespace ) {
		// Freespace region already exists.
		RETURN(0);
	}

	// Create a name for this freespace region.
	strncpy(free_name, &container->name[strlen(AIX_NAME)+1], EVMS_VOLUME_NAME_SIZE);
	strncat(free_name, "/", EVMS_VOLUME_NAME_SIZE - strlen(free_name));
	strncat(free_name, "Freespace", EVMS_VOLUME_NAME_SIZE - strlen(free_name));

	// Create a dummy lv_entries structure and initialize.
	free_lv = aix_engine->engine_alloc(sizeof(lv_entries));
	if ( ! free_lv ) {
		LOG_CRITICAL("Memory error creating LV structure for Freespace region for container %s\n", container->name);
		RETURN(ENOMEM);
	}

	free_lv->lvname         = -1;
	free_lv->lv_state       = LVM_LVDEFINED;
	free_lv->mirror         = 1;

	// Create the region.
	free_region = aix_allocate_region(free_lv, free_name, 0);
	if ( ! free_region ) {
		LOG_CRITICAL("Memory error creating Freespace region for container %s\n", container->name);
		RETURN(ENOMEM);
	}
	free_region->data_type = FREE_SPACE_TYPE;

	// Place the freespace region in the container.
	aix_add_object_to_list(free_region, container->objects_produced, AppendToList);
	free_region->producing_container = container;
	c_data->freespace = free_region;

	RETURN(0);
}


/* aix_update_freespace_region
 *
 *	As PVs are added/removed from a container and as regions are created
 *	and deleted, the amount of freespace in the container changes. This
 *	function will update the freespace region to the current number of
 *	free partitions on the container.
 */
int aix_update_freespace_region( storage_container_t * container )
{
	aix_container_data_t    * c_data = container->private_data;
	storage_object_t        * free_region = c_data->freespace;
	aix_region_data_t       * r_data = free_region->private_data;
	int                     rc;

	LOG_ENTRY;

	aix_clear_region_child_list(free_region);

	r_data->lv->num_lps = aix_calculate_free_lps(container);
	free_region->size = r_data->lv->num_lps * PP_SIZE_TO_VSECTORS(c_data->vgda->vg_head->pp_size);

	aix_deallocate_lp_map(free_region);
	rc = aix_allocate_lp_map(free_region);
	if (rc) {
		LOG_SERIOUS("Error creating updated LP map for freespace region %s.\n", free_region->name);
		RETURN(rc);
	}

	RETURN(0);
}




/* aix_discover_regions_in_container
 *
 *	Examine the LV array in the VG metadata, and create a new region
 *	for each valid LV entry.
 */
static int aix_discover_regions_in_container( storage_container_t * container )
{
	aix_container_data_t    * c_data = container->private_data;
	aix_vgda_t              * vgda = c_data->vgda;
	storage_object_t        * region;
	unsigned int            i;

	LOG_ENTRY;

	// Look for every valid entry in the LV array. Create a new region
	// for that entry if one does not yet exist.
	for ( i = 0; i < AIX_MAX_LVS; i++ ) {
		if ( vgda->lv_array[i].lv_state != LVM_LVUNDEF &&
		     ! c_data->regions[i] ) {
			region = aix_allocate_region(&vgda->lv_array[i],
						     vgda->lv_names->name[i],
						     vgda->vg_head->pp_size);
			if ( ! region ) {
				LOG_CRITICAL("Memory error creating region %s\n", vgda->lv_names->name[i]);
				continue;
			}

			LOG_DEBUG("Discovered AIX Region %s\n", vgda->lv_names->name[i]);

			aix_add_object_to_list(region, container->objects_produced, AppendToList);
			region->producing_container = container;
			c_data->regions[i] = region;

		}
	}

	RETURN(0);
}

/* aix_discover_regions
 *
 *	For each container, create a freespace region, create all the
 *	necessary real regions, and set up the LP maps.
 */
int aix_discover_regions( void )
{
	storage_container_t     * container;
	int                     rc;

	LOG_ENTRY;

	DLIST_FOR_EACH(container, aix_container_list, rc) {

		LOG_DETAILS("Searching for regions in container %s\n", container->name);

		// Create all real regions
		aix_discover_regions_in_container(container);

		// Build the LP maps
		aix_build_lp_maps(container, NULL, NULL, FALSE);
	}

	RETURN(0);
}


/* aix_export_regions
 *
 *	Go through all existing containers and export all regions.
 *      When a region is successfully exported to the evms engine
 *      we will update device mapper status in the region storage
 *      object.
 */
int aix_export_regions( dlist_t regions )
{
	storage_container_t     * container;
	aix_container_data_t    * c_data;
	storage_object_t        * region;
	aix_region_data_t       * r_data;
	unsigned int            i, count = 0;
	int                     rc;

	LOG_ENTRY;

	DLIST_FOR_EACH(container, aix_container_list, rc) {

		c_data = container->private_data;

		for ( i = 0; i < AIX_MAX_LVS; i++ ) {
			if ( c_data->regions[i] ) {
				region = c_data->regions[i];
				r_data = region->private_data;
				if ( ! (r_data->flags & AIX_REGION_EXPORTED) ) {
					aix_add_object_to_list(region, regions, AppendToList);
					r_data->flags |= AIX_REGION_EXPORTED;
					aix_update_LV_devmap_info(region);
					count++;
				}
			}
		}
	}



	RETURN(count);
}


/* aix_allocate_partitions_to_region
 */
int aix_allocate_partitions_to_region(  storage_object_t        * region,
					storage_container_t     * container,
					aix_region_options_t    * region_opts,
					unsigned short            operation)
{
	aix_container_data_t    * c_data = container->private_data;
	aix_region_data_t       * r_data = region->private_data;
	pv_header               * pv;
	pp_entries              * pp_map;
	unsigned long           lps_allocated = 0;
	unsigned int            i, j;

	LOG_ENTRY;


	// For now, this implements a simple, linear, start-from-the-front PP allocation
	// scheme. This will change soon to support other allocation types.

	// Check every active PV in this container.
	for ( i = 0; i < AIX_MAX_PVS && lps_allocated < region_opts->num_lps; i++ ) {
		if ( c_data->vgda->pv_headers[i] ) {
			pv = c_data->vgda->pv_headers[i];
			pp_map = c_data->vgda->pp_array[i];

			LOG_DEBUG("i:%d pv:%d pp_map:%d \n", i, pv->pv_num, pp_map->lv_index);

			// Check every PP entry in this PV.
			for ( j = 0; j < pv->pp_count && lps_allocated < region_opts->num_lps; j++ ) {

				LOG_DEBUG("region:[%s] pp_map[%d].state:%d\n", region->name,j,pp_map[j].pp_state);

				if ( pp_map[j].pp_state == LVM_PPFREE ) {
					LOG_DEBUG("Adding partition %d to region [%s] state:%d\n",j,region->name,pp_map[j].pp_state);
					memset(&(pp_map[j]), 0, sizeof(pp_entries));
					pp_map[j].lv_index      = r_data->lv->lvname +1;
					pp_map[j].lp_num        = lps_allocated +1;
					pp_map[j].copy          = LVM_PRIMARY;
					pp_map[j].pp_state      = LVM_PPALLOC;
					lps_allocated++;
				}
			}
		}
	}


	region->flags |= SCFLAG_DIRTY;

	if ( lps_allocated != region_opts->num_lps) {
		// This should never happen, as long as we check ahead of time
		// for sufficient space for the allocation scheme. Need to
		// clean up all entries that have been marked so far.
		MESSAGE("Not enough partitions available to allocate to the new region!\n");
		LOG_DEBUG("Need %d PPs.  Have only %d available.\n", region_opts->num_lps, lps_allocated);
		RETURN(EINVAL);
	}

	RETURN(0);
}


/* aix_deallocate_partitions_from_region
 */
int aix_deallocate_partitions_from_region( storage_object_t * region, int count )
{
	aix_region_data_t   * r_data = region->private_data;
	aix_container_data_t * c_data = region->producing_container->private_data;
	lv_entries      * lv = r_data->lv;
	aix_lp_map_entry_t  * lp_map;
	aix_pv_data_t       * pv;
	storage_object_t    * object;
	unsigned long       pp_number;
	unsigned int        j, pe_size;


	LOG_ENTRY;

	lp_map = r_data->lp_map;

	if ( lp_map ) {
		
		// Walk through each LP in this map.
		for ( j = 0; j < r_data->lv->num_lps; j++ ) {

			if ( lp_map[j].object ) {

				object    = lp_map[j].object;
				pp_number = lp_map[j].pp_number;

				pv = object->consuming_private_data;

				// Make sure this PV/PP belongs to this
				// region, and zero the PP entry.
				if ( pv->pp_map[pp_number].lv_index-1 == lv->lvname ) {

					if (pv->pp_map[pp_number].lp_num-1 >= count) {
						memset(&(pv->pp_map[pp_number]), 0, sizeof(pp_entries));
					}

				} else {
					LOG_SERIOUS("LP/PP map inconsistency detected!\n");
					LOG_SERIOUS("Region %s (%d), Mirror ?, LP %d maps to Object %s, PP %d.\n",
						    region->name, lv->lvname, j, object->name, pp_number );
					LOG_SERIOUS("Object %s, PP %d maps to Region %d, Copy %d, LP %d.\n",
						    object->name, pp_number, pv->pp_map[pp_number].lv_index-1,
						    pv->pp_map[pp_number].copy, pv->pp_map[pp_number].lp_num-1);
				}
			}
		}

		r_data->lv->num_lps = count;
		pe_size = (1 << c_data->vgda->vg_head->pp_size);
		region->size = ((r_data->lv->num_lps * pe_size) / AIX_SECTOR_SIZE) / r_data->lv->mirror;
	}

	RETURN(0);
}

/* Function: aix_check_lv_size
 *
 *	Make sure that lv_size is a multiple of the group's PE size. If it
 *	isn't, round up to the next PE size multiple.
 */
int aix_check_lv_size(  u_int32_t       * lv_size,
			u_int32_t       pe_size )
{
	int rc = 0;

	LOG_ENTRY;

	if ( *lv_size % pe_size ) {
		LOG_WARNING("LV Size (%ld) is not a multiple of the PE size %ld\n", *lv_size, pe_size);
		*lv_size = round_up(*lv_size, pe_size);
		LOG_WARNING("Rounding LV Size up to %ld\n", *lv_size);
		rc = -1;
	}

	RETURN(rc);
}

/* Function: aix_compare_lv_size_and_extents
 *
 *	Verify that the specified lv_size and extents match. If one is not
 *	set, calculate it from the other. If neither are set, return error.
 *
 *	The aix_check_lv_size function should be called before this to
 *	make sure lv_size is a multiple of the group's pe_size.
 */
int aix_compare_lv_size_and_extents(    u_int32_t       * lv_size,
					u_int32_t       * extents,
					u_int32_t       pe_size )
{
	int rc = 0;

	LOG_ENTRY;

	if ( *lv_size ) {
		if ( *extents ) {
			if ( *lv_size != *extents * pe_size ) {
				LOG_ERROR("Mismatch in LV Size and Extents\n");
				LOG_ERROR("LV Size: %ld\n", *lv_size);
				LOG_ERROR("Extents: %ld (total size: %ld)\n", *extents, *extents * pe_size);
				LOG_ERROR("Please specify only LV Size or only Extents\n");
				rc = EINVAL;
			}
		}
		else {
			*extents = *lv_size / pe_size;
		}
	}
	else if ( *extents ) {
		*lv_size = *extents * pe_size;
	}
	else {
		LOG_ERROR("Must specify either LV Size or Extents\n");
		rc = EINVAL;
	}

	RETURN(rc);
}

/* aix_expand_allocate_partitions_to_region
 */
int aix_expand_allocate_partitions_to_region(   storage_object_t        * region,
						storage_container_t     * container,
						aix_region_options_t    * region_opts,
						unsigned short            operation)
{
	aix_container_data_t    * c_data = container->private_data;
	aix_region_data_t       * r_data = region->private_data;
	pv_header               * pv;
	pp_entries              * pp_map;
	unsigned long           lps_allocated = r_data->lv->num_lps;
	unsigned long       original_lps   = r_data->lv->num_lps;
	unsigned long       pe_size = 0;
	unsigned int            i, j, k;

	LOG_ENTRY;


	aix_reallocate_lp_map(region, lps_allocated + region_opts->num_lps);

	// Check every active PV in this container.
	for ( i = 0; i < AIX_MAX_PVS && lps_allocated < region_opts->num_lps + original_lps; i++ ) {
		if ( c_data->vgda->pv_headers[i] ) {
			pv = c_data->vgda->pv_headers[i];
			pp_map = c_data->vgda->pp_array[i];

			LOG_DEBUG("i:%d pv:%d pp_map:%d \n", i, pv->pv_num, pp_map->lv_index);

			// Check every PP entry in this PV.
			for ( j = 0; j < pv->pp_count && lps_allocated < region_opts->num_lps + original_lps; j++ ) {
				if ( pp_map[j].pp_state == LVM_PPFREE ) {
					LOG_DEBUG("Adding partition %d to region [%s] \n",j,region->name);
					memset(&(pp_map[j]), 0, sizeof(pp_entries));
					pp_map[j].lv_index      = r_data->lv->lvname +1;
					pp_map[j].lp_num        = lps_allocated +1;
					pp_map[j].copy          = LVM_PRIMARY;
					pp_map[j].pp_state      = LVM_PPALLOC;
					lps_allocated++;
				}
			}
			if (operation == AIX_EXPAND_REGION) {

				for (k = 0; k < AIX_MAX_LVS; k++) {
					if (c_data->vgda->lv_array[k].lvname == r_data->lv->lvname) {
						c_data->vgda->lv_array[k].num_lps += region_opts->num_lps;
						break;
					}
				}

				r_data->lv->num_lps = lps_allocated;
				pe_size = (1 << c_data->vgda->vg_head->pp_size);
				region->size = ((r_data->lv->num_lps * pe_size) / AIX_SECTOR_SIZE) / r_data->lv->mirror;
				region->flags |= SCFLAG_DIRTY;
			}
		}
	}

	if ( lps_allocated != r_data->lv->num_lps) {
		// This should never happen, as long as we check ahead of time
		// for sufficient space for the allocation scheme. Need to
		// clean up all entries that have been marked so far.
		MESSAGE("Not enough partitions available to allocate to the new region! %d\n",__LINE__);
		RETURN(EINVAL);
	}

	RETURN(0);
}

/* aix_check_physical_layout
 */

int aix_check_physical_layout(storage_object_t  * object,
			      int               * AllOnOnePV)

{

	aix_region_data_t  * r_data = object->private_data;
	int  i;

	LOG_ENTRY;

	*(AllOnOnePV) = TRUE;

	// We'll start with the 2nd LP mapping, if there's only one LP then it has to be contained on one PV

	for (i = 1; i < AIX_MAX_LVS && i < r_data->lv->num_lps; i++) {
		if (r_data->lp_map) {
			if (r_data->lp_map[i].object != r_data->lp_map[i-1].object){
				//LOG_DEBUG("Comparision found mismatch object:%s vs object-1:%s \n", r_data->lp_map[0][i].object->name, r_data->lp_map[0][i-1].object->name);
				*(AllOnOnePV) = FALSE;
			}
		}
	}

	RETURN(0);
}

/* aix_add_mirrors_to_object
 */

int aix_add_mirrors_to_object(storage_object_t     * object,
			      storage_object_t     * mirror_object,
			      storage_object_t     * mirror_object2,
			      int                    AllOnOnePV,
			      storage_container_t * container)

{

	aix_region_data_t     * r_data = object->private_data;
	aix_container_data_t  * c_data = container->private_data;
	storage_object_t      * tmp_object;
	int answer = NO_MIRROR;
	int rc, num_lps;
	char *choice_text[3] = {"No", "Continue", NULL};
	char *mirror_msg = "You are attempting to create a mirrored object in a Container on the same physical volume or partition as the original.\
                        A disk failure will result in the loss of ALL copies of the data!";

	LOG_ENTRY;

	LOG_DEBUG("AllOnOnePv:%d numpvs:%d mir1 name:%s region_object->name:%s \n", AllOnOnePV, c_data->vgda->vg_head->numpvs,
		  mirror_object->name,
		  r_data->lp_map[0].object->name);


	// Only query if the PV > 1. If it's one, the user has been asked whether to continue previously.

	if (AllOnOnePV && c_data->vgda->vg_head->numpvs > 1 && (mirror_object == r_data->lp_map[0].object || mirror_object2 == r_data->lp_map[0].object)) {
		aix_engine->user_message(aix_plugin, &answer, choice_text, mirror_msg);
	} else {
		answer = MIRROR_CONTINUE;
	}

	if (answer != MIRROR_CONTINUE) {
		return(EINVAL);
	}

	if (r_data->lv->mirror < AIX_MAX_MIRRORS) {
		num_lps = r_data->lv->num_lps / r_data->lv->mirror;
		r_data->lv->mirror += LVM_ADDMIRROR;
		tmp_object = mirror_object;

		rc = aix_allocate_mirror_partitions_to_region(object, tmp_object, container, num_lps, AllOnOnePV);
	} else {
		return(EINVAL);
	}


	if (mirror_object2 && r_data->lv->mirror < AIX_MAX_MIRRORS) {
		r_data->lv->mirror += LVM_ADDMIRROR;
		tmp_object = mirror_object2;

		rc = aix_allocate_mirror_partitions_to_region(object, tmp_object, container, num_lps, AllOnOnePV);
	}


	if (rc) {
		MESSAGE("Error allocating new partitions to the region.\n");
		RETURN(rc);
	}

	// Update the freespace region and rebuild the LP maps.
	rc = aix_update_freespace_region(container);
	if (rc) {

		MESSAGE("Error updating freespace for container %s.\n", container->name);
		RETURN(rc);
	}

	aix_build_lp_maps(container, mirror_object, mirror_object2, TRUE);

	aix_connect_mapped_lvs(container, object, mirror_object, mirror_object2, AllOnOnePV);

	AIX_dump_expand_region(object,__LINE__);



	RETURN(0);

}

/* aix_allocate_mirror_partitions_to_region
 */
int aix_allocate_mirror_partitions_to_region( storage_object_t      * region,
					      storage_object_t      * mirror_object,
					      storage_container_t   * container,
					      int                     num_lps,
					      int                     AllOnOnePV)

{
	aix_container_data_t   * c_data = container->private_data;
	aix_region_data_t      * r_data = region->private_data;
	storage_object_t       * object;
	pv_header       * pv;
	pp_entries      * pp_map;
	unsigned long       lps_allocated = r_data->lv->num_lps;
	unsigned long       original_lps   = r_data->lv->num_lps;
	unsigned long       pe_size = 0;
	unsigned int        i, j, k;

	LOG_ENTRY;


	aix_reallocate_lp_map(region, lps_allocated * r_data->lv->mirror);

	// Check every active PV in this container.
	for ( i = 0; i < AIX_MAX_PVS && lps_allocated < num_lps + original_lps; i++ ) {
		if ( c_data->vgda->pv_headers[i] ) {
			pv = c_data->vgda->pv_headers[i];
			pp_map = c_data->vgda->pp_array[i];
			object = aix_find_pv_by_number(container, pv->pv_num);

			if (AllOnOnePV && object != mirror_object) {
				continue;
			}

			LOG_DEBUG("i:%d pv:%d pp_map:%d \n", i, pv->pv_num, pp_map->lv_index);

			// Check every PP entry in this PV.
			for ( j = 0; j < pv->pp_count && lps_allocated < num_lps + original_lps; j++ ) {
				if ( pp_map[j].pp_state == LVM_PPFREE ) {
					LOG_DEBUG("Adding partition %d to region [%s] \n",j,region->name);
					memset(&(pp_map[j]), 0, sizeof(pp_entries));
					pp_map[j].lv_index  = r_data->lv->lvname +1;
					pp_map[j].lp_num    = lps_allocated +1;

					switch (r_data->lv->mirror) {
					case AIX_DEFAULT_MIRRORING:
						pp_map[j].copy = LVM_PRIMARY;
						break;
					case AIX_FIRST_MIRROR:
						pp_map[j].copy = LVM_SCNDOF2;
						break;

					case AIX_MAX_MIRRORS:
						pp_map[j].copy = LVM_TERTOF3;
						break;
					}

					pp_map[j].pp_state  = LVM_PPALLOC;
					lps_allocated++;
				}
			}

			for (k = 0; k < AIX_MAX_LVS; k++) {
				if (c_data->vgda->lv_array[k].lvname == r_data->lv->lvname) {
					c_data->vgda->lv_array[k].num_lps += num_lps;
					break;
				}
			}

			r_data->lv->num_lps = lps_allocated;
			pe_size = (1 << c_data->vgda->vg_head->pp_size);
			region->size = ((r_data->lv->num_lps * pe_size) / AIX_SECTOR_SIZE) / r_data->lv->mirror;
			region->flags |= SCFLAG_DIRTY;
		}
	}

	if ( lps_allocated != r_data->lv->num_lps) {
		// This should never happen, as long as we check ahead of time
		// for sufficient space for the allocation scheme. Need to
		// clean up all entries that have been marked so far.
		MESSAGE("Not enough partitions available to allocate to the new region! %d\n",__LINE__);
		RETURN(EINVAL);
	}

	RETURN(0);
}
/* aix_build_mirror_lp_maps
 *
 *	For each region in this container, set up the logical-to-physical
 *	remapping table.
 */
int aix_build_lp_maps( storage_container_t * container, storage_object_t * mirror_object,storage_object_t * mirror_object2, int expand)
{
	aix_container_data_t    * c_data = container->private_data;
	aix_vgda_t              * vgda = c_data->vgda;
	storage_object_t        * object=NULL;
	storage_object_t        * region=NULL;
	aix_region_data_t       * r_data=NULL;
	pp_entries              * pp_map=NULL;
	pv_header               * pv_head=NULL;
	unsigned int            lv_index = 0;
	unsigned int            lp_num = 0;
	unsigned int            mirror_copy = 0;
	unsigned int            current_freespace_lp = 0;
	unsigned int            i, j, k, rc;

	LOG_ENTRY;

	// Search every PVs PP map
	for ( i = 0; i < AIX_MAX_PVS; i++ ) {
		if ( ! vgda->pv_headers[i] ||
		     ! vgda->pp_array[i] ) {
			continue;
		}

		if (mirror_object) {
			LOG_DEBUG("mirror_object:%s \n", mirror_object->name);
		}

		pv_head = vgda->pv_headers[i];
		pp_map = vgda->pp_array[i];
		object = aix_find_pv_by_number(container, pv_head->pv_num);


		LOG_DEBUG("find_pv returned object:%s for pv_NUM:%d \n", object->name, pv_head->pv_num);

		// Check every PP entry on this PV.
		for ( j = 0; j < pv_head->pp_count; j++ ) {
			if ( pp_map[j].pp_state ) {
				// Valid PP. Add it to the correct LV's map.
				lv_index    = pp_map[j].lv_index -1;
				lp_num      = pp_map[j].lp_num -1;
				region      = c_data->regions[lv_index];
				if ( region ) {
					r_data = region->private_data;
				}

				// Determine the value for mirror_copy
				switch (pp_map[j].copy) {
				case LVM_PRIMARY:
				case LVM_PRIMOF2:
				case LVM_PRIMOF3:
					mirror_copy = 0;
					LOG_DEBUG("Mirror copy 1\n");

					if (!mirror_object && r_data->lv->mirror >= AIX_FIRST_MIRROR) {
						mirror_object = aix_find_pv_by_number(container, pp_map[j].fst_alt_vol);
					}

					if (!mirror_object2 && r_data->lv->mirror == AIX_MAX_MIRRORS) {
						mirror_object2 = aix_find_pv_by_number(container, pp_map[j].snd_alt_vol);
					}
					break;
				case LVM_SCNDOF2:
				case LVM_SCNDOF3:
					mirror_copy = 1;
					LOG_DEBUG("Mirror copy 2\n");
					if (!mirror_object) {
						mirror_object = object;
					}
					break;
				case LVM_TERTOF3:
					mirror_copy = 2;
					LOG_DEBUG("Mirror copy 3\n");
					if (!mirror_object2) {
						mirror_object2 = object;
					}
					break;
				default:
					LOG_SERIOUS("ERROR: Invalid value for PP map copy field.\n");
					LOG_SERIOUS("       PV %d : PP entry %d : copy = %d\n",
						    pv_head->pv_num, j, pp_map[j].copy);
					continue;
				}


				if ( region ) {
					r_data = region->private_data;
					r_data->lp_map[lp_num].pp_number = j;

					if ( object ) {

						LOG_DEBUG("region:%s lp_num:%d object:%s mirror_copy:%d \n", region->name, lp_num, object->name, mirror_copy);


						switch (mirror_copy) {
						case 0:
							r_data->lp_map[lp_num].object = object;
							rc = aix_append_region_to_object(region, object, expand);
							LOG_DEBUG("Case:0 aix_append rc:%d \n", rc);
							break;
						case 1:
							r_data->lp_map[lp_num].object = mirror_object;
							rc = aix_append_region_to_object(region, mirror_object, expand);
							LOG_DEBUG("Case:1 aix_append rc:%d \n", rc);
							break;
						case 2:
							r_data->lp_map[lp_num].object = mirror_object2;
							rc = aix_append_region_to_object(region, mirror_object2, expand);
							LOG_DEBUG("Case:2 aix_append rc:%d \n", rc);
							break;
						default:
							r_data->lp_map[lp_num].object = object;
							rc = aix_append_region_to_object(region, object, expand);
							LOG_DEBUG("Case:Default aix_append rc:%d \n", rc);
							break;
						}


					}
					else {
						r_data->lp_map[lp_num].flags |= AIX_LP_MISSING;
					}
				} else {
					LOG_SERIOUS("Possible metadata inconsistency detected!\n");
					LOG_SERIOUS("PP %d on PV %s belongs to lv_index %d, lp_num %d, copy %d\n",
						    j, object->name, lv_index, lp_num, mirror_copy);
					LOG_SERIOUS("but no region exists with lv_index %d.\n", lv_index);
					// Should we delete this PP entry?
				}

			} else {
				// Unassigned PP. Add it to the freespace.
				r_data = c_data->freespace->private_data;

				if (! r_data->lp_map) {
					// No freespace allocated.  We thought
					// all PPs were in use, but this one got
					// hosed somehow.  Now we need freespace.
					// Assume the worst case: all remaining
					// PPs are bad and go to freespace.
					r_data->lv->num_lps = pv_head->pp_count - j;
					for (k = i+1; k < AIX_MAX_PVS; k++ ) {
						if ( ! vgda->pv_headers[k] ||
						     ! vgda->pp_array[k] ) {
							continue;
						}
						r_data->lv->num_lps += vgda->pv_headers[k]->pp_count;
					}
					aix_allocate_lp_map(c_data->freespace);
				} else {
					if (r_data->lv->num_lps <= current_freespace_lp) {
						aix_reallocate_lp_map(c_data->freespace,
								      r_data->lv->num_lps + pv_head->pp_count - j);
					}
				}
				if (r_data->lp_map) {
					r_data->lp_map[current_freespace_lp].pp_number = j;
					if ( object ) {
						r_data->lp_map[current_freespace_lp].object = object;
						aix_append_region_to_object(c_data->freespace, object, expand);
					}
					else {
						r_data->lp_map[current_freespace_lp].flags |= AIX_LP_MISSING;
					}
					current_freespace_lp++;
				} else {
					LOG_SERIOUS("Error allocating an lp_map for the free space region.\n");
				}
			}
		}
	}



	RETURN(0);
}
int aix_connect_mapped_lvs(storage_container_t * container,
			   storage_object_t    * region,
			   storage_object_t    * mirror_object,
			   storage_object_t    * mirror_object2,
			   int               AllOnOnePV)
{
	aix_container_data_t * c_data = container->private_data;
	aix_vgda_t           * vgda   = c_data->vgda;
	aix_region_data_t    * r_data = region->private_data;
	pp_entries           * pp_map=NULL;
	pp_entries           * pp_map_mirror=NULL;
	pp_entries           * pp_map_mirror2=NULL;
	pv_header            * pv_head=NULL;
	aix_pv_data_t        * mir_pv_data = NULL;
	aix_pv_data_t        * mir_pv_data2 = NULL;
	storage_object_t     * mirror_object_tmp = NULL;
	int                 h = 0, i, j, k = 0;
	int found_free = FALSE;

	LOG_ENTRY;

	if (r_data->lv->mirror < 2) {
		RETURN(0);
	}

	mir_pv_data  = mirror_object->consuming_private_data;

	if (mirror_object2) {
		mir_pv_data2 = mirror_object2->consuming_private_data;
	}
	else if (!mirror_object2 && r_data->lv->mirror == AIX_MAX_MIRRORS) { // We're adding a mirror and need to get the other object
		mirror_object_tmp = aix_get_mirror_object(c_data, r_data);
		mir_pv_data  = mirror_object_tmp->consuming_private_data;
		mirror_object2 = mirror_object;
		mirror_object = mirror_object_tmp;
		mir_pv_data  = mirror_object->consuming_private_data;
		mir_pv_data2 = mirror_object2->consuming_private_data;
	}

	for (i = 0; i < AIX_MAX_PVS; i++) {
		if (vgda->pv_headers[i] && vgda->pp_array[i]) {

			pv_head = vgda->pv_headers[i];
			pp_map = vgda->pp_array[i];
			pp_map_mirror = mir_pv_data->pp_map;
			if (mirror_object2) {
				pp_map_mirror2 = mir_pv_data2->pp_map;
			}


			for (j = 0, k = 0, h = 0; j < pv_head->pp_count; j++) {
				if (pp_map[j].lv_index-1 == r_data->lv->lvname){

					found_free = FALSE;

					while (!found_free  && k < mir_pv_data->pv_head->pp_count) {
						if ((pp_map_mirror[k].lv_index-1 == r_data->lv->lvname) &&
						    (!AllOnOnePV || (r_data->lp_map[pp_map[j].lp_num-1].object != mirror_object))) {

							pp_map[j].fst_alt_part = r_data->lp_map[pp_map_mirror[k].lp_num-1].pp_number;
							pp_map[j].fst_alt_vol  = mir_pv_data->pv_head->pv_num;

							pp_map_mirror[k].fst_alt_part = r_data->lp_map[pp_map[j].lp_num-1].pp_number;
							pp_map_mirror[k].fst_alt_vol  = pv_head->pv_num;

							found_free = TRUE;

							if (r_data->lv->mirror == AIX_MAX_MIRRORS) {

								found_free = FALSE;

								while (!found_free  && h < mir_pv_data2->pv_head->pp_count) {
									if ((pp_map_mirror2[h].lv_index-1 == r_data->lv->lvname) &&
									    (!AllOnOnePV || (r_data->lp_map[pp_map[j].lp_num-1].object != mirror_object2))) {

										pp_map[j].snd_alt_part = r_data->lp_map[pp_map_mirror2[h].lp_num-1].pp_number;
										pp_map[j].snd_alt_vol  = mir_pv_data2->pv_head->pv_num;

										pp_map_mirror[k].snd_alt_part = r_data->lp_map[pp_map_mirror2[h].lp_num-1].pp_number;
										pp_map_mirror[k].snd_alt_vol  = mir_pv_data2->pv_head->pv_num;

										pp_map_mirror2[h].fst_alt_part = r_data->lp_map[pp_map[j].lp_num-1].pp_number;
										pp_map_mirror2[h].fst_alt_vol  = pv_head->pv_num;

										pp_map_mirror2[h].snd_alt_part = r_data->lp_map[pp_map_mirror[k].lp_num-1].pp_number;
										pp_map_mirror2[h].snd_alt_vol  = mir_pv_data->pv_head->pv_num;

										found_free = TRUE;

									}
									h++;
								}
							}

							LOG_DEBUG("object:%s lp_map.pp_number:%d\n",r_data->lp_map[pp_map[j].lp_num-1].object->name,
								  r_data->lp_map[pp_map[j].lp_num-1].pp_number);
							LOG_DEBUG("lp_num:%d pp_number:%d lv_index:%d pp_state:%d fap:%d fav:%d sap:%d sav:%d  \n", pp_map[j].lp_num-1,j,
								  pp_map[j].lv_index,
								  pp_map[j].pp_state,
								  pp_map[j].fst_alt_part,
								  pp_map[j].fst_alt_vol,
								  pp_map[j].snd_alt_part,
								  pp_map[j].snd_alt_vol);
						}
						k++;
					}

				}
			}
		}
	}

	RETURN(0);

}
int aix_get_free_pps_for_object(storage_object_t  *object,
				aix_region_data_t * r_data)
{
	int num_free_pps = 0;
	int i;
	aix_pv_data_t     * pv_data = object->consuming_private_data;
	pp_entries        * pp_map  = pv_data->pp_map;

	LOG_ENTRY;
	if (pv_data->pv_head) {
		for (i = 0; i < pv_data->pv_head->pp_count; i++) {
			if ((pp_map[i].lv_index == 0) &&
			    (pp_map[i].pp_state == LVM_PPFREE)) {
				num_free_pps++;
			}
		}
	}

	LOG_DEBUG("Found %d free pps for object:%s \n", num_free_pps, object->name);


	RETURN(num_free_pps);
}
storage_object_t * aix_get_mirror_object(aix_container_data_t *c_data, aix_region_data_t * r_data)
{
	aix_vgda_t          * vgda   = c_data->vgda;
	pv_header           * pv_head;
	pp_entries          * pp_map;
	int i,j;

	LOG_ENTRY;


	for (i = 0; i < AIX_MAX_PVS; i++) {
		pv_head = vgda->pv_headers[i];
		pp_map = vgda->pp_array[i];

		for (j = 0; j < pv_head->pp_count; j++) {
			if (pp_map[j].lv_index-1 == r_data->lv->lvname) {
				if (pp_map[j].copy == LVM_SCNDOF2 || pp_map[j].copy == LVM_SCNDOF3) {
					return(r_data->lp_map[pp_map[j].lp_num-1].object);
				}
			}

		}
	}


	return(NULL);
}


int  aix_split_mirror_region(storage_object_t * object, option_array_t * options)
{

	aix_container_data_t     * c_data = object->producing_container->private_data;
	storage_container_t * container = object->producing_container;
	aix_region_data_t     * r_data = object->private_data;
	aix_region_options_t    region_opts;
	lv_entries            * lv;
	storage_object_t      * new_region;
//	dlist_t                 new_region_list;    DLM ... no idea what this is doing here
	int                     rc;



	LOG_ENTRY;

	aix_dump_region(object,__LINE__);

	memset(&region_opts, 0, sizeof(aix_region_options_t));
	region_opts.num_lps = r_data->lv->num_lps / r_data->lv->mirror;

	rc = aix_create_region_parse_options(options, &region_opts);
	if (rc) {
		MESSAGE("Error parsing and verifying options.\n");
		MESSAGE("Unable to create new region in container %s.\n", container->name);
		RETURN(rc);
	}

	lv = aix_initialize_new_lv_entry(container, &region_opts);
	if ( ! lv ) {
		MESSAGE("Error initializing the LV metadata.\n");
		MESSAGE("Unable to create new region in container %s.\n", container->name);
		RETURN(rc);
	}

	// Allocate and initialize the new region.
	new_region = aix_allocate_region(lv, region_opts.name, c_data->vgda->vg_head->pp_size);
	if ( ! new_region ) {
		aix_clear_lv_entry(lv);
		MESSAGE("Error allocating new region.\n");
		MESSAGE("Unable to create new region in container %s.\n", container->name);
		RETURN(ENOMEM);
	}

	rc = aix_split_partitions_on_region(object, container, &region_opts, new_region);


	// Add the new region to the container and update all info.
	aix_add_object_to_list(new_region, container->objects_produced, AppendToList);
	new_region->producing_container = container;
	c_data->regions[lv->lvname] = new_region;
	c_data->vgda->vg_head->numlvs++;

	// Update the freespace region and rebuild the LP maps.
	rc = aix_update_freespace_region(container);
	if (rc) {
// Ack!!! We've already allocated PPs at this point, and
// added the new region to the container. Lots to fix up!
		MESSAGE("Error updating freespace for container %s.\n", container->name);
		RETURN(rc);
	}

	//aix_build_lp_maps(container, NULL, NULL, FALSE);

	// Update the container timestamps, mark it dirty,
	// and return the new region.
	aix_update_container_timestamp(container);
	container->flags |= SCFLAG_DIRTY;
	new_region->flags |= SCFLAG_DIRTY;

// I have no idea why the following line is needed when the new_region_list is not
// even a created dlist!!  DLM
//	aix_add_object_to_list(new_region, new_region_list, AppendToList);

	aix_dump_region(new_region,__LINE__);



	RETURN(0);

}
/* aix_split_partitions_on_region
 */
int aix_split_partitions_on_region(     storage_object_t        * region,
					storage_container_t     * container,
					aix_region_options_t    * region_opts,
					storage_object_t    * new_region)
{
	aix_container_data_t    * c_data = container->private_data;
	aix_region_data_t       * r_data = region->private_data;
	aix_region_data_t       * r_data_new = new_region->private_data;
	aix_pv_data_t       * pv_data;
	pv_header               * pv;
	pp_entries              * pp_map, * new_pp_map;
	unsigned long           lps_allocated = 0;
	unsigned int            i, j, mirror_copy;

	LOG_ENTRY;

        // check that we have a valid mirror ...
	// 1 == no mirror ... one copy of lps
	// 2 == single mirror copy
	// 3 == 2 mirror copies
	// so we can only split a mirror when lv-mirror == 2 or 3
	if ( r_data->lv->mirror != 2 &&
	     r_data->lv->mirror != 3 ) {
		RETURN(EINVAL);
	}

// For now, this implements a simple, linear, start-from-the-front PP allocation
// scheme. This will change soon to support other allocation types.

	switch (r_data->lv->mirror) {
	case 2:
		mirror_copy = LVM_SCNDOF2;
		break;

	case 3:
		mirror_copy     = LVM_TERTOF3;
		break;
	default:
		mirror_copy     = -1;
		break;
	}

	// Check every active PV in this container.
	for ( i = 0; i < AIX_MAX_PVS && lps_allocated < region_opts->num_lps; i++ ) {
		if ( c_data->vgda->pv_headers[i] ) {
			pv = c_data->vgda->pv_headers[i];
			pp_map = c_data->vgda->pp_array[i];

			LOG_DEBUG("i:%d pv:%d pp_map:%d \n", i, pv->pv_num, pp_map->lv_index);

			// Check every PP entry in this PV.
			for ( j = 0; j < pv->pp_count && lps_allocated < region_opts->num_lps; j++ ) {

				LOG_DEBUG("region:[%s] pp_map[%d].copy:%d\n", region->name,j,pp_map[j].copy);

				if ( pp_map[j].lv_index == r_data->lv->lvname+1 && pp_map[j].copy == mirror_copy) {
					LOG_DEBUG("Moving pp:%d region [%s] to region [%s] old lv:%d new lv:%d\n",j, region->name,
						  new_region->name, pp_map[j].lv_index, r_data_new->lv->lvname+1);
					LOG_DEBUG("region object:%s pp_num:%d\n", r_data->lp_map[pp_map[j].lp_num-1].object->name,
						  r_data->lp_map[pp_map[j].lp_num-1].pp_number);

					r_data_new->lp_map[lps_allocated].pp_number = r_data->lp_map[pp_map[j].lp_num-1].pp_number;
					r_data_new->lp_map[lps_allocated].object    = r_data->lp_map[pp_map[j].lp_num-1].object;

					LOG_DEBUG("Clearing r_data mirror posn:%d \n", pp_map[j].lp_num-1);

					memset(&r_data->lp_map[pp_map[j].lp_num-1], 0, sizeof(aix_lp_map_entry_t));

					aix_append_region_to_object(new_region, r_data_new->lp_map[lps_allocated].object, FALSE);

					pv_data = r_data_new->lp_map[lps_allocated].object->consuming_private_data;
					new_pp_map = c_data->vgda->pp_array[pv_data->pv_head->pv_num-1];


					new_pp_map[r_data_new->lp_map[lps_allocated].pp_number].lv_index        = r_data_new->lv->lvname +1;
					new_pp_map[r_data_new->lp_map[lps_allocated].pp_number].lp_num          = lps_allocated +1;
					new_pp_map[r_data_new->lp_map[lps_allocated].pp_number].copy            = LVM_PRIMARY;
					new_pp_map[r_data_new->lp_map[lps_allocated].pp_number].pp_state        = LVM_PPALLOC;
					new_pp_map[r_data_new->lp_map[lps_allocated].pp_number].fst_alt_part = 0;
					new_pp_map[r_data_new->lp_map[lps_allocated].pp_number].fst_alt_vol  = 0;
					new_pp_map[r_data_new->lp_map[lps_allocated].pp_number].snd_alt_part = 0;
					new_pp_map[r_data_new->lp_map[lps_allocated].pp_number].snd_alt_vol  = 0;


					memset(&(pp_map[j]), 0, sizeof(pp_entries));

					lps_allocated++;
				}

				if ( pp_map[j].lv_index == r_data->lv->lvname+1 && pp_map[j].copy != mirror_copy) {

					if (mirror_copy == LVM_SCNDOF2) {
						pp_map[j].fst_alt_part = 0;
						pp_map[j].fst_alt_vol  = 0;

					}

					if (mirror_copy == LVM_TERTOF3) {
						pp_map[j].snd_alt_part = 0;
						pp_map[j].snd_alt_vol  = 0;

					}
				}
			}
		}
	}

	r_data->lv->num_lps = r_data->lv->num_lps - r_data_new->lv->num_lps;
	r_data->lv->mirror--;

	region->flags |= SCFLAG_DIRTY;

	if ( lps_allocated != region_opts->num_lps) {
		// This should never happen, as long as we check ahead of time
		// for sufficient space for the allocation scheme. Need to
		// clean up all entries that have been marked so far.
		MESSAGE("Not enough partitions available to allocate to the new region! %d\n",__LINE__);
		RETURN(EINVAL);
	}

	RETURN(0);
}
