/*
 *
 *   Copyright (c) International Business Machines  Corp., 2000
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: bbroptions.c
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <plugin.h>

#include "bbr_seg.h"
#include "options.h"

/*
 *  Returns count of options for specified task
 */
int BBR_GetOptionCount(task_context_t * task)
{
    int count;

    LOGENTRY();

    switch (task->action) {
	
    case EVMS_Task_Assign_Plugin:
    case EVMS_Task_Create:
    case EVMS_Task_Expand:
    case EVMS_Task_Shrink:
        count = BBR_OPTION_COUNT;
        break;

    default:
        count = 0;
        break;
    }

    LOGEXIT();
    return count;
}

/*
 * Validate the objects in the selected_objects dlist in the task context.
 * Remove from the selected objects lists any objects which are not
 * acceptable.
 *
 * For unacceptable objects, create a declined_handle_t structure with the
 * reason why it is not acceptable, and add it to the declined_objects dlist.
 * Modify the accepatble_objects dlist in the task context as necessary
 * based on the selected objects and the current settings of the options.
 *
 * Modify any option settings as necessary based on the selected objects.
 * Return the appropriate task_effect_t settings if the object list(s),
 * minimum or maximum objects selected, or option settings have changed.
 */
int BBR_SetObjects( task_context_t * context,
                    dlist_t          declined_objects,
                    task_effect_t  * effect )
{

    int rc = EINVAL;

    LOGENTRY();

    if (context) {

        switch (context->action) {

	case EVMS_Task_Assign_Plugin:
        case EVMS_Task_Create:
        case EVMS_Task_Expand:
        case EVMS_Task_Shrink:
	    rc = 0;
        default:

            LOG_ERROR("context->action is unknown or unsupported\n");
            break;
        }
    }

    LOGEXITRC();
    return rc;
}

int BBR_SetOption( task_context_t * context,
                   u_int32_t        index,
                   value_t        * value,
                   u_int32_t      * info )
{
    int rc=EINVAL;

    LOGENTRY();

    // a measure of protection ...
    if (context == NULL) {
        LOGEXITRC();
        return rc;
    }

    switch (context->action) {

    case EVMS_Task_Assign_Plugin:
    case EVMS_Task_Create:
    case EVMS_Task_Expand:
    case EVMS_Task_Shrink:
	rc = 0;
	break;
    default:
        break;
    }


    LOGEXITRC();

    return rc;
}

/*
 *  Called from get_acceptable_create_objects to prune objects we
 *  cant layer BBR on top of.
 */
static BOOLEAN prune_noncreatable_objects( ADDRESS   Object,
                                           TAG       ObjectTag,
                                           ADDRESS   ObjectHandle,
                                           ADDRESS   Parameters,
                                           BOOLEAN  *FreeMemory,
                                           uint     *Error )
{
    storage_object_t * obj = (storage_object_t *) Object;

    *FreeMemory = FALSE;           // tells dlist not to free any memory
    *Error      = DLIST_SUCCESS;   // tells dlist we were successful

    if (obj) {

        if ( obj->plugin  != BBR_PluginRecord_Ptr ) {

                return FALSE;

        }

    }

    return TRUE;           // tells dlist to prune the segment from the list
}


/*
 * Validate the objects in the source and target dlists in the context
 * record.  Remove from the selected objects lists any objects which are no
 * longer acceptable.  Return all acceptable objects in the parameter
 * dlists.  Also, for any object which is removed from the selected dlists,
 * or is otherwise not acceptable, create a declined_handle_t struct with
 * reason why not acceptable and add to the declined_objects dlist.
 */
static int get_acceptable_create_objects( task_context_t * context )
{
    int   rc = EINVAL;
    uint  count;

    LOGENTRY();

    rc = GetListSize(context->acceptable_objects, &count);
    if (rc) {
        count = 0;
    }

    if (count == 0) {


        rc = BBREngFncs->get_object_list( SEGMENT | DISK,
                                          DATA_TYPE,
                                          NULL,
                                          NULL,
                                          VALID_INPUT_OBJECT,
                                          &context->acceptable_objects );

        if (rc == DLIST_SUCCESS) {

            PruneList(context->acceptable_objects, 
		      prune_noncreatable_objects, NULL);

        }

    }
    else {
        LOG_ERROR("context already has acceptable objects\n");
    }

    LOGEXITRC();
    return rc;
}




/*
 *  Called from get_acceptable_shrink_objects
 */
static BOOLEAN prune_unshrinkable_objects( ADDRESS   Object,
                                           TAG       ObjectTag,
                                           ADDRESS   ObjectHandle,
                                           ADDRESS   Parameters,
                                           BOOLEAN  *FreeMemory,
                                           uint     *Error )
{
    int                        rc;
    storage_object_t          *child=NULL;
    dlist_t                    shrink_points=CreateList();
    struct plugin_functions_s *fncs;


    *FreeMemory = FALSE;           // tells dlist not to free any memory
    *Error      = DLIST_SUCCESS;   // tells dlist we were successful

    if ( (Object)&&
         (shrink_points)) {

        child = GET_BBR_CHILD( ((storage_object_t *)Object) );

        if ( child ) {

            fncs = 
		(struct plugin_functions_s *)child->plugin->functions.plugin;

            rc = fncs->can_shrink(child, NULL, shrink_points);
            if (rc == 0) {
                DestroyList(&shrink_points,FALSE);
                return FALSE;
            }
        }

        DestroyList(&shrink_points,FALSE);

    }

    return TRUE;           // tells dlist to prune the segment from the list
}


/*
 * Validate the objects in the source and target dlists in the context
 * record.  Remove from the selected objects lists any objects which are no
 * longer acceptable.  Return all acceptable objects in the parameter
 * dlists.  Also, for any object which is removed from the selected dlists,
 * or is otherwise not acceptable, create a declined_handle_t struct with
 * reason why not acceptable and add to the declined_objects dlist.
 */
static int get_acceptable_shrink_objects( task_context_t * context  )
{
    int   rc = EINVAL;
    uint  count;


    LOGENTRY();

    rc = GetListSize(context->acceptable_objects, &count);
    if (rc) {
        count = 0;
    }

    if (count == 0) {

        rc = BBREngFncs->get_object_list( SEGMENT,
                                          DATA_TYPE,
                                          BBR_PluginRecord_Ptr,
                                          NULL,
                                          VALID_INPUT_OBJECT,
                                          &context->acceptable_objects );

        if (rc == DLIST_SUCCESS) {

            PruneList(context->acceptable_objects, 
		      prune_unshrinkable_objects, NULL);

        }

    }
    else {
        LOG_ERROR("context already has acceptable objects\n");
    }


    LOGEXITRC();
    return rc;
}

/*
 *  Called from get_acceptable_expand_objects
 */
static BOOLEAN prune_unexpandable_objects( ADDRESS   Object,
                                           TAG       ObjectTag,
                                           ADDRESS   ObjectHandle,
                                           ADDRESS   Parameters,
                                           BOOLEAN  *FreeMemory,
                                           uint     *Error )
{
    int                        rc;
    storage_object_t          *child=NULL;
    dlist_t                    expand_points=CreateList();
    struct plugin_functions_s *fncs;


    *FreeMemory = FALSE;           // tells dlist not to free any memory
    *Error      = DLIST_SUCCESS;   // tells dlist we were successful

    if ( (Object)&&
         (expand_points)) {

        child = GET_BBR_CHILD( ((storage_object_t *)Object) );

        if ( child ) {

            fncs = 
		(struct plugin_functions_s *)child->plugin->functions.plugin;

            rc = fncs->can_expand(child, NULL, expand_points);
            if (rc == 0) {
                DestroyList(&expand_points,FALSE);
                return FALSE;
            }
        }

        DestroyList(&expand_points,FALSE);

    }

    return TRUE;           // tells dlist to prune the segment from the list
}

/*
 * Validate the objects in the source and target dlists in the context
 * record.  Remove from the selected objects lists any objects which are no
 * longer acceptable.  Return all acceptable objects in the parameter
 * dlists.  Also, for any object which is removed from the selected dlists,
 * or is otherwise not acceptable, create a declined_handle_t struct with
 * reason why not acceptable and add to the declined_objects dlist.
 */
static int get_acceptable_expand_objects( task_context_t * context  )
{
    int   rc = EINVAL;
    uint  count;


    LOGENTRY();

    rc = GetListSize(context->acceptable_objects, &count);
    if (rc) {
        count = 0;
    }

    if (count == 0) {

        rc = BBREngFncs->get_object_list( SEGMENT,
                                          DATA_TYPE,
                                          BBR_PluginRecord_Ptr,
                                          NULL,
					  VALID_INPUT_OBJECT,
                                          &context->acceptable_objects );

        if (rc == DLIST_SUCCESS) {

            PruneList(context->acceptable_objects, 
		      prune_unexpandable_objects, NULL);

        }

    }
    else {
        LOG_ERROR("context already has acceptable objects\n");
    }


    LOGEXITRC();
    return rc;
}



/*
 * Validate the objects in the source and target dlists in the context
 * record.  Remove from the selected objects lists any objects which are no
 * longer acceptable.  Return all acceptable objects in the parameter
 * dlists.  Also, for any object which is removed from the selected dlists,
 * or is otherwise not acceptable, create a declined_handle_t struct with
 * reason why not acceptable and add to the declined_objects dlist.
 */
int BBR_InitTask( task_context_t * context )
{
    int rc = EINVAL;

    LOGENTRY();


    if (context) {

        switch (context->action) {
	    
	case EVMS_Task_Assign_Plugin:

            context->min_selected_objects = 1;
            context->max_selected_objects = 1;

	    rc = get_acceptable_create_objects( context );
	    break;

        case EVMS_Task_Create:

            context->min_selected_objects = 1;
            context->max_selected_objects = 1;

	    rc = get_acceptable_create_objects( context );
	    break;

        case EVMS_Task_Expand:

            context->min_selected_objects = 1;
            context->max_selected_objects = 1;

	    rc = get_acceptable_expand_objects( context );
            break;

        case EVMS_Task_Shrink:

            context->min_selected_objects = 1;
            context->max_selected_objects = 1;

	    rc = get_acceptable_shrink_objects( context );
	    break;
	case BBR_FUNCTION_ENABLE:
	case BBR_FUNCTION_DISABLE:
            context->min_selected_objects = 0;
            context->max_selected_objects = 0;
	    context->option_descriptors->count = 0;
	    rc = 0;
	    break;
	    
        default:
               LOG_ERROR("context->action is unknown or unsupported\n");
               break;
        }
    }


    LOGEXITRC();
    return rc;
}


/*
 * Returns information about a BBR object
 */
int BBR_GetInfo( storage_object_t  * object, char *name,  
		 extended_info_array_t * * info)
{
    int rc = EINVAL;
    extended_info_array_t   *Info;
    BBR_Private_Data        *pdata = (BBR_Private_Data *) object->private_data;


    LOGENTRY();

    // a measure of protection ...
    if ((info == NULL)||(pdata->signature != EVMS_BBR_SIGNATURE)) {
        LOGEXITRC();
        return rc;
    }

    rc    = ENOMEM;  // init to failed calloc
    *info = NULL;     // init to no info returned

    if ( object->object_type == SEGMENT ) {

        Info = BBREngFncs->engine_alloc( sizeof(extended_info_array_t) + 
					 (BBR_INFO_COUNT * 
					  sizeof(extended_info_t) ) );
        if (Info) {

            Info->count = BBR_INFO_COUNT;

            SET_STRING_FIELD( Info->info[BBR_INFO_NAME_INDEX].name, "Name" );
            SET_STRING_FIELD( Info->info[BBR_INFO_NAME_INDEX].title, "Name" );
            SET_STRING_FIELD( Info->info[BBR_INFO_NAME_INDEX].desc, 
			      "This is the unique name the system gave the segment." );
            Info->info[BBR_INFO_NAME_INDEX].type               = EVMS_Type_String;
            Info->info[BBR_INFO_NAME_INDEX].unit               = EVMS_Unit_None;
            SET_STRING_FIELD( Info->info[BBR_INFO_NAME_INDEX].value.s, object->name );
            Info->info[BBR_INFO_NAME_INDEX].collection_type    = EVMS_Collection_None;
            memset( &Info->info[BBR_INFO_NAME_INDEX].group, 0, sizeof(group_info_t));

            SET_STRING_FIELD( Info->info[BBR_INFO_UUID_INDEX].name, "Uuid" );
            SET_STRING_FIELD( Info->info[BBR_INFO_UUID_INDEX].title, "Uuid" );
            SET_STRING_FIELD( Info->info[BBR_INFO_UUID_INDEX].desc, 
			      "This is the unique id the system gave the segment." );
            Info->info[BBR_INFO_UUID_INDEX].type               = EVMS_Type_String;
            Info->info[BBR_INFO_UUID_INDEX].unit               = EVMS_Unit_None;
            SET_STRING_FIELD( Info->info[BBR_INFO_UUID_INDEX].value.s, object->uuid );
            Info->info[BBR_INFO_UUID_INDEX].collection_type    = EVMS_Collection_None;
            memset( &Info->info[BBR_INFO_UUID_INDEX].group, 0, sizeof(group_info_t));

            SET_STRING_FIELD( Info->info[BBR_INFO_SIZE_INDEX].name, "Size" );
            SET_STRING_FIELD( Info->info[BBR_INFO_SIZE_INDEX].title, "Size" );
            SET_STRING_FIELD( Info->info[BBR_INFO_SIZE_INDEX].desc, "This is the size of the storage object after reserving space for metadata.");
            Info->info[BBR_INFO_SIZE_INDEX].type               = EVMS_Type_Unsigned_Int64;
            Info->info[BBR_INFO_SIZE_INDEX].unit               = EVMS_Unit_None;
            Info->info[BBR_INFO_SIZE_INDEX].value.ui64         = object->size;
            Info->info[BBR_INFO_SIZE_INDEX].collection_type    = EVMS_Collection_None;
            memset( &Info->info[BBR_INFO_SIZE_INDEX].group, 0, sizeof(group_info_t));

            SET_STRING_FIELD( Info->info[BBR_INFO_BLKS_INDEX].name, "Reserve Blocks" );
            SET_STRING_FIELD( Info->info[BBR_INFO_BLKS_INDEX].title, "Blocks" );
            SET_STRING_FIELD( Info->info[BBR_INFO_BLKS_INDEX].desc, "This is the number of replacement blocks BBR is reserving for this storage object.");
            Info->info[BBR_INFO_BLKS_INDEX].type               = EVMS_Type_Unsigned_Int64;
            Info->info[BBR_INFO_BLKS_INDEX].unit               = EVMS_Unit_None;
            Info->info[BBR_INFO_BLKS_INDEX].value.ui64         = pdata->replacement_blocks_needed;
            Info->info[BBR_INFO_BLKS_INDEX].collection_type    = EVMS_Collection_None;
            memset( &Info->info[BBR_INFO_BLKS_INDEX].group, 0, sizeof(group_info_t));

            SET_STRING_FIELD( Info->info[BBR_INFO_BLKSZ_INDEX].name, "Block Size" );
            SET_STRING_FIELD( Info->info[BBR_INFO_BLKSZ_INDEX].title, "Block Size" );
            SET_STRING_FIELD( Info->info[BBR_INFO_BLKSZ_INDEX].desc, "This value tells you the size of a replacement block.");
            Info->info[BBR_INFO_BLKSZ_INDEX].type               = EVMS_Type_Unsigned_Int64;
            Info->info[BBR_INFO_BLKSZ_INDEX].unit               = EVMS_Unit_None;
            Info->info[BBR_INFO_BLKSZ_INDEX].value.ui64         = pdata->block_size;
            Info->info[BBR_INFO_BLKSZ_INDEX].collection_type    = EVMS_Collection_None;
            memset( &Info->info[BBR_INFO_BLKSZ_INDEX].group, 0, sizeof(group_info_t));

            SET_STRING_FIELD( Info->info[BBR_INFO_TBLSZ_INDEX].name, "Table Size" );
            SET_STRING_FIELD( Info->info[BBR_INFO_TBLSZ_INDEX].title, "Size of BBR table" );
            SET_STRING_FIELD( Info->info[BBR_INFO_TBLSZ_INDEX].desc, "This is the number of sectors being used by the BBR remapping table.");
            Info->info[BBR_INFO_TBLSZ_INDEX].type               = EVMS_Type_Unsigned_Int64;
            Info->info[BBR_INFO_TBLSZ_INDEX].unit               = EVMS_Unit_None;
            Info->info[BBR_INFO_TBLSZ_INDEX].value.ui64         = pdata->bbr_table_size_in_sectors;
            Info->info[BBR_INFO_TBLSZ_INDEX].collection_type    = EVMS_Collection_None;
            memset( &Info->info[BBR_INFO_TBLSZ_INDEX].group, 0, sizeof(group_info_t));

            *info = Info;

            rc = 0;
        }
        else {
            LOG_ERROR("unable to malloc memory for extended info array\n");
        }
    }

    LOGEXITRC();
    return rc;
}


/*
 *
 */
int BBR_GetPluginInfo( char * descriptor_name, extended_info_array_t * * info )
{
    int rc = EINVAL;
    extended_info_array_t   *Info;
    char                     version_string[64];
    char                     required_engine_api_version_string[64];
    char                     required_plugin_api_version_string[64];


    LOGENTRY();

    // a measure of protection ...
    if (info == NULL) {
        LOGEXITRC();
        return rc;
    }

    rc    = ENOMEM;  // init to failed calloc
    *info = NULL;     // init to no info returned

    Info = BBREngFncs->engine_alloc( sizeof(extended_info_array_t) + 
				     (BBR_PLUGIN_INFO_COUNT * 
				      sizeof(extended_info_t))  );
    if (Info) {

        Info->count = BBR_PLUGIN_INFO_COUNT;

        sprintf(version_string, "%d.%d.%d",
                MAJOR_VERSION,
                MINOR_VERSION,
                PATCH_LEVEL );

        sprintf(required_engine_api_version_string, "%d.%d.%d",
                BBR_PluginRecord_Ptr->required_engine_api_version.major,
                BBR_PluginRecord_Ptr->required_engine_api_version.minor,
                BBR_PluginRecord_Ptr->required_engine_api_version.patchlevel );

        sprintf(required_plugin_api_version_string, "%d.%d.%d",
                BBR_PluginRecord_Ptr->required_plugin_api_version.plugin.major,
                BBR_PluginRecord_Ptr->required_plugin_api_version.plugin.minor,
                BBR_PluginRecord_Ptr->required_plugin_api_version.plugin.patchlevel );

        SET_STRING_FIELD( Info->info[BBR_PLUGIN_INFO_SNAME_INDEX].name, "Short Name" );
        SET_STRING_FIELD( Info->info[BBR_PLUGIN_INFO_SNAME_INDEX].title, "Short Name" );
        SET_STRING_FIELD( Info->info[BBR_PLUGIN_INFO_SNAME_INDEX].desc, "A short name given to this plugin.");
        Info->info[BBR_PLUGIN_INFO_SNAME_INDEX].type               = EVMS_Type_String;
        Info->info[BBR_PLUGIN_INFO_SNAME_INDEX].unit               = EVMS_Unit_None;
        SET_STRING_FIELD( Info->info[BBR_PLUGIN_INFO_SNAME_INDEX].value.s, BBR_PluginRecord_Ptr->short_name );
        Info->info[BBR_PLUGIN_INFO_SNAME_INDEX].collection_type    = EVMS_Collection_None;
        memset( &Info->info[BBR_PLUGIN_INFO_SNAME_INDEX].group, 0, sizeof(group_info_t));

        SET_STRING_FIELD( Info->info[BBR_PLUGIN_INFO_LNAME_INDEX].name, "Long Name" );
        SET_STRING_FIELD( Info->info[BBR_PLUGIN_INFO_LNAME_INDEX].title, "Long Name" );
        SET_STRING_FIELD( Info->info[BBR_PLUGIN_INFO_LNAME_INDEX].desc, "A long name given to this plugin.");
        Info->info[BBR_PLUGIN_INFO_LNAME_INDEX].type               = EVMS_Type_String;
        Info->info[BBR_PLUGIN_INFO_LNAME_INDEX].unit               = EVMS_Unit_None;
        SET_STRING_FIELD( Info->info[BBR_PLUGIN_INFO_LNAME_INDEX].value.s, BBR_PluginRecord_Ptr->long_name );
        Info->info[BBR_PLUGIN_INFO_LNAME_INDEX].collection_type    = EVMS_Collection_None;
        memset( &Info->info[BBR_PLUGIN_INFO_LNAME_INDEX].group, 0, sizeof(group_info_t));

        SET_STRING_FIELD( Info->info[BBR_PLUGIN_INFO_TYPE_INDEX].name, "Type" );
        SET_STRING_FIELD( Info->info[BBR_PLUGIN_INFO_TYPE_INDEX].title, "Plugin Type" );
        SET_STRING_FIELD( Info->info[BBR_PLUGIN_INFO_TYPE_INDEX].desc, "There are various types of plugins; each responsible for some kind of storage object.");
        Info->info[BBR_PLUGIN_INFO_TYPE_INDEX].type               = EVMS_Type_String;
        Info->info[BBR_PLUGIN_INFO_TYPE_INDEX].unit               = EVMS_Unit_None;
        SET_STRING_FIELD( Info->info[BBR_PLUGIN_INFO_TYPE_INDEX].value.s, "Segment Manager");
        Info->info[BBR_PLUGIN_INFO_TYPE_INDEX].collection_type    = EVMS_Collection_None;
        memset( &Info->info[BBR_PLUGIN_INFO_TYPE_INDEX].group, 0, sizeof(group_info_t));

        SET_STRING_FIELD( Info->info[BBR_PLUGIN_INFO_VERSION_INDEX].name, "Version" );
        SET_STRING_FIELD( Info->info[BBR_PLUGIN_INFO_VERSION_INDEX].title, "Plugin Version" );
        SET_STRING_FIELD( Info->info[BBR_PLUGIN_INFO_VERSION_INDEX].desc, "This is the version number of the plugin.");
        Info->info[BBR_PLUGIN_INFO_VERSION_INDEX].type               = EVMS_Type_String;
        Info->info[BBR_PLUGIN_INFO_VERSION_INDEX].unit               = EVMS_Unit_None;
        SET_STRING_FIELD( Info->info[BBR_PLUGIN_INFO_VERSION_INDEX].value.s, version_string );
        Info->info[BBR_PLUGIN_INFO_VERSION_INDEX].collection_type    = EVMS_Collection_None;
        memset( &Info->info[BBR_PLUGIN_INFO_VERSION_INDEX].group, 0, sizeof(group_info_t));

        SET_STRING_FIELD( Info->info[BBR_PLUGIN_INFO_REQENGSVCVERSION_INDEX].name, "Required Engine Services Version" );
        SET_STRING_FIELD( Info->info[BBR_PLUGIN_INFO_REQENGSVCVERSION_INDEX].title, "Required Engine Services Version" );
        SET_STRING_FIELD( Info->info[BBR_PLUGIN_INFO_REQENGSVCVERSION_INDEX].desc, "This is the version of the Engine services that this plug-in requires. It will not run on older versions of the Engine services.");
        Info->info[BBR_PLUGIN_INFO_REQENGSVCVERSION_INDEX].type               = EVMS_Type_String;
        Info->info[BBR_PLUGIN_INFO_REQENGSVCVERSION_INDEX].unit               = EVMS_Unit_None;
        SET_STRING_FIELD( Info->info[BBR_PLUGIN_INFO_REQENGSVCVERSION_INDEX].value.s, required_engine_api_version_string );
        Info->info[BBR_PLUGIN_INFO_REQENGSVCVERSION_INDEX].collection_type    = EVMS_Collection_None;
        memset( &Info->info[BBR_PLUGIN_INFO_REQENGSVCVERSION_INDEX].group, 0, sizeof(group_info_t));

        SET_STRING_FIELD( Info->info[BBR_PLUGIN_INFO_REQPLUGAPIVERSION_INDEX].name, "Required Engine Plug-in API Version" );
        SET_STRING_FIELD( Info->info[BBR_PLUGIN_INFO_REQPLUGAPIVERSION_INDEX].title, "Required Engine Plug-in API Version" );
        SET_STRING_FIELD( Info->info[BBR_PLUGIN_INFO_REQPLUGAPIVERSION_INDEX].desc, "This is the version of the Engine plug-in API that this plug-in requires. It will not run on older versions of the Engine plug-in API.");
        Info->info[BBR_PLUGIN_INFO_REQPLUGAPIVERSION_INDEX].type               = EVMS_Type_String;
        Info->info[BBR_PLUGIN_INFO_REQPLUGAPIVERSION_INDEX].unit               = EVMS_Unit_None;
        SET_STRING_FIELD( Info->info[BBR_PLUGIN_INFO_REQPLUGAPIVERSION_INDEX].value.s, required_plugin_api_version_string );
        Info->info[BBR_PLUGIN_INFO_REQPLUGAPIVERSION_INDEX].collection_type    = EVMS_Collection_None;
        memset( &Info->info[BBR_PLUGIN_INFO_REQPLUGAPIVERSION_INDEX].group, 0, sizeof(group_info_t));

        *info = Info;

        rc = 0;
    }


    LOGEXITRC();
    return rc;
}
