/*
 *   Copyright (c) International Business Machines  Corp., 2000
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Local Disk Manager plugin.
 */

#define _GNU_SOURCE

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <mntent.h>
#include <dirent.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <errno.h>
#include <glob.h>

#include <plugin.h>
#include <ldm_dpc.h>
#include "localdskmgr.h"
#include "cache.h"

engine_functions_t *EngFncs = NULL;
char * base_directory = NULL;
int    base_directory_len;
char * sysfs_mount_point;

const char * scan;
const char * const * directories;
int directories_count;
const char * const * includes;
int include_count;
const char * const default_legacy_includes[] = {
	"sd?",
	"hd?",
	"dasd?"
};
const char * const default_sysfs_includes[] = {
	"*?",
};
const char * const * excludes;
int exclude_count;
glob_t dev_names_glob;
int glob_flags;
static char pattern[PATH_MAX];


BOOLEAN version_2_4 = FALSE;

/**
 * round_down_to_hard_sector
 * @size:	Count of sectors to be rounded down
 * @hs:		Hard-sector-size, in bytes
 *
 * Given a size in vsectors, round it down to be a multiple of the sector size
 * of a given storage object.
 **/
static inline sector_count_t round_down_to_hard_sector(sector_count_t size,
						       u_int32_t hs)
{
	if (hs > EVMS_VSECTOR_SIZE) {
		sector_count_t hs_in_vsectors = ((sector_count_t)hs) >> EVMS_VSECTOR_SIZE_SHIFT;
		return size & ~(hs_in_vsectors - 1);
	} else {
		return size;
	}
}

/**
 * round_up_to_hard_sector
 * @size:	Count of sectors to be rounded up
 * @hs:		Hard-sector-size, in bytes
 *
 * Given a size in vsectors, round it up to be a multiple of the sector size
 * of a given storage object.
 **/
static inline sector_count_t round_up_to_hard_sector(sector_count_t size,
						     u_int32_t hs)
{
	if (hs  > EVMS_VSECTOR_SIZE) {
		sector_count_t hs_in_vsectors = ((sector_count_t)hs) >> EVMS_VSECTOR_SIZE_SHIFT;
		sector_count_t temp_size = size + hs_in_vsectors - 1;
		return temp_size & ~(hs_in_vsectors - 1);
	} else {
		return size;
	}
}

/**
 * is_2_4
 *
 * Determine if we are running on a 2.4 kernel.
 **/
static void is_2_4(void)
{
	int fd;
	char needle[] = "version";
	char haystack[256];
	char *ver;
	int major, minor, patch;

	LOG_PROC_ENTRY();

	fd = open("/proc/version", O_RDONLY);
	if (fd < 0) {
		version_2_4 = TRUE;
		LOG_PROC_EXIT_BOOLEAN(TRUE);
		return;
	}
	read(fd, haystack, 256);
	close(fd);
	ver = strstr(haystack, needle);
	if (!ver) {
		version_2_4 = TRUE;
		LOG_PROC_EXIT_BOOLEAN(TRUE);
		return;
	}
	sscanf(ver, "%*s %d.%d.%d", &major, &minor, &patch);

	LOG_DEBUG("Kernel version is: %d.%d.%d\n", major, minor, patch);
	if (major == 2 && minor == 4) {
		version_2_4 = TRUE;
		LOG_DEBUG("All I/O will be done in at least 1K chunks\n");
	}
	LOG_PROC_EXIT_BOOLEAN(version_2_4);
}

/**
 * where_is_sysfs
 *
 * Is sysfs mounted. If so, return the mount point.
 **/
static BOOLEAN where_is_sysfs(char ** mount_name)
{
	BOOLEAN result = FALSE;
	FILE * mount_records;
	struct mntent * mount_entry;

	LOG_PROC_ENTRY();

	if (mount_name != NULL) {
		*mount_name = NULL;
	}

	mount_records = setmntent(MOUNTED, "r");
	if (mount_records == NULL) {
		mount_records = setmntent("/proc/mounts", "r");
	}
	if (mount_records == NULL) {
		LOG_ERROR("Could not get list of mounted devices.\n");
		LOG_PROC_EXIT_BOOLEAN(result);
		return result;
	}
	while (!result && (mount_entry = getmntent(mount_records)) != NULL) {
		if (strcmp(mount_entry->mnt_type, "sysfs") == 0) {
			result = TRUE;
			if (mount_name != NULL) {
				*mount_name = strdup( mount_entry->mnt_dir );
			}
		}
	}

	endmntent(mount_records);

	LOG_PROC_EXIT_BOOLEAN(result);
	return result;
}

static void get_legacy_config() {

	LOG_PROC_ENTRY();

	scan = "/dev/";
	EngFncs->get_config_string("legacy_devices.scan", &scan);

	directories_count = 0;
	directories = NULL;
	EngFncs->get_config_string_array("legacy_devices.directories",
					 &directories_count, &directories);
	include_count = 0;
	includes = NULL;
	EngFncs->get_config_string_array("legacy_devices.include",
					 &include_count, &includes);
	if (includes == NULL) {
		includes = default_legacy_includes;
		include_count = sizeof(default_legacy_includes) / sizeof(default_legacy_includes[0]);
	}

	exclude_count = 0;
	excludes = NULL;
	EngFncs->get_config_string_array("legacy_devices.exclude",
					 &exclude_count, &excludes);

	LOG_PROC_EXIT_VOID();
}

static void get_sysfs_config() {

	LOG_PROC_ENTRY();

	include_count = 0;
	includes = NULL;
	EngFncs->get_config_string_array("sysfs_devices.include",
					 &include_count, &includes);
	if (includes == NULL) {
		includes = default_sysfs_includes;
		include_count = sizeof(default_sysfs_includes) / sizeof(default_sysfs_includes[0]);
	}

	exclude_count = 0;
	excludes = NULL;
	EngFncs->get_config_string_array("sysfs_devices.exclude",
					 &exclude_count, &excludes);

	LOG_PROC_EXIT_VOID();
}

static int LD_setup(engine_functions_t * engine_function_table)
{
	/* save info we get from the engine */
	EngFncs = engine_function_table;

	LOG_PROC_ENTRY();

	if (where_is_sysfs(&sysfs_mount_point)) {
		get_sysfs_config();
	} else {
		get_legacy_config();
	}

	/* For I/O path need to know if version is 2.4. */
	is_2_4();

	LOG_PROC_EXIT_INT(0);
	return 0;
}

/**
 * close_dev
 **/
static int close_dev(void * object,
		     TAG object_tag,
		     void * object_handle,
		     void * parameters)
{
	storage_object_t * obj = object;
	int * fd = obj->private_data;

	LOG_PROC_ENTRY();

	/* Safety check. */
	if (object_tag == DISK_TAG) {
		if (fd) {
			EngFncs->close_object(obj, *fd);
			obj->private_data = NULL;
		}
	}

	LOG_PROC_EXIT_INT(DLIST_SUCCESS);
	return DLIST_SUCCESS;
}

/**
 * LD_cleanup
 *
 * Find any disks & close the device that was opended during discovery.
 * Used by both Legacy and Sysfs versions.
 **/
static void LD_cleanup(void)
{
	int rc;
	dlist_t disk_list;

	LOG_PROC_ENTRY();

	/* Get a list of disks that are managed by this plug-in. */
	rc = EngFncs->get_object_list(DISK,0, LD_PluginRecord, NULL, 0, &disk_list);
	if (!rc) {
		/* Close and dev handles that might be open. */
		ForEachItem(disk_list, close_dev, NULL, TRUE);
		DestroyList(&disk_list, FALSE);
	}

	destroy_cache();

	if (base_directory != NULL) {
		free(base_directory);
		base_directory = NULL;
	}
	if (sysfs_mount_point != NULL) {
		free(sysfs_mount_point);
		sysfs_mount_point = NULL;
	}

	LOG_PROC_EXIT_VOID();
}

static void filter_out_excludes(char * pattern, int path_len, int new_globs_index)
{
	int rc;
	int i;
	glob_t exclude_glob = {0};

	LOG_PROC_ENTRY();

	for (i = 0; i < exclude_count; i++) {
			
		strcpy(pattern + path_len, excludes[i]);

		rc = glob(pattern, glob_flags, NULL, &exclude_glob);

		if (rc == 0) {
			glob_flags |= GLOB_APPEND;

		} else {
			if (rc != GLOB_NOMATCH) {
				LOG_WARNING("glob() of pattern %s failed with error %s\n", pattern,
					    (rc == GLOB_NOSPACE) ? "GLOB_NOSPACE" :
					    (rc == GLOB_ABEND) ? "GLOB_ABEND" :
					    "(unknown)");
			}
		}
	}

	for (i = 0; i < exclude_glob.gl_pathc; i++) {
		int j;

		for (j = new_globs_index; j < dev_names_glob.gl_pathc; j++) {
			if (strcmp(exclude_glob.gl_pathv[i], dev_names_glob.gl_pathv[j]) == 0) {
				int k;

				LOG_DEBUG("Removing %s.\n", dev_names_glob.gl_pathv[j]);
				free(dev_names_glob.gl_pathv[j]);

				/* Scoot up all following entries. */
				for (k = j+1; k < dev_names_glob.gl_pathc; k++) {
					dev_names_glob.gl_pathv[k-1] = dev_names_glob.gl_pathv[k];
				}
				dev_names_glob.gl_pathc--;

				break;
			}
		}
	}

	if (exclude_glob.gl_pathc >= 0) {
		globfree(&exclude_glob);
	}

	LOG_PROC_EXIT_VOID();
}

static void filter_out_non_block_devices(int new_globs_index)
{
	int i;
	struct stat statbuf;
	int status;

	LOG_PROC_ENTRY();

	i = new_globs_index;
	while (i < dev_names_glob.gl_pathc) {
			
		status = stat(dev_names_glob.gl_pathv[i], &statbuf);

		if (status == 0) {
			if (!S_ISBLK(statbuf.st_mode)) {
				int j;

				LOG_DEBUG("Removing %s.\n", dev_names_glob.gl_pathv[i]);
				free(dev_names_glob.gl_pathv[i]);

				/* Scoot up all following entries. */
				for (j = i+1; j < dev_names_glob.gl_pathc; j++) {
					dev_names_glob.gl_pathv[j-1] = dev_names_glob.gl_pathv[j];
				}
				dev_names_glob.gl_pathc--;
				dev_names_glob.gl_pathv[dev_names_glob.gl_pathc] = NULL;

				/* Leave "i" as it is so we check the new
				 * entry at the current index.
				 */
				continue;
			}

		} else {
			LOG_WARNING("stat(%s) failed with error code %d: %s\n", dev_names_glob.gl_pathv[i], errno, strerror(errno));
		}

		i++;
	}

	LOG_PROC_EXIT_VOID();
}

static void get_dev_names(const char * dir) {

	int rc;
	int i;
	int path_len;
	int new_globs_index;

	LOG_PROC_ENTRY();
	LOG_DEBUG("Get device names in directory %s\n", dir);

	strcpy(pattern, dir);
	path_len = strlen(pattern);
	if (pattern[path_len-1] != '/') {
		pattern[path_len] = '/';
		pattern[path_len+1] = '\0';
		path_len++;
	}

	new_globs_index = dev_names_glob.gl_pathc;

	for (i = 0; i < include_count; i++) {
			
		strcpy(pattern + path_len, includes[i]);

		rc = glob(pattern, glob_flags, NULL, &dev_names_glob);

		if (rc == 0) {
			glob_flags |= GLOB_APPEND;

		} else {
			if (rc != GLOB_NOMATCH) {
				LOG_WARNING("glob() of pattern %s failed with error %s\n", pattern,
					    (rc == GLOB_NOSPACE) ? "GLOB_NOSPACE" :
					    (rc == GLOB_ABEND) ? "GLOB_ABEND" :
					    "(unknown)");
			}
		}
	}

	filter_out_excludes(pattern, path_len, new_globs_index);

	if (sysfs_mount_point == NULL) {
		filter_out_non_block_devices(new_globs_index);
	}

	LOG_PROC_EXIT_VOID();
}

static char dir_pattern[PATH_MAX];

static void process_dir(char * name) {
	
	int i;
	glob_t dirs_glob;

	LOG_PROC_ENTRY();

	/* Process entries in this directory. */
	get_dev_names(name);

	/* Get a list of this directory's subdirectories. */
	strcpy(dir_pattern, name);
	strcat(dir_pattern, "*/");

	if (glob(dir_pattern, 0, NULL, &dirs_glob) == 0) {

		/* Process the subdirectories. */
		for (i = 0; i < dirs_glob.gl_pathc; i++) {
			int status;
			struct stat statbuf;

			status = stat(dirs_glob.gl_pathv[i], &statbuf);
			if (status == 0) {
				if (S_ISDIR(statbuf.st_mode)) {
					process_dir(dirs_glob.gl_pathv[i]);
				}
			}
		}

		globfree(&dirs_glob);
	}

	LOG_PROC_EXIT_VOID();
}

static char dir_path[PATH_MAX];

static int get_legacy_devs() {

	int base_len;
	int rc = 0;
	char * pch;

	LOG_PROC_ENTRY();

	memset(&dev_names_glob, 0, sizeof(dev_names_glob));
	glob_flags = 0;

	/* Make sure the user-specified directory ends with a '/'. */
	base_len = strlen(scan);
	if (scan[base_len-1] != '/') {
		pch = malloc(base_len + 2);
		if (pch) {
			strcpy(pch, scan);
			strcpy(pch + base_len, "/");
			scan = pch;
			base_len += 2;
		}
	}

	base_directory = strdup(scan);
	base_directory_len = strlen(base_directory);

	if (directories_count == 0) {
		get_dev_names(base_directory);

	} else {
		int i;

		strcpy(dir_path, base_directory);

		for (i = 0; i < directories_count; i++) {
			int len;

			strcpy(dir_path + base_directory_len, directories[i]);

			len = strlen(dir_path);
			if (dir_path[len-1] != '/') {
				strcpy(dir_path + len, "/");
			}

			process_dir(dir_path);
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

static int get_sysfs_devs() {

	int rc = 0;

	LOG_PROC_ENTRY();

	memset(&dev_names_glob, 0, sizeof(dev_names_glob));
	glob_flags = 0;

	strcpy(dir_path, sysfs_mount_point);
	strcat(dir_path, "/block/");

	base_directory = strdup(dir_path);
	base_directory_len = strlen(base_directory);

	LOG_DEBUG("Scanning %s\n", dir_path);
	get_dev_names(dir_path);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

static int get_sysfs_size(char * name, u_int64_t * p_size)
{
	int rc = 0;
	int fd;
	char * size_file = malloc(strlen(name) + 6);
	char size_str[24];

	LOG_PROC_ENTRY();

	if (size_file != NULL) {
		strcpy(size_file, name);
		strcat(size_file, "/size");
		fd = open(size_file, O_RDONLY);
		if (fd > 0) {
			int bytes_read;

			bytes_read = read(fd, size_str, 24);

			if (bytes_read > 0) {
				/* Size is already in sectors. */
				*p_size = strtoull(size_str, NULL, 10);

			} else {
				if (bytes_read == 0) {
					LOG_ERROR("No bytes read from %s.\n", size_file);
				}

				rc = errno;
				LOG_ERROR("read(%d) returned error %d: %s\n", rc, strerror(rc));
			}

			close(fd);

		} else {
			rc = errno;
			LOG_ERROR("open(%s) returned error %d: %s\n", size_file, rc, strerror(rc));
		}

		free(size_file);
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

static int get_legacy_size(char * name, u_int64_t * p_size)
{
	int rc = 0;
	int fd;

	LOG_PROC_ENTRY();

	fd = open(name, O_RDONLY);

	if (fd > 0) {

		/* Ioctl to get size. (returns bytes) */
		rc = ioctl(fd, BLKGETSIZE64, p_size);
		if (rc == 0) {
			*p_size >>= EVMS_VSECTOR_SIZE_SHIFT;

		} else {
			rc = errno;
			LOG_DETAILS("ioctl to get the size returned error code %d: %s.\n",
				  rc, strerror(rc));
		}

		close(fd);

	} else {
		rc = errno;
		LOG_DETAILS("open(%s) returned error %d: %s\n", name, rc, strerror(rc));
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

static int get_sysfs_major_minor(char * name, u_int32_t * p_major, u_int32_t * p_minor)
{

	int rc = 0;
	int fd;
	char * dev_file = malloc(strlen(name) + 5);
	char dev_str[16];
	dev_t dev;

	LOG_PROC_ENTRY();

	if (dev_file != NULL) {
		strcpy(dev_file, name);
		strcat(dev_file, "/dev");
		fd = open(dev_file, O_RDONLY);
		if (fd > 0) {
			int bytes_read;

			bytes_read = read(fd, dev_str, 16);

			if (bytes_read > 0) {
				dev = strtoul(dev_str, NULL, 16);
				*p_major = major(dev);
				*p_minor = minor(dev);

			} else {
				if (bytes_read == 0) {
					LOG_ERROR("No bytes read from %s.\n", dev_file);
				}

				rc = errno;
				LOG_ERROR("read(%d) returned error %d: %s\n", rc, strerror(rc));
			}

			close(fd);

		} else {
			rc = errno;
			LOG_ERROR("open(%s) returned error %d: %s\n", dev_file, rc, strerror(rc));
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

static int get_legacy_major_minor(char * name, u_int32_t * p_major, u_int32_t * p_minor)
{

	int rc = 0;
	struct stat statbuf;

	LOG_PROC_ENTRY();

	rc = stat(name, &statbuf);
	if (rc == 0) {
		*p_major = major(statbuf.st_rdev);
		*p_minor = minor(statbuf.st_rdev);

	} else {
		rc = errno;
		LOG_ERROR("stat(%s) returned error code %d: %s\n", rc, strerror(rc));
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * LD_discover
 **/
static storage_object_t working_object;

static int LD_discover(dlist_t input_list,
		       dlist_t output_list,
		       BOOLEAN final_call)
{
	int rc;
	int i;
	int fd;
	u_int32_t hardsector_size;
	struct hd_geometry geometry;
	struct stat statbuf;
	storage_object_t * ld;
	void * trash;
	uint count;

	LOG_PROC_ENTRY();

	if (sysfs_mount_point != NULL) {
		rc = get_sysfs_devs();

	} else {
		rc = get_legacy_devs();
	}

	if (rc == 0) {
		int tmp_rc;

		for (i = 0; i < dev_names_glob.gl_pathc; i++) {
			char * full_node_path = dev_names_glob.gl_pathv[i];

			/* A block device for a disk must have a size. */
			if (sysfs_mount_point != NULL) {
				tmp_rc = get_sysfs_size(full_node_path, &working_object.size);

			} else {
				tmp_rc = get_legacy_size(full_node_path, &working_object.size);
			}

			if (tmp_rc != 0) {
				continue;
			}

			if (working_object.size == 0) {
				LOG_DEBUG("%s has a size of zero.  It's not a disk.\n", full_node_path);
				continue;
			}

			/* Check the major:minor of the device. */
			if (sysfs_mount_point != NULL) {
				tmp_rc = get_sysfs_major_minor(full_node_path, &working_object.dev_major, &working_object.dev_minor);

			} else {
				tmp_rc = get_legacy_major_minor(full_node_path, &working_object.dev_major, &working_object.dev_minor);
			}

			if (tmp_rc != 0) {
				continue;
			}

			/* Exclude floppy(2 & 47), md(9), and lvm1(58) devices. */
			if (working_object.dev_major ==  2 ||
			    working_object.dev_major == 47 ||
			    working_object.dev_major ==  9 ||
			    working_object.dev_major == 58) {
				LOG_DEBUG("%s has a major number that is not for a hard drive.\n", full_node_path);
				continue;
			}

			memset(working_object.name, 0, sizeof(working_object.name));
			strncpy(working_object.name, full_node_path + base_directory_len, EVMS_NAME_SIZE);

			/*
			 * Even though we are opening a local object, use the
			 * Engine's open_object() service so that the dev node
			 * gets created in the /dev/evms/.nodes tree.
			 */
			fd = EngFncs->open_object(&working_object, O_DIRECT|O_RDWR|O_SYNC);
			if (fd < 0) {
				LOG_DEBUG("Unable to open %s.  Error code is %d: %s\n", dev_names_glob.gl_pathv[i], errno, strerror(errno));
				fd = 0;
				continue;
			}

			/*
			 * Ioctl to get geometry.
			 * Don't need the Engine service since we know this
			 * is a local open.
			 */
			tmp_rc = ioctl(fd, HDIO_GETGEO, &geometry);
			if (tmp_rc != 0) {
				LOG_DEBUG("ioctl to get geometry failed with error code %d: %s.\n", errno, strerror(errno));
				close(fd);
				fd = 0;
				continue;
			}

			/* A disk's geometry must start at offset 0. */
			if (geometry.start != 0) {
				LOG_DEBUG("Geometry says the start is not at offset zero.  Therefore it is not considered a disk.\n");
				close(fd);
				fd = 0;
				continue;
			}

			/* Ioctl to get hardsector size. If this returns an
			 * error, fall back on the default sector-size (which
			 * is the same default used in the kernel - see
			 * include/linux/blkdev.h::get_hardsect_size).
			 */
			tmp_rc = ioctl(fd, BLKSSZGET, &hardsector_size);
			if (tmp_rc != 0) {
				LOG_DEBUG("ioctl to get hardsector size failed with error code %d: %s.\n",
					  errno, strerror(errno));
				hardsector_size = EVMS_VSECTOR_SIZE;
			}

			/* fstat to get block size. */
			tmp_rc = fstat(fd,  &statbuf);
			if (tmp_rc) {
				LOG_DEBUG("fstat to get block size failed with error code %d: %s.\n",
					  errno, strerror(errno));
				close(fd);
				fd = 0;
				continue;
			}

			/* Passed all checks.  Get a logical_disk structure. */
			rc = EngFncs->allocate_logical_disk(working_object.name, &ld);
			if (rc) {
				LOG_SERIOUS("Error code %d from Engine's allocate_logical_disk.  Cannot create a disk object.\n", rc);
				LOG_SERIOUS("Engine's allocate_logical_disk(%s) failed with error code %d: %s.\n",
					    full_node_path + base_directory_len, rc, EngFncs->strerror(rc));
				close(fd);
				fd = 0;
				break;
			}

			/* Allocate and fill in private data. */
			ld->private_data = EngFncs->engine_alloc(sizeof(int));
			if (!ld->private_data) {
				LOG_SERIOUS("Unable to allocate private data for disk %s.\n", ld->name);
				EngFncs->free_logical_disk(ld);
				close(fd);
				fd = 0;
				rc = ENOMEM;
				break;
			}

			*(int *)ld->private_data = fd;
			/* Leave the file descriptor for the disk open. */

			/* Initialize the logical disk structure */
			ld->plugin			= LD_PluginRecord;
			ld->data_type			= DATA_TYPE;
			ld->flags		       |= SOFLAG_ACTIVE;
			ld->size			= working_object.size;
			ld->dev_major			= working_object.dev_major;
			ld->dev_minor			= working_object.dev_minor;
			ld->geometry.cylinders		= geometry.cylinders;
			ld->geometry.heads		= geometry.heads;
			ld->geometry.sectors_per_track	= geometry.sectors;
			ld->geometry.bytes_per_sector	= hardsector_size;
			ld->geometry.block_size		= statbuf.st_blksize;

			/*
			 * Fill in the boot cylinder limit (LBA of 1st sector
			 * above boot cylinder) for this drive.
			 */
			if (ld->geometry.cylinders < 1024) {

				/*
				 * If the drive is too small then the limit is
				 * the size of the drive.
				 */
				ld->geometry.boot_cylinder_limit = ld->size;

			} else {
				/* Else the limit is calculated. */
				ld->geometry.boot_cylinder_limit =
					ld->geometry.heads * ld->geometry.sectors_per_track * 1023;
			}

			/* Insert the logical disk into discovered list */
			rc = InsertObject (output_list, ld, DISK_TAG, NULL, AppendToList, TRUE, &trash);
			if (rc) {
				LOG_SERIOUS("Error code %d when inserting new logical disk %s into the output list.  Deleting the disk.\n", rc, ld->name);
				free(ld->private_data);
				EngFncs->free_logical_disk(ld);
				close(fd);
				fd = 0;
				break;
			}

			/* Successful */
			LOG_DETAILS("New Logical Device\n");
			LOG_DETAILS("name:       %s\n", ld->name);
			LOG_DETAILS("size:       %"PRIu64"\n", ld->size);
			LOG_DETAILS("geometry:\n");
			LOG_DETAILS("  cylinders:   %"PRIu64"\n", ld->geometry.cylinders);
			LOG_DETAILS("  heads:       %d\n", ld->geometry.heads);
			LOG_DETAILS("  sectors:     %d\n", ld->geometry.sectors_per_track);
			LOG_DETAILS("  sector size: %d (bytes)\n", ld->geometry.bytes_per_sector);
		}
	}

	GetListSize(output_list, &count);
	LOG_DEBUG("Exporting %d Objects\n", count);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * LD_read
 *
 * Used by both Legacy and Sysfs versions.
 **/
static int LD_read(storage_object_t * ld,
		   lsn_t offset,
		   sector_count_t count,
		   void * buffer)
{
	void * local_buffer = NULL, * local_buffer_aligned;
	lsn_t local_offset;
	sector_count_t local_count;
	u_int32_t hss, local_buffer_bytes;
	u_int32_t offset_diff, offset_diff_bytes;
	u_int32_t alignment_base, alignment_diff;
	int * fd_p = ld->private_data;
	int fd, rc = EINVAL;

	LOG_PROC_ENTRY();

        if (offset + count > ld->size) {
		LOG_ERROR("Read request at sector %"PRIu64" for %"PRIu64
			  " sectors goes off end of disk.\n", offset, count);
		LOG_ERROR("  Disk: %s   Size: %"PRIu64"\n", ld->name, ld->size);
		goto out;
	}

	rc = read_from_cache(ld, offset, count, buffer);
	if (!rc) {
		/* Found in the cache. */
		goto out;
	}

	/* Not in the cache. Have to go to disk. */
	LOG_DEBUG("Read %"PRIu64" sectors from disk %s at sector %"PRIu64".\n",
		  count, ld->name, offset);

	if (!fd_p) {
		/* Private data must have gotten lost.
		 * Just allocate a new one.
		 */
		fd_p = EngFncs->engine_alloc(sizeof(int));
		if (!fd_p) {
			rc = ENOMEM;
			goto out;
		}
		ld->private_data = fd_p;
	}

	fd = *fd_p;
	if (fd <= 0) {
		/* Need to open the disk. */
		fd = EngFncs->open_object(ld, O_DIRECT|O_RDONLY|O_SYNC);
		if (fd > 0) {
			*fd_p = fd;
		}
	}

	if (fd < 0) {
		/* Couldn't open the disk. */
		rc = EIO;
		goto out;
	}

	/* Grab the hard-sector-size. */
	hss = ld->geometry.bytes_per_sector;
	if (version_2_4 && hss < 1024) {
		/* For O_DIRECT */
		hss = 1024;
	}

	/* Round down starting offset to hard-sector-size. */
	local_offset = round_down_to_hard_sector(offset, hss);

	/* Difference between real offset and local offset. */
	offset_diff = offset - local_offset;
	offset_diff_bytes = offset_diff << EVMS_VSECTOR_SIZE_SHIFT;

	/* Round up total count of sectors to hard-sector-size. */
	local_count = round_up_to_hard_sector(count + offset_diff, hss);

	/* Allocate the buffer that will actually perform the I/O. This needs
	 * to allocate enough extra to compensate for the alignment that will
	 * occur next.
	 */
	local_buffer_bytes = local_count << EVMS_VSECTOR_SIZE_SHIFT;
	local_buffer = EngFncs->engine_alloc(local_buffer_bytes + hss - 1);
	if (!local_buffer) {
		rc = ENOMEM;
		goto out;
	}

	/* Align the buffer on hard-sector-size. */
	alignment_base = (unsigned long)local_buffer & (hss - 1);
	alignment_diff = (alignment_base + hss - 1) & ~(hss - 1);
	alignment_diff -= alignment_base;
	local_buffer_aligned = local_buffer + alignment_diff;

	/* Sanity check */
	if (((unsigned long)local_buffer_aligned % hss) != 0) {
		LOG_DEBUG("BUFFER NOT ALIGNED!!: buf: %p, hss: %u\n",
			  local_buffer_aligned, hss);
	}

	LOG_DEBUG("read(Fd: %d, Buf: %p, Size: %u, Offset: %"PRIu64"\n",
		  fd, local_buffer_aligned, local_buffer_bytes,
		  local_offset << EVMS_VSECTOR_SIZE_SHIFT);

	rc = EngFncs->read_object(ld, fd, local_buffer_aligned, local_buffer_bytes,
				  local_offset << EVMS_VSECTOR_SIZE_SHIFT);
	if (rc < 0) {
		rc = EIO;
		goto out;
	}

	/* Copy the data back to the caller's buffer. */
	memcpy(buffer, local_buffer_aligned+offset_diff_bytes,
	       count << EVMS_VSECTOR_SIZE_SHIFT);
	write_to_cache(ld, offset, count, buffer);
	rc = 0;

out:
	if (local_buffer) {
		EngFncs->engine_free(local_buffer);
	}
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * LD_write
 *
 * Used by both Legacy and Sysfs versions.
 **/
static int LD_write(storage_object_t * ld,
		    lsn_t offset,
		    sector_count_t count,
		    void * buffer)
{
	void * local_buffer = NULL, * local_buffer_aligned;
	lsn_t local_offset;
	sector_count_t local_count;
	u_int32_t hss, local_buffer_bytes;
	u_int32_t offset_diff, offset_diff_bytes;
	u_int32_t alignment_base, alignment_diff;
	int * fd_p = ld->private_data;
	int fd, rc = EINVAL;

	LOG_PROC_ENTRY();

	if (offset + count > ld->size) {
		LOG_ERROR("Write request at sector %"PRIu64" for %"PRIu64
			  " sectors goes off end of disk.\n", offset, count);
		LOG_ERROR("  Disk: %s   Size: %"PRIu64"\n", ld->name, ld->size);
		goto out;
	}

	LOG_DEBUG("Write %"PRIu64" sectors to disk %s at sector %"PRIu64".\n",
		  count, ld->name, offset);

	if (!fd_p) {
		/* Private data must have gotten lost.
		 * Just allocate a new one.
		 */
		fd_p = EngFncs->engine_alloc(sizeof(int));
		if (!fd_p) {
			rc = ENOMEM;
			goto out;
		}
		ld->private_data = fd_p;
	}

	fd = *fd_p;
	if (fd <= 0) {
		/* Need to open the disk. */
		fd = EngFncs->open_object(ld, O_DIRECT|O_RDWR|O_SYNC);
		if (fd > 0) {
			*fd_p = fd;
		}
	}

	if (fd < 0) {
		/* Couldn't open the disk. */
		rc = EIO;
		goto out;
	}

	/* Grab the hard-sector-size. */
	hss = ld->geometry.bytes_per_sector;
	if (version_2_4 && hss < 1024) {
		/* For O_DIRECT */
		hss = 1024;
	}

	/* Round down starting offset to hard-sector-size. */
	local_offset = round_down_to_hard_sector(offset, hss);
	
	/* Difference between real offset and local offset. */
	offset_diff = offset - local_offset;
	offset_diff_bytes = offset_diff << EVMS_VSECTOR_SIZE_SHIFT;
	
	/* Round up total count of sectors to hard-sector-size. */
	local_count = round_up_to_hard_sector(count + offset_diff, hss);
	
	/* Allocate the buffer that will actually perform the I/O. This needs
	 * to allocate enough extra to compensate for the alignment that will
	 * occur next.
	 */
	local_buffer_bytes = local_count << EVMS_VSECTOR_SIZE_SHIFT;
	local_buffer = EngFncs->engine_alloc(local_buffer_bytes + hss - 1);
	if (!local_buffer) {
		rc = ENOMEM;
		goto out;
	}
	
	/* Align the buffer on hard-sector-size. */
	alignment_base = (unsigned long)local_buffer & (hss - 1);
	alignment_diff = (alignment_base + hss - 1) & ~(hss - 1);
	alignment_diff -= alignment_base;
	local_buffer_aligned = local_buffer + alignment_diff;

	/* Sanity check */
	if (((unsigned long)local_buffer_aligned % hss) != 0) {
		LOG_DEBUG("BUFFER NOT ALIGNED!!: buf: %p, hss: %u\n",
			  local_buffer_aligned, hss);
	}

	if (local_count != count) {
		/* Padding for alignment read first */
		rc = EngFncs->read_object(ld, fd, local_buffer_aligned,
					  local_buffer_bytes,
					  local_offset << EVMS_VSECTOR_SIZE_SHIFT);
		if (rc < 0) {
			rc = EIO;
			goto out;
		}
	}
	
	/* Put user data at the right place in the buffer */
	memcpy(local_buffer_aligned+offset_diff_bytes, buffer,
	       count << EVMS_VSECTOR_SIZE_SHIFT);
	
	rc = EngFncs->write_object(ld, fd, local_buffer_aligned,
				   local_buffer_bytes,
				   local_offset << EVMS_VSECTOR_SIZE_SHIFT);
	if (rc < 0) {
		rc = EIO;
		goto out;
	}

	/*
	 * The cache is too simple to do real caching.  It's really only a read
	 * cache.  A write, which should not happen during discovery, means the
	 * contents of the cache may not be up to date.  Purge the cache and
	 * start caching all over again.
	 */
	purge_cache();
	rc = 0;

out:
	if (local_buffer) {
		EngFncs->engine_free(local_buffer);
	}
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * LD_add_sectors_to_kill_list
 *
 * Used by both Legacy and Sysfs versions.
 **/
static int LD_add_sectors_to_kill_list( storage_object_t * ld,
					lsn_t lsn,
					sector_count_t count)
{
	int rc = EINVAL;

	LOG_PROC_ENTRY();

	if (lsn + count > ld->size) {
		LOG_ERROR("Sector offset plus count goes beyond the end of the disk.\n");
	} else {
		rc = EngFncs->add_sectors_to_kill_list(ld, lsn, count);
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * LD_commit_changes
 *
 * Disk manager doesn't do anything during commit. Just return success.
 **/
static int LD_commit_changes(storage_object_t * ld, commit_phase_t phase)
{
	LOG_PROC_ENTRY();
	LOG_PROC_EXIT_INT(0);
	return 0;
}

/**
 * LD_get_info
 *
 * Returns DISK specific information
 **/
static int LD_get_info(storage_object_t * object,
		       char * descriptor_name,
		       extended_info_array_t ** info)
{
	extended_info_array_t * Info;
	uint count = 0;
	int rc = EINVAL;

	LOG_PROC_ENTRY();

	if (!info) {
		goto out;
	}
	*info = NULL;
	
        if (descriptor_name) {
		goto out;
	}
	
	Info = EngFncs->engine_alloc(sizeof(extended_info_array_t) +
				     (LD_INFO_COUNT - 1) * sizeof(extended_info_t));
	if (!Info) {
		rc = ENOMEM;
		goto out;
	}
		
	Info->count = LD_INFO_COUNT;
		
	SET_STRING_FIELD(Info->info[LD_INFO_NAME_INDEX].name, "Name");
	SET_STRING_FIELD(Info->info[LD_INFO_NAME_INDEX].title, "Name");
	SET_STRING_FIELD(Info->info[LD_INFO_NAME_INDEX].desc, "EVMS name for the DISK storage object");
	Info->info[LD_INFO_NAME_INDEX].type = EVMS_Type_String;
	SET_STRING_FIELD(Info->info[LD_INFO_NAME_INDEX].value.s, object->name);

	SET_STRING_FIELD(Info->info[LD_INFO_SIZE_INDEX].name, "Size");
	SET_STRING_FIELD(Info->info[LD_INFO_SIZE_INDEX].title, "Size");
	SET_STRING_FIELD(Info->info[LD_INFO_SIZE_INDEX].desc, "Size of the disk in sectors");
	Info->info[LD_INFO_SIZE_INDEX].type = EVMS_Type_Unsigned_Int64;
	Info->info[LD_INFO_SIZE_INDEX].unit = EVMS_Unit_Sectors;
	Info->info[LD_INFO_SIZE_INDEX].value.ui64 = object->size;

	SET_STRING_FIELD(Info->info[LD_INFO_CYLINDERS_INDEX].name, "Cyl");
	SET_STRING_FIELD(Info->info[LD_INFO_CYLINDERS_INDEX].title, "Cylinders");
	SET_STRING_FIELD(Info->info[LD_INFO_CYLINDERS_INDEX].desc, "Drive geometry -- number of cylinders");
	Info->info[LD_INFO_CYLINDERS_INDEX].type = EVMS_Type_Unsigned_Int64;
	Info->info[LD_INFO_CYLINDERS_INDEX].value.ui64 = object->geometry.cylinders;

	SET_STRING_FIELD(Info->info[LD_INFO_HEADS_INDEX].name, "Heads");
	SET_STRING_FIELD(Info->info[LD_INFO_HEADS_INDEX].title, "Heads");
	SET_STRING_FIELD(Info->info[LD_INFO_HEADS_INDEX].desc, "Drive geometry -- number of heads");
	Info->info[LD_INFO_HEADS_INDEX].type = EVMS_Type_Unsigned_Int32;
	Info->info[LD_INFO_HEADS_INDEX].value.ui32 = object->geometry.heads;

	SET_STRING_FIELD(Info->info[LD_INFO_SECTORS_INDEX].name, "Sectors");
	SET_STRING_FIELD(Info->info[LD_INFO_SECTORS_INDEX].title, "Sectors");
	SET_STRING_FIELD(Info->info[LD_INFO_SECTORS_INDEX].desc, "Drive geometry -- sectors per track");
	Info->info[LD_INFO_SECTORS_INDEX].type = EVMS_Type_Unsigned_Int32;
	Info->info[LD_INFO_SECTORS_INDEX].value.ui32 = object->geometry.sectors_per_track;

	SET_STRING_FIELD(Info->info[LD_INFO_SECTOR_SIZE_INDEX].name, "SectorSize");
	SET_STRING_FIELD(Info->info[LD_INFO_SECTOR_SIZE_INDEX].title, "Sector Size");
	SET_STRING_FIELD(Info->info[LD_INFO_SECTOR_SIZE_INDEX].desc, "Number of bytes per sector");
	Info->info[LD_INFO_SECTOR_SIZE_INDEX].type = EVMS_Type_Unsigned_Int32;
	Info->info[LD_INFO_SECTOR_SIZE_INDEX].unit = EVMS_Unit_Bytes;
	Info->info[LD_INFO_SECTOR_SIZE_INDEX].value.ui32 = object->geometry.bytes_per_sector;

	SET_STRING_FIELD(Info->info[LD_INFO_BLOCK_SIZE_INDEX].name, "BlockSize");
	SET_STRING_FIELD(Info->info[LD_INFO_BLOCK_SIZE_INDEX].title, "Block Size");
	SET_STRING_FIELD(Info->info[LD_INFO_BLOCK_SIZE_INDEX].desc, "Number of bytes per block");
	Info->info[LD_INFO_BLOCK_SIZE_INDEX].type = EVMS_Type_Unsigned_Int64;
	Info->info[LD_INFO_BLOCK_SIZE_INDEX].unit = EVMS_Unit_Bytes;
	Info->info[LD_INFO_BLOCK_SIZE_INDEX].value.ui64 = object->geometry.block_size;

	SET_STRING_FIELD(Info->info[LD_INFO_BOOT_LIMIT_INDEX].name, "BootLimit");
	SET_STRING_FIELD(Info->info[LD_INFO_BOOT_LIMIT_INDEX].title, "Boot Cylinder Limit");
	SET_STRING_FIELD(Info->info[LD_INFO_BOOT_LIMIT_INDEX].desc, "LBA of the first sector above the boot cylinder limit for this drive");
	Info->info[LD_INFO_BOOT_LIMIT_INDEX].type = EVMS_Type_Unsigned_Int64;
	Info->info[LD_INFO_BOOT_LIMIT_INDEX].value.ui64 = object->geometry.boot_cylinder_limit;

	GetListSize(object->parent_objects, &count);

	SET_STRING_FIELD(Info->info[LD_INFO_SEGMENTS_INDEX].name, "Segments");
	SET_STRING_FIELD(Info->info[LD_INFO_SEGMENTS_INDEX].title, "Segments");
	SET_STRING_FIELD(Info->info[LD_INFO_SEGMENTS_INDEX].desc, "Number of segments discovered on the drive (metadata, data, freespace)");
	Info->info[LD_INFO_SEGMENTS_INDEX].type = EVMS_Type_Unsigned_Int32;
	Info->info[LD_INFO_SEGMENTS_INDEX].value.ui32 = count;

	*info = Info;
	rc = 0;

out:
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * LD_get_plugin_info
 *
 * Returns plug-in specific information
 **/
static int LD_get_plugin_info(char * descriptor_name,
			      extended_info_array_t ** info)
{
	int rc = EINVAL;
	extended_info_array_t * Info;
	char version_string[64];
	char required_engine_api_version_string[64];
	char required_plugin_api_version_string[64];

	LOG_PROC_ENTRY();

	if (!info) {
		goto out;
	}
	*info = NULL;

	if (descriptor_name) {
		goto out;
	}

	Info = EngFncs->engine_alloc(sizeof(extended_info_array_t) +
				     6 * sizeof(extended_info_t));
	if (!Info) {
		rc = ENOMEM;
		goto out;
	}

	Info->count = 6;

	sprintf(version_string, "%d.%d.%d",
		MAJOR_VERSION, MINOR_VERSION, PATCH_LEVEL);

	sprintf(required_engine_api_version_string, "%d.%d.%d",
		LD_PluginRecord->required_engine_api_version.major,
		LD_PluginRecord->required_engine_api_version.minor,
		LD_PluginRecord->required_engine_api_version.patchlevel);

	sprintf(required_plugin_api_version_string, "%d.%d.%d",
		LD_PluginRecord->required_plugin_api_version.plugin.major,
		LD_PluginRecord->required_plugin_api_version.plugin.minor,
		LD_PluginRecord->required_plugin_api_version.plugin.patchlevel);

	SET_STRING_FIELD(Info->info[0].name, "Short Name");
	SET_STRING_FIELD(Info->info[0].title, "Short Name");
	SET_STRING_FIELD(Info->info[0].desc, "A short name given to this plugin");
	Info->info[0].type = EVMS_Type_String;
	SET_STRING_FIELD(Info->info[0].value.s, LD_PluginRecord->short_name);

	SET_STRING_FIELD(Info->info[1].name, "Long Name");
	SET_STRING_FIELD(Info->info[1].title, "Long Name");
	SET_STRING_FIELD(Info->info[1].desc, "A long name given to this plugin");
	Info->info[1].type = EVMS_Type_String;
	SET_STRING_FIELD(Info->info[1].value.s, LD_PluginRecord->long_name);

	SET_STRING_FIELD(Info->info[2].name, "Type");
	SET_STRING_FIELD(Info->info[2].title, "Plug-in Type");
	SET_STRING_FIELD(Info->info[2].desc, "There are various types of plugins, each responsible for some kind of storage object.");
	Info->info[2].type = EVMS_Type_String;
	SET_STRING_FIELD(Info->info[2].value.s, "Device Manager");

	SET_STRING_FIELD(Info->info[3].name, "Version");
	SET_STRING_FIELD(Info->info[3].title, "Plugin Version");
	SET_STRING_FIELD(Info->info[3].desc, "Version number of this plug-in");
	Info->info[3].type = EVMS_Type_String;
	SET_STRING_FIELD(Info->info[3].value.s, version_string);

	SET_STRING_FIELD(Info->info[4].name, "Required Engine Services Version");
	SET_STRING_FIELD(Info->info[4].title, "Required Engine Services Version");
	SET_STRING_FIELD(Info->info[4].desc, "Version of the Engine services that this plug-in requires. It will not run on older versions of the Engine services.");
	Info->info[4].type = EVMS_Type_String;
	SET_STRING_FIELD(Info->info[4].value.s, required_engine_api_version_string);

	SET_STRING_FIELD(Info->info[5].name, "Required Plug-in API Version");
	SET_STRING_FIELD(Info->info[5].title, "Required Plug-in API Version");
	SET_STRING_FIELD(Info->info[5].desc, "Version of the Engine plug-in API that this plug-in requires. It will not run on older versions of the Engine plug-in API.");
	Info->info[5].type = EVMS_Type_String;
	SET_STRING_FIELD(Info->info[5].value.s, required_plugin_api_version_string);

	*info = Info;
	rc = 0;

out:
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * LD_direct_plugin_communication
 *
 * Device-manager-specific commands for starting and stopping the cache, and
 * opening and closing a disk.
 **/
static int LD_direct_plugin_communication(void * object,
					  BOOLEAN target_kernel_plugin,
					  void * arg)
{
	int rc = 0, fd = -1;
	storage_object_t * ld = object;
	ldm_dpc_t * dcp = arg;

	LOG_PROC_ENTRY();
	if (ld != NULL) {
		LOG_DEBUG("Action targeted at %s.\n", ld->name);
	}

	switch (dcp->function) {

	case START_CACHE:
		LOG_DEBUG("START_CACHE\n");
		initialize_cache();
		break;

	case STOP_CACHE:
		LOG_DEBUG("STOP_CACHE\n");
		destroy_cache();
		break;

       	case LDM_CLOSE_DISK:
		LOG_DEBUG("LDM_CLOSE_DISK\n");
		fd = *(int *)ld->private_data;
		if (fd >= 0) {			
			rc = EngFncs->close_object(ld, fd);
			if (!rc) {
				*(int *)ld->private_data = -1;
			}
		}
		break;

	case LDM_OPEN_DISK:
		LOG_DEBUG("LDM_OPEN_DISK\n");
		fd = *(int *)ld->private_data;
		if (fd < 0) {	
			fd = EngFncs->open_object(ld, O_DIRECT|O_RDWR|O_SYNC);
			if (fd >= 0) {	
				*(int *)ld->private_data = fd;
				rc = 0;
			} else {
				rc = errno;
			}
		}
		break;

	default:
		LOG_ERROR("%d is not a valid function code.\n", dcp->function);
		rc = EINVAL;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

static plugin_functions_t ft_sysfs = {
	.setup_evms_plugin		= LD_setup,
	.cleanup_evms_plugin		= LD_cleanup,
	.discover			= LD_discover,
	.add_sectors_to_kill_list	= LD_add_sectors_to_kill_list,
	.commit_changes			= LD_commit_changes,
	.get_info			= LD_get_info,
	.get_plugin_info		= LD_get_plugin_info,
	.read				= LD_read,
	.write				= LD_write,
	.direct_plugin_communication	= LD_direct_plugin_communication
};

plugin_record_t plugin_record = {
	.id = SetPluginID(EVMS_OEM_IBM, EVMS_DEVICE_MANAGER, 2),
	.version = {
		.major		= MAJOR_VERSION,
		.minor		= MINOR_VERSION,
		.patchlevel	= PATCH_LEVEL
	},
	.required_engine_api_version = {
		.major		= 9,
		.minor		= 0,
		.patchlevel	= 0
	},
	.required_plugin_api_version = {
		.plugin = {
			.major		= 10,
			.minor		= 0,
			.patchlevel	= 0
		}
	},
	.short_name = "LocalDskMgr",
	.long_name = "Local Disk Manager",
	.oem_name = "IBM",
	.functions = {
		.plugin = &ft_sysfs
	},
	.container_functions = NULL
};

plugin_record_t * LD_PluginRecord = &plugin_record;

plugin_record_t * evms_plugin_records[] = { &plugin_record,
					    NULL };

