/*
 *   Copyright (c) International Business Machines  Corp., 2001
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or 
 *   (at your option) any later version.
 * 
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software 
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: lvmregmgr
 * File: lvm_move.c
 *
 * Description: This file contains all functions related to moving extents.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <plugin.h>
#include "lvmregmgr.h"

static int disable_move_message = FALSE;

/**
 * lvm_move_extent_setup
 * @le:		Logical extent to move.
 * @dst_pe:	Physical extent to move the LE to.
 *
 * Setup the specified LE to be moved from its current PE to the new PE. Make
 * sure the destination PE is available before calling this function.
 **/
static int lvm_move_extent_setup(lvm_logical_extent_t * le,
				 lvm_physical_extent_t * dst_pe)
{
	lvm_logical_volume_t * volume = le->volume;
	lvm_volume_group_t * group = volume->group;

	LOG_ENTRY;

	le->new_pe = dst_pe;
	dst_pe->new_le = le;
	dst_pe->pv->move_extents++;
	group->move_extents++;
	volume->flags |= LVM_LV_FLAG_MOVE_PENDING;

	lvm_append_region_to_segment(volume->region, dst_pe->pv->segment);

	RETURN(0);
}

/**
 * lvm_commit_start_copy_progress
 *
 * Set-up the copy-progress indicator.
 **/
static int lvm_commit_start_copy_progress(copy_job_t * copy_job,
					  progress_t * progress)
{
	int title_size = (EVMS_NAME_SIZE+5)*3 + 30;
	lvm_logical_extent_t * le = copy_job->context;

	LOG_ENTRY;

	progress->id = 0;
	progress->title = lvm_engine->engine_alloc(title_size);
	snprintf(progress->title, title_size,
		 "LVM: Copying LE %s:%d from %s:%d to %s:%d", 
		 le->volume->region->name, le->number,
		 le->pe->pv->segment->name, le->pe->number,
		 le->new_pe->pv->segment->name, le->new_pe->number);
	progress->description = NULL;
	progress->type = DISPLAY_PERCENT;
	progress->count = 0;
	progress->total_count = copy_job->src.len / copy_job->chunk_size + 1;
	progress->remaining_seconds = 0;
	progress->plugin_private_data = le;

	copy_job->context = progress;

	lvm_engine->progress(progress);

	RETURN(0);
}

/**
 * lvm_commit_update_copy_progress
 *
 * Status function that will be called periodically during the copying of
 * an extent.
 **/
static void lvm_commit_update_copy_progress(u_int64_t chunks_copied,
					    void * context)
{
	progress_t * progress = context;
	progress->count++;
	lvm_engine->progress(progress);
}

/**
 * lvm_commit_end_copy_progress
 *
 * Tear-down the copy-progress indicator.
 **/
static int lvm_commit_end_copy_progress(progress_t * progress)
{
	LOG_ENTRY;

	progress->count = progress->total_count;
	lvm_engine->progress(progress);
	lvm_engine->engine_free(progress->title);

	RETURN(0);
}

/**
 * lvm_commit_copy_extent
 *
 * Copy the data from the source PE to the destination PE.
 *
 * Offline implementation:
 * Setup an engine copy job and submit to the engine's copy-service.
 *
 * Online implementation:
 * To be written...
 **/
static int lvm_commit_copy_extent(lvm_physical_extent_t * src_pe,
				  lvm_physical_extent_t * dst_pe)
{
	copy_job_t copy_job;
	progress_t copy_progress;
	u_int32_t pe_size = src_pe->pv->pv->pe_size;
	int rc;

	LOG_ENTRY;

	copy_job.src.obj = src_pe->pv->segment;
	copy_job.src.start = src_pe->sector;
	copy_job.src.len = pe_size;
	copy_job.trg.obj = dst_pe->pv->segment;
	copy_job.trg.start = dst_pe->sector;
	copy_job.trg.len = pe_size;
	copy_job.chunk_size = min(pe_size, LVM_COPY_CHUNK_SIZE_DEFAULT);
	copy_job.context = src_pe->le;

	lvm_commit_start_copy_progress(&copy_job, &copy_progress);
	rc = lvm_engine->copy(&copy_job, lvm_commit_update_copy_progress);
	lvm_commit_end_copy_progress(&copy_progress);

	RETURN(rc);
}

/**
 * lvm_commit_move_extent
 *
 * Move the specified LE from its current PE to its new PE. Copy the data
 * from the old PE to the new PE, update the mappings for both PEs, write
 * out the new metadata, and reactivate the volume with the new mapping.
 **/
static int lvm_commit_move_extent(lvm_logical_extent_t * le)
{
	lvm_physical_extent_t * src_pe = le->pe;
	lvm_physical_extent_t * dst_pe = le->new_pe;
	int rc = 0;

	LOG_ENTRY;

	if (!dst_pe) {
		RETURN(0);
	}

	/* Copy the data from the source to the destination. */
	if (src_pe) {
		rc = lvm_commit_copy_extent(src_pe, dst_pe);
		if (rc) {
			RETURN(rc);
		}
	}

	/* Update the source's metadata. */
	if (src_pe) {
		src_pe->pe.lv_num = 0;
		src_pe->pe.le_num = 0;
		src_pe->pv->pv->pe_allocated--;
		if (!lvm_volume_is_on_pv(le->volume, src_pe->pv)) {
			src_pe->pv->pv->lv_cur--;
			lvm_remove_region_from_segment(le->volume->region,
						       src_pe->pv->segment);
		}
		src_pe->le = NULL;
	}

	/* Update the destination's metadata. */
	if (!lvm_volume_is_on_pv(le->volume, dst_pe->pv)) {
		dst_pe->pv->pv->lv_cur++;
	}
	dst_pe->pe.lv_num = le->volume->number;
	dst_pe->pe.le_num = le->number;
	dst_pe->pv->pv->pe_allocated++;
	dst_pe->new_le = NULL;
	dst_pe->le = le;
	dst_pe->pv->move_extents--;
	dst_pe->pv->group->move_extents--;

	/* Update volume's map. */
	le->pe = dst_pe;
	le->new_pe = NULL;

	/* Write the metadata to disk. */
	if (src_pe) {
		lvm_write_pv(src_pe->pv);
		lvm_write_pe_map(src_pe->pv);
	}
	lvm_write_pv(dst_pe->pv);
	lvm_write_pe_map(dst_pe->pv);

	/* Re-activate the volume with the new mapping. */
	rc = lvm_plugin->functions.plugin->activate(le->volume->region);

	RETURN(rc);
}

/**
 * lvm_volume_is_busy
 *
 * Check if the specified volume is open/in-use/mounted. If so, ask the user
 * if they'd like to unmount and retry.
 *
 * Return: TRUE: skip processing this volume.
 *         FALSE: proceed with this volume.
 **/
static int lvm_volume_is_busy(lvm_logical_volume_t * volume,
			      int prompt_user)
{
	char * choices[] = {"Skip", "Retry", NULL};
	char * mount_point;
	int mounted, answer;
	int rc = FALSE;

	LOG_ENTRY;

	if (!volume->region->volume) {
		/* If the region does not have an EVMS volume, assume
		 * the user knows what he is doing.
		 */
		goto out;
	}

	/* This loop will not exit unless the volume is unmounted, or the
	 * user decides to skip this volume.
	 */
	while ((mounted = lvm_engine->is_mounted(volume->region->volume->name,
						 &mount_point))) {
		answer = 0;
		if (prompt_user) {
			lvm_engine->user_message(lvm_plugin, &answer, choices,
					 "Region \"%s\" has extents scheduled "
					 "to be moved. However, this region is "
					 "part of volume \"%s\", which is "
					 "mounted at %s. Please unmount the "
					 "volume and choose \"%s\" to continue "
					 "with the move, or choose \"%s\" to "
					 "skip the move at this time (the move "
					 "will be attempted again the next time "
					 "changes are saved).",
					 volume->region->name,
					 volume->region->volume->name,
					 mount_point, choices[1], choices[0]);
		}

		if (answer == 0) {
			break;
		}
	}

	rc = (mounted && answer == 0);

out:
	RETURN(rc);
}

/**
 * lvm_commit_move_extents
 *
 * Move all extents that are waiting to be moved in this container.
 **/
int lvm_commit_move_extents(lvm_volume_group_t * group)
{
	lvm_logical_volume_t * volume;
	int i, j, rc = 0;

	LOG_ENTRY;

	for (i = 1; i < MAX_LV; i++) {
		volume = group->volume_list[i];
		if (volume && (volume->flags & LVM_LV_FLAG_MOVE_PENDING)) {

			rc = lvm_volume_is_busy(volume, TRUE);
			if (rc) {
				continue;
			}

			for (j = 0; j < volume->lv->lv_allocated_le; j++) {
				if (volume->le_map[j].new_pe) {
					rc |= lvm_commit_move_extent(&volume->le_map[j]);
				}
			}

			if (!rc) {
				volume->flags &= ~LVM_LV_FLAG_MOVE_PENDING;
			}
		}
	}

	/* Need to rebuild the freespace, since the number and
	 * location of unused extents has probably changed.
	 */
	rc = lvm_update_freespace_volume(group);
	if (rc) {
		LOG_SERIOUS("Error updating freespace for container %s\n",
			    group->container->name);
	}

	RETURN(rc);
}


/********** Move-Extent Task Functions **********/


/**
 * lvm_move_extent_init_task
 *
 * Initialize the option descriptor for a MOVE_EXTENT task.
 * Option 0: Logical Extent
 * Option 1: Physical Volume
 * Option 2: Physical Extent
 **/
int lvm_move_extent_init_task(task_context_t * context)
{
	option_desc_array_t * od = context->option_descriptors;
	lvm_logical_volume_t * volume = context->object->private_data;
	lvm_volume_group_t * group = volume->group;
	lvm_physical_volume_t * pv_entry;
	u_int32_t i, index, count = 0;

	LOG_ENTRY;

	/* Option 0: logical extent number.
	 * Generate a list of all the extents in this
	 * volume that are not already being moved.
	 */
	index = LVM_OPTION_MOVE_EXTENT_LE_INDEX;
	SET_STRING(od->option[index].name, LVM_OPTION_MOVE_EXTENT_LE_STR);
	SET_STRING(od->option[index].title, "Logical Extent");
	SET_STRING(od->option[index].tip, "Logical extent to move");
	od->option[index].type = EVMS_Type_Unsigned_Int32;
	od->option[index].flags = EVMS_OPTION_FLAGS_NO_INITIAL_VALUE;
	od->option[index].constraint_type = EVMS_Collection_List;
	od->option[index].constraint.list = lvm_engine->engine_alloc(sizeof(value_list_t) +
								     sizeof(value_t) *
								     volume->lv->lv_allocated_le);
	for (i = 0; i < volume->lv->lv_allocated_le; i++) {
		if (!lvm_le_is_scheduled_for_move(&volume->le_map[i])) {
			od->option[index].constraint.list->value[count].ui32 = i;
			count++;
		}
	}
	od->option[index].constraint.list->count = count;

	/* Option 1: new physical volume.
	 * Generate a list of all PVs that have unused extents.
	 */
	count = 0;
	index = LVM_OPTION_MOVE_EXTENT_PV_INDEX;
	SET_STRING(od->option[index].name, LVM_OPTION_MOVE_EXTENT_PV_STR);
	SET_STRING(od->option[index].title, "Physical Volume");
	SET_STRING(od->option[index].tip, "Physical volume to move this logical extent to.");
	od->option[index].type = EVMS_Type_String;
	od->option[index].max_len = EVMS_NAME_SIZE;
	od->option[index].flags = EVMS_OPTION_FLAGS_NO_INITIAL_VALUE;
	od->option[index].constraint_type = EVMS_Collection_List;
	od->option[index].constraint.list = lvm_engine->engine_alloc(sizeof(value_list_t) +
								     sizeof(value_t) * MAX_PV);
	for (i = 0; i < MAX_PV; i++) {
		pv_entry = group->pv_list[i];
		if (pv_entry && lvm_pv_has_available_extents(pv_entry)) {
			SET_STRING(od->option[index].constraint.list->value[count].s,
				   pv_entry->segment->name);
			count++;
		}
	}
	od->option[index].constraint.list->count = count;
	od->option[index].value.s = lvm_engine->engine_alloc(EVMS_NAME_SIZE+1);

	/* Option 2: new physical extent.
	 * Set up a list option, but can't fill it in yet. Have to wait
	 * until the user chooses which PV to move this extent to.
	 */
	index = LVM_OPTION_MOVE_EXTENT_PE_INDEX;
	SET_STRING(od->option[index].name, LVM_OPTION_MOVE_EXTENT_PE_STR);
	SET_STRING(od->option[index].title, "Physical Extent");
	SET_STRING(od->option[index].tip, "Physical extent to move this logical extent to.");
	od->option[index].type = EVMS_Type_Unsigned_Int32;
	od->option[index].flags = EVMS_OPTION_FLAGS_NO_INITIAL_VALUE |
				  EVMS_OPTION_FLAGS_INACTIVE;
	od->option[index].constraint_type = EVMS_Collection_List;

	od->count = index + 1;
	RETURN(0);
}

/**
 * lvm_move_extent_set_option
 *
 * Verify and set an option during a MOVE_EXTENT task.
 **/
int lvm_move_extent_set_option(task_context_t * context,
			       u_int32_t index,
			       value_t * value,
			       task_effect_t * effect)
{
	option_desc_array_t * od = context->option_descriptors;
	lvm_logical_volume_t * volume = context->object->private_data;
	lvm_physical_volume_t * pv_entry;
	u_int32_t i, count = 0;
	int rc = EINVAL;

	LOG_ENTRY;
	LOG_EXTRA("Setting option %d\n", index);

	*effect = 0;

	switch (index) {

	case LVM_OPTION_MOVE_EXTENT_LE_INDEX:
		/* Make sure the LE is in range. */
		if (lvm_le_is_valid(volume, value->ui32) &&
		    !lvm_le_is_scheduled_for_move(&volume->le_map[value->ui32])) {
			od->option[index].value.ui32 = value->ui32;
			rc = 0;
		}
		break;

	case LVM_OPTION_MOVE_EXTENT_PV_INDEX:
		/* Make sure a valid PV was selected, and
		 * that the PV has unused PEs available.
		 */
		pv_entry = lvm_get_pv_for_name(value->s, volume->group);
		if (pv_entry && lvm_pv_has_available_extents(pv_entry)) {
			strncpy(od->option[index].value.s, value->s, EVMS_NAME_SIZE);

			/* Set up the list of possible PEs on this PV. */
			index = LVM_OPTION_MOVE_EXTENT_PE_INDEX;
			lvm_engine->engine_free(od->option[index].constraint.list);
			od->option[index].constraint.list = lvm_engine->engine_alloc(sizeof(value_list_t) +
										     sizeof(value_t) *
										     pv_entry->pv->pe_total);
			for (i = 0; i < pv_entry->pv->pe_total; i++) {
				if (lvm_pe_is_available(&pv_entry->pe_map[i])) {
					od->option[index].constraint.list->value[count].ui32 = i;
					count++;
				}
			}
			od->option[index].constraint.list->count = count;
			od->option[index].flags &= ~EVMS_OPTION_FLAGS_INACTIVE;
			od->option[index].flags |= EVMS_OPTION_FLAGS_NO_INITIAL_VALUE;

			*effect |= EVMS_Effect_Reload_Options;
			rc = 0;
		}
		break;

	case LVM_OPTION_MOVE_EXTENT_PE_INDEX:
		/* Get the selected PV from the option descriptor,
		 * and make sure the selected PE is available.
		 */
		pv_entry = lvm_get_pv_for_name(od->option[LVM_OPTION_MOVE_EXTENT_PV_INDEX].value.s,
					       volume->group);
		if (pv_entry &&
		    lvm_pe_is_valid(pv_entry, value->ui32) &&
		    lvm_pe_is_available(&pv_entry->pe_map[value->ui32])) {
			od->option[index].value.ui32 = value->ui32;
			rc = 0;
		}
		break;

	default:
		break;
	}

	RETURN(rc);
}

/**
 * lvm_move_extent_parse_options
 **/
static int lvm_move_extent_parse_options(option_array_t * options,
					 lvm_logical_volume_t * volume,
					 lvm_physical_volume_t ** pv_entry,
					 u_int32_t * le,
					 u_int32_t * pe)
{
	int i;

	LOG_ENTRY;

	for (i = 0; i < options->count; i++) {
		/* If only the name-based index is specified, get the number. */
		if (!options->option[i].is_number_based) {
			if (!strcmp(options->option[i].name, LVM_OPTION_MOVE_EXTENT_LE_STR)) {
				options->option[i].number = LVM_OPTION_MOVE_EXTENT_LE_INDEX;
			} else if (!strcmp(options->option[i].name, LVM_OPTION_MOVE_EXTENT_PV_STR)) {
				options->option[i].number = LVM_OPTION_MOVE_EXTENT_PV_INDEX;
			} else if (!strcmp(options->option[i].name, LVM_OPTION_MOVE_EXTENT_PE_STR)) {
				options->option[i].number = LVM_OPTION_MOVE_EXTENT_PE_INDEX;
			} else {
				continue;
			}
		}

		LOG_EXTRA("Parsing option %d\n", options->option[i].number);

		/* Use the number-based index to record the option. */
		switch (options->option[i].number) {
		case LVM_OPTION_MOVE_EXTENT_LE_INDEX:
			*le = options->option[i].value.ui32;
			break;
		case LVM_OPTION_MOVE_EXTENT_PV_INDEX:
			*pv_entry = lvm_get_pv_for_name(options->option[i].value.s,
							volume->group);
			break;
		case LVM_OPTION_MOVE_EXTENT_PE_INDEX:
			*pe = options->option[i].value.ui32;
			break;
		default:
			break;
		}
	}

	RETURN(0);
}

/**
 * lvm_move_extent_verify_options
 *
 * Verify that the specified options are all valid for moving an extent.
 **/
int lvm_move_extent_verify_options(lvm_logical_volume_t * volume,
				   lvm_physical_volume_t * pv_entry,
				   u_int32_t le,
				   u_int32_t pe)
{
	int rc = EINVAL;

	LOG_ENTRY;

	if (!lvm_le_is_valid(volume, le)) {
		LOG_ERROR("LE %d is out of range for region %s.\n",
			  le, volume->region->name);
	} else if (lvm_le_is_scheduled_for_move(&volume->le_map[le])) {
		LOG_ERROR("LE %d on region %s is already scheduled for move.\n",
			  le, volume->region->name);
	} else if (!pv_entry) {
		LOG_ERROR("Invalid PV specified.\n");
	} else if (!lvm_pe_is_valid(pv_entry, pe)) {
		LOG_ERROR("PE %d is out of range for PV %s.\n",
			  pe, pv_entry->segment->name);
	} else if (!lvm_pe_is_available(&pv_entry->pe_map[pe])) {
		LOG_ERROR("PE %d on PV %s is not available for move.\n",
			  pe, pv_entry->segment->name);
	} else {
		rc = 0;
	}

	RETURN(rc);
}

/**
 * lvm_move_extent
 **/
int lvm_move_extent(lvm_logical_volume_t * volume,
		    option_array_t * options)
{
	lvm_physical_volume_t * pv_entry = NULL;
	u_int32_t le_num = 0, pe_num = 0;
	char * choices[] = {"Continue",
			    "Continue All",
			    "Don't Continue",
			    NULL};
	int rc, answer = 0;

	LOG_ENTRY;

	/* Parse and verify the options for move. */
	lvm_move_extent_parse_options(options, volume,
				      &pv_entry, &le_num, &pe_num);
	rc = lvm_move_extent_verify_options(volume, pv_entry,
					    le_num, pe_num);
	if (rc) {
		LOG_ERROR("Error verifying move-extent options.\n");
		RETURN(rc);
	}

	if (!disable_move_message) {
		lvm_engine->user_message(lvm_plugin, &answer, choices,
					 "You have chosen to move LE %s:%d from "
					 "PE %s:%d to PE %s:%d. This move cannot "
					 "be performed safely while region %s is "
					 "in use. Please be sure %s is unmounted "
					 "and not in use before saving these "
					 "changes.",
					 volume->region->name, le_num,
					 volume->le_map[le_num].pe ? volume->le_map[le_num].pe->pv->segment->name : "missing-pv",
					 volume->le_map[le_num].pe ? volume->le_map[le_num].pe->number : 0, 
					 pv_entry->segment->name, pe_num,
					 volume->region->name,
					 volume->region->name);
		if (answer == 1) {
			disable_move_message = TRUE;
		} else if (answer == 2) {
			RETURN(E_CANCELED);
		}
	}

	/* Prepare the extents to be moved. */
	lvm_move_extent_setup(&volume->le_map[le_num], &pv_entry->pe_map[pe_num]);

	/* Rebuild the freespace. */
	rc = lvm_update_freespace_volume(volume->group);
	if (rc) {
		LOG_ERROR("Error updating freespace for container %s\n",
			  volume->group->container->name);
	}

	volume->group->container->flags |= SCFLAG_DIRTY;
	lvm_engine->set_changes_pending();

	RETURN(rc);
}


/********** Move-PV Task Functions **********/

/**
 * lvm_can_move_pv_stripes
 *
 * Helper-function for lvm_can_move_pv for determining whether striped-regions
 * on the specified PV can be moved.
 **/
static int lvm_can_move_pv_stripes(lvm_physical_volume_t * source_pv,
				   u_int32_t * free_extents,
				   int maintain_stripes)
{
	storage_object_t * region, * segment;
	lvm_logical_volume_t * volume;
	lvm_physical_volume_t * pv_entry;
	u_int32_t free_extents_striped[MAX_PV+1] = {0};
	u_int32_t extents_per_pv;
	int i, j, rc, rc2;

	LOG_ENTRY;

	DLIST_FOR_EACH(region, source_pv->segment->parent_objects, rc) {
		if (region->data_type != DATA_TYPE) {
			/* Don't care about freespace regions. */
			continue;
		}

		volume = region->private_data;
		if (volume->lv->lv_stripes <= 1) {
			/* Only concerned about striped regions. */
			continue;
		}

		/* Make sure this region hasn't already been moved. */
		if (volume->flags & LVM_LV_FLAG_MOVE_PENDING) {
			LOG_WARNING("Region %s has extents waiting to be "
				    "moved.\n", region->name);
			LOG_WARNING("Please save pending moves before "
				    "performing \"Move-PV\"\n");
			rc = EINVAL;
			goto out;
		}

		/* Make a local copy of the free-extents array, since
		 * we're going to mangle it.
		 */
		memcpy(free_extents_striped, free_extents,
		       sizeof(free_extents_striped));

		if (maintain_stripes != MAINTAIN_STRIPES_OFF) {
			/* If we are trying to maintain striping, zero the
			 * entries for the PVs that this region already maps
			 * to, so we don't consider them.
			 */
			DLIST_FOR_EACH(segment, region->child_objects, rc2) {
				pv_entry = lvm_get_pv_for_segment(segment);
				free_extents_striped[pv_entry->number] = 0;
			}
		}

		if (maintain_stripes == MAINTAIN_STRIPES_STRICT) {
			/* Count the number of extents on this
			 * region that map to this PV.
			 */
			extents_per_pv = 0;
			for (i = 0; i < volume->lv->lv_allocated_le; i++) {
				if (volume->le_map[i].pe &&
				    volume->le_map[i].pe->pv == source_pv) {
					extents_per_pv++;
				}
			}

			/* Search for a target PV with this many free PEs. */
			for (i = 0; i <= MAX_PV; i++) {
				if (free_extents_striped[i] >= extents_per_pv) {
					free_extents_striped[i] -= extents_per_pv;
					free_extents[i] -= extents_per_pv;
					break;
				}
			}
			if (i > MAX_PV) {
				/* Couldn't find a free PE for this LE. */
				LOG_WARNING("Not enough free PEs to move "
					    "region %s.\n", region->name);
				rc = EINVAL;
				goto out;
			}
		} else {
			for (i = 0, j = 0; i < volume->lv->lv_allocated_le; i++) {
				if (volume->le_map[i].pe &&
				    volume->le_map[i].pe->pv == source_pv) {
					/* This LE needs to be moved.
					 * Find a free PE.
					 */
					for ( ; j <= MAX_PV; j++) {
						if (free_extents_striped[j]) {
							free_extents_striped[j]--;
							free_extents[j]--;
							break;
						}
					}
					if (j > MAX_PV) {
						/* No free PE for this LE. */
						LOG_WARNING("Not enough free PEs to "
							    "move region %s.\n",
							    region->name);
						rc = EINVAL;
						goto out;
					}
				}
			}
		}
	}

	rc = 0;

out:
	RETURN(rc);
}

/**
 * lvm_can_move_pv_linears
 *
 * Helper-function for lvm_can_move_pv for determining whether linear-regions
 * on the specified PV can be moved.
 **/
static int lvm_can_move_pv_linears(lvm_physical_volume_t * source_pv,
				   u_int32_t * free_extents)
{
	storage_object_t * region;
	lvm_logical_volume_t * volume;
	int rc, i, j = 0;

	LOG_ENTRY;

	DLIST_FOR_EACH(region, source_pv->segment->parent_objects, rc) {
		if (region->data_type != DATA_TYPE) {
			/* Don't care about freespace regions. */
			continue;
		}

		volume = region->private_data;
		if (volume->lv->lv_stripes > 1) {
			/* Only concerned about linear regions. */
			continue;
		}

		/* Make sure this region hasn't already been moved. */
		if (volume->flags & LVM_LV_FLAG_MOVE_PENDING) {
			LOG_WARNING("Region %s has extents waiting to be "
				    "moved.\n", region->name);
			LOG_WARNING("Please save pending moves before "
				    "performing \"Move-PV\"\n");
			rc = EINVAL;
			goto out;
		}

		for (i = 0; i < volume->lv->lv_allocated_le; i++) {
			if (volume->le_map[i].pe &&
			    volume->le_map[i].pe->pv == source_pv) {
				/* This LE needs to be moved. Find a free PE. */
				for ( ; j <= MAX_PV; j++) {
					if (free_extents[j]) {
						free_extents[j]--;
						break;
					}
				}
				if (j > MAX_PV) {
					/* No free PEs for this LE. */
					LOG_WARNING("Not enough free PEs to "
						    "move region %s.\n",
						    region->name);
					rc = EINVAL;
					goto out;
				}
			}
		}
	}

	rc = 0;

out:
	RETURN(rc);
}

/**
 * lvm_can_move_pv
 * @source_pv:		The PV from which all PEs should be moved.
 * @target_pvs:		List of PVs to use in considering how to move source_pv.
 *			If the first entry is NULL, use all PVs in the group.
 *			If the first entry is non-NULL, then this argument must
 *			point to an array of MAX_PV+1 entries (not all non-NULL).
 * @maintain_stripes:	If 0, don't bother maintaining striped regions. This
 *			means a striped region may wind up with multiple stripes
 *			on the same PV, but allows the most flexibility in
 *			moving the desired PV.
 *			If 1, maintain stripes "loosely". This will ensure that
 *			moved extents don't wind up on PVs that the striped
 *			region already uses. However, the moved extents may not
 *			all end up on the same PV.
 *			If 2, maintain stripes "strictly". This will ensure that
 *			all moved extents end up on the same PV, thus
 *			maintaining the same stripe configuration. This option
 *			is the most restrictive, and may not always allow the PV
 *			to be moved.
 *
 * Can all of the extents on this PV be moved to other PVs in the container?
 * This is only a check to see if the move is possible. No extents are actually
 * setup for a real move.
 **/
int lvm_can_move_pv(lvm_physical_volume_t * source_pv,
		    lvm_physical_volume_t * target_pvs[],
		    int maintain_stripes)
{
	lvm_volume_group_t * group = source_pv->group;
	u_int32_t free_extents[MAX_PV+1] = {0};
	int i, rc;

	LOG_ENTRY;

	/* There is nothing to move if the source PV has no allocated PEs. */
	if (source_pv->pv->pe_allocated == 0) {
		LOG_DETAILS("No extents allocated on PV %s.\n",
			    source_pv->segment->name);
		rc = EINVAL;
		goto out;
	}

	/* If no target PVs were specified, use all PVs in the group. */
	if (!target_pvs[0]) {
		target_pvs = group->pv_list;
	}

	/* Set up an array with the number of free-extents per target-PV. */
	for (i = 0; i <= MAX_PV; i++) {
		if (target_pvs[i] && target_pvs[i] != source_pv) {
			free_extents[target_pvs[i]->number] = lvm_pv_num_available_extents(target_pvs[i]);
		}
	}

	/* Check striped regions first. */
	rc = lvm_can_move_pv_stripes(source_pv, free_extents, maintain_stripes);
	if (rc) {
		goto out;
	}

	/* Now check linear regions. */
	rc = lvm_can_move_pv_linears(source_pv, free_extents);
	if (rc) {
		goto out;
	}

out:
	RETURN(rc);
}

/**
 * lvm_parse_maintain_stripes
 **/
static int lvm_parse_maintain_stripes(char * string)
{
	int rc = MAINTAIN_STRIPES_OFF;

	LOG_ENTRY;

	if (strcasecmp(string, "strict") == 0) {
		rc = MAINTAIN_STRIPES_STRICT;
	} else if (strcasecmp(string, "loose") == 0) {
		rc = MAINTAIN_STRIPES_LOOSE;
	}

	RETURN(rc);
}

/**
 * lvm_move_pv_init_task
 *
 * Initialize the option descriptor for a MOVE_PV task.
 * Option 0: List of Target PVs
 * Option 1: Flag to maintain striped regions
 **/
int lvm_move_pv_init_task(task_context_t * context)
{
	option_desc_array_t * od = context->option_descriptors;
	lvm_volume_group_t * group = context->container->private_data;
	lvm_physical_volume_t * dummy_pv = NULL;
	int i, rc, index;

	LOG_ENTRY;

	/* Populate the acceptable-objects list with all PVs that could
	 * be moved.
	 */
	for (i = 0; i <= MAX_PV; i++) {
		if (group->pv_list[i]) {
			rc = lvm_can_move_pv(group->pv_list[i], &dummy_pv,
					     MAINTAIN_STRIPES_OFF);
			if (!rc) {
				lvm_add_object_to_list(group->pv_list[i]->segment,
						       context->acceptable_objects);
			}
		}
	}

	if (ListEmpty(context->acceptable_objects)) {
		/* No PVs can be moved. */
		RETURN(EINVAL);
	}

	context->min_selected_objects = 1;
	context->max_selected_objects = 1;

	/* Setup the option descriptor. */

	/* Option 0: List of Target PVs
	 * We don't know which PVs can be targets until we have selected a
	 * source PV. Just allocate the empty list right now.
	 */
	index = LVM_OPTION_MOVE_PV_TARGET_LIST_IDX;
	SET_STRING(od->option[index].name, LVM_OPTION_MOVE_PV_TARGET_LIST_STR);
	SET_STRING(od->option[index].title, "Target PVs");
	SET_STRING(od->option[index].tip, "List of PVs that should be used as "
					  "targets for moving the desired source "
					  "PV. Defaults to all remaining PVs in "
					  "the container.");
	od->option[index].type = EVMS_Type_String;
	od->option[index].min_len = 1;
	od->option[index].max_len = EVMS_NAME_SIZE;
	od->option[index].flags = EVMS_OPTION_FLAGS_NOT_REQUIRED |
				  EVMS_OPTION_FLAGS_AUTOMATIC |
				  EVMS_OPTION_FLAGS_VALUE_IS_LIST;
	od->option[index].constraint_type = EVMS_Collection_List;
	od->option[index].constraint.list = lvm_engine->engine_alloc(sizeof(value_list_t) +
								     sizeof(value_t) * MAX_PV);
	od->option[index].value.list = lvm_engine->engine_alloc(sizeof(value_list_t) +
								sizeof(value_t) * MAX_PV);
	/* Don't need to allocate space for the strings yet. That will
	 * happen when one of the acceptable-objects is selected.
	 */


	/* Option 1: Maintain Striping
	 * No, Loose, or Strict
	 */
	index = LVM_OPTION_MOVE_PV_MAINTAIN_STRIPES_IDX;
	SET_STRING(od->option[index].name, LVM_OPTION_MOVE_PV_MAINTAIN_STRIPES_STR);
	SET_STRING(od->option[index].title, "Maintain Stripes");
	SET_STRING(od->option[index].tip, "No: Don't bother to maintain true striping.\n"
					  "Loose: Don't move extents to PVs that are already "
					  "used by a striped region.\n"
					  "Strict: Maintain true striping.");
	od->option[index].type = EVMS_Type_String;
	od->option[index].min_len = 1;
	od->option[index].max_len = EVMS_NAME_SIZE;
	od->option[index].flags = EVMS_OPTION_FLAGS_NOT_REQUIRED |
				  EVMS_OPTION_FLAGS_AUTOMATIC;
	od->option[index].constraint_type = EVMS_Collection_List;
	od->option[index].constraint.list = lvm_engine->engine_alloc(sizeof(value_list_t) +
								     sizeof(value_t) * 2);
	SET_STRING(od->option[index].constraint.list->value[0].s, "no");
	SET_STRING(od->option[index].constraint.list->value[1].s, "loose");
	SET_STRING(od->option[index].constraint.list->value[2].s, "strict");
	od->option[index].constraint.list->count = 3;
	SET_STRING(od->option[index].value.s, "no");

	index++;
	od->count = index;

	RETURN(0);
}

/**
 * lvm_move_pv_set_objects
 *
 * Verify that the selected PV is valid for Move-PV. If it is, finish
 * initializing the option-descriptor.
 **/
int lvm_move_pv_set_objects(task_context_t * context,
			    dlist_t declined_objects,
			    task_effect_t * effect)
{
	option_desc_array_t * od = context->option_descriptors;
	lvm_volume_group_t * group = context->container->private_data;
	lvm_physical_volume_t * source_pv, * dummy_pv = NULL;
	int rc, i, j = 0;

	LOG_ENTRY;

	/* Get the PV from the selected-objects list. */
	source_pv = lvm_get_selected_segment(context->selected_objects);
	if (!source_pv) {
		rc = EINVAL;
		goto out;
	}

	/* Make sure this PV can be moved. */
	rc = lvm_can_move_pv(source_pv, &dummy_pv, MAINTAIN_STRIPES_OFF);
	if (rc) {
		goto out;
	}

	/* Fill in the initial Target-PVs list. */
	od->option[LVM_OPTION_MOVE_PV_TARGET_LIST_IDX].constraint.list->count = 0;
	od->option[LVM_OPTION_MOVE_PV_TARGET_LIST_IDX].value.list->count = 0;
	for (i = 0; i <= MAX_PV; i++) {
		if (group->pv_list[i] &&
		    group->pv_list[i] != source_pv) {
			SET_STRING(od->option[LVM_OPTION_MOVE_PV_TARGET_LIST_IDX].constraint.list->value[j].s,
				   group->pv_list[i]->segment->name);
			SET_STRING(od->option[LVM_OPTION_MOVE_PV_TARGET_LIST_IDX].value.list->value[j].s,
				   group->pv_list[i]->segment->name);
			od->option[LVM_OPTION_MOVE_PV_TARGET_LIST_IDX].constraint.list->count++;
			od->option[LVM_OPTION_MOVE_PV_TARGET_LIST_IDX].value.list->count++;
			j++;
		}
	}

	/* Reset the Maintain-Stripes option to the default. */
	SET_STRING(od->option[LVM_OPTION_MOVE_PV_MAINTAIN_STRIPES_IDX].value.s, "no");

	*effect |= EVMS_Effect_Reload_Options;

out:
	RETURN(rc);
}

/**
 * lvm_move_pv_set_option
 *
 * Verify that the specified option, along with the currently-selected
 * object and the remaining options present in the option-descriptor, are
 * valid for performing a PV-Move.
 **/
int lvm_move_pv_set_option(task_context_t * context,
			   u_int32_t index,
			   value_t * value,
			   task_effect_t * effect)
{
	option_desc_array_t * od = context->option_descriptors;
	lvm_volume_group_t * group = context->container->private_data;
	lvm_physical_volume_t * source_pv;
	lvm_physical_volume_t * target_pvs[MAX_PV+1] = {NULL};
	value_list_t * pv_list;
	char * maintain_stripes_string;
	int maintain_stripes;
	int i, rc;

	LOG_ENTRY;

	LOG_EXTRA("Setting option %d\n", index);

	/* Get the PV from the selected-objects list. */
	source_pv = lvm_get_selected_segment(context->selected_objects);
	if (!source_pv) {
		rc = EINVAL;
		goto out;
	}

	switch (index) {

	case LVM_OPTION_MOVE_PV_TARGET_LIST_IDX:
		pv_list = value->list;
		maintain_stripes_string = od->option[LVM_OPTION_MOVE_PV_MAINTAIN_STRIPES_IDX].value.s;
		break;

	case LVM_OPTION_MOVE_PV_MAINTAIN_STRIPES_IDX:
		pv_list = od->option[LVM_OPTION_MOVE_PV_TARGET_LIST_IDX].value.list;
		maintain_stripes_string = value->s;
		break;

	default:
		rc = EINVAL;
		goto out;
	}

	rc = lvm_parse_pv_list_option(pv_list, target_pvs, group);
	if (rc) {
		goto out;
	}

	maintain_stripes = lvm_parse_maintain_stripes(maintain_stripes_string);

	rc = lvm_can_move_pv(source_pv, target_pvs, maintain_stripes);
	if (rc) {
		*effect |= EVMS_Effect_Reload_Options;
		goto out;
	}

	switch (index) {

	case LVM_OPTION_MOVE_PV_TARGET_LIST_IDX:
		for (i = 0; i < pv_list->count; i++) {
			SET_STRING(od->option[index].value.list->value[i].s, pv_list->value[i].s);
		}
		for (; i <= MAX_PV; i++) {
			lvm_engine->engine_free(od->option[index].value.list->value[i].s);
			od->option[index].value.list->value[i].s = NULL;
		}
		od->option[index].value.list->count = pv_list->count;
		break;

	case LVM_OPTION_MOVE_PV_MAINTAIN_STRIPES_IDX:
		SET_STRING(od->option[index].value.s, maintain_stripes_string);
		break;

	default:
		break;
	}

	rc = 0;

out:
	RETURN(rc);
}

/**
 * lvm_move_pv_parse_options
 **/
static int lvm_move_pv_parse_options(option_array_t * options,
				     lvm_physical_volume_t * pv_entries[],
				     lvm_volume_group_t * group,
				     int * maintain_stripes)
{
	int i, rc = 0;

	LOG_ENTRY;

	/* Default value. */
	*maintain_stripes = MAINTAIN_STRIPES_OFF;

	for (i = 0; i < options->count; i++) {
		/* If only the name-based index is specified, get the number. */
		if (!options->option[i].is_number_based) {
			if (!strcmp(options->option[i].name, LVM_OPTION_MOVE_PV_TARGET_LIST_STR)) {
				options->option[i].number = LVM_OPTION_MOVE_PV_TARGET_LIST_IDX;
			}
			else if (!strcmp(options->option[i].name, LVM_OPTION_MOVE_PV_MAINTAIN_STRIPES_STR)) {
				options->option[i].number = LVM_OPTION_MOVE_PV_MAINTAIN_STRIPES_IDX;
			}
			else {
				continue;
			}
		}

		LOG_EXTRA("Parsing option %d\n", options->option[i].number);

		/* Use the number-based index to record the option. */
		switch (options->option[i].number) {
		case LVM_OPTION_MOVE_PV_TARGET_LIST_IDX:
			rc = lvm_parse_pv_list_option(options->option[i].value.list,
						      pv_entries, group);
			if (rc) {
				goto out;
			}
			break;
		case LVM_OPTION_MOVE_PV_MAINTAIN_STRIPES_IDX:
			*maintain_stripes = lvm_parse_maintain_stripes(options->option[i].value.s);
			break;
		default:
			break;
		}
	}

out:
	RETURN(rc);
}

/**
 * lvm_move_pv_display_warning
 **/
static int lvm_move_pv_display_warning(lvm_physical_volume_t * source_pv)
{
	storage_object_t * region;
	lvm_logical_volume_t * volume;
	char * volume_string;
	char * choices[] = {"Continue",
			    "Don't Continue",
			    NULL};
	int rc, rc2, count, answer = 0;

	LOG_ENTRY;

	GetListSize(source_pv->segment->parent_objects, &count);

	volume_string = lvm_engine->engine_alloc(count*(EVMS_NAME_SIZE+1) + 1);
	if (!volume_string) {
		RETURN(ENOMEM);
	}

	count = 0;

	DLIST_FOR_EACH(region, source_pv->segment->parent_objects, rc) {
		if (region->data_type == DATA_TYPE) {
			volume = region->private_data;
			rc2 = lvm_volume_is_busy(volume, FALSE);
			if (rc2) {
				count++;
				if (region->volume) {
					strncat(volume_string, region->volume->name, EVMS_NAME_SIZE);
				} else {
					strncat(volume_string, region->name, EVMS_NAME_SIZE);
				}
				strncat(volume_string, "\n", 1);
			}
		}
	}

	if (count) {
		lvm_engine->user_message(lvm_plugin, &answer, choices,
				 "You have chosen to move PV %s. Currently, "
				 "all move operations must be performed "
				 "off-line. The following volumes and/or "
				 "regions must be unmounted before saving "
				 "these changes:\n\n"
				 "%s",
				 source_pv->segment->name, volume_string);
	}

	rc = answer ? E_CANCELED : 0;

	lvm_engine->engine_free(volume_string);

	RETURN(rc);
}

/**
 * lvm_move_pv_stripes
 *
 * Move all extents (belonging to striped regions) from the source PV to
 * PEs on PVs in the target list.
 **/
static int lvm_move_pv_stripes(lvm_physical_volume_t * source_pv,
			       lvm_physical_volume_t * target_pvs[],
			       int maintain_stripes)
{
	storage_object_t * region, * segment;
	lvm_logical_volume_t * volume;
	lvm_physical_volume_t * target_pvs_striped[MAX_PV+1] = {NULL};
	lvm_physical_volume_t * pv_entry;
	int extents_per_pv;
	int i, j = 0, k = 0;
	int rc, rc2;

	LOG_ENTRY;

	/* If no target PVs were specified, use the group's whole list. */
	if (!target_pvs[0]) {
		target_pvs = source_pv->group->pv_list;
	}

	/* Check every region that maps to the source PV. */
	DLIST_FOR_EACH(region, source_pv->segment->parent_objects, rc) {
		if (region->data_type != DATA_TYPE) {
			/* Don't care about freespace regions. */
			continue;
		}

		volume = region->private_data;
		if (volume->lv->lv_stripes <= 1) {
			/* Only move LEs on striped regions. */
			continue;
		}

		/* Need a local copy of the target PV array, since we're
		 * going to mangle it.
		 */
		memcpy(target_pvs_striped, target_pvs,
		       sizeof(target_pvs_striped));

		if (maintain_stripes != MAINTAIN_STRIPES_OFF) {
			/* It we are maintaining any form of striping, we need
			 * to prune the list of target PVs.
			 */
			DLIST_FOR_EACH(segment, region->child_objects, rc2) {
				pv_entry = lvm_get_pv_for_segment(segment);
				for (i = 0; i <= MAX_PV; i++) {
					if (target_pvs_striped[i] == pv_entry) {
						target_pvs_striped[i] = NULL;
						break;
					}
				}
			}
		}

		if (maintain_stripes == MAINTAIN_STRIPES_STRICT) {
			/* Count the number of extents on this region
			 * that map to this PV.
			 */
			extents_per_pv = 0;
			for (i = 0; i < volume->lv->lv_allocated_le; i++) {
				if (volume->le_map[i].pe &&
				    volume->le_map[i].pe->pv == source_pv) {
					extents_per_pv++;
				}
			}

			/* Search for a target PV with this many free
			 * extents.
			 */
			for (i = 0; i <= MAX_PV; i++) {
				if (!target_pvs_striped[i] ||
				    target_pvs_striped[i] == source_pv ||
				    lvm_pv_num_available_extents(target_pvs_striped[i]) < extents_per_pv) {
					continue;
				}

				/* This PV has enough free extents. Search its
				 * PE map and setup the moves.
				 */
				for (j = 0, k = 0; j < volume->lv->lv_allocated_le; j++) {
					if (!volume->le_map[j].pe ||
					    volume->le_map[j].pe->pv != source_pv) {
						continue;
					}

					/* Find an available PE on this target PV. */
					for ( ; k < target_pvs_striped[i]->pv->pe_total; k++) {
						if (lvm_pe_is_available(&(target_pvs_striped[i]->pe_map[k]))) {
							break;
						}
					}

					if (k < target_pvs_striped[i]->pv->pe_total) {
						lvm_move_extent_setup(&(volume->le_map[j]),
								      &(target_pvs_striped[i]->pe_map[k]));
					} else {
						/* Wasn't able to find a free PE! Based
						 * on the above checks, this should
						 * really never happen.
						 */
						LOG_SERIOUS("BUG! Error finding free PEs "
							    "for region %s.\n", region->name);
						RETURN(EINVAL);
					}
				}

				break;
			}

			if (i > MAX_PV) {
				/* Wasn't able to find a PV with free PEs! If we
				 * checked everything ahead of time, this should
				 * not happen.
				 */
				LOG_SERIOUS("BUG! Error finding PV with free PEs "
					    "for region %s.\n", region->name);
				RETURN(EINVAL);
			}

		} else {
			/* MAINTAIN_STRIPES_LOOSE or MAINTAIN_STRIPES_OFF */

			/* Check every LE on this region. */
			for (i = 0; i < volume->lv->lv_allocated_le; i++) {
				if (!volume->le_map[i].pe ||
				    volume->le_map[i].pe->pv != source_pv) {
					/* Only move LEs that map to the source PV. */
					continue;
				}

				/* This LE needs to be moved. */
				for ( ; j <= MAX_PV; j++, k = 0) {
					if (!target_pvs_striped[j] ||
					    target_pvs_striped[j] == source_pv ||
					    !lvm_pv_has_available_extents(target_pvs_striped[j])) {
						/* Can't use this PV as the target. */
						continue;
					}

					/* Find an available PE on this target PV. */
					for ( ; k < target_pvs_striped[j]->pv->pe_total; k++) {
						if (lvm_pe_is_available(&(target_pvs_striped[j]->pe_map[k]))) {
							break;
						}
					}

					if (k < target_pvs_striped[j]->pv->pe_total) {
						lvm_move_extent_setup(&(volume->le_map[i]),
								      &(target_pvs_striped[j]->pe_map[k]));
						break;
					} else {
						/* Wasn't able to find a free PE! Based
						 * on the above checks, this should
						 * really never happen.
						 */
						LOG_SERIOUS("BUG! Error finding free PEs "
							    "for region %s.\n", region->name);
						RETURN(EINVAL);
					}
				}

				if (j > MAX_PV) {
					/* Wasn't able to find a PV with free PEs! If we
					 * checked everything ahead of time, this should
					 * not happen.
					 */
					LOG_SERIOUS("BUG! Error finding PV with free PEs "
						    "for region %s.\n", region->name);
					RETURN(EINVAL);
				}
			}
		}
	}

	RETURN(0);
}

/**
 * lvm_move_pv_linears
 *
 * Move all extents (belonging to linear regions) from the source PV to
 * PEs on PVs in the target list.
 **/
static int lvm_move_pv_linears(lvm_physical_volume_t * source_pv,
			       lvm_physical_volume_t * target_pvs[])
{
	storage_object_t * region;
	lvm_logical_volume_t * volume;
	int i, j = 0, k = 0, rc;

	LOG_ENTRY;

	/* If no target PVs were specified, use the group's whole list. */
	if (!target_pvs[0]) {
		target_pvs = source_pv->group->pv_list;
	}

	/* Check every region that maps to the source PV. */
	DLIST_FOR_EACH(region, source_pv->segment->parent_objects, rc) {
		if (region->data_type != DATA_TYPE) {
			/* Don't care about freespace regions. */
			continue;
		}

		volume = region->private_data;
		if (volume->lv->lv_stripes > 1) {
			/* Only move LEs on linear regions. */
			continue;
		}

		/* Check every LE on this region. */
		for (i = 0; i < volume->lv->lv_allocated_le; i++) {
			if (!volume->le_map[i].pe ||
			    volume->le_map[i].pe->pv != source_pv) {
				/* Only move LEs that map to the source PV. */
				continue;
			}

			/* This LE needs to be moved. */
			for ( ; j <= MAX_PV; j++, k = 0) {
				if (!target_pvs[j] ||
				    target_pvs[j] == source_pv ||
				    !lvm_pv_has_available_extents(target_pvs[j])) {
					/* Can't use this PV as the target. */
					continue;
				}

				/* Find an available PE on this target PV. */
				for ( ; k < target_pvs[j]->pv->pe_total; k++) {
					if (lvm_pe_is_available(&(target_pvs[j]->pe_map[k]))) {
						break;
					}
				}

				if (k < target_pvs[j]->pv->pe_total) {
					lvm_move_extent_setup(&(volume->le_map[i]),
							      &(target_pvs[j]->pe_map[k]));
					break;
				} else {
					/* Wasn't able to find a free PE! Based
					 * on the above checks, this should
					 * really never happen.
					 */
					LOG_SERIOUS("BUG! Error finding free PEs "
						    "for region %s.\n", region->name);
					RETURN(EINVAL);
				}
			}

			if (j > MAX_PV) {
				/* Wasn't able to find a PV with free PEs! If we
				 * checked everything ahead of time, this should
				 * not happen.
				 */
				LOG_SERIOUS("BUG! Error finding PV with free PEs "
					    "for region %s.\n", region->name);
				RETURN(EINVAL);
			}
		}
	}

	RETURN(0);
}

/**
 * lvm_move_pv
 *
 * Move all extents from the selected PV to remaining PVs in the container.
 **/
int lvm_move_pv(lvm_volume_group_t * group,
		dlist_t objects,
		option_array_t * options)
{
	lvm_physical_volume_t * source_pv;
	lvm_physical_volume_t * target_pvs[MAX_PV+1] = {NULL};
	int maintain_stripes, rc;

	LOG_ENTRY;

	/* Get the PV from the selected-objects list. */
	source_pv = lvm_get_selected_segment(objects);
	if (!source_pv) {
		rc = EINVAL;
		goto out;
	}

	/* Parse the option array. */
	rc = lvm_move_pv_parse_options(options, target_pvs,
				       group, &maintain_stripes);
	if (rc) {
		goto out;
	}

	/* Test that the move is possible. */
	rc = lvm_can_move_pv(source_pv, target_pvs, maintain_stripes);
	if (rc) {
		goto out;
	}

	/* Display warning message */
	rc = lvm_move_pv_display_warning(source_pv);
	if (rc) {
		goto out;
	}

	/* Move striped regions first. */
	rc = lvm_move_pv_stripes(source_pv, target_pvs, maintain_stripes);
	if (rc) {
		goto out;
	}

	/* Now move linear regions. */
	rc = lvm_move_pv_linears(source_pv, target_pvs);
	if (rc) {
		goto out;
	}

	/* Rebuild the freespace. */
	rc = lvm_update_freespace_volume(group);
	if (rc) {
		LOG_ERROR("Error updating freespace for container %s\n",
			  group->container->name);
	}

	group->container->flags |= SCFLAG_DIRTY;
	lvm_engine->set_changes_pending();

out:
	RETURN(rc);
}

