/*
 *   Copyright (c) International Business Machines  Corp., 2001
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: mdregmgr
 * File: mdregmgr.h
 */

#ifndef _MD_REGION_MANAGER_H_
#define _MD_REGION_MANAGER_H_ 1

/*
 * RAID superblock.
 *
 * The RAID superblock maintains some statistics on each RAID configuration.
 * Each real device in the RAID set contains it near the end of the device.
 * Some of the ideas are copied from the ext2fs implementation.
 *
 * We currently use 4096 bytes as follows:
 *
 *	word offset	function
 *
 *	   0  -    31	Constant generic RAID device information.
 *        32  -    63   Generic state information.
 *	  64  -   127	Personality specific information.
 *	 128  -   511	12 32-words descriptors of the disks in the raid set.
 *	 512  -   911	Reserved.
 *	 912  -  1023	Disk specific descriptor.
 */

/*
 * If x is the real device size in bytes, we return an apparent size of:
 *
 *	y = (x & ~(MD_RESERVED_BYTES - 1)) - MD_RESERVED_BYTES
 *
 * and place the 4kB superblock at offset y.
 */
#define MD_RESERVED_BYTES		(64 * 1024)
#define MD_RESERVED_SECTORS		(MD_RESERVED_BYTES / 512)
#define MD_RESERVED_BLOCKS		(MD_RESERVED_BYTES / BLOCK_SIZE)

#define MD_NEW_SIZE_SECTORS(x)		((x & ~(MD_RESERVED_SECTORS - 1)) - MD_RESERVED_SECTORS)
#define MD_NEW_SIZE_BLOCKS(x)		((x & ~(MD_RESERVED_BLOCKS - 1)) - MD_RESERVED_BLOCKS)

#define MD_SB_BYTES			4096
#define MD_SB_WORDS			(MD_SB_BYTES / 4)
#define MD_SB_BLOCKS			(MD_SB_BYTES / BLOCK_SIZE)
#define MD_SB_SECTORS			(MD_SB_BYTES / 512)

/*
 * The following are counted in 32-bit words
 */
#define	MD_SB_GENERIC_OFFSET		0
#define MD_SB_PERSONALITY_OFFSET	64
#define MD_SB_DISKS_OFFSET		128
#define MD_SB_DESCRIPTOR_OFFSET		992

#define MD_SB_GENERIC_CONSTANT_WORDS	32
#define MD_SB_GENERIC_STATE_WORDS	32
#define MD_SB_GENERIC_WORDS		(MD_SB_GENERIC_CONSTANT_WORDS + MD_SB_GENERIC_STATE_WORDS)
#define MD_SB_PERSONALITY_WORDS		64
#define MD_SB_DESCRIPTOR_WORDS		32
#define MD_SB_DISKS			27
#define MD_SB_DISKS_WORDS		(MD_SB_DISKS*MD_SB_DESCRIPTOR_WORDS)
#define MD_SB_RESERVED_WORDS		(1024 - MD_SB_GENERIC_WORDS - MD_SB_PERSONALITY_WORDS - MD_SB_DISKS_WORDS - MD_SB_DESCRIPTOR_WORDS)
#define MD_SB_EQUAL_WORDS		(MD_SB_GENERIC_WORDS + MD_SB_PERSONALITY_WORDS + MD_SB_DISKS_WORDS)


/*
 * Device "operational" state bits
 */
#define MD_DISK_FAULTY		0 /* disk is faulty / operational */
#define MD_DISK_ACTIVE		1 /* disk is running or spare disk */
#define MD_DISK_SYNC		2 /* disk is in sync with the raid set */
#define MD_DISK_REMOVED		3 /* disk has kind of been removed, but not really or it would not be here */
#define MD_DISK_NEW		4 /* disk has just been added to the raid set */
#define MD_DISK_PENDING_ACTIVE	5 /* disk was spare, but should be activated */

typedef struct mdp_device_descriptor_s {
	u_int32_t number;	/* 0 Device number in the entire set	      */
	u_int32_t major;	/* 1 Device major number		      */
	u_int32_t minor;	/* 2 Device minor number		      */
	u_int32_t raid_disk;	/* 3 The role of the device in the raid set   */
	u_int32_t state;	/* 4 Operational state			      */
	u_int32_t reserved[MD_SB_DESCRIPTOR_WORDS - 5];
} mdp_disk_t;

static inline int descriptor_removed(mdp_disk_t *disk) {
	if ( (disk->major == 0) &&  (disk->minor == 0) &&
	     (disk->state & (1 << MD_DISK_REMOVED)))
		return 1;
	else
		return 0;

}

#define MD_SB_MAGIC		0xa92b4efc

/*
 * Superblock state bits
 */
#define MD_SB_CLEAN		0
#define MD_SB_ERRORS		1

typedef struct mdp_superblock_s {
	/*
	 * Constant generic information
	 */
	u_int32_t md_magic;		/*  0 MD identifier 			      */
	u_int32_t major_version;	/*  1 major version to which the set conforms */
	u_int32_t minor_version;	/*  2 minor version ...			      */
	u_int32_t patch_version;	/*  3 patchlevel version ...		      */
	u_int32_t gvalid_words;		/*  4 Number of used words in this section    */
	u_int32_t set_uuid0;		/*  5 Raid set identifier		      */
	u_int32_t ctime;		/*  6 Creation time			      */
	u_int32_t level;		/*  7 Raid personality			      */
	u_int32_t size;			/*  8 Apparent size of each individual disk   */
	u_int32_t nr_disks;		/*  9 total disks in the raid set	      */
	u_int32_t raid_disks;		/* 10 disks in a fully functional raid set    */
	u_int32_t md_minor;		/* 11 preferred MD minor device number	      */
	u_int32_t not_persistent;	/* 12 does it have a persistent superblock    */
	u_int32_t set_uuid1;		/* 13 Raid set identifier #2		      */
	u_int32_t set_uuid2;		/* 14 Raid set identifier #3		      */
	u_int32_t set_uuid3;		/* 15 Raid set identifier #4		      */
	u_int32_t gstate_creserved[MD_SB_GENERIC_CONSTANT_WORDS - 16];

	/*
	 * Generic state information
	 */
	u_int32_t utime;		/*  0 Superblock update time		      */
	u_int32_t state;		/*  1 State bits (clean, ...)		      */
	u_int32_t active_disks;		/*  2 Number of currently active disks	      */
	u_int32_t working_disks;	/*  3 Number of working disks		      */
	u_int32_t failed_disks;		/*  4 Number of failed disks		      */
	u_int32_t spare_disks;		/*  5 Number of spare disks		      */
	u_int32_t sb_csum;		/*  6 checksum of the whole superblock        */
#if __BYTE_ORDER == __BIG_ENDIAN
	u_int32_t events_hi;		/*  7 high-order of superblock update count   */
	u_int32_t events_lo;		/*  8 low-order of superblock update count    */
#else
	u_int32_t events_lo;		/*  7 low-order of superblock update count    */
	u_int32_t events_hi;		/*  8 high-order of superblock update count   */
#endif
	u_int32_t gstate_sreserved[MD_SB_GENERIC_STATE_WORDS - 9];

	/*
	 * Personality information
	 */
	u_int32_t layout;		/*  0 the array's physical layout	      */
	u_int32_t chunk_size;		/*  1 chunk size in bytes		      */
	u_int32_t root_pv;		/*  2 LV root PV */
	u_int32_t root_block;		/*  3 LV root block */
	u_int32_t pstate_reserved[MD_SB_PERSONALITY_WORDS - 4];

	/*
	 * Disks information
	 */
	mdp_disk_t disks[MD_SB_DISKS];

	/*
	 * Reserved
	 */
	u_int32_t reserved[MD_SB_RESERVED_WORDS];

	/*
	 * Active descriptor
	 */
	mdp_disk_t this_disk;

}mdp_super_t;

#define MD_SAVED_SB_INFO_BYTES		1024
#define MD_SAVED_SB_INFO_WORDS		(MD_SAVED_SB_INFO_BYTES / 4)
#define MD_SAVED_SB_INFO_SECTS		(MD_SAVED_SB_INFO_BYTES / 512)
#define MD_SAVED_SB_SECTOR_OFFSET	(MD_RESERVED_SECTORS - MD_SAVED_SB_INFO_SECTS)

typedef struct mdp_saved_super_s {
	/*
	 * Constant generic information
	 */
	u_int32_t md_magic;		/*  0 MD identifier			*/
	u_int32_t major_version;	/*  1 major version when created	*/
	u_int32_t minor_version;	/*  2 minor version when created	*/
	u_int32_t patch_version;	/*  3 patchlevel version		*/
	u_int32_t set_uuid0;		/*  4 MD set identifier #1		*/
	u_int32_t set_uuid1;		/*  5 MD set identifier #2		*/
	u_int32_t set_uuid2;		/*  6 MD set identifier #3		*/
	u_int32_t set_uuid3;		/*  7 MD set identifier #4		*/
	u_int32_t ctime;		/*  8 Creation time			      */
	u_int32_t level;		/*  9 Raid personality			      */
	u_int32_t size;			/* 10 Apparent size of each individual disk   */
	u_int32_t md_minor;		/* 11 preferred MD minor device number	      */

	u_int32_t this_disk_number;	/* 12 original this disk number		*/
	u_int32_t this_disk_major;	/* 13 original this disk major		*/
	u_int32_t this_disk_minor;	/* 14 original this disk minor		*/
	u_int32_t sb_csum;		/* 15 checksum				*/
	u_int32_t reserved[MD_SAVED_SB_INFO_WORDS - 16];
}mdp_saved_super_t;

static inline u_int64_t md_event(mdp_super_t *sb) {
	u_int64_t ev = sb->events_hi;
	return (ev<<32)| sb->events_lo;
}

// Datatypes and Defs used in mdregmgr.c
#define SECTOR_SIZE_BITS            9
#define PAGE_SIZE_SECTORS           16      // for 8k pages
#define PE_SIZE_ALIGNMENT_SECTORS   128     // for 64k aligned PEs
#define MD_NAME_SPACE               EVMS_DEV_NODE_PATH "md"
#define MD_DEV_DIRECTORY            "md/"
#define MAX_MD_DEVICES              27
#define MAX_MD_MINORS               255
#define MD_MIN_CHUNK_SIZE           4       // in kilobytes
#define MD_MAX_CHUNK_SIZE           4096    // in kilobytes = 4MB
#define MD_DEFAULT_CHUNK_SIZE       32      // in kilobytes
#define BLOCK_SIZE                  1024    // needed by md size macros

typedef u_int8_t boolean;

#define MD_RESERVED       0UL
#define LINEAR            1UL
#define RAID0             2UL
#define RAID1             3UL
#define RAID5             4UL
#define TRANSLUCENT       5UL
#define HSM               6UL
#define MULTIPATH         7UL
#define MAX_PERSONALITY   8UL

#define INVALID_LEVEL    99UL

static inline int pers_to_level (int pers) {
	switch (pers) {
	case MULTIPATH:     return -4;
	case HSM:           return -3;
	case TRANSLUCENT:   return -2;
	case LINEAR:        return -1;
	case RAID0:         return 0;
	case RAID1:         return 1;
	case RAID5:         return 5;
	}
	return INVALID_LEVEL;
}

static inline int level_to_pers (int level) {
	switch (level) {
	case -3: return HSM;
	case -2: return TRANSLUCENT;
	case -1: return LINEAR;
	case 0:  return RAID0;
	case 1:  return RAID1;
	case 4:
	case 5:  return RAID5;
	}
	return MD_RESERVED;
}
// md_volume_t flags
#define MD_DISCOVERED	(1 << 0)
#define MD_DEGRADED	(1 << 1)
#define MD_CORRUPT	(1 << 2)
#define MD_DIRTY	(1 << 3)
#define MD_ACTIVE	(1 << 4)  // This md volume is active in the kernel
#define MD_SAVED_SB_EXISTS	(1 << 6)	// Indicates that saved SB info block exists
#define MD_USE_OLD_DEV		(1 << 7)	// The user has chosen to restore the original dev maj/minor
#define MD_DELAY_ACTIVATE	(1 << 8)	// Delay activation for certain operation, such as Replace

// commit_flags
#define MD_COMMIT_USE_DISK	(1 << 0)	// Set this flag in modifies if you want commit
						// to preserve the existing on disk SB.state
#define MD_COMMIT_SAVE_SB	(1 << 1)  	// need to write a backup SB info because the array
						// was created by other MD tools

#define MD_COMMIT_RESTORE_SAVED_SB (1 << 2)	// Set this flag in modifies to restore original SB info
						// per user's request

typedef struct md_volume_s {
	storage_object_t * region;
	storage_object_t * child_object[MAX_MD_DEVICES];
	mdp_super_t      * super_array[MAX_MD_DEVICES];
	mdp_saved_super_t * saved_super_array[MAX_MD_DEVICES];
	char               name[128];
	u_int32_t          commit_flag;	 //if set, use SB state from disk, else use from Master SB
	u_int32_t          uuid;
	u_int32_t          nr_disks;
	u_int32_t          flags;
	u_int32_t          personality;	 // uses the positive defines, not level
	mdp_super_t *      super_block;	 // master superblock
	struct md_volume_s * next;
	dlist_t            removed_disks;
	dlist_t            added_disks;
	dlist_t            activated_disks;
	dlist_t            deactivated_disks;
	void             * private_data;
} md_volume_t;


#define MD_MODIFY_TAG 99  // tag for evms_md_kdev_t in DLIST

// Global data used by mdregmgr
extern engine_functions_t   * EngFncs;
extern plugin_record_t      * linear_plugin;
extern plugin_record_t        linear_plugin_record;
extern plugin_record_t      * raid1_plugin;
extern plugin_record_t        raid1_plugin_record;
extern plugin_record_t      * raid0_plugin;
extern plugin_record_t        raid0_plugin_record;
extern plugin_record_t      * raid5_plugin;
extern plugin_record_t        raid5_plugin_record;
extern plugin_record_t      * my_plugin;
extern md_volume_t          * volume_list_head;	// List of real volume groups, indexed by vg_number.
extern char                   message_buffer[1024];


// Log message macros
#define LOG_CRITICAL(msg, args...)  EngFncs->write_log_entry(CRITICAL,  my_plugin, "%s: " msg, __FUNCTION__ , ## args)
#define LOG_SERIOUS(msg, args...)   EngFncs->write_log_entry(SERIOUS,   my_plugin, "%s: " msg, __FUNCTION__ , ## args)
#define LOG_ERROR(msg, args...)     EngFncs->write_log_entry(ERROR,     my_plugin, "%s: " msg, __FUNCTION__ , ## args)
#define LOG_WARNING(msg, args...)   EngFncs->write_log_entry(WARNING,   my_plugin, "%s: " msg, __FUNCTION__ , ## args)
#define LOG(msg, args...)           EngFncs->write_log_entry(DEFAULT,   my_plugin, "%s: " msg, __FUNCTION__ , ## args)
#define LOG_DETAILS(msg, args...)   EngFncs->write_log_entry(DETAILS,   my_plugin, "%s: " msg, __FUNCTION__ , ## args)
#define LOG_DEBUG(msg, args...)     EngFncs->write_log_entry(DEBUG,     my_plugin, "%s: " msg, __FUNCTION__ , ## args)
#define LOG_EXTRA(msg, args...)     EngFncs->write_log_entry(EXTRA,     my_plugin, "%s: " msg, __FUNCTION__ , ## args)
#define LOG_PROC(msg, args...)      EngFncs->write_log_entry(ENTRY_EXIT,    my_plugin, "%s: " msg, __FUNCTION__ , ## args)
#define LOG_EVERYTHING(msg, args...)    EngFncs->write_log_entry(EVERYTHING,    my_plugin, "%s: " msg, __FUNCTION__ , ## args)
#define MESSAGE(msg, args...)       EngFncs->user_message(my_plugin, NULL, NULL, msg, ##args)

#define SET_STRING(a,b) a = EngFncs->engine_alloc( strlen(b)+1 );if (a ) { strcpy(a, b); } else { rc = ENOMEM; LOG_EXIT(rc);}

#define LOG_ENTRY       my_plugin = MY_PLUGIN; LOG_PROC("Enter.\n")
#define LOG_EXIT(x)     LOG_PROC("Exit. rc = %d\n", x)
#define RETURN(x)       {LOG_EXIT(x); return(x);}

#ifndef min
	#define min(a,b) (((a)<(b))?(a):(b))
#endif
#ifndef max
	#define max(a,b) (((a)>(b))?(a):(b))
#endif

extern int md_can_replace_child(storage_object_t *region,
			    storage_object_t *child,
			    storage_object_t *new_child);
extern int md_replace_child(storage_object_t *region,
			    storage_object_t *child,
			    storage_object_t *new_child);
extern long calc_log2( long arg );
/* Macro: SET_POWER2_LIST
 */
// Set a contraint list that increments each item by a power of 2.
// b is the starting value, c is ending value (both must be power of 2).
#define SET_POWER2_LIST(a, b, c) \
{ \
    if ( ((b) & ((b)-1)) || ((c) & ((c)-1)) ) { \
        RETURN(EINVAL); \
    } \
    (a) = EngFncs->engine_alloc(calc_log2(c)*sizeof(value_t)+1); \
    if (a) { \
        long i = b; \
        (a)->count = 0; \
        while ( i <= c ) { \
            (a)->value[(a)->count].ui32 = i; \
            (a)->count++; \
            i <<= 1; \
        } \
    } else { \
        RETURN(ENOMEM); \
    } \
}


#define MD_CHUNK_ALIGN_NEW_SIZE_SECTORS(chunk_size_in_sectors, size_in_sectors) (MD_NEW_SIZE_SECTORS(size_in_sectors) & ~(chunk_size_in_sectors - 1))
#define MD_CHUNK_ALIGN_NEW_SIZE_BLOCKS(chunk_size_in_blocks, size_in_blocks) (MD_NEW_SIZE_BLOCKS(size_in_blocks) & ~(chunk_size_in_blocks - 1))

#define VSECTORS_TO_BLOCKS(sectors) ((sector_count_t) sectors / (BLOCK_SIZE >> EVMS_VSECTOR_SIZE_SHIFT))
#define BLOCKS_TO_VSECTORS(blocks)  ((sector_count_t) blocks  * (BLOCK_SIZE >> EVMS_VSECTOR_SIZE_SHIFT))

/* ioctls */
#define MD_MAJOR 9

#define MINORBITS	8
#define MINORMASK	((1U << MINORBITS) - 1)
#define MINOR(dev)	((dev) & MINORMASK)
#define MKDEV(ma,mi)	(((ma) << MINORBITS) | (mi))

/* status */
#define RAID_VERSION		_IOR (MD_MAJOR, 0x10, mdu_version_t)
#define GET_ARRAY_INFO		_IOR (MD_MAJOR, 0x11, mdu_array_info_t)
#define GET_DISK_INFO		_IOR (MD_MAJOR, 0x12, mdu_disk_info_t)
#define PRINT_RAID_DEBUG	_IO (MD_MAJOR, 0x13)
#define RAID_AUTORUN		_IO (MD_MAJOR, 0x14)

/* configuration */
#define CLEAR_ARRAY		_IO (MD_MAJOR, 0x20)
#define ADD_NEW_DISK		_IOW (MD_MAJOR, 0x21, mdu_disk_info_t)
#define HOT_REMOVE_DISK		_IO (MD_MAJOR, 0x22)
#define SET_ARRAY_INFO		_IOW (MD_MAJOR, 0x23, mdu_array_info_t)
#define SET_DISK_INFO		_IO (MD_MAJOR, 0x24)
#define WRITE_RAID_INFO		_IO (MD_MAJOR, 0x25)
#define UNPROTECT_ARRAY		_IO (MD_MAJOR, 0x26)
#define PROTECT_ARRAY		_IO (MD_MAJOR, 0x27)
#define HOT_ADD_DISK		_IO (MD_MAJOR, 0x28)
#define SET_DISK_FAULTY		_IO (MD_MAJOR, 0x29)
#define HOT_GENERATE_ERROR	_IO (MD_MAJOR, 0x2a)

/* usage */
#define RUN_ARRAY		_IOW (MD_MAJOR, 0x30, mdu_param_t)
#define START_ARRAY		_IO (MD_MAJOR, 0x31)
#define STOP_ARRAY		_IO (MD_MAJOR, 0x32)
#define STOP_ARRAY_RO		_IO (MD_MAJOR, 0x33)
#define RESTART_ARRAY_RW	_IO (MD_MAJOR, 0x34)

typedef struct mdu_version_s {
	int major;
	int minor;
	int patchlevel;
} mdu_version_t;

typedef struct mdu_array_info_s {
	/*
	 * Generic constant information
	 */
	int major_version;
	int minor_version;
	int patch_version;
	int ctime;
	int level;
	int size;
	int nr_disks;
	int raid_disks;
	int md_minor;
	int not_persistent;

	/*
	 * Generic state information
	 */
	int utime;		/*  0 Superblock update time		      */
	int state;		/*  1 State bits (clean, ...)		      */
	int active_disks;	/*  2 Number of currently active disks	      */
	int working_disks;	/*  3 Number of working disks		      */
	int failed_disks;	/*  4 Number of failed disks		      */
	int spare_disks;	/*  5 Number of spare disks		      */

	/*
	 * Personality information
	 */
	int layout;		/*  0 the array's physical layout	      */
	int chunk_size;	/*  1 chunk size in bytes		      */

} mdu_array_info_t;

typedef struct mdu_disk_info_s {
	/*
	 * configuration/status of one particular disk
	 */
	int number;
	int major;
	int minor;
	int raid_disk;
	int state;

} mdu_disk_info_t;

typedef struct mdu_start_info_s {
	/*
	 * configuration/status of one particular disk
	 */
	int major;
	int minor;
	int raid_disk;
	int state;

} mdu_start_info_t;

typedef struct mdu_param_s
{
	int			personality;	/* 1,2,3,4 */
	int			chunk_size;	/* in bytes */
	int			max_fault;	/* unused for now */
} mdu_param_t;

#define EVMS_MD_ADD		1
#define EVMS_MD_REMOVE		2
#define EVMS_MD_ACTIVATE	3
#define EVMS_MD_DEACTIVATE	4

/**
 * structure definition to use with MD_ADD, MD_REMOVE, MD_ACTIVATE
 **/
struct evms_md_kdev {
	u_int32_t major;
	u_int32_t minor;
	storage_object_t * object;
};

typedef struct evms_md_kdev evms_md_kdev_t;

#include "md_discover.h"
#include "md_dlist.h"
#include "md_io.h"
#include "md_create.h"
#include "md_info.h"

#endif

