/*
 *   Copyright (c) International Business Machines  Corp., 2001
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: mdregmgr
 * File: md_discover.c
 *
 * Description: This file contains all functions related to the initial
 *              discovery of MD physical volumes, volume groups, and logical
 *              volumes.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <plugin.h>
#include <time.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#define MY_PLUGIN my_plugin
#include "md.h"
#include "linear_discover.h"
#include "raid0_discover.h"
#include "raid1_discover.h"
#include "raid5_discover.h"


// function to check if the uuids of 2 super block are equal,
// returns 1 if they are equal, 0 otherwise.
int md_uuids_equal(mdp_super_t * sb1, mdp_super_t * sb2)
{
	LOG_ENTRY;
	if (sb1->set_uuid0 == sb2->set_uuid0 &&
		sb1->set_uuid1 == sb2->set_uuid1 &&
		sb1->set_uuid2 == sb2->set_uuid2 &&
		sb1->set_uuid3 == sb2->set_uuid3 ) {
		RETURN(1);
	} else {
		RETURN(0);
	}
}

// function to return the event counts as a 64 bit field.
u_int64_t md_get_event(mdp_super_t * sb)
{
	u_int64_t event;
	LOG_ENTRY;

	event = (u_int64_t) (sb->events_hi);
	event <<= 32;
	event += (u_int64_t) (sb->events_lo);
	RETURN(event);
}


/* returns csum of superblock, super->csum is left unchanged */
int calc_sb_csum(mdp_super_t *super)
{
	unsigned int  oldcsum = super->sb_csum;
	unsigned long long newcsum = 0;
	unsigned long csum;
	int i;
	unsigned int *superc = (int*) super;
	super->sb_csum = 0;

	for (i=0; i<MD_SB_BYTES/4; i++)
		newcsum+= superc[i];
	csum = (newcsum& 0xffffffff) + (newcsum>>32);
	super->sb_csum = oldcsum;
	return csum;
}

/* returns csum of saved SB INFO, saved_sb->csum is left unchanged */
int calc_saved_sb_csum(mdp_saved_super_t *saved_sb)
{
	unsigned int  oldcsum = saved_sb->sb_csum;
	unsigned long long newcsum = 0;
	unsigned long csum;
	int i;
	unsigned int *superc = (int*) saved_sb;
	saved_sb->sb_csum = 0;

	for (i=0; i<MD_SAVED_SB_INFO_WORDS; i++)
		newcsum+= superc[i];
	csum = (newcsum& 0xffffffff) + (newcsum>>32);
	saved_sb->sb_csum = oldcsum;
	return csum;
}

/* Function: md_validate_disk_sb
 *
 *	This function validates one RAID superblock for generic plausibility
 */

int md_validate_disk_sb(mdp_super_t *sb)
{
	int old_csum, new_csum;

	if ( ! ( sb->md_magic == MD_SB_MAGIC &&
			 sb->major_version == 0 &&
			 sb->minor_version == 90 ) ) {
		LOG_EXTRA("Bad signature or version\n");
		RETURN(ENXIO);
	}

	old_csum = sb->sb_csum;

	new_csum = calc_sb_csum(sb);

	if (new_csum != old_csum) {
		LOG("Bad Checksum\n");
		RETURN(ENXIO);
	}
	RETURN(0);
}

/* Function: md_validate_saved_sb
 *
 *	This function validates one RAID superblock for generic plausibility
 */

int md_validate_saved_sb(mdp_saved_super_t *saved_sb)
{
	int old_csum, new_csum;

	if ( ! ( saved_sb->md_magic == MD_SB_MAGIC &&
			 saved_sb->major_version == 0 &&
			 saved_sb->minor_version == 90 ) ) {
		LOG_EXTRA("Bad signature or version\n");
		RETURN(ENXIO);
	}

	old_csum = saved_sb->sb_csum;

	new_csum = calc_saved_sb_csum(saved_sb);

	if (new_csum != old_csum) {
		LOG("Bad Checksum\n");
		RETURN(ENXIO);
	}
	RETURN(0);
}

/* Function:  md_read_metadata_from_disk
 *
 *	This function reads MD superblock or "backup" MD info from disk.
 */
int  md_read_metadata_from_disk(
	storage_object_t *object, 
	u_int64_t location, 
	void * buffer, 
	u_int32_t sects)
{
	int rc = 0;
	LOG_ENTRY;

	if (object->data_type != DATA_TYPE) {
		LOG_DETAILS("Object not data type, skipping %s\n", object->name);
		RETURN(rc);
	}

	if (object->size <= MD_RESERVED_SECTORS) {
		LOG_DETAILS("Object too small for MD, skipping %s\n", object->name);
		RETURN(rc);
	}

	LOG_DEBUG("Looking for MD metadata at %"PRIu64" on %s\n",location, object->name);

	if ( READ(object, location, sects, (char*)buffer )) {
		LOG_SERIOUS("Error reading MD superblock from object %s\n", object->name);
		RETURN(EIO);
	}

	RETURN (rc);

}

/* function: md_save_sb
 *
 *	Save some information from the original superblock,
 *	use the last MD_SAVED_SB_INFO_SECTS of reserved MD sectors.
 */
int md_save_sb(md_volume_t * volume, int i)
{
	u_int64_t location;
	mdp_super_t *sb;
	mdp_saved_super_t *saved_sb = NULL;
	storage_object_t *object = volume->child_object[i];
	int rc = 0;

	LOG_ENTRY;

	sb = volume->super_array[i];

	//must have MD_COMMIT_SAVE_SB flag set to continue
	if ( !(volume->commit_flag & MD_COMMIT_SAVE_SB) ||
	     (sb->disks[i].state & (1<<MD_DISK_NEW)) ) {
	   	RETURN(0);
	}
	

	if (volume->saved_super_array[i]) {
		saved_sb = volume->saved_super_array[i];
		LOG_DEBUG("skipping %s, saved SB info already exists.\n", object->name);
		RETURN(0);

	} else {

		saved_sb = EngFncs->engine_alloc(MD_SB_BYTES);
		if (!saved_sb)
			RETURN(ENOMEM);
		
		saved_sb->md_magic = sb->md_magic;
		saved_sb->major_version = sb->major_version;
		saved_sb->minor_version = sb->minor_version;
		saved_sb->patch_version = sb->patch_version;
		saved_sb->set_uuid0 = sb->set_uuid0;
		saved_sb->set_uuid1 = sb->set_uuid1;
		saved_sb->set_uuid2 = sb->set_uuid2;
		saved_sb->set_uuid3 = sb->set_uuid3;
		saved_sb->ctime = sb->ctime;
		saved_sb->level = sb->level;
		saved_sb->size = sb->size;
		saved_sb->md_minor = sb->md_minor;

		saved_sb->this_disk_number = sb->disks[i].number;
		saved_sb->this_disk_major = sb->disks[i].major;
		saved_sb->this_disk_minor = sb->disks[i].minor;

		saved_sb->sb_csum = calc_saved_sb_csum(saved_sb);
	
		//Location of MD superblock
		location = MD_NEW_SIZE_SECTORS(object->size);
		location += MD_SAVED_SB_SECTOR_OFFSET;

		LOG_DEBUG("Writing MD backup SB at %"PRIu64" on %s\n", location, object->name);
		if ( WRITE(object, location, MD_SAVED_SB_INFO_SECTS,
			   (char*)saved_sb)) {
			LOG_SERIOUS("Error writing saved SB info to %s\n", object->name);
			rc = EIO;
			goto out_free;
		}
		RETURN(0);
	}

out_free:
	if (saved_sb)
		EngFncs->engine_free(saved_sb);
	RETURN(rc);
}

/* write all of the super block for an MD region to disk */
int  md_write_sbs_to_disk(md_volume_t * volume) {
	int rc = 0;
	u_int64_t location;
	int i,j;
	int nr_disks;
//	mdp_super_t *super;
	time_t utime;
	mdp_super_t *sb;
	LOG_ENTRY;

	// update utime
	utime = time(NULL);
	volume->super_block->utime = utime;
	volume->super_block->events_lo++;
	if (volume->super_block->events_lo == 0) {
		volume->super_block->events_hi++;
	}
	
	if ( md_is_region_active(volume->region) &&
	     !(volume->super_block->state & (1 << MD_SB_CLEAN)) ) {

		/* The array was active when we read the superblock.
		 * If we did not find any problems, mark the array clean.
		 */

		if (!(volume->flags & (MD_CORRUPT | MD_DEGRADED))) {
			volume->super_block->state |= (1 << MD_SB_CLEAN);
		}
	}

	if (volume->commit_flag & MD_COMMIT_RESTORE_SAVED_SB) {
		if (volume->region)
			md_deactivate_region(volume->region);
		else
			LOG_ERROR("No region for %s to deactivate\n", volume->name);
	}

	for (i = 0, nr_disks = 0; (i < MAX_MD_DEVICES) && (nr_disks < volume->nr_disks); i++) {

		sb = volume->super_array[i];
		if (sb != NULL) {
			
			// First, we might have to write some saved superblock
			rc = md_save_sb(volume,i);
			if (rc)
				RETURN(rc);

			// copy superblock form master, and fix up this_disk
			memcpy(sb, volume->super_block, MD_SB_BYTES);


			if (sb->disks[i].state & (1<<MD_DISK_NEW)) {
				sb->events_lo = 0; // null out event counter
				sb->events_hi = 0; // as signal to kernel on new devices
			}
			// turn off new bit in master SB
			for (j = 0; j < MAX_MD_DEVICES; j++) {
				sb->disks[j].state &= ~(1<<MD_DISK_NEW);
			}
			volume->super_block->disks[i].state &= ~(1<<MD_DISK_NEW);

			sb->this_disk = sb->disks[i];
			if (volume->commit_flag & MD_COMMIT_RESTORE_SAVED_SB) {
				LOG("%s should have old dev number: (%d:%d)\n",
				    volume->child_object[i]->name,
				    sb->this_disk.major,
				    sb->this_disk.minor);
			}


			// see if we need to read the current superblock state from disk.
/*			if (volume->commit_flag & MD_COMMIT_USE_DISK) {
		LOG_DETAILS("reading state info for disk %d of region %s\n",i,volume->name);
		rc = md_check_for_pv(volume->child_object[i], &super);
		if (!rc) {
			if (md_uuids_equal(super, volume->super_block)) {
			volume->super_array[i]->state = super->state;
			}
			EngFncs->engine_free(super);
		} else {
			LOG_ERROR("Error reading state info for disk %d of region %s\n",i,volume->name);
		}
		} */
			location = MD_NEW_SIZE_SECTORS(volume->child_object[i]->size);
			LOG("Writing MD Superblock at %"PRIu64" on %s\n",location, volume->child_object[i]->name);


			/*        volume->super_array[i]->sb_csum = 0;
				rc = EngFncs->calculate_checksum((char *)volume->super_array[i],
								 MD_SB_BYTES,
								 0,
								 &volume->super_array[i]->sb_csum);*/
			sb->sb_csum = calc_sb_csum(volume->super_array[i]);

			if ( WRITE(volume->child_object[i], location, MD_SB_SECTORS, (char*)sb)) {
				LOG_SERIOUS("Error writing MD SUperBlock from object %s\n", volume->child_object[i]->name);
				RETURN(EIO);
			}
			volume->region->flags |= SOFLAG_NEEDS_ACTIVATE;

			nr_disks++;
		}
	}
	volume->commit_flag &= ~MD_COMMIT_SAVE_SB; //only write MD saved SB info once
	volume->commit_flag &= ~MD_COMMIT_RESTORE_SAVED_SB;
	RETURN (rc);

}


// the input object has been validated as a pv for this personality.
// using the list head for this personality, search to see if other PVs
// for this volume already exist, if so add this supreblock to the list for that volume
// else create a new volume structure for it and add it to the volume list

int md_find_volume_for_object(storage_object_t * object, mdp_super_t *md_super_buffer) {

	int i;
	md_volume_t * volume = volume_list_head;
	mdp_super_t * tmp_super = NULL;
	storage_object_t * tmp_object = NULL;
	mdp_saved_super_t      * saved_sb = NULL;

	LOG_ENTRY;
	while (volume) {
		if (md_uuids_equal(volume->super_block, md_super_buffer)) {
			if (md_get_event(md_super_buffer) > md_get_event(volume->super_block)) {
				memcpy(volume->super_block,
					   md_super_buffer,
					   MD_SB_BYTES);
			}
			if (volume->flags & MD_DISCOVERED) {
				// The volume has beed discovered.  The
				// volume may have been exported in degrade
				// mode and this object is late for the show,
				// or this may be a stale object that has
				// been resurrected.
				RETURN(EEXIST);
			}
			// check to make sure slot is empty
			if (!volume->super_array[md_super_buffer->this_disk.number]) {
				volume->super_array[md_super_buffer->this_disk.number] = md_super_buffer;
				volume->child_object[md_super_buffer->this_disk.number] = object;
			} else {
				if (md_get_event(md_super_buffer) > md_get_event(volume->super_array[md_super_buffer->this_disk.number])) {
					// new buffer is newer, use it.
					// save off the old super and object
					volume->flags |= MD_DIRTY;
					tmp_super = volume->super_array[md_super_buffer->this_disk.number];
					tmp_object = volume->child_object[md_super_buffer->this_disk.number];
					// fill in the new info
					volume->super_array[md_super_buffer->this_disk.number] = md_super_buffer;
					volume->child_object[md_super_buffer->this_disk.number] = object;
				} else if (md_get_event(md_super_buffer) < md_get_event(volume->super_array[md_super_buffer->this_disk.number])) {
					//existing buffer the correct one, reassign this one into tmp
					volume->flags |= MD_DIRTY;
					tmp_super = md_super_buffer;
					tmp_object = object;
				} else {
					// event counters are equal, bad news.
					MESSAGE("Multiple superblocks found for region %s index %d\n",volume->name,md_super_buffer->this_disk.number);
					MESSAGE("Object 1 %s, Object 2 %s\n",object->name,volume->child_object[md_super_buffer->this_disk.number]->name);
					volume->flags |= MD_CORRUPT;
					// don't want to write this out on commit.
					tmp_super = md_super_buffer;
					tmp_object = object;
				}
				// stick in next available slot, figure out later
				for (i = 0; i< MAX_MD_DEVICES; i++) {
					if (volume->super_array[i] == NULL) {
						volume->super_array[i] = tmp_super;
						volume->child_object[i] = tmp_object;
						// fix up the this_disk field
						memcpy(&tmp_super->this_disk, &tmp_super->disks[i], sizeof(tmp_super->this_disk));
						break;
					}
				}
			}

			volume->nr_disks++;
			break;
		}
		volume = volume->next;

	}
	if (!volume) {
		// add new entry
		if ( !(volume = EngFncs->engine_alloc(sizeof(md_volume_t) ) )) {
			LOG_CRITICAL("Memory error creating buffer to read super block.\n");
			RETURN(ENOMEM);
		}

		volume->super_array[md_super_buffer->this_disk.number] = md_super_buffer;
		volume->child_object[md_super_buffer->this_disk.number] = object;
		volume->personality = level_to_pers(md_super_buffer->level);
		volume->nr_disks = 1;
		volume->removed_disks = CreateList();
		volume->added_disks = CreateList();
		volume->activated_disks = CreateList();
		volume->deactivated_disks = CreateList();
		volume->commit_flag = MD_COMMIT_USE_DISK; // use on disk state for 'discovered' volumes
		volume->super_block = EngFncs->engine_alloc(MD_SB_BYTES);
		memcpy(volume->super_block,
			   md_super_buffer,
			   MD_SB_BYTES);
		md_add_volume_to_list(volume);
		sprintf(volume->name, "md/md%d",md_super_buffer->md_minor);
	}
	
	//Now try to get the saved SB info
	if (!md_check_for_saved_sb(object, &saved_sb))
		volume->saved_super_array[md_super_buffer->this_disk.number] = saved_sb;

	RETURN (0);
}

/* Function: md_check_for_pv
 *
 *	This function is used during discovery as an argument to PruneList.
 *	When this function is called, one object is passed in. The first
 *	sector of that object is examined to determine if it is an MD PV.
 *      If it is a PV, then a 0 return code is returned along with the super_block
 *      buffer allocated.
 */
int  md_check_for_pv(storage_object_t * object,mdp_super_t ** md_super_buffer) {
	int rc = 0;
	LOG_ENTRY;


	if ( !(*md_super_buffer = EngFncs->engine_alloc(MD_SB_BYTES))) {
		LOG_CRITICAL("Memory error creating buffer to read super block.\n");
		RETURN(ENOMEM);
	}

	// Read the first block and look for a PV signature.
	if ( md_read_metadata_from_disk(object,
					MD_NEW_SIZE_SECTORS(object->size),
                                        *md_super_buffer,
					MD_SB_SECTORS) ) {
		LOG_SERIOUS("I/O error on object %s.\n", object->name);
		EngFncs->engine_free(*md_super_buffer);
		RETURN(EIO);
	}
	
	rc = md_validate_disk_sb(*md_super_buffer);
	if (rc) {
		LOG_DEBUG("%s is not a a valid MD object\n", object->name);
		EngFncs->engine_free(*md_super_buffer);
	}
	RETURN(rc);
}

/* Function: md_check_for_saved_sb
 *
 *	This function is used during discovery.  It's called after we know that
 * the object is an MD device.  If a saved SB info block exists, then return 0
 * along with the allocated saved_sb.
 *
 */
int md_check_for_saved_sb(storage_object_t * object, mdp_saved_super_t ** saved_sb) {
	int rc = 0;
	LOG_ENTRY;

	if ( !(*saved_sb = EngFncs->engine_alloc(MD_SAVED_SB_INFO_BYTES))) {
		LOG_CRITICAL("No memory to read saved SB info.\n");
		RETURN(ENOMEM);
	}

	if ( md_read_metadata_from_disk(
		object,
		MD_NEW_SIZE_SECTORS(object->size) + MD_SAVED_SB_SECTOR_OFFSET,
		*saved_sb,
		MD_SAVED_SB_INFO_SECTS) ) {
		
		LOG_SERIOUS("I/O error on object %s.\n", object->name);
		rc = EIO;
		goto out_free;
	}
	
	rc = md_validate_saved_sb(*saved_sb);
	if (!rc) {
		LOG_DEBUG("Found saved SB info on %s\n", object->name);
		goto out;
	} else {
		LOG_EXTRA("%s does not have a valid MD saved SB info\n", object->name);
		goto out_free;	
	}

out_free:
	EngFncs->engine_free(*saved_sb);
	*saved_sb = NULL;
out:
	RETURN(rc);
}

int md_can_restore_saved_sb(storage_object_t *region)
{
	int i;
	logical_volume_t *volume;
	md_volume_t * md_volume;
	mdp_saved_super_t *saved_sb;
	mdp_super_t *sb;

	LOG_ENTRY;

	if (region == NULL)
		RETURN(FALSE);

	volume = region->volume;
	if (volume == NULL)
		RETURN(FALSE);
	if ((volume->flags & VOLFLAG_COMPATIBILITY) == 0)
		RETURN(FALSE);

	md_volume = (md_volume_t *) region->private_data;
	if (md_volume == NULL)
		RETURN(FALSE);

	if (md_volume->flags & MD_USE_OLD_DEV)
		RETURN(FALSE);

	/*
	 * If we found backup superblock in all devices of the MD array,
	 * we will provide an option to convert it back.
	 */
	md_volume->flags |= MD_SAVED_SB_EXISTS;
	for (i = 0; i < MAX_MD_DEVICES; i++) {
		if (md_volume->child_object[i]) {
			sb = md_volume->super_array[i];
			saved_sb = md_volume->saved_super_array[i];

			// All devices must contain a saved SB info block
			if (saved_sb == NULL) {
				md_volume->flags &= ~MD_SAVED_SB_EXISTS;
				break;
			}

			if (md_volume->super_block->disks[i].major == saved_sb->this_disk_major) {
				md_volume->flags &= ~MD_SAVED_SB_EXISTS;
				break;
			}

			// Validate the constant part of MD superblock
			if ((sb->set_uuid0 == saved_sb->set_uuid0) &&
			    (sb->set_uuid1 == saved_sb->set_uuid1) &&
			    (sb->set_uuid2 == saved_sb->set_uuid2) &&
			    (sb->set_uuid3 == saved_sb->set_uuid3) &&
			    (sb->ctime == saved_sb->ctime) &&
			    (sb->level == saved_sb->level) &&
			    (sb->size == saved_sb->size) &&
			    (sb->md_minor == saved_sb->md_minor)) {
				LOG_DEBUG("%s contains a valid MD saved info block\n",
					  md_volume->child_object[i]->name);
			} else {
				md_volume->flags &= ~MD_SAVED_SB_EXISTS;
				break;
			}
		}
	}

	if (md_volume->flags & MD_SAVED_SB_EXISTS)
		RETURN(TRUE);
	RETURN(FALSE);
}

/* Function: md_restore_saved_sb
 * 	This function restores saved superblock info to the volume master superblock.
 */
int md_restore_saved_sb(md_volume_t *volume)
{
	int i;
	u_int32_t number;
	mdp_saved_super_t *saved_sb;
	mdp_super_t *sb = volume->super_block;

	LOG_ENTRY;

	for (i = 0; i < MAX_MD_DEVICES; i++) {
		if ((volume->child_object[i] != NULL) &&
		    (volume->super_array[i] != NULL) &&
		    (volume->saved_super_array[i] != NULL)) {
			sb->disks[i].major = volume->saved_super_array[i]->this_disk_major;
			sb->disks[i].minor = volume->saved_super_array[i]->this_disk_minor;
		}
	}

	number = sb->this_disk.number;

	saved_sb  = volume->saved_super_array[number];

	sb->major_version = saved_sb->major_version;
	sb->minor_version = saved_sb->minor_version;
	sb->patch_version = saved_sb->patch_version;
	sb->this_disk.major = saved_sb->this_disk_major;
	sb->this_disk.minor = saved_sb->this_disk_minor;

	volume->commit_flag |= MD_COMMIT_RESTORE_SAVED_SB;
	volume->flags |= MD_USE_OLD_DEV;
	RETURN(0);
}


/* Function: md_discover_volume
 *
 *	This function is the entry point into the first half of discovery,
 *	which examines all objects to find PVs and assigns them to the
 *	appropriate groups.
 */
int md_discover_volumes( dlist_t input_list, dlist_t output_list) {
	void * waste;
	storage_object_t * object;
	int tag;
	int rc = 0;
	mdp_super_t      * md_super_buffer;

	LOG_ENTRY;
	LOG_DETAILS("Searching for MD Super Blocks.\n");

	// A buffer for md_check_object_for_pv to use for reading
	// the PV metadata.

	while (!(rc = BlindExtractObject(input_list, (TAG *)&tag, NULL, (void *)&object))) {
		if (object->data_type == DATA_TYPE) {
			rc = md_check_for_pv(object, &md_super_buffer);
			if (rc) {
				// wasn't ours, put it on the output list now.
				InsertObject(output_list, object, tag, NULL, AppendToList, FALSE, &waste);
			} else {
				if ( md_find_volume_for_object(object, md_super_buffer)) {
					EngFncs->engine_free(md_super_buffer);
					LOG_WARNING("Error finding volume minor %d for PV %s\n", md_super_buffer->md_minor, object->name);
					// Object was not clamed for a volume.
					// Put it on the output list.
					InsertObject(output_list, object, tag, NULL, AppendToList, FALSE, &waste);
				}
			}
		} else {
			LOG_DETAILS("Skipping object %s because not DATA_TYPE\n",object->name);
		}

	}
	if ((rc == DLIST_EMPTY) || (rc == DLIST_END_OF_LIST)) {
		rc = 0;
	} else {
		LOG_WARNING("Error processing input list rc = %d\n", rc);
	}
	RETURN(rc);
}

// Get information for MD device from kernel
//Caller is responsible for freeing superblock memory on 0 rc, otherwise memory not allocated.

#define BLKGETSIZE64 _IOR(0x12,114,sizeof(u_int64_t)) 
int md_get_kernel_info(storage_object_t * region, mdu_array_info_t* md_info) {
	int rc=0;
	int fd;
	u_int64_t size = 0;

	LOG_ENTRY;

	fd = EngFncs->open_object(region, O_RDWR);
	if (fd <= 0) {
		LOG_WARNING("Error getting handle for volume %d.\n",errno);
	} else {
		// the arg to stop_array is a not used.
		// just send ioctl to the md device to stop
		if (EngFncs->ioctl_object(region, fd, GET_ARRAY_INFO, md_info)) {
			rc = errno;
			region->flags &= ~SOFLAG_ACTIVE;
			LOG_WARNING("Ioctl error (%d).\n", rc);
		} else {
			rc = EngFncs->ioctl_object(region, fd, BLKGETSIZE64, &size);
			if (!rc && size != 0) {
				region->flags |= SOFLAG_ACTIVE;
			} else {
				// Get info succeeded but can't get size of md region
				// probably due to inactive array in kernel
				// Just stop it since it really isn't active and will cause up problems later.
				LOG_WARNING("Ioctl error getting size of md device rc=%d size = %"PRIu64".\n", rc, size);
				region->flags &= ~SOFLAG_ACTIVE;
//				md_deactivate_region(region);
				rc = EBUSY;
			}
		}
		EngFncs->close_object(region, fd);
	}
	RETURN(rc);
}

int md_is_region_active(storage_object_t * region)
{
	MD_CHECK_ACTIVE(region);
	return (region->flags & SOFLAG_ACTIVE);
}


static BOOLEAN  md_namespace_registered = FALSE;

int md_register_name_space(void) {

	int rc = 0;

	LOG_ENTRY;

	if (!md_namespace_registered) {
		rc = EngFncs->register_name(MD_NAME_SPACE);

		if (rc == 0) {
			md_namespace_registered = TRUE;
		} else {
			LOG_SERIOUS("Error registering the MD name space \"%s\".\n", MD_NAME_SPACE);
		}
	}

	RETURN(rc);
}


static BOOLEAN md_final_call = FALSE;

int md_discover_final_call( dlist_t input_list, dlist_t output_list, int *count ) {

	int rc = 0;

	if (!md_final_call) {
		md_discover_volumes(input_list, output_list);
		LOG_DETAILS("PV discovery complete.\n");

		raid5_discover_regions(output_list, count, TRUE);
		LOG_DETAILS("RAID4/5 volume discovery complete.\n");

		DeleteAllItems(input_list, FALSE);
		AppendList(input_list, output_list);

		md_discover_volumes(input_list, output_list);
		LOG_DETAILS("PV discovery complete.\n");

		raid1_discover_regions(output_list, count, TRUE);
		LOG_DETAILS("RAID1 volume discovery complete.\n");

		DeleteAllItems(input_list, FALSE);
		AppendList(input_list, output_list);

		md_discover_volumes(input_list, output_list);
		LOG_DETAILS("PV discovery complete.\n");

		raid0_discover_regions(output_list, count, TRUE);
		LOG_DETAILS("RAID0 volume discovery complete.\n");

		DeleteAllItems(input_list, FALSE);
		AppendList(input_list, output_list);

		md_discover_volumes(input_list, output_list);
		LOG_DETAILS("PV discovery complete.\n");

		linear_discover_regions(output_list, count, TRUE);
		LOG_DETAILS("LINEAR volume discovery complete.\n");

		md_final_call = TRUE;

	} else {
		/*
		 * Final discovery has been done.  Just move the
		 * objects from the input list to the output list.
		 */
		CopyList(output_list, input_list, AppendToList);
	}

	RETURN(rc);
}

int md_activate_region(storage_object_t * region)
{
	int fd;
	int rc=0;
	unsigned long dev;
	mdu_array_info_t md_info;
	md_volume_t  *volume = (md_volume_t *)region->private_data;

	LOG_ENTRY;

	if (volume->flags & MD_DELAY_ACTIVATE) {
		LOG("Skipping activation of %s region for now.\n", region->name);
		RETURN(0);
	}

	if (volume->flags & MD_USE_OLD_DEV) {
		LOG("The original dev nodes have been restored for %s, will not re-activate\n",region->name);
		RETURN(0);
	}

	if (region->flags & SOFLAG_ACTIVE) {
		LOG("%s is currently active, deactivating...\n", region->name);
		rc = md_deactivate_region(region);
		if (rc) {
			RETURN(rc);
		}
	}

	fd = EngFncs->open_object(region, O_RDWR );
	if ( fd <= 0 ) {
		LOG_WARNING("Error getting handle for object %s rc=%d.\n",region->name,errno);
	} else {
		// the arg to start array is a dev_t of any disk in the array
		// just use the this_disk in the reference super since it should
		// always be valid.
		volume->super_block->this_disk = volume->super_block->disks[volume->super_block->this_disk.number];
		dev = MKDEV(volume->super_block->this_disk.major,
					volume->super_block->this_disk.minor);
		LOG_DETAILS("Starting array %s major,%d minor %d\n",
					volume->name,
					volume->super_block->this_disk.major,
					volume->super_block->this_disk.minor);

		if (EngFncs->ioctl_object(region, fd, START_ARRAY, dev)) {
			rc = errno;
			LOG_WARNING("Ioctl error MD Activate failed on object %s (%d).\n",
						region->name, rc);
			EngFncs->close_object(region, fd);
		} else {
			// activate succeded, turn on active flag 
			EngFncs->close_object(region, fd);
			rc = md_get_kernel_info(region, &md_info);
			//region->flags |= SOFLAG_ACTIVE;
		}
	}

	if (!rc) {
		volume->super_block->state &= ~(1 << MD_SB_CLEAN);
		region->flags &= ~SOFLAG_NEEDS_ACTIVATE;
	}

	RETURN(rc);

}

int md_deactivate_region(storage_object_t * region)
{
	int fd;
	int rc=0;
	unsigned long arg= 0;

	LOG_ENTRY;
	if (!region) {
		RETURN(EFAULT);
	}
	

	fd = EngFncs->open_object(region, O_RDWR );
	if ( fd <= 0 ) {
		LOG_WARNING("Error getting handle for object %s rc=%d.\n",region->name,errno);
	} else {
		// the arg to stop_array is a not used.
		// just send ioctl to the md device to stop
		if (EngFncs->ioctl_object(region, fd, STOP_ARRAY, arg)) {
			rc = errno;
			LOG_ERROR("Ioctl error MD Deactivate failed on object %s (%d).\n",
						region->name, rc);
		} else {
			if (md_is_region_active(region)) {
				rc = EBUSY;
				LOG_ERROR("Ioctl error MD Deactivate failed on object %s (%d).\n",region->name, rc);
			}
		}
		EngFncs->close_object(region, fd);
	}
	
	if (!rc || (rc == ENODEV)) {
		region->flags &= ~(SOFLAG_ACTIVE | SOFLAG_NEEDS_DEACTIVATE);
		rc = 0;
	}

	RETURN(rc);

}
