/*
 *   Copyright (c) International Business Machines  Corp., 2000-2003
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: snap_create.c
 *
 * Functions to help in the creation of snapshot volumes.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <plugin.h>
#include "snapshot.h"

/* Only for is_2_4(). */
#include <fcntl.h>
#include <unistd.h>

/* List of plugins whose compatibility volumes cannot be snapshotted. */
static char * reject_plugins[] = {
	"LocalDskMgr",
	"MDLinearRegMgr",
	"MDRaid0RegMgr",
	"MDRaid1RegMgr",
	"MDRaid5RegMgr",
	NULL
};

/**
 * make_parent_and_child
 *
 * Make "parent" a parent of "child", and make "child" a child of "parent".
 **/
void make_parent_and_child(storage_object_t * parent,
			   storage_object_t * child)
{
	void * handle;

	LOG_ENTRY;

	InsertObject(parent->child_objects, child,
		     get_tag_for_object(child),
		     NULL, InsertAtStart, 0, &handle);
	InsertObject(child->parent_objects, parent,
		     get_tag_for_object(parent),
		     NULL, InsertAtStart, 0, &handle);

	LOG_EXIT(0);
}

/**
 * unmake_parent_and_child
 *
 * Opposite of make_parent_and_child().
 **/
void unmake_parent_and_child(storage_object_t * parent,
			     storage_object_t * child)
{
	LOG_ENTRY;

	if (parent && child) {
		DeleteObject(parent->child_objects, child);
		DeleteObject(child->parent_objects, parent);
	}

	LOG_EXIT(0);
}

/**
 * add_snapshot_to_origin
 *
 * Add the specified snapshot to the specified origin's list of snapshots.
 * The snapshot state should be up-to-date before calling this.
 **/
void add_snapshot_to_origin(snapshot_volume_t * snap_volume,
			    snapshot_volume_t * org_volume)
{
	LOG_ENTRY;
	LOG_DEBUG("Adding snapshot %s to origin %s.\n",
		  snap_volume->parent->name, org_volume->parent->name);

	snap_volume->origin = org_volume;
	snap_volume->parent->associated_object = org_volume->parent;
	snap_volume->next = org_volume->next;
	org_volume->next = snap_volume;
	org_volume->count++;
	if (is_active(snap_volume)) {
		org_volume->active_count++;
	}

	LOG_EXIT(0);
}

/**
 * remove_snapshot_from_origin
 *
 * Remove the specified snapshot from its origin's list of snapshots.
 **/
void remove_snapshot_from_origin(snapshot_volume_t * snap_volume)
{
	snapshot_volume_t * org_volume = snap_volume->origin;
	snapshot_volume_t * this_volume;

	LOG_ENTRY;

	for (this_volume = org_volume;
	     this_volume; this_volume = this_volume->next) {
		if (this_volume->next == snap_volume) {
			this_volume->next = snap_volume->next;
			org_volume->count--;
			if (is_active(snap_volume)) {
				org_volume->active_count--;
			}
			break;
		}
	}
	snap_volume->next = NULL;
	snap_volume->origin = NULL;

	LOG_EXIT(0);
}

/**
 * get_snapshot_state
 *
 * Get Device-Mapper's state for this snapshot. If the device is active,
 * retrieve the percent-full value.
 **/
int get_snapshot_state(snapshot_volume_t * snap_volume)
{
	storage_object_t * snap_parent = snap_volume->parent;
	char * info = NULL;
	int percent;
	int rc;

	LOG_ENTRY;
	LOG_DEBUG("Checking state of snapshot %s.\n", snap_parent->name);

	if (!is_active(snap_volume)) {
		EngFncs->dm_update_status(snap_volume->sibling);
		EngFncs->dm_update_status(snap_parent);
	}

	if (is_active(snap_volume)) {
		rc = EngFncs->dm_get_info(snap_parent, &info);
		if (!rc) {
			if (sscanf(info, "%d%%", &percent) == 1) {
				snap_volume->percent_full = percent;
			} else if (!strcmp(info, "Invalid")) {
				snap_volume->percent_full = 0;
				snap_volume->flags |= SNAPSHOT_INVALID;
			} else {
				snap_volume->percent_full = -1;
			}

			EngFncs->engine_free(info);
		}
	}

	RETURN(0);
}

/**
 * get_origin_state
 *
 * Get Device-Mapper's state for this origin. If the origin is still inactive
 * after checking the kernel state, we need to grab the child's device
 * numbers.
 **/
int get_origin_state(snapshot_volume_t * org_volume)
{
	LOG_ENTRY;
	LOG_DEBUG("Checking state of origin %s.\n", org_volume->parent->name);

	if (!is_active(org_volume)) {
		EngFncs->dm_update_status(org_volume->parent);
	}

	RETURN(0);
}

/**
 * skip_dev_path
 *
 * Given a name of a volume, return a pointer that skips over the "/dev/evms"
 * at the front of the name.
 **/
char * skip_dev_path(char * volume_name)
{
	char * name = strstr(volume_name, EVMS_DEV_NODE_PATH);
	if (name) {
		name += strlen(EVMS_DEV_NODE_PATH);
	} else {
		name = volume_name;
	}
	return name;
}

/**
 * get_origin_object_name
 *
 * Generate the name for an origin object based on the origin child object
 * name. This simply appends a "#origin#" at the end of the name.
 **/
static void get_origin_object_name(char * child_name, char * parent_name)
{
	LOG_ENTRY;

	strncpy(parent_name, child_name, EVMS_NAME_SIZE);
	strncat(parent_name, "#origin#", EVMS_NAME_SIZE-strlen(parent_name));

	LOG_EXIT(0);
}

/**
 * allocate_snapshot
 *
 * Allocate a new EVMS feature-object and all necessary private data for a
 * snapshot. The metadata must already be allocated and fully initialized,
 * and the child object must already have an initialized feature header.
 **/
snapshot_volume_t * allocate_snapshot(storage_object_t * snap_child,
				      snapshot_metadata_t * metadata)
{
	snapshot_volume_t * volume = NULL;
	storage_object_t * snap_parent = NULL;
	storage_object_t * snap_sibling = NULL;
	int rc;

	LOG_ENTRY;
	LOG_DEBUG("Allocating snapshot %s.\n",
		  snap_child->feature_header->object_name);

	/* Allocate the private data and the EVMS object. */
	volume = EngFncs->engine_alloc(sizeof(*volume));
	if (!volume) {
		LOG_ERROR("Memory error allocating private data for snapshot "
			  "%s.\n", snap_child->feature_header->object_name);
		RETURN(NULL);
	}

	rc = EngFncs->allocate_evms_object(snap_child->feature_header->object_name,
					   &snap_parent);
	if (rc) {
		EngFncs->engine_free(volume);
		LOG_ERROR("Memory error allocating object for snapshot %s.\n",
			  snap_child->feature_header->object_name);
		RETURN(NULL);
	}

	snap_sibling = EngFncs->engine_alloc(sizeof(storage_object_t));
	if (!snap_sibling) {
		EngFncs->free_evms_object(snap_parent);
		EngFncs->engine_free(volume);
		LOG_ERROR("Memory error allocating sibling object for snapshot "
			  "%s.\n", snap_child->feature_header->object_name);
		RETURN(NULL);
	}

	/* Intialize the private data. */
	volume->parent = snap_parent;
	volume->child = snap_child;
	volume->sibling = snap_sibling;
	volume->metadata = metadata;
	volume->flags |= SNAPSHOT;

	/* Initialize the object. */
	snap_parent->data_type = DATA_TYPE;
	snap_parent->plugin = snapshot_plugin;
	snap_parent->flags |= SOFLAG_MUST_BE_TOP;
	snap_parent->size = metadata->total_chunks * metadata->chunk_size;
	snap_parent->geometry = snap_child->geometry;
	snap_parent->private_data = volume;
	if (! (metadata->flags & SNAPSHOT_WRITEABLE)) {
		snap_parent->flags |= SOFLAG_READ_ONLY;
	}

	make_parent_and_child(snap_parent, snap_child);

	/* Initialize the sibling. */
	snap_sibling->object_type = EVMS_OBJECT;
	snap_sibling->data_type = DATA_TYPE;
	snap_sibling->plugin = snapshot_plugin;
	snap_sibling->size = snap_child->feature_header->feature_data1_start_lsn;
	snap_sibling->node = snap_parent->node;
	strncpy(snap_sibling->name, snap_child->name, EVMS_NAME_SIZE);
	strncat(snap_sibling->name, "#snap#",
		EVMS_NAME_SIZE-strlen(snap_sibling->name));

	RETURN(volume);
}

/**
 * deallocate_snapshot
 *
 * Free the memory for the specified snapshot volume. This includes the
 * metadata, the parent object, and the volume itself. The child object
 * is not freed, but the parent object is removed from its parent list.
 * The snapshot should be removed from its origin's list before calling
 * this function.
 **/
void deallocate_snapshot(snapshot_volume_t * snap_volume)
{
	LOG_ENTRY;

	if (!snap_volume) {
		goto out;
	}

	/* Free the metadata. */
	EngFncs->engine_free(snap_volume->metadata);

	EngFncs->engine_free(snap_volume->sibling);

	if (snap_volume->parent) {
		unmake_parent_and_child(snap_volume->parent,
					snap_volume->child);
		EngFncs->free_evms_object(snap_volume->parent);
	}
	
	EngFncs->engine_free(snap_volume);

out:
	LOG_EXIT(0);
}

/**
 * is_2_4
 *
 * On 2.4 kernels, device sizes are stored in kilobytes, not sectors. When
 * creating an origin of a device with an odd number of sectors, we need to
 * round-down the origin parent size to the nearest kilobyte, or the kernel
 * will reject the creation of the origin parent. If we somehow can't
 * determine which version the kernel is, assume it's 2.4. Ugly!!!
 **/
static int is_2_4(void)
{
	char needle[]="version";
	char haystack[256];
	char * ver;
	int fd, major, minor, patch;

	fd = open("/proc/version", O_RDONLY);
	if (fd < 0) {
		return TRUE;
	}
	read(fd, haystack, 256);
	close(fd);
	ver = strstr(haystack, needle);
	if (!ver) {
		return TRUE;
	}
	sscanf(ver, "%*s %d.%d.%d", &major, &minor, &patch);

	return (major == 2 && minor == 4);
}

/**
 * allocate_origin
 *
 * Allocate a new EVMS feature-object and all necessary private data for an
 * origin. The org_child must already point to a logical_volume.
 **/
snapshot_volume_t * allocate_origin(storage_object_t * org_child)
{
	snapshot_volume_t * org_volume = NULL;
	storage_object_t * org_parent = NULL;
	char origin_name[EVMS_NAME_SIZE + 1] = {0};
	int rc;

	LOG_ENTRY;
	LOG_DEBUG("Allocating origin %s.\n", org_child->volume->name);


	/* Allocate the private data and the EVMS object. */
	org_volume = EngFncs->engine_alloc(sizeof(*org_volume));
	if (!org_volume) {
		LOG_ERROR("Memory error allocating private data for origin "
			  "%s.\n", org_child->volume->name);
		RETURN(NULL);
	}

	get_origin_object_name(org_child->name, origin_name);
	rc = EngFncs->allocate_evms_object(origin_name, &org_parent);
	if (rc) {
		EngFncs->engine_free(org_volume);
		LOG_ERROR("Memory error allocating object for origin %s.\n",
			  org_child->volume->name);
		RETURN(NULL);
	}

	/* Initialize the private data. */
	org_volume->parent = org_parent;
	org_volume->child = org_child;
	org_volume->flags |= SNAPSHOT_ORIGIN;

	/* Initialize the object. */
	org_parent->data_type = DATA_TYPE;
	org_parent->plugin = snapshot_plugin;
	org_parent->flags |= SOFLAG_MUST_BE_TOP;
	org_parent->size = org_child->size;
	org_parent->volume = org_child->volume;
	org_parent->geometry = org_child->geometry;
	org_parent->private_data = org_volume;

	/* Round-down the paren't size to nearest 1k if this is a 2.4 kernel. */
	if (is_2_4()) {
		org_parent->size &= ~(1);
	}

	make_parent_and_child(org_parent, org_child);

	/* Add the new object as the top object in the volume stack. */
	org_parent->volume->object = org_parent;

	RETURN(org_volume);
}

/**
 * deallocate_origin
 *
 * Free the memory for the specified origin volume. This includes the parent
 * object and the volume itself. The child object is not freed, but the parent
 * object is removed from its parent list, and the EVMS volume is pointed back
 * at the child object.
 **/
void deallocate_origin(snapshot_volume_t * org_volume)
{
	LOG_ENTRY;

	if (!org_volume) {
		LOG_EXIT(0);
		return;
	}

	/* Place the child object back as the top object in the volume stack. */
	if (org_volume->child) {
		org_volume->child->volume->object = org_volume->child;
	}

	if (org_volume->parent) {
		unmake_parent_and_child(org_volume->parent,
					org_volume->child);
		EngFncs->free_evms_object(org_volume->parent);
	}

	EngFncs->engine_free(org_volume);

	LOG_EXIT(0);
}

/**
 * allocate_feature_header
 *
 * Allocate and initialize a feature header for snapshotting.
 **/
evms_feature_header_t * allocate_feature_header(storage_object_t * snap_child,
						char * snap_name)
{
	evms_feature_header_t * fh;

	LOG_ENTRY;

	fh = EngFncs->engine_alloc(EVMS_FEATURE_HEADER_SECTORS *
				   EVMS_VSECTOR_SIZE);
	if (!fh) {
		LOG_ERROR("Memory error allocating feature header for "
			  "snapshot %s.\n", snap_name);
		RETURN(NULL);
	}

	fh->signature = EVMS_FEATURE_HEADER_SIGNATURE;
	fh->flags |= EVMS_FEATURE_ACTIVE;
	fh->feature_id = snapshot_plugin->id;
	fh->feature_data1_start_lsn = snap_child->size - 3;
	fh->feature_data1_size = 1;
	strncpy(fh->object_name, snap_name, EVMS_VOLUME_NAME_SIZE);

	RETURN(fh);
}

/**
 * allocate_metadata
 *
 * Allocate and initialize a snapshot metadata sector.
 **/
snapshot_metadata_t * allocate_metadata(char * org_vol_name,
					u_int64_t org_size,
					u_int64_t snap_size,
					u_int32_t chunk_size,
					int writeable)
{
	snapshot_metadata_t * metadata;

	LOG_ENTRY;

	metadata = EngFncs->engine_alloc(EVMS_VSECTOR_SIZE);
	if (!metadata) {
		LOG_ERROR("Memory error allocating metadata for snapshot.\n");
		RETURN(NULL);
	}

	metadata->signature = SNAPSHOT_SIGNATURE;
	metadata->version = snapshot_plugin->version;
	metadata->origin_size = org_size;
	metadata->chunk_size = chunk_size;
	metadata->total_chunks = calculate_data_chunks(snap_size,
						       chunk_size);
	strncpy(metadata->origin_volume,
		org_vol_name+strlen(EVMS_DEV_NODE_PATH),
		EVMS_NAME_SIZE);
	if (writeable) {
		metadata->flags |= SNAPSHOT_WRITEABLE;
	}

	RETURN(metadata);
}

/**
 * calculate_data_chunks
 *
 * Calculate the number of data chunks that a snapshot-child object can hold,
 * based on the size of that object and the snapshot chunk size.
 *
 * - total_chunks is the number of full chunks on the objects, minus a chunk
 *   for the header at the beginning.
 * - exceptions_per_cow_table is the number of data chunks that each metadata
 *   chunk can address.
 * - stride is the number of chunks in one set of metadata and data.
 * - cow_tables is the number of strides that can fit on the object, plus one
 *   if part of a stride can fit at the end.
 * - data_chunks is the total number of chunks minus the number of cow tables.
 **/
u_int64_t calculate_data_chunks(u_int64_t snap_child_size,
				u_int32_t chunk_size)
{
	u_int64_t data_chunks;
	u_int64_t total_chunks;
	u_int64_t cow_tables;
	u_int32_t exceptions_per_cow_table;
	u_int32_t stride;

	LOG_ENTRY;

	total_chunks = snap_child_size / chunk_size - 1;
	exceptions_per_cow_table = chunk_size * EVMS_VSECTOR_SIZE /
				   SIZEOF_DISK_EXCEPTION;
	stride = exceptions_per_cow_table + 1;
	cow_tables = total_chunks / stride + 
		     (total_chunks % stride) ? 1 : 0;
	data_chunks = total_chunks - cow_tables;

	RETURN(data_chunks);
}

/**
 * verify_origin
 *
 * Only other Device-Mapper devices can be snapshotted. This is because the
 * origin device's map must be replaced with an origin map, which obviously
 * won't work for non-Device-Mapper devices. Thus, when creating a new snapshot,
 * verify that the specified origin belongs to a plugin that uses Device-Mapper.
 **/
int verify_origin(storage_object_t * org_child)
{
	plugin_record_t * plugin;
	int rc, i;

	LOG_ENTRY;
	LOG_DEBUG("Verifying that %s can be used as a snapshot origin.\n",
		  org_child->name);

	/* Only create snapshots of volumes, not objects. */
	if (!org_child->volume) {
		LOG_ERROR("Only volumes can be snapshotted. Object %s is "
			  "not a volume.\n", org_child->name);
		RETURN(EINVAL);
	}

	/* Search the list of disallowed plugins. */
	for (i = 0; reject_plugins[i]; i++) {
		rc = EngFncs->get_plugin_by_name(reject_plugins[i], &plugin);
		if (rc) {
			continue;
		}

		if (org_child->plugin == plugin) {
			LOG_ERROR("Cannot create snapshots of "
				  "volumes from the %s plugin.\n",
				  plugin->short_name);
			RETURN(EINVAL);
		}
	}

	RETURN(0);
}

