/*
 *   Copyright (c) International Business Machines  Corp., 2000-2003
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: snap_discover.c
 *
 * Functions to help in the discovery of snapshot volumes.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <plugin.h>
#include "snapshot.h"

/**
 * read_metadata
 *
 * Read the snapshot metadata from the specified object. The location and
 * size of the metadata is in the object's feature header.
 **/
static int read_metadata(storage_object_t * object,
			 snapshot_metadata_t ** metadata)
{
	void * data;
	int rc;

	LOG_ENTRY;

	/* Allocate an I/O buffer. */
	data = EngFncs->engine_alloc(EVMS_VSECTOR_SIZE);
	if (!data) {
		LOG_ERROR("Memory error creating buffer to read snapshot "
			  "metadata from %s.\n", object->name);
		RETURN(ENOMEM);
	}

	rc = READ(object, object->feature_header->feature_data1_start_lsn,
		  object->feature_header->feature_data1_size, data);
	if (rc) {
		EngFncs->engine_free(data);
		LOG_ERROR("I/O error reading snapshot metadata from %s.\n",
			  object->name);
		RETURN(rc);
	}

	*metadata = data;

	RETURN(0);
}

/**
 * check_metadata_signature
 *
 * Check for a valid signature in the snapshot metadata. The metadata is
 * assumed to be in disk-order.
 **/
static inline int check_metadata_signature(snapshot_metadata_t * metadata)
{
	LOG_ENTRY;
	if (DISK_TO_CPU32(metadata->signature) != SNAPSHOT_SIGNATURE) {
		LOG_WARNING("No snapshot metadata found.\n");
		RETURN(EINVAL);
	}
	RETURN(0);
}

/**
 * check_metadata_crc
 *
 * Check the CRC in the metadata. Still assumed to be disk-order.
 **/
static inline int check_metadata_crc(snapshot_metadata_t * metadata)
{
	u_int32_t org_crc, new_crc;

	LOG_ENTRY;

	org_crc = DISK_TO_CPU32(metadata->crc);
	metadata->crc = 0;
	new_crc = EngFncs->calculate_CRC(EVMS_INITIAL_CRC,
					 metadata, sizeof(*metadata));
	if (new_crc != org_crc) {
		LOG_ERROR("Incorrect CRC found in snapshot metadata.\n");
		RETURN(EINVAL);
	}

	RETURN(0);
}

/**
 * calculate_metadata_crc
 *
 * Calculate a new CRC for the specified metadata, which should be in
 * disk-order.
 **/
void calculate_metadata_crc(snapshot_metadata_t * metadata)
{
	LOG_ENTRY;
	metadata->crc = 0;
	metadata->crc = EngFncs->calculate_CRC(EVMS_INITIAL_CRC, metadata,
					       sizeof(*metadata));
	metadata->crc = DISK_TO_CPU32(metadata->crc);
	LOG_EXIT(0);
}

/**
 * convert_metadata
 *
 * Endian-convert the snapshot metadata.
 **/
static void convert_metadata(snapshot_metadata_t * metadata)
{
	LOG_ENTRY;
	metadata->signature		= DISK_TO_CPU32(metadata->signature);
	metadata->crc			= DISK_TO_CPU32(metadata->crc);
	metadata->version.major		= DISK_TO_CPU32(metadata->version.major);
	metadata->version.minor		= DISK_TO_CPU32(metadata->version.minor);
	metadata->version.patchlevel	= DISK_TO_CPU32(metadata->version.patchlevel);
	metadata->flags			= DISK_TO_CPU32(metadata->flags);
	metadata->origin_size		= DISK_TO_CPU64(metadata->origin_size);
	metadata->total_chunks		= DISK_TO_CPU32(metadata->total_chunks);
	metadata->chunk_size		= DISK_TO_CPU32(metadata->chunk_size);
	LOG_EXIT(0);
}

/**
 * check_metadata_version
 *
 * Check that the metadata version matches this version of the snapshot
 * plugin. Metadata is now assumed to be in cpu-order.
 **/
static inline int check_metadata_version(snapshot_metadata_t * metadata)
{
	LOG_ENTRY;
	if (metadata->version.major != snapshot_plugin->version.major) {
		LOG_ERROR("Version of metadata does not match snapshot plugin version.\n");
		LOG_ERROR("Metadata: %d.%d.%d\n",
			  metadata->version.major,
			  metadata->version.minor,
			  metadata->version.patchlevel);
		LOG_ERROR("Plugin: %d.%d.%d\n",
			  snapshot_plugin->version.major,
			  snapshot_plugin->version.minor,
			  snapshot_plugin->version.patchlevel);
		RETURN(EINVAL);
	}
	RETURN(0);
}

/**
 * get_snapshot_metadata
 *
 * Read the snapshot metadata from the specified object. Verify the signature,
 * CRC, and version number. If everything checks out, endian-convert the
 * metadata.
 **/
int get_snapshot_metadata(storage_object_t * snap_child,
			  snapshot_metadata_t ** metadata)
{
	int rc;

	LOG_ENTRY;
	LOG_DEBUG("Getting snapshot metadata from %s\n", snap_child->name);

	rc = read_metadata(snap_child, metadata);
	if (rc) {
		RETURN(rc);
	}

	rc = check_metadata_signature(*metadata);
	if (rc) {
		RETURN(rc);
	}

	rc = check_metadata_crc(*metadata);
	if (rc) {
		RETURN(rc);
	}

	convert_metadata(*metadata);

	rc = check_metadata_version(*metadata);
	if (rc) {
		RETURN(rc);
	}

	RETURN(0);
}

/**
 * write_snapshot_metadata
 *
 * Write the specified snapshot's metadata to disk.
 **/
int write_snapshot_metadata(snapshot_volume_t * snap_volume)
{
	int rc;

	LOG_ENTRY;

	/* Convert from cpu-order to disk-order and get a new CRC. */
	convert_metadata(snap_volume->metadata);
	calculate_metadata_crc(snap_volume->metadata);

	rc = WRITE(snap_volume->child,
		   snap_volume->child->feature_header->feature_data1_start_lsn,
		   snap_volume->child->feature_header->feature_data1_size,
		   snap_volume->metadata);
	if (rc) {
		LOG_ERROR("I/O error writing snapshot metadata to %s.\n",
			  snap_volume->child->name);
	}

	/* Convert back to cpu-order. */
	convert_metadata(snap_volume->metadata);

	RETURN(rc);
}

/**
 * erase_snapshot_metadata
 *
 * Erase the snapshot metadata from the specified object. The object should
 * still have the feature-header attached. The engine takes care of erasing
 * the feature-headers from disk. This also clears the DM header sector.
 **/
void erase_snapshot_metadata(storage_object_t * snap_child)
{
	LOG_ENTRY;
	KILL_SECTORS(snap_child,
		     snap_child->feature_header->feature_data1_start_lsn,
		     snap_child->feature_header->feature_data1_size);
	KILL_SECTORS(snap_child, 0, 1);
	LOG_EXIT(0);
}

/**
 * find_origin
 * @origin_name: Name of the EVMS volume for the desired origin. The
 *		 /dev/evms portion of the name should be removed first.
 *
 * Get the global list of volumes, and search for the specified volume name.
 * If found, return the underlying storage object.
 **/
storage_object_t * find_origin(char * origin_name)
{
	dlist_t volumes;
	logical_volume_t * volume;
	storage_object_t * origin = NULL;
	TAG tag;
	int rc = 0;

	LOG_ENTRY;
	LOG_DEBUG("Searching for origin volume %s.\n", origin_name);

	rc = EngFncs->get_volume_list(NULL, NULL, &volumes);
	if (rc) {
		LOG_ERROR("Error getting volume list from engine.\n");
		RETURN(NULL);
	}

	while (BlindExtractObject(volumes, &tag, NULL, (void**)&volume) == 0) {
		if (!strncmp(skip_dev_path(volume->name), origin_name,
			     EVMS_VOLUME_NAME_SIZE)) {
			origin = volume->object;
			break;
		}
	}
	DestroyList(&volumes, FALSE);

	if (!origin) {
		LOG_ERROR("Cannot find origin volume %s.\n", origin_name);
	}

	RETURN(origin);
}

/**
 * check_snapshot_header
 *
 * Read the DM snapshot header from the first sector of the snapshot child
 * object and check the signature and valid bit. This sector is not stored
 * in endian-neutral format.
 **/
int check_snapshot_header(snapshot_volume_t * snap_volume)
{
	storage_object_t * snap_child = snap_volume->child;
	dm_snapshot_header_t * header;
	int rc;

	LOG_ENTRY;

	/* Allocate an I/O buffer. */
	header = EngFncs->engine_alloc(EVMS_VSECTOR_SIZE);
	if (!header) {
		LOG_ERROR("Memory error creating buffer to read DM snapshot "
			  "header from %s.\n", snap_child->name);
		RETURN(ENOMEM);
	}

	/* Read the first sector. */
	rc = READ(snap_child, 0, 1, header);
	if (rc) {
		EngFncs->engine_free(header);
		LOG_ERROR("I/O error reading DM snapshot header from %s.\n",
			  snap_child->name);
		RETURN(rc);
	}

	if (header->magic == SNAPSHOT_DM_MAGIC) {
		if (header->valid == 0) {
			/* This snapshot has been initialized by DM,
			 * and is now marked full or disabled.
			 */
			snap_volume->flags |= SNAPSHOT_INVALID;
		}
	}

	EngFncs->engine_free(header);

	RETURN(0);
}

