/*
 *   Copyright (c) International Business Machines  Corp., 2000-2003
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: snapshot.c
 *
 * This module provides support for snapshotting volumes. This function is
 * implemented as an associative feature so that it can be applied at any time
 * to an existing volume. The 'snapshot' volume is created on top of an
 * existing, unused storage object, and becomes an EVMS volume.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <plugin.h>
#include "snapshot.h"

engine_functions_t * EngFncs;
plugin_record_t * snapshot_plugin;

/**
 * snap_setup_evms_plugin
 *
 * Make a local copy of the engine functions.
 **/
static int snap_setup_evms_plugin(engine_functions_t * functions)
{
	EngFncs = functions;
	snapshot_plugin = &Snapshot_Plugin;
	LOG_ENTRY;
	RETURN(0);
}

/**
 * snap_cleanup_evms_plugin
 *
 * Get the list of all objects owned by snapshotting, and delete the snapshot
 * private data. Do not actually free the objects - the engine will handle that.
 **/
void snap_cleanup_evms_plugin(void)
{
	dlist_t tmp_list;
	storage_object_t * object;
	snapshot_volume_t * volume;
	int rc = 0;

	LOG_ENTRY;

	rc = EngFncs->get_object_list(EVMS_OBJECT, DATA_TYPE,
				      snapshot_plugin, NULL, 0, &tmp_list);
	if (!rc) {
		while (ExtractObject(tmp_list, EVMS_OBJECT_TAG,
				     NULL, (void**)&object) == 0) {
			volume = object->private_data;
			if (is_origin(volume)) {
				deallocate_origin(volume);
			} else {
				deallocate_snapshot(volume);
			}
		}
		DestroyList(&tmp_list, FALSE);
	}

	LOG_EXIT(0);
	return;
}

/**
 * snap_can_delete
 *
 * Only allow deletes of snapshot objects, not origins. The origins will be
 * automatically deleted when their last snapshot is deleted.
 **/
static int snap_can_delete(storage_object_t * object)
{
	snapshot_volume_t * volume = object->private_data;

	LOG_ENTRY;
	if (is_origin(volume)) {
		RETURN(EINVAL);
	}
	RETURN(0);
}

/**
 * snap_can_set_volume
 *
 * Origins cannot have their volume removed. Origins may have their volume
 * set if they are being renamed.
 **/
static int snap_can_set_volume(storage_object_t * object,
			       BOOLEAN creating)
{
	snapshot_volume_t * volume = object->private_data;

	LOG_ENTRY;
	if (is_origin(volume) && !creating) {
		RETURN(EINVAL);
	}
	RETURN(0);
}

/**
 * snap_discover
 * @input_objects:	List of one object, which will be the snapshot child.
 * @output_objects:	List to add the newly created snapshot object to.
 * @final_call:		Ignored in snapshotting.
 *
 * Read the snapshot metadata from the child object. Find the origin volume.
 * Create a new snapshot object. Create a new origin object (if this is the
 * first snapshot of that origin) and insert it in the origin volume's
 * stack.
 **/
static int snap_discover(dlist_t input_objects,
			 dlist_t output_objects,
			 BOOLEAN final_call)
{
	storage_object_t * org_child, * snap_child;
	snapshot_volume_t * org_volume, * snap_volume = NULL;
	snapshot_metadata_t * metadata = NULL;
	void * handle;
	int rc = EINVAL;

	LOG_ENTRY;

	/* Get the snapshot child object from the input list. */
	snap_child = dlist_get_current_item(input_objects);
	if (!snap_child) {
		LOG_ERROR("Discovery input list empty or corrupt.\n");
		goto error;
	}

	/* Read and verify the snapshot metadata. */
	rc = get_snapshot_metadata(snap_child, &metadata);
	if (rc) {
		LOG_ERROR("Error getting snapshot metadata from %s.\n",
			  snap_child->name);
		goto error;
	}

	/* Look up the origin volume. */
	org_child = find_origin(metadata->origin_volume);
	if (!org_child) {
		goto error;
	}

	/* Verify the origin volume has the correct size. */
	if (org_child->volume->vol_size != metadata->origin_size) {
		goto error;
	}

	/* Allocate a new snapshot. */
	snap_volume = allocate_snapshot(snap_child, metadata);
	if (!snap_volume) {
		goto error;
	}

	if (org_child->plugin == snapshot_plugin) {
		/* Already have at least one snapshot on this origin. */
		org_volume = org_child->private_data;
		org_child = org_volume->child;
	} else {
		/* Need a new origin. */
		org_volume = allocate_origin(org_child);
		if (!org_volume) {
			goto error;
		}
	}

	/* Update the status of the snapshot and origin from the kernel. */
	get_snapshot_state(snap_volume);
	get_origin_state(org_volume);

	/* Link the snapshot and the origin together. */
	add_snapshot_to_origin(snap_volume, org_volume);

	/* If the origin is not already active, it needs to be activated. */
	if (!is_active(org_volume)) {
		schedule_for_activate(org_volume);
	}

	/* If the snapshot is not already active, and the snapshot will become
	 * a volume, and it isn't marked invalid in the metadata, it needs to
	 * be activated.
	 */
	if (!is_active(snap_volume) &&
	    !(snap_volume->child->feature_header->flags & EVMS_VOLUME_DATA_OBJECT)) {
		check_snapshot_header(snap_volume);
		if (!(snap_volume->flags & SNAPSHOT_INVALID)) {
			schedule_for_activate(snap_volume);
		}
	}

	/* Display a warning if the snapshot is full/disabled. */
	if (snap_volume->flags & SNAPSHOT_INVALID) {
		MESSAGE("Snapshot object %s discovered in full/disabled state. "
			"Remove and recreate the volume %s to reset the "
			"snapshot of %s", snap_volume->parent->name,
			snap_volume->child->feature_header->volume_name,
			org_volume->parent->volume->name);
	}

	/* Put the snapshot parent on the output list. */
	InsertObject(output_objects, snap_volume->parent,
		     get_tag_for_object(snap_volume->parent),
		     NULL, InsertAtStart, 0, &handle);

	RETURN(0);

error:
	deallocate_snapshot(snap_volume);
	EngFncs->engine_free(metadata);
	RETURN(EVMS_FEATURE_FATAL_ERROR);
}

/**
 * snap_create
 * @objects:		List of one object, which will be the new snapshot child.
 * @options:		Array of creation options
 *			- object name
 *			- volume to snapshot
 *			- chunk size
 *			- read-only or writeable
 * @new_objects:	List to output the new snapshot object to.
 *
 * Create a new snapshot object.
 **/
static int snap_create(dlist_t objects,
		       option_array_t * options,
		       dlist_t new_objects)
{
	storage_object_t * org_child, * snap_child;
	snapshot_volume_t * org_volume, * snap_volume = NULL;
	evms_feature_header_t * feature_header = NULL;
	snapshot_metadata_t * metadata = NULL;
	char * org_vol_name, * snap_name;
	u_int32_t chunk_size;
	int writeable, rc = EINVAL;
	void * handle;

	LOG_ENTRY;

	/* Get the snapshot child object from the input list. */
	snap_child = dlist_get_current_item(objects);
	if (!snap_child) {
		LOG_ERROR("Input list corrupt.\n");
		goto error;
	}

	/* Parse and verify the options for creating the new snapshot. */
	parse_creation_options(options, &org_vol_name, &snap_name,
			       &chunk_size, &writeable);

	rc = verify_creation_options(org_vol_name, snap_name, chunk_size);
	if (rc) {
		goto error;
	}

	/* Find the specified origin volume, and check that
	 * it is valid for using as a snapshot origin.
	 */
	org_child = find_origin(skip_dev_path(org_vol_name));
	if (!org_child) {
		rc = EINVAL;
		goto error;
	}

	rc = verify_origin(org_child);
	if (rc) {
		goto error;
	}

	/* Allocate a new feature header and metadata for the snapshot. */
	feature_header = allocate_feature_header(snap_child, snap_name);
	if (!feature_header) {
		rc = ENOMEM;
		goto error;
	}
	snap_child->feature_header = feature_header;

	metadata = allocate_metadata(org_vol_name, org_child->volume->vol_size,
				     snap_child->size, chunk_size, writeable);
	if (!metadata) {
		rc = ENOMEM;
		goto error;
	}

	/* Allocate the new snapshot */
	snap_volume = allocate_snapshot(snap_child, metadata);
	if (!snap_volume) {
		rc = ENOMEM;
		goto error;
	}

	if (org_child->plugin == snapshot_plugin) {
		/* Already have at least one snapshot on this origin. */
		org_volume = org_child->private_data;
		org_child = org_volume->child;
	} else {
		/* Need a new origin. */
		org_volume = allocate_origin(org_child);
		if (!org_volume) {
			LOG_ERROR("Memory error allocating origin %s.\n",
				  org_vol_name);
			goto error;
		}
	}

	/* Link the snapshot and the origin together. */
	add_snapshot_to_origin(snap_volume, org_volume);

	/* If the origin is not already active, it needs to be activated. The
	 * snapshot that was just created should not be activated until it is
	 * made into a volume.
	 */
	if (!is_active(org_volume)) {
		schedule_for_activate(org_volume);
	}

	schedule_for_commit(snap_volume);

	InsertObject(new_objects, snap_volume->parent,
		     get_tag_for_object(snap_volume->parent),
		     NULL, AppendToList, FALSE, &handle);

	RETURN(0);

error:
	if (snap_child) {
		snap_child->feature_header = NULL;
	}
	deallocate_snapshot(snap_volume);
	EngFncs->engine_free(feature_header);
	EngFncs->engine_free(metadata);
	RETURN(rc);
}

/**
 * snap_delete
 *
 * Delete the specified snapshot object.
 *
 * If the object is currently active, all of the private data must be left
 * intact so it can be deactivated later. At deactivation time, all of the
 * private data will be deleted.
 *
 * If the object is currently inactive, everything can be deleted now.
 *
 * Either way, queue up kill-sectors to erase the metadata from disk.
 **/
static int snap_delete(storage_object_t * snap_parent,
		       dlist_t child_objects)
{
	snapshot_volume_t * snap_volume = snap_parent->private_data;
	snapshot_volume_t * org_volume = snap_volume->origin;
	storage_object_t * snap_child = snap_volume->child;
	void * handle;
	int snap_active = is_active(snap_volume);
	int org_active = is_active(org_volume);
	int rc = 0;

	LOG_ENTRY;
	LOG_DEBUG("Deleting snapshot %s.\n", snap_parent->name);

	/* Don't allow origins to be deleted. */
	rc = snap_can_delete(snap_parent);
	if (rc) {
		RETURN(rc);
	}

	unmake_parent_and_child(snap_parent, snap_child);
	remove_snapshot_from_origin(snap_volume);
	EngFncs->free_evms_object(snap_parent);

	if (snap_active) {
		/* Wait until deactivation to delete all private data. */
		schedule_for_delete(snap_volume);
	} else {
		/* Deactivate all private data now. */
		EngFncs->engine_free(snap_volume->metadata);
		EngFncs->engine_free(snap_volume->sibling);
		EngFncs->engine_free(snap_volume);
	}

	/* If the origin has no more snapshots, it can be deleted. */
	if (org_volume->count == 0) {
		LOG_DEBUG("Deleting origin %s.\n", org_volume->parent->name);
		unmake_parent_and_child(org_volume->parent, org_volume->child);
		org_volume->child->volume->object = org_volume->child;
		EngFncs->free_evms_object(org_volume->parent);
		if (org_active) {
			schedule_for_delete(org_volume);
		} else {
			EngFncs->engine_free(org_volume);
		}
	}

	erase_snapshot_metadata(snap_child);

	InsertObject(child_objects, snap_child,
		     get_tag_for_object(snap_child),
		     NULL, InsertAtStart, 0, &handle);

	RETURN(rc);
}

/**
 * snap_set_volume
 *
 * When a volume is added to a snapshot object, the first sector needs to be
 * zeroed so Device-Mapper will know it is a new snapshot.
 *
 * When a volume is added to an origin, it means the origin volume is being
 * renamed. The metadata for all of its snapshots must be updated to reflect
 * the new name.
 **/
static void snap_set_volume(storage_object_t * object,
			    BOOLEAN creating)
{
	snapshot_volume_t * volume = object->private_data;
	snapshot_volume_t * snap_volume;

	LOG_ENTRY;
	LOG_DEBUG("%s volume on %s.\n",
		  creating ? "Adding" : "Removing", object->name);

	if (is_origin(volume)) {
		if (creating) {
			for (snap_volume = volume->next;
			     snap_volume; snap_volume = snap_volume->next) {
				strncpy(snap_volume->metadata->origin_volume,
					skip_dev_path(object->volume->name),
					EVMS_VOLUME_NAME_SIZE);
				schedule_for_commit(snap_volume);
			}
		}
	} else {
		if (creating) {
			KILL_SECTORS(volume->child, 0, 1);
			schedule_for_commit(volume);
			schedule_for_activate(volume);
		} else {
			schedule_for_deactivate(volume);
		}
	}

	LOG_EXIT(0);
}

/**
 * snap_add_sectors_to_kill_list
 *
 * We don't really expect to get called on this API.
 **/
static int snap_add_sectors_to_kill_list(storage_object_t * object,
					 lsn_t lsn,
					 sector_count_t count)
{
	LOG_ENTRY;
	LOG_WARNING("WARNING!!! Called for Kill-Sectors!\n");
	RETURN(0);
}

/**
 * snap_commit_changes
 *
 * Write out the metadata sector for the specified snapshot object. Clearing
 * the snapshot header sector is handled in set_volume.
 **/
static int snap_commit_changes(storage_object_t * object,
			       commit_phase_t phase)
{
	snapshot_volume_t * volume = object->private_data;
	int rc = 0;

	LOG_ENTRY;

	/* Only have one copy of metadata. */
	if (phase != FIRST_METADATA_WRITE) {
		RETURN(rc);
	}

	LOG_DEBUG("Committing changes for %s.\n", object->name);

	/* Only snapshots have metadata. */
	if (!is_origin(volume)) {
		rc = write_snapshot_metadata(volume);
	}

	if (!rc) {
		commit_complete(volume);
	}

	RETURN(rc);
}

/**
 * snap_activate
 *
 * Activate a snapshot or an origin. Origins are now expected to be activated
 * independently, and prior to activating any of their snapshots.
 **/
static int snap_activate(storage_object_t * object)
{
	snapshot_volume_t * volume = object->private_data;
	int rc = 0;

	LOG_ENTRY;

	if (is_active(volume)) {
		/* Re-activating isn't allowed. */
		goto out;
	}

	LOG_DEBUG("Activating %s.\n", object->name);

	if (is_origin(volume)) {
		rc = activate_origin(volume);
	} else {
		if (is_active(volume->origin)) {
			rc = activate_snapshot(volume);
		} else {
			LOG_ERROR("Cannot activate snapshot %s with an "
				  "inactive origin.\n", object->name);
			rc = EINVAL;
		}
	}

out:
	if (!rc) {
		activate_complete(volume);
	}
	RETURN(rc);
}

/**
 * snap_deactivate
 * 
 * Deactivate a snapshot or an origin. If it is marked for a pending delete,
 * finish freeing all the private data. All snapshots must be deactivated
 * before the origin can be deactivated.
 **/
static int snap_deactivate(storage_object_t * object)
{
	snapshot_volume_t * volume = object->private_data;

	LOG_ENTRY;

	if (!is_active(volume)) {
		RETURN(0);
	}

	LOG_DEBUG("Deactivating %s.\n", object->name);

	if (is_origin(volume)) {
		if (volume->active_count) {
			LOG_ERROR("Cannot deactivate origin %s with %d active "
				  "snapshots.\n", object->name,
				  volume->active_count);
			RETURN(EINVAL);
		}
		deactivate_origin(volume);
	} else {
		deactivate_snapshot(volume);
	}

	deactivate_complete(volume);

	if (is_delete_pending(volume)) {
		if (!is_origin(volume)) {
			EngFncs->engine_free(volume->metadata);
			EngFncs->engine_free(volume->sibling);
		}
		EngFncs->engine_free(volume);
	}

	RETURN(0);
}

/**
 * snap_get_option_count
 *
 * Return the number of available options for the specified task.
 **/
static int snap_get_option_count(task_context_t * context)
{
	int count = 0;

	LOG_ENTRY;

	switch (context->action) {

	case EVMS_Task_Create:
		count = SNAP_OPTION_WRITEABLE_INDEX + 1;
		break;

	case SNAPSHOT_FUNCTION_ROLLBACK:
	case SNAPSHOT_FUNCTION_RESET:
		count = 0;
		break;

	default:
		count = -1;
	}

	RETURN(count);
}

/** 
 * snap_init_task
 *
 * Prepare an option descriptor for the specified task.
 **/
static int snap_init_task(task_context_t * context)
{
	int rc;

	LOG_ENTRY;

	switch (context->action) {

	case EVMS_Task_Create:
		rc = init_task_create(context);
		break;

	case SNAPSHOT_FUNCTION_ROLLBACK:
	case SNAPSHOT_FUNCTION_RESET:
		rc = 0;
		break;

	default:
		rc = EINVAL;
		break;
	}

	RETURN(rc);
}

/**
 * snap_set_option
 *
 * Set an option for the specified task.
 **/
static int snap_set_option(task_context_t * context,
			   u_int32_t index,
			   value_t * value,
			   task_effect_t * effect)
{
	int rc;

	LOG_ENTRY;

	switch (context->action) {

	case EVMS_Task_Create:
		rc = set_option_create(context, index, value, effect);
		break;

	case SNAPSHOT_FUNCTION_ROLLBACK:
	case SNAPSHOT_FUNCTION_RESET:
		rc = 0;
		break;

	default:
		rc = EINVAL;
		break;
	}

	RETURN(rc);
}

/**
 * snap_set_objects
 *
 * Verify a selected object for the specified task.
 **/
static int snap_set_objects(task_context_t * context,
			    dlist_t declined_objects,
			    task_effect_t * effect)
{
	int rc;

	LOG_ENTRY;

	switch (context->action) {

	case EVMS_Task_Create:
		rc = set_objects_create(context, declined_objects, effect);
		break;

	case SNAPSHOT_FUNCTION_ROLLBACK:
	case SNAPSHOT_FUNCTION_RESET:
		rc = 0;
		break;

	default:
		rc = EINVAL;
		break;
	}

	RETURN(rc);
}

/**
 * snap_get_info
 **/
static int snap_get_info(storage_object_t * object,
			 char * name,
			 extended_info_array_t ** info_array)
{
	extended_info_array_t * info;
	snapshot_volume_t * volume = object->private_data;
	snapshot_volume_t * snap_volume;
	int i = 0;

	LOG_ENTRY;

	if (name) {
		LOG_ERROR("No support for extra information about "
			  "\"%s\" for object %s.\n", name, object->name);
		RETURN(EINVAL);
	}

	if (is_origin(volume)) {
		/* Origin */
		info = EngFncs->engine_alloc(sizeof(extended_info_array_t) +
					     sizeof(extended_info_t) * volume->count);
		if (!info) {
			RETURN(ENOMEM);
		}

		for (i = 0, snap_volume = volume->next;
		     snap_volume; snap_volume = snap_volume->next, i++) {
			if (is_active(snap_volume)) {
				SET_STRING(info->info[i].name, "SnapShot");
				SET_STRING(info->info[i].title, "SnapShotted on");
				SET_STRING(info->info[i].desc, "Snapshots of this volume");
				info->info[i].type = EVMS_Type_String;
				SET_STRING(info->info[i].value.s, snap_volume->parent->volume->name);
			} else {
				SET_STRING(info->info[i].name, "iSnapShot");
				SET_STRING(info->info[i].title, "Inactive Snapshot on");
				SET_STRING(info->info[i].desc, "This object is set up to snapshot this volume. If the object is made into a volume the snapshot will be reset and activated");
				info->info[i].type = EVMS_Type_String;
				SET_STRING(info->info[i].value.s, snap_volume->parent->name);
			}
		}
	} else {
		/* Snapshot */
		info = EngFncs->engine_alloc(sizeof(extended_info_array_t) +
					     sizeof(extended_info_t) * 4);
		if (!info) {
			RETURN(ENOMEM);
		}

		get_snapshot_state(volume);

		if (is_active(volume)) {
			SET_STRING(info->info[i].name, "Original");
			SET_STRING(info->info[i].title, "SnapShot of");
			SET_STRING(info->info[i].desc, "Indicates which volume this volume is a snapshot of.");
			info->info[i].type = EVMS_Type_String;
			SET_STRING(info->info[i].value.s, volume->origin->parent->volume->name);
		} else {
			SET_STRING(info->info[i].name, "iOriginal");
			SET_STRING(info->info[i].title, "Inactive SnapShot of");
			SET_STRING(info->info[i].desc, "The volume which this object is a snapshot of.  Making this object a volume will reset and activate the snapshot.");
			info->info[i].type = EVMS_Type_String;
			SET_STRING(info->info[i].value.s, volume->origin->parent->volume->name);
		}
		i++;

		SET_STRING(info->info[i].name, "ChunkSize");
		SET_STRING(info->info[i].title, "Chunk Size");
		SET_STRING(info->info[i].desc, "The size of the chunks which are copied to this snapshot");
		info->info[i].type = EVMS_Type_Unsigned_Int32;
		info->info[i].unit = EVMS_Unit_Sectors;
		info->info[i].value.ui32 = volume->metadata->chunk_size;
		i++;

		SET_STRING(info->info[i].name, "Writeable");
		SET_STRING(info->info[i].title, "Writeable");
		SET_STRING(info->info[i].desc, "Whether snapshot is writeable or read-only");
		info->info[i].type = EVMS_Type_Boolean;
		info->info[i].value.bool = (volume->metadata->flags & SNAPSHOT_WRITEABLE);
		i++;

		SET_STRING(info->info[i].name, "State");
		SET_STRING(info->info[i].title, "State");
		SET_STRING(info->info[i].desc, "Current state of the snapshot.");
		info->info[i].type = EVMS_Type_String;
		if (!is_active(volume)) {
			SET_STRING(info->info[i].value.s, "Inactive");
		} else if (volume->flags & SNAPSHOT_INVALID) {
			SET_STRING(info->info[i].value.s, "Disabled/Full");
		} else {
			SET_STRING(info->info[i].value.s, "Active");
		}
		i++;

		SET_STRING(info->info[i].name, "PercentFull");
		SET_STRING(info->info[i].title, "Percent Full");
		SET_STRING(info->info[i].desc, "Percentage of the snapshot device that has been used to save data from the origin.");
		if (volume->percent_full < 0) {
			info->info[i].type = EVMS_Type_String;
			SET_STRING(info->info[i].value.s, "Unknown");
		} else {
			info->info[i].type = EVMS_Type_Unsigned_Int;
			info->info[i].unit = EVMS_Unit_Percent;
			info->info[i].value.ui = volume->percent_full;
		}
		i++;
	}

	info->count = i;
	*info_array = info;
	RETURN(0);
}

/**
 * snap_get_plugin_info
 *
 * Return information about the Snapshot plugin. There is no "extra"
 * information about Snapshot, so "name" should always be NULL.
 **/
static int snap_get_plugin_info(char * name,
				extended_info_array_t ** info_array)
{
	extended_info_array_t * info = NULL;
	char buffer[50] = {0};
	int  i = 0;

	LOG_ENTRY;

	if (name) {
		LOG_ERROR("No support for extra plugin information about "
			  "\"%s\"\n", name);
		RETURN(EINVAL);
	}

	info = EngFncs->engine_alloc(sizeof(extended_info_array_t) +
				     sizeof(extended_info_t)*5);
	if (!info) {
		LOG_ERROR("Error allocating memory for plugin info array\n");
		RETURN(ENOMEM);
	}

	/* Short Name */
	SET_STRING(info->info[i].name, "ShortName");
	SET_STRING(info->info[i].title, "Short Name");
	SET_STRING(info->info[i].desc, "A short name given to this plugin");
	info->info[i].type = EVMS_Type_String;
	SET_STRING(info->info[i].value.s, snapshot_plugin->short_name);
	i++;

	/* Long Name */
	SET_STRING(info->info[i].name, "LongName");
	SET_STRING(info->info[i].title, "Long Name");
	SET_STRING(info->info[i].desc, "A long name given to this plugin");
	info->info[i].type = EVMS_Type_String;
	SET_STRING(info->info[i].value.s, snapshot_plugin->long_name);
	i++;

	/* Plugin Type */
	SET_STRING(info->info[i].name, "Type");
	SET_STRING(info->info[i].title, "Plugin Type");
	SET_STRING(info->info[i].desc, "There are various types of plugins; each responsible for some kind of storage object.");
	info->info[i].type = EVMS_Type_String;
	SET_STRING(info->info[i].value.s, "Associative Feature");
	i++;

	/* Plugin Version */
	SET_STRING(info->info[i].name, "Version");
	SET_STRING(info->info[i].title, "Plugin Version");
	SET_STRING(info->info[i].desc, "This is the version number of the plugin.");
	info->info[i].type = EVMS_Type_String;
	snprintf(buffer, 50, "%d.%d.%d",
		 snapshot_plugin->version.major,
		 snapshot_plugin->version.minor,
		 snapshot_plugin->version.patchlevel);
	SET_STRING(info->info[i].value.s, buffer);
	i++;

	/* Required Engine Services Version */
	SET_STRING(info->info[i].name, "Required_Engine_Version");
	SET_STRING(info->info[i].title, "Required Engine Services Version");
	SET_STRING(info->info[i].desc, "This is the version of the Engine services that this plug-in requires.  It will not run on older versions of the Engine services.");
	info->info[i].type = EVMS_Type_String;
	snprintf(buffer, 50, "%d.%d.%d",
		 snapshot_plugin->required_engine_api_version.major,
		 snapshot_plugin->required_engine_api_version.minor,
		 snapshot_plugin->required_engine_api_version.patchlevel);
	SET_STRING(info->info[i].value.s, buffer);
	i++;

	/* Required Plug-in API Version */
	SET_STRING(info->info[i].name, "Required_Plugin_Version");
	SET_STRING(info->info[i].title, "Required Plug-in API Version");
	SET_STRING(info->info[i].desc, "This is the version of the Engine plug-in API that this plug-in requires.  It will not run on older versions of the Engine plug-in API.");
	info->info[i].type = EVMS_Type_String;
	snprintf(buffer, 50, "%d.%d.%d",
		 snapshot_plugin->required_plugin_api_version.plugin.major,
		 snapshot_plugin->required_plugin_api_version.plugin.minor,
		 snapshot_plugin->required_plugin_api_version.plugin.patchlevel);
	SET_STRING(info->info[i].value.s, buffer);
	i++;

	info->count = i;
	*info_array = info;

	RETURN(0);
}

/**
 * snap_read
 *
 * Reads are only allowed to the origin volume. Reads from the snapshot
 * would essentially be "around" the snapshot kernel code, and hence would
 * not reflect the correct state of the snapshot.
 **/
static int snap_read(storage_object_t * object,
		     lsn_t lsn,
		     sector_count_t count,
		     void * buffer)
{
	snapshot_volume_t * volume = object->private_data;
	int rc = ENOSYS;

	LOG_ENTRY;

	if (is_origin(volume)) {
		rc = READ(volume->child, lsn, count, buffer);
	} else {
		LOG_ERROR("No reading from snapshot volumes within the engine!\n");
	}

	RETURN(rc);
}

/**
 * snap_write
 *
 * No writes are allowed to snapshot or origin objects within the engine.
 * These writes would essentially be "around" the snapshot kernel code, and
 * thus not actually be snapshotted.
 **/
static int snap_write(storage_object_t * object,
		      lsn_t lsn,
		      sector_count_t count,
		      void * buffer)
{
	LOG_ENTRY;
	LOG_ERROR("No writing to snapshot or origin volumes within the engine!\n");
	RETURN(ENOSYS);
}

/**
 * snap_get_plugin_functions
 **/
static int snap_get_plugin_functions(storage_object_t * object,
				     function_info_array_t ** functions)
{
#if 0
	int rc = 0;
	function_info_array_t * fia;
	int function_count;
	snapshot_volume_t * snap_volume;

	LOG_ENTRY;

	/*
	 * If object is NULL, that means the user is asking for private actions
	 * on the plug-in.  We don't have any private actions that are global
	 * for the plug-in.
	 */
	if (object == NULL) {
		RETURN(ENOSYS);
	}

	snap_volume = (snapshot_volume_t *) object->private_data;

	fia = EngFncs->engine_alloc(sizeof(function_info_array_t) + sizeof(function_info_t) * (SNAPSHOT_FUNCTION_COUNT - 1));

	if (fia != NULL) {
		function_count = 0;

		/*
		 * Our functions are only allowed if the Engine is opened for
		 * writing.
		 */
		if (EngFncs->get_engine_mode() == ENGINE_READWRITE) {

			/*
			 * If snap_volume->snapshot_org == NULL then this is
			 * the snapshot object placed on the original.  There
			 * are no funtions available on an original.
			 */
			if (snap_volume->snapshot_org != NULL) {

				/*
				 * If the snapshot can be rolled back, add the
				 * function to the list.
				 */
				if (can_rollback(snap_volume) == 0) {
					fia->info[function_count].function = SNAPSHOT_FUNCTION_ROLLBACK;
					SET_STRING(fia->info[function_count].name, "rollback");
					SET_STRING(fia->info[function_count].title, "Roll back");
					SET_STRING(fia->info[function_count].verb, "Roll back");
					SET_STRING(fia->info[function_count].help, "Roll back the contents of the snapshot to the original volume.");

					function_count++;
				}

				/*
				 * If the snapshot is active. i.e., it is part
				 * of a volume, and it is not marked for
				 * rollback, then the snapshot can be reset.
				 */
				if ((object->volume != NULL) &&
				    (!(snap_volume->meta_data.flags & SNAPSHOT_ROLLBACK)))  {
					// Only allow reset on disabled snapshots for now.
					// The call to can_rollback() above got
					// the currenct state of the flags.
					if (snap_volume->meta_data.flags & SNAPSHOT_DISABLED) {
						fia->info[function_count].function = SNAPSHOT_FUNCTION_RESET;
						SET_STRING(fia->info[function_count].name, "reset");
						SET_STRING(fia->info[function_count].title, "Reset");
						SET_STRING(fia->info[function_count].verb, "Reset");
						SET_STRING(fia->info[function_count].help, "Reset the snapshot so that it keeps the current contents of the original volume.");

						function_count++;
					}
				}
			}
		}

		fia->count = function_count;
		*functions = fia;

	} else {
		LOG_CRITICAL("Error allocating memory for an action info array.\n");
		rc = ENOMEM;
	}

	RETURN(rc);
#endif
	return ENOSYS;
}

/**
 * snap_plugin_function
 **/
static int snap_plugin_function(storage_object_t * object,
				task_action_t action,
				dlist_t objects,
				option_array_t * options)
{
#if 0
	int rc = 0;
	snapshot_volume_t * snap_volume = (snapshot_volume_t *) object->private_data;

	LOG_ENTRY;

	if ((action < EVMS_Task_Plugin_Function) ||
	    (action >= EVMS_Task_Plugin_Function + SNAPSHOT_FUNCTION_COUNT)) {
		LOG_ERROR("Action code 0x%x is out of range.\n", action);
		RETURN(EINVAL);
	}

	if (snap_volume == NULL) {
		LOG_ERROR("Object %s does not have snapshot private data.\n", object->name);
		RETURN(EINVAL);
	}

	/*
	 * If snap_volume->snapshot_org == NULL then this is the snapshot object
	 * placed on the original.  There are no funtions available on an
	 * original.
	 */
	if (snap_volume->snapshot_org == NULL) {
		LOG_ERROR("Cannot perform snapshot functions on object %s.\n", object->name);
		RETURN(EINVAL);
	}

	switch (action) {
	case SNAPSHOT_FUNCTION_ROLLBACK:
		if (can_rollback(snap_volume) == 0) {
			int    answer = 0;
			char * choice_text[3] = { "Don't Rollback", "Rollback", NULL};
			EngFncs->user_message(snapshot_plugin, &answer,choice_text,
					      "***WARNING*** Selecting Rollback will cause the Snapshot "
					      "to be copied back to the original volume.  All changes "
					      "made to the original volume (%s) since the Snapshot was started "
					      "will be lost.\n\n",
					      snap_volume->snapshot_org->parent->volume->name);
			if (answer) {
				mark_for_rollback(snap_volume);
			}

		} else {
			LOG_ERROR("Cannot rollback the snapshot because the original volume \"%s\" is mounted on \"%s\".\n",
				  snap_volume->snapshot_org->parent->volume->name,
				  snap_volume->snapshot_org->parent->volume->mount_point);
			rc = EINVAL;
		}
		break;

	case SNAPSHOT_FUNCTION_RESET:
		get_snap_state(snap_volume);

		/*
		 * If the snapshot is active. i.e., it is part
		 * of a volume, and it is not marked for
		 * rollback, then the snapshot can be reset.
		 */
		if ((object->volume != NULL) &&
		    (!(snap_volume->meta_data.flags & SNAPSHOT_ROLLBACK)))  {
			/*
			 * If the snapshot is currently active, confirm the reset.
			 */
			if (!(snap_volume->meta_data.flags & SNAPSHOT_DISABLED)) {

//				int    answer = 0;
//				char * choice_text[3] = { "Don't Reset", "Reset", NULL};
//				EngFncs->user_message(snapshot_plugin, &answer,choice_text,
//						      "***WARNING*** Selecting Reset will cause all the data saved in the Snapshot to be lost.");
//				if (answer == 0) {
//					RETURN(0);
//				}

				// For now, don't allow resets of active
				// snapshots until we figure out how to handle
				// resetting async snapshots.
				LOG_ERROR("Reset is not allowed on active snapshots.\n");
				RETURN(EBUSY);
			}

			/*
			 * SnapSetVolume() has all the code needed to
			 * start a new snapshot...
			 */
			snap_set_volume(object, TRUE);
			/*
			 * except the code to rediscover the volume.
			 */
			EngFncs->rediscover_volume(object->volume, FALSE);
			EngFncs->set_changes_pending();
		}
		break;

	default:
		break;
	}

	RETURN(rc);
#endif
	return ENOSYS;
}

/**
 * can_rollback
 **/
int can_rollback(snapshot_volume_t * snap_volume)
{
#if 0
	snapshot_volume_t * org_volume = snap_volume->snapshot_org;
	int rc = EINVAL;

	LOG_ENTRY;

	/* Must be the only snapshot. */
	if (snap_volume->snapshot_org->snapshot_next == snap_volume &&
	    snap_volume->snapshot_next == NULL) {
		/* And not dirty. */
		if (!(snap_volume->parent->flags & SOFLAG_DIRTY)) {
			get_snap_state(snap_volume);
			/* And not disabled or already rolled back. */
			if (!(snap_volume->meta_data.flags & (SNAPSHOT_DISABLED|SNAPSHOT_ROLLBACK_COMP))) {
				/* And must be a volume. */
				if (snap_volume->parent->volume) {
					/* Snapshot can not be mounted
					 * and original can not be mounted
					 */
					if (!EngFncs->is_mounted(snap_volume->parent->volume->name, NULL) &&
					    !EngFncs->is_mounted(org_volume->parent->volume->name, NULL)) {
						/* Ok, to rollback. */
						rc = 0;
					}
				}
			}
		}
	}
	RETURN(rc);
#endif
	return 0;
}

/* Snapshot function table and plugin record. */
static plugin_functions_t snapshot_functions = {
	.setup_evms_plugin		= snap_setup_evms_plugin,
	.cleanup_evms_plugin		= snap_cleanup_evms_plugin,
	.can_delete			= snap_can_delete,
	.can_set_volume			= snap_can_set_volume,
	.discover			= snap_discover,
	.create				= snap_create,
	.delete				= snap_delete,
	.set_volume			= snap_set_volume,
	.add_sectors_to_kill_list	= snap_add_sectors_to_kill_list,
	.commit_changes			= snap_commit_changes,
	.activate			= snap_activate,
	.deactivate			= snap_deactivate,
	.get_option_count		= snap_get_option_count,
	.init_task			= snap_init_task,
	.set_option			= snap_set_option,
	.set_objects			= snap_set_objects,
	.get_info			= snap_get_info,
	.get_plugin_info		= snap_get_plugin_info,
	.read				= snap_read,
	.write				= snap_write,
	.get_plugin_functions		= snap_get_plugin_functions,
	.plugin_function		= snap_plugin_function
};

plugin_record_t Snapshot_Plugin = {
	.id = SetPluginID(IBM_OEM_ID,
			  EVMS_ASSOCIATIVE_FEATURE,
			  SNAPSHOT_FEATURE_ID),
	.version = {
		.major		= MAJOR_VERSION,
		.minor		= MINOR_VERSION,
		.patchlevel	= PATCH_LEVEL
	},
	.required_engine_api_version = {
		.major		= 9,
		.minor		= 0,
		.patchlevel	= 0
	},
	.required_plugin_api_version = {
		.plugin = {
			.major		= 10,
			.minor		= 0,
			.patchlevel	= 0
		}
	},
	.short_name = "Snapshot",
	.long_name = "Snapshot Feature",
	.oem_name = "IBM",
	.functions = {
		.plugin = &snapshot_functions
	},
	.container_functions = NULL
};

plugin_record_t * evms_plugin_records[] = {
	&Snapshot_Plugin,
	NULL
};

