/*
 *   Copyright (c) International Business Machines  Corp., 2000-2003
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifndef __SNAPSHOT_H__
#define __SNAPSHOT_H__ 1

#define SNAPSHOT_FEATURE_ID		104
#define SNAPSHOT_SIGNATURE		0x536e4170	/* SnAp */
#define SNAPSHOT_ORIGIN_SIGNATURE	0x4f724967	/* OrIg */

/* Status flags */
#define SNAPSHOT			(1 << 0)
#define SNAPSHOT_ORIGIN			(1 << 1)
#define SNAPSHOT_INVALID		(1 << 2)
#define SNAPSHOT_WRITEABLE 		(1 << 3)
#define SNAPSHOT_DELETE_PENDING		(1 << 4)

/* Private ioctl commands */
#define SNAPSHOT_QUERY_PERCENT_FULL	1
#define SNAPSHOT_START_ROLLBACK		2
#define SNAPSHOT_CHECK_STATE		3

/* Snapshot chunk sizes can range from 8kB to 1MB. The
 * default size is 64kB. These values are in sectors.
 */
#define SNAPSHOT_DEFAULT_CHUNK_SIZE	128
#define SNAPSHOT_MIN_CHUNK_SIZE		16
#define SNAPSHOT_MAX_CHUNK_SIZE		2048

/* These are dependent on the definitions from
 * drivers/md/dm-exception-store.c in the kernel.
 */
#define SNAPSHOT_DM_MAGIC		0x70416e53
#define SNAPSHOT_DM_DISK_VERSION	1
#define SIZEOF_DISK_EXCEPTION		16

/**
 * struct snapshot_metadata
 * @signature:		0  : SNAPSHOT_SIGNATURE
 * @CRC:		4  :
 * @version:		8  : Major, minor, patchlevel
 * @flags:		20 : SNAPSHOT_*
 * @original_volume:	24 : Name of volume being snapshotted
 * @original_size:	152: In sectors
 * @total_chunks:	160: Number of data chunks this snapshot can hold
 * @chunk_size:		168: In sectors
 *
 * On-disk metadata sector for EVMS Snapshot feature.
 */
typedef struct snapshot_metadata {
	u_int32_t		signature;
	u_int32_t		crc;
	struct evms_version	version;
	u_int32_t		flags;
	u_int8_t		origin_volume[128];
	u_int64_t		origin_size;
	u_int64_t		total_chunks;
	u_int32_t		chunk_size;
} snapshot_metadata_t;

/**
 * struct dm_snapshot_header
 * @magic:	SNAPSHOT_DM_MAGIC
 * @valid:	Non-zero if the snapshot is valid.
 * @version:	SNAPSHOT_DM_DISK_VERSION
 * @chunk_size:	In sectors
 *
 * Header defined in drivers/md/dm-exception-store.c in the kernel. This
 * structure is found on the first sector of the snapshot object (only after
 * the snapshot has been initialized by Device-Mapper).
 **/
typedef struct dm_snapshot_header {
	u_int32_t magic;
	int valid;
	u_int32_t version;
	u_int32_t chunk_size;
} dm_snapshot_header_t;

/**
 * struct snapshot_volume
 * @signature:		SNAPSHOT or SNAPSHOT_ORIGIN
 * @parent:		Our exported object
 * @child:		Our imported object
 * @sibling:		A secondary child object. Hidden from the engine.
 * @origin:		The volume being snapshotted
 * @next:		List of snapshots for one original
 * @metadata:		Copy of the on-disk metadata
 * @flags:		SNAPSHOT_*
 * @count:		Number of snapshots for this origin.
 * @active_count:	Number of activated snapshots for this origin.
 *
 * The in-memory representation for both a snapshot and an original.
 */
typedef struct snapshot_volume {
	storage_object_t	* parent;
	storage_object_t	* child;
	storage_object_t	* sibling;
	struct snapshot_volume	* next;
	struct snapshot_volume	* origin;
	snapshot_metadata_t	* metadata;
	unsigned int		flags;
	unsigned int		count;
	unsigned int		active_count;
	int			percent_full;
} snapshot_volume_t;

/* Log message macros. */
#define MESSAGE(msg, args...)		EngFncs->user_message(snapshot_plugin, NULL, NULL, msg , ## args)
#define LOG_CRITICAL(msg, args...)	EngFncs->write_log_entry(CRITICAL,	snapshot_plugin, "%s: " msg, __FUNCTION__ , ## args)
#define LOG_SERIOUS(msg, args...)	EngFncs->write_log_entry(SERIOUS,	snapshot_plugin, "%s: " msg, __FUNCTION__ , ## args)
#define LOG_ERROR(msg, args...)		EngFncs->write_log_entry(ERROR,		snapshot_plugin, "%s: " msg, __FUNCTION__ , ## args)
#define LOG_WARNING(msg, args...)	EngFncs->write_log_entry(WARNING,	snapshot_plugin, "%s: " msg, __FUNCTION__ , ## args)
#define LOG(msg, args...)		EngFncs->write_log_entry(DEFAULT,	snapshot_plugin, "%s: " msg, __FUNCTION__ , ## args)
#define LOG_DETAILS(msg, args...)	EngFncs->write_log_entry(DETAILS,	snapshot_plugin, "%s: " msg, __FUNCTION__ , ## args)
#define LOG_DEBUG(msg, args...)		EngFncs->write_log_entry(DEBUG,		snapshot_plugin, "%s: " msg, __FUNCTION__ , ## args)
#define LOG_EXTRA(msg, args...)		EngFncs->write_log_entry(EXTRA,		snapshot_plugin, "%s: " msg, __FUNCTION__ , ## args)
#define LOG_PROC(msg, args...)		EngFncs->write_log_entry(ENTRY_EXIT,	snapshot_plugin, "%s: " msg, __FUNCTION__ , ## args)
#define LOG_EVERYTHING(msg, args...)	EngFncs->write_log_entry(EVERYTHING,	snapshot_plugin, "%s: " msg, __FUNCTION__ , ## args)

#define	LOG_ENTRY			LOG_PROC("Enter.\n")
#define LOG_EXIT(x) 			LOG_PROC("Exit. rc = %d\n", x)
#define RETURN(x)			do { LOG_EXIT(x); return(x); } while (0)

/* Macro: SET_STRING
 *
 * Allocate a string for "a" and copy "b" to "a".
 */
#define SET_STRING(a,b)					\
{							\
	EngFncs->engine_free(a);			\
	(a) = EngFncs->engine_alloc(strlen(b)+1);	\
	if (a) {					\
		strcpy(a, b);				\
	} else {					\
		RETURN(ENOMEM);				\
	}						\
}

/* Macro: SET_POWER2_LIST
 *
 * Set a contraint list that increments each item by a power of 2.
 * b is the starting value, c is ending value (both must be power of 2).
 */
#define SET_POWER2_LIST(a, b, c)					\
{									\
	if (((b) & ((b)-1)) || ((c) & ((c)-1))) {			\
		RETURN(EINVAL);						\
	}								\
	(a) = EngFncs->engine_alloc(calc_log2(c)*sizeof(value_list_t));	\
	if (a) {							\
		long i = (b);						\
		(a)->count = 0;						\
		while (i <= (c)) {					\
			(a)->value[(a)->count].ui32 = i;		\
			(a)->count++;					\
			i <<= 1;					\
		}							\
	} else {							\
		RETURN(ENOMEM);						\
	}								\
}

#define READ(object, start, len, buffer)  object->plugin->functions.plugin->read(object, start, len, buffer)
#define WRITE(object, start, len, buffer) object->plugin->functions.plugin->write(object, start, len, buffer)
#define KILL_SECTORS(object, start, len)  object->plugin->functions.plugin->add_sectors_to_kill_list(object, start, len)

static inline int is_active(snapshot_volume_t * volume)
{
	return (volume->parent->flags & SOFLAG_ACTIVE);
}

static inline int is_origin(snapshot_volume_t * volume)
{
	return (volume->flags & SNAPSHOT_ORIGIN);
}

static inline int is_delete_pending(snapshot_volume_t * volume)
{
	return (volume->flags & SNAPSHOT_DELETE_PENDING);
}

static inline void schedule_for_delete(snapshot_volume_t * volume)
{
	volume->flags |= SNAPSHOT_DELETE_PENDING;
}

static inline void schedule_for_activate(snapshot_volume_t * volume)
{
	volume->parent->flags |= SOFLAG_NEEDS_ACTIVATE;
}

static inline void activate_complete(snapshot_volume_t * volume)
{
	volume->parent->flags &= ~SOFLAG_NEEDS_ACTIVATE;
}

static inline void schedule_for_deactivate(snapshot_volume_t * volume)
{
	volume->parent->flags |= SOFLAG_NEEDS_DEACTIVATE;
}

static inline void deactivate_complete(snapshot_volume_t * volume)
{
	volume->parent->flags &= ~SOFLAG_NEEDS_DEACTIVATE;
}

static inline void schedule_for_commit(snapshot_volume_t * volume)
{
	volume->parent->flags |= SOFLAG_DIRTY;
}

static inline void commit_complete(snapshot_volume_t * volume)
{
	volume->parent->flags &= ~SOFLAG_DIRTY;
}

extern engine_functions_t * EngFncs;
extern plugin_record_t * snapshot_plugin;
extern plugin_record_t Snapshot_Plugin;

#include "snap_activate.h"
#include "snap_create.h"
#include "snap_discover.h"
#include "snap_options.h"

#endif

