/*
 *
 *   Copyright (c) International Business Machines  Corp., 2002, 2003
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or 
 *   (at your option) any later version.
 * 
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software 
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */ 

#include <time.h>
#include <glib.h>
#include <ncurses.h>
#include <panel.h>
#include <frontend.h>

#include "common.h"
#include "window.h"
#include "menu.h"
#include "dialog.h"
#include "views.h"
#include "callbacks.h"
#include "save.h"

/**
 *	save_button_activated - invoked to save changes in the engine to disk
 *	@item: the save button that invoked this function
 *
 *	This routine is invoked when the user activates the Save button in 
 *	either the quit or save confirmation dialogs.
 */
int save_button_activated(struct menu_item *item)
{
	int rc, *event_rc;
	struct dialog_window *dialog = item->user_data;

	enable_message_queuing();
	rc = evms_commit_changes();
	disable_message_queuing();
	display_queued_messages();

	update_status(_("Save"), rc);
	event_rc = dialog->user_data;
	if (rc == 0) {
		/*
		 * If we have an event return code to set, set it so that the
		 * application quits the main event loop and exits nicely. 
		 * Don't bother refreshing the views as we are exiting anyways.
		 */
		if (event_rc != NULL)
			*event_rc = ESHUTDOWN;
		else
			refresh_views();
	} else if (event_rc != NULL) {
		/*
		 * Commit failed so set the error so that we don't exit the app.
		 */
		*event_rc = EIO;
	}
	dialog->status = DLG_STATUS_CLOSING;

	return 0;
}

/**
 *	cancel_save_button_activated - invoked when user does not wish to save changes
 *	@item: the menu item that was activated
 *
 *	This routine is invoked when the Cancel button (in the save confirmation dialog)
 *	or the "Quit without Saving" button (in the exit dialog) are activated.
 */
int cancel_save_button_activated(struct menu_item *item)
{
	int *event_rc;
	struct dialog_window *dialog = item->user_data;

	event_rc = dialog->user_data;
	if (event_rc != NULL)
		*event_rc = ESHUTDOWN;
	dialog->status = DLG_STATUS_CLOSING;
	return 0;
}

/**
 *	cancel_exit_button_activated - invoked when user cancels application exit
 *	@item: the menu item that was activated
 *
 *	This routine is invoked when the Cancel button in the exit dialog is activated.
 */
int cancel_exit_button_activated(struct menu_item *item)
{
	int *event_rc;
	struct dialog_window *dialog = item->user_data;

	event_rc = dialog->user_data;
	if (event_rc != NULL)
		*event_rc = 0;
	dialog->status = DLG_STATUS_CLOSING;
	return 0;
}

/**
 *	show_commit_on_exit_dialog - invoked to prompt user that changes should be saved before exiting
 *	@void
 *
 *	This routine is invoked to present the user with a dialog to allow him/her/it to save
 *	pending changes before exiting the application.
 */
int show_commit_on_exit_dialog(void)
{
	int rc;
	struct dialog_window *dialog;

	dialog = create_dialog_window(
				_("Unsaved Changes Detected"),
				NULL,
				(dialog_event_handler)NULL,
				(dialog_show_func)NULL,
				(dialog_delete_func)NULL,
				(dialog_delete_cb)NULL,
				_("_Save"),
				(menuitem_activate_cb)save_button_activated,
				_("_Quit without Saving"),
				(menuitem_activate_cb)cancel_save_button_activated,
				&rc);

	print_centered(dialog->win, getmaxy(dialog->win) / 2,
			_("Save changes before quitting?"));

	set_menu_item_sensitivity(dialog->next_button, TRUE);
	set_menu_item_activate_cb(dialog->cancel_button,
				(menuitem_activate_cb)cancel_exit_button_activated);
	set_horizontal_menu_focus(dialog->menu, dialog->next_button);
	process_modal_dialog(dialog);
	return rc;
}

/**
 *	show_save_confirmation_dialog - invoked to confirm a save operation
 *	@title: the window title
 *	@cancel_label: the label text for the cancel button
 *	@prompt: the prompt to display in the window
 *
 *	This routine is invoked to present the user a confirmation dialog for a
 *	save operation.
 */
void show_save_confirmation_dialog(char *title, char *cancel_label, char *prompt)
{
	struct dialog_window *dialog;
	dialog = create_dialog_window(
				title,
				NULL,
				(dialog_event_handler)NULL,
				(dialog_show_func)NULL,
				(dialog_delete_func)NULL,
				(dialog_delete_cb)NULL,
				_("_Save"),
				(menuitem_activate_cb)save_button_activated,
				cancel_label,
				(menuitem_activate_cb)cancel_save_button_activated,
				NULL);

	print_centered(dialog->win, getmaxy(dialog->win) / 2, prompt);

	set_menu_item_sensitivity(dialog->next_button, TRUE);
	set_menu_item_visibility(dialog->cancel_button, FALSE);
	set_horizontal_menu_focus(dialog->menu, dialog->next_button);
	process_modal_dialog(dialog);
}

/**
 *	on_quit_menuitem_activated - callback invoked when user has selected to quit app
 *	@item: the menu item that caused this function to get invoked
 *
 *	This routine is invoked by the Actions->Quit menu item to attempt to save
 *	any pending changes.
 */
int on_quit_menuitem_activated(struct menu_item *item)
{
	int rc;

	if (evms_changes_pending()) 
		rc = show_commit_on_exit_dialog();
	else
		rc = ESHUTDOWN;
	return rc;
}

/**
 *	on_save_menuitem_activated - callback innvoked when user has selected to save changes
 *	@item: the menu item that caused this function to get invoked
 *
 *	This routine is invoked by the Actions->Save menu item to attempt to save
 *	any pending changes.
 */
int on_save_menuitem_activated(struct menu_item *item)
{
	if (evms_changes_pending())
		show_save_confirmation_dialog(_("Save Changes Confirmation"),
						_("_Cancel"),
						_("Do you wish to save changes now or cancel?"));
	else
		print_statusbar_text("There are currently no changes to save.");
	return 0;
}
